unit CodeCalculTopo;
{$INCLUDE CompilationParameters.inc}


// spar solver
{$DEFINE SPAR_SOLVER}
{$UNDEF SPAR_SOLVER}


//****************************************************
// Projet     : GHTopo
// Modules    : Code de calcul topo
// Licence    : General Public License - WITHOUT WARRANTY
// Auteur     : JP CASSOU
// OS         : Windows 9x - Linux
// Langage    : Free-Pascal 2.6.x sous Lazarus
//-----------------------------------------------------
// Module     : ToporobotClasses:
//            : Structure des données d'un dossier Toporobot
// Compatibilité intégrale avec le standard Toporobot 1994.
// 22/08/2012 : Refonte du noyau de GHTopo (ToporobotClasses)
//              en séparant le SGBD et le code de calcul
// 24/08/2012 : Code de calcul validé.
// 14/02/2013 : Réacculturation au code de GHTopo. Petites corrections
// 03/10/2013 : Support du lasermètre Stanley TLM330
// 04/10/2013 : Conversions UTF8 <> ANSI fixed
// 05/11/2013 : Support de la notion de secteurs
// 10/11/2013 : Support de la nouvelle structure de données compilées
// 09/04/2014 : Support des pourcentages (mode 370) corrigé
// 05/05/2014 : Le calcul des antennes ne nécessite plus la création de noeuds
// 23/05/2014 : Optimisation dans le calcul des contours (4x moins de calcul trigo)
// 26/05/2014 : Activation/désactivation du mode débogage des matrices
// 28/05/2014 : Quelques fix, dont un calcul oublié.
// 04/08/2014 : Les entrées sont exclues de la liste des entités,
//              ce qui fixe par la même occasion de nombreux bugs.
// 08/08/2014 : Optimisation dans GetNoeud() augmentant considérablement la vitesse
//              de la phase de découpage en branches.
// 06/09/2014 : Déplacement de la fonction CalculerVisee dans l'unité Common
// 08/10/2014 : Liste des écarts
// 18/12/2014 : CalculerAccroissements() : Nettoyage des branches à accroissement nul
// 06/02/2015 : Les visées en antenne reprennent le couple série/point d'accrochage
//              sous la forme E.Entite_serie = -NumSerie; E.Entite_St = NumPoint
// 07/06/2016 : Les listes de jonctions et de branches passent aux génériques
// 28/08/2016 : Nettoyage de code dans la gestion des branches
// 26/09/2016 : MAJ dans les fonctions de correction d'erreurs systématiques instrumentales
// 07/10/2016 : Mise en place d'un callback pour affichage de la progression (refonte de l'interface)
// 29/03/2017 : TCodeCalculTopo hérite maintenant de TConteneurCodeCalcul situé dans unitConteneurCodeCalcul.pas
// 13/09/2017 : Décomposition de la phase de résolution du système matriciel:
//              La factorisation de B fournit une matrice L via la fonction MakeLMatrix()
//              Cette matrice L est ensuite utilisée par la fonction RemonterSysteme() qui retourne les coordonnées des noeuds
//              Avantage: La méthode des pondérations forfaitaires implique une matrice de compensation B identique
//              pour les trois axes; elle ne doit être construite qu'une seule fois
//              Sa factorisation L est également identique pour les trois axes, ce qui permet de la sortir de la boucle (X, Y, Z)
//              La factorisation de B est l'étape la plus longue du calcul (trois niveaux de boucles)
// 14/09/2017 : La recherche de l'étendue des termes non nuls des matrices B et L (aka R) accélère très fortement
//              la vitesse de calcul qui devient presque instantanée même avec de très gros réseaux
//              (testé avec la fusion de Citon I et II, 1940 noeuds: 4 secondes pour le calcul matriciel, contre 19 sans recherche d'étendue)
// 18/09/2017 : Mémorisation LOL des matrices: Fonctionnel et testé avec succès mais plus lent que la méthode tableaux rectangulaires 2D
// 06/08/2018 : Mémorisation LOT des matrices: idem LOL, un peu plus rapide
// 08/08/2018 : Unification des objets matrices: les TMatrix deviennent des objets TMatriceCreuses
//              -> la directive USE_MATRICES_CREUSES est inutile
// 09/08/2018 : Suppression de certains vecteurs inutiles
//              La fonction MakeVectorSecondMembre() construit les trois seconds membres d'un coup
//              La matrice R est détruite par MakeVectorSecondMembre()
// 02/09/2018 : Support de l'erreur de tourillon (composante verticale de l'erreur de centrage)
//-----------------------------------------------------



{$DEFINE PLUTON_II_DEBUG_MODE}      // débug du code de calcul
{$UNDEF PLUTON_II_DEBUG_MODE}
interface

uses
  {$IFDEF LINUX}
  cthreads,
  {$endif}
  {$INCLUDE SelectionLangues.inc} // insère les unités en fonction de la langue
  StructuresDonnees,



  unitConteneurCodeCalcul,
  ToporobotClasses2012,
  unitobjetserie,
  UnitEntitesExtended,
  UnitClassPalette,
  Common,
  Classes, SysUtils, Graphics, math,
  // choix du modèle de stockage des matrices
  unitMatricesCreusesArray2D,             // OK
  //unitMatricesCreuses_LOL,              // OK
  //unitMatricesCreuses_LOT,                // OK, un peu plus rapide que LOL,
  //unitMatricesCreuses_HASH_TABLE,
  {$IFDEF SPAR_SOLVER}
  SolverHugeSystem,
  {$ENDIF SPAR_SOLVER}
  Forms; // pour Application


// Noeuds
type TNoeud = record
  IDNoeud : integer;
  XN      : double;
  YN      : double;
  ZN      : double;
end;

// LISTE DES BRANCHES


//Table des Noeuds
type TTableNoeuds = array of TNoeud;
// Thread pour traitement d'un axe
Type

{ TMyThreadForAxis }

 TThreadForAxis = class(TThread)
  private
    FMyAxis: TAxeCoordonnees;
    FProc  : TProcedureOfObject;
    procedure ExecProc();
  protected
    procedure Execute; override;
  public
    Constructor Create(CreateSuspended : boolean);
    function Initialiser(const X: TAxeCoordonnees; const MyProc: TProcedureOfObject): boolean;
  end;




// Code de calcul
type

{ TCodeDeCalcul }

TCodeDeCalcul = class(TConteneurCodeCalcul)
  strict private
    R_Matrix: TMatriceCreuse;
    B_Matrix: TMatriceCreuse;
    L_Matrix: TMatriceCreuse;

    {$IFDEF SPAR_SOLVER}
    FMySolver : TMySolver;
    {$ENDIF SPAR_SOLVER}

    // index extrêmes de la matrice B
    FBMatrix_LowIndex  : array of Integer;
    FBMatrix_HighIndex : array of Integer;
    // matrice de pondération diagonale
    W_Mtx    : TVector;
    //A        : TVectorOfTPoint3Df;
    VecteurSecondMembre: TArrayPoints3Df;
    // matrices temporaires
    S_Vector : TArrayPoints3Df; // vecteur local complémentaire
    XX       : TArrayPoints3Df;


    procedure InitialiseMatrices(const m, n: integer);
    procedure FinalizeMatrices();
    //procedure MakeAVector();
    procedure MakeBMatrix();
    procedure MakeVectorSecondMembre();
    function  MakeFactorisedLMatrix(out QL: TMatriceCreuse): boolean;
    procedure MakeWMatrix();
    procedure MakeRMatrix();

    procedure SolveAxis(const Axe: TAxeCoordonnees);
    procedure SolveAxisX();
    procedure SolveAxisY();
    procedure SolveAxisZ();
    function  ExtractValueFromTPoint3Df(const Axe: TAxeCoordonnees; const P: TPoint3Df): double; inline;
    {$IFDEF USE_METHODE_GRADIENT_CONJUGUE}
    // produit matriciel B.V pour la méthode du gradient conjugué
    function  ProduitMatB_by_V(const QV: TVector; var QR: TVector): boolean;
    function  ProduitScal(const QA, QB: TVector; const QN: integer): double;
    {$ENDIF USE_METHODE_GRADIENT_CONJUGUE}
  private
    // fichier de sortie TOP
    FFichierTOP: string;


    // fichier de débogage des matrices
    //FDocumentToporobot: TToporobotStructure2012;
    //FTableJonctions   : TTableJonctionsXYZ;
    // gestion table des branches
    //FTableBranches    : TTableBranchesXYZ;
    //Table des Noeuds
    FTableNoeuds : TTableNoeuds;
    procedure CalculContoursGaleriesExtended(const MyBDDEntites: TBDDEntites; const FichierTOP: string; const DoGenerateFile: boolean);
    function  CalculerCoordNoeuds(): boolean;
    function  EnvoyerJonctionsVersBDDEntites(const MyBDDEntites: TBDDEntites): boolean;

    function GetMaxNode(): integer;


    procedure recopierTablesEntreesReseauxSecteursCodesExpes(const MyBDDEntites: TBDDEntites);
    procedure RemonterSysteme(const QL: TMatriceCreuse; const Axe: TAxeCoordonnees);
    procedure RepartirEcarts();
    // pour la coupe développée
  public
    // gestion des noeuds
    function  GetNoeud(const Serie, Station: integer): integer;
    // lancer le calcul
    function  CalculComplet(const MyBDDEntites: TBDDEntites; const UseDebug: boolean): boolean;
    // générer un fichier de synthèse (i.e. un TAB simplifié avec les coordonnées)
    function  GenererFichierSyntheseAvecCoordonnees(const MyBDDEntites: TBDDEntites; const FileName: string): boolean;
    // si utilisation d'un solveur externe
    {$IFDEF SPAR_SOLVER}
    function  PreparerSolver(const Axe: TAxeCoordonnees): boolean;
    {$ENDIF SPAR_SOLVER}
end;

implementation
const
  FMT_NDSER_PT = '%d-%d';

const
  PALIERS_INDICATION: integer = 100;
  NBSTEPS = 7;


{ TMyThreadForAxis }



constructor TThreadForAxis.Create(CreateSuspended: boolean);
begin
  FreeOnTerminate := true;
  inherited Create(CreateSuspended);
end;

function TThreadForAxis.Initialiser(const X: TAxeCoordonnees; const MyProc: TProcedureOfObject): boolean;
begin
  FMyAxis := X;
  FProc   := MyProc;
end;

procedure TThreadForAxis.ExecProc();
begin
  FProc();
end;

procedure TThreadForAxis.Execute;
begin
  AfficherMessage(Format('%s.Execute: %s', [ClassName, chr(88 + ord(FMyAxis))]));

  //while (not Terminated) and ([any condition required]) do
  while (not Terminated) do
  begin
    Synchronize(FProc);
    //pass;
    (*...
    [here goes the code of the main thread loop]
    ...
    if NewStatus <> fStatusText then
      begin
        fStatusText := newStatus;
        Synchronize(@Showstatus);
      end;
     //*)
  end;
end;


//******************************************************************************
(* TCodeDeCalcul *)
procedure TCodeDeCalcul.InitialiseMatrices(const m, n: integer);
begin
  AfficherMemoryUsage();
  // matrices temporaires
  SetLength(XX, 0);
  SetLength(XX, n + 1);
  SetLength(S_Vector, 0);
  SetLength(S_Vector, m +  1);
  //SetLength(A, 0);
  //SetLength(A,  m + 1);
  SetLength(VecteurSecondMembre, 0);
  SetLength(VecteurSecondMembre, n + 1);
  //solveur
  {$IFDEF SPAR_SOLVER}
    FMySolver := TMySolver.Create;
  {$ENDIF SPAR_SOLVER}

  R_Matrix := TMatriceCreuse.Create;
  R_Matrix.Initialiser(m, n, 'R');
  B_Matrix := TMatriceCreuse.Create;
  B_Matrix.Initialiser(n+1, n+1, 'B');
  L_Matrix := TMatriceCreuse.Create;
  L_Matrix.Initialiser(n+1, n+1, 'L');
  AfficherMemoryUsage();
end;

procedure TCodeDeCalcul.FinalizeMatrices();
begin
  try
    //R_Matrix.Finaliser();
    B_Matrix.Finaliser();
    L_Matrix.Finaliser();
    AfficherMemoryUsage();
  finally
    //R_Matrix.Free;
    B_Matrix.Free;
    L_Matrix.Free;
  end;
  //SetLength(A    , 0);
  SetLength(VecteurSecondMembre, 0);
  SetLength(W_mtx, 0);
  SetLength(XX   , 0);
  SetLength(S_Vector, 0);
  {$IFDEF SPAR_SOLVER}
    try
      FMySolver.Finaliser();

    finally
      FMySolver.Free;
    end;
  {$ENDIF SPAR_SOLVER}
end;
function TCodeDeCalcul.GetMaxNode(): integer;
var
  Branche: TBrancheXYZ;
  i      : integer;
  M      , QNbBranches: integer;
begin
  M := Low(Integer);
  QNbBranches := GetNbBranches();
  for i := 1 to QNbBranches - 1 do
  begin
    Branche:= GetBranche(i);
    if (Branche.NoeudDepart > M)  then M := Branche.NoeudDepart;
    if (Branche.NoeudArrivee > M) then M := Branche.NoeudArrivee;
  end;
  Result:=M;
  AfficherMemoryUsage();
end;
// matrice de connexion (ici, matrice d'incidence)
procedure TCodeDeCalcul.MakeRMatrix();
var
  i,j, QNbBranches, QNbNodes: integer;
  Branche: TBrancheXYZ;
  EWE: String;
  WU : integer;
  YaAutoloop: boolean;
begin

  QNbBranches := GetNbBranches();
  QNbNodes    := GetMaxNode();
  for i := 1 to QNbBranches - 1 do
  begin
    YaAutoloop  := false;
    if (i MOD 100 = 0) then
    begin
      EWE := Format('Matrice d''incidence: ' + rsLINEOFNB, [i, QNbBranches]);
      if (Assigned(self.ProcDispProgression)) then
          self.ProcDispProgression(EWE, i, 0, QNbBranches);
    end;
    Branche := GetBranche(i);
    for j := 1 to QNbNodes do
    begin
      // /!\ La matrice d'incidence des graphes orientés ne supporte pas les boucles, par construction
      // cf https://fr.wikipedia.org/wiki/Matrice_d%27incidence
      // En raison de la quantité considérable de code à élaborer au vu du très faible nombre d'occurrences de boucles simples,
      // on fait le choix de ne pas supporter cette configuration.
      // Toutefois, on signale la série incriminée.
      if ((Branche.NoeudDepart = j) and (Branche.NoeudArrivee = j)) then
      begin
        EWE := Format('[i=%d] La branche %d (%d->%d) appartenant à la série %d se ferme sur elle-même', [i, Branche.NoeudDepart, Branche.NoeudArrivee, Branche.NoBranche, Branche.NoSerie]);
        AfficherMessageErreur(EWE);
        //R_Matrix.SetValeur(i, j, 2, True);
        //YaAutoloop := true;
        //break;
      end;
      if      (Branche.NoeudDepart  = j) then R_Matrix.SetValeur(i,j, -1, True)
      else if (Branche.NoeudArrivee = j) then R_Matrix.SetValeur(i,j,  1, True)
                                         else R_Matrix.SetValeur(i,j,  0, True);

      //R_Matrix.SetValeur(i,j, beuh(Branche, j), True);
    end;
    if (YaAutoloop) then Continue;
  end;
  R_Matrix.RemoveAllZeroes();
  AfficherMemoryUsage();
  //R_Matrix.Lister();
  //R_Matrix.CreerRepresentationMatriceEnImage(GetGHTopoDirectory()+ 'R_Matrix.bmp');
end;
// matrice de compensation = Rt.W.R
procedure TCodeDeCalcul.MakeBMatrix();
var
  i,j,k: integer;
  ww, R1, R2: double;
  //f: TextFile;
  q, fi, li, QNbNodes: integer;
  QTo, QT1: TDateTime;
begin
  QTo := Now();
  AfficherMessage(GetResourceString(rsBUILDMATRICE));
  QNbNodes     := GetMaxNode();
  // rechercher les index mini et maxi des valeurs non nulles de RMatri
  SetLength(FBMatrix_LowIndex,0);
  SetLength(FBMatrix_HighIndex,0);
  SetLength(FBMatrix_LowIndex  , QNbNodes + 1);
  SetLength(FBMatrix_HighIndex , QNbNodes + 1);
  AfficherMemoryUsage();
  AfficherMessage('---> '+ GetResourceString(rsFIND_SUM_LIMITS));
  Application.ProcessMessages;
  for i:=1 to QNbNodes do
  begin
    //if (i MOD 20 = 0) then AfficherMessage(Format(GetResourceString(rsLINEOFNB),[i, QNbNodes]));
    for j:= 1 to R_Matrix.MaxRow do //QNbBranches - 1 do
    begin
      if (Abs(R_Matrix.GetValeur(j, i)) > 0.00) then
      begin
        FBMatrix_LowIndex[i] := j;
        Break;
      end;
    end;
  end;

  Application.ProcessMessages;
  for i:=1 to QNbNodes do
  begin
    //if (i MOD 20 = 0) then AfficherMessage(Format(GetResourceString(rsLINEOFNB),[i, QNbNodes]));
    for j := R_Matrix.MaxRow downto 1 do     //for j := QNbBranches - 1 downto 1 do
    begin
      // Ne pas utiliser la fonction IsZero (lenteur)
      if (Abs(R_Matrix.GetValeur(j,i)) > 0.00) then
      begin
        FBMatrix_HighIndex[i] := j;
        Break;
      end;
    end;
  end;
  AfficherMessage('---> '+ GetResourceString(rsPRODUIT_MAT_Rt_W_R));
  Application.ProcessMessages;
  for i := 1 to QNbNodes do
  begin
    if (i MOD 100 = 0) then
    begin
      Application.ProcessMessages;
      if (Assigned(self.ProcDispProgression)) then
          self.ProcDispProgression(Format('Matrice de compensation: ' + rsLINEOFNB, [i, QNbNodes]), i, 0, QNbNodes);
      AfficherMessage(Format('--> Ligne %d / %d', [i, QNbNodes]));
    end;
    for j := 1 to i do
    begin
      ww := 0;
      for k := FBMatrix_LowIndex[i] to FBMatrix_HighIndex[i] do
      begin
        // Les comparaisons à zéro de réels sont plus coûteuses que les multiplications
        //R1 := R_Matrix.GetValeur(k,i);
        //R2 := R_Matrix.GetValeur(k,j);
        //if (not(IsZero(R1) OR IsZero(R2))) then ww += R1 * W_mtx[k] * R2;
        ww += R_Matrix.GetValeur(k,i) * W_mtx[k] * R_Matrix.GetValeur(k,j);
      end;
      B_Matrix.SetValeur(i, j, WW, true);//B[i,j]:= ww;
    end;
  end;
  Application.ProcessMessages;
  // remplissage de la symétrie
  for i:=1 to QNbNodes - 1 do
    for j:=1+i to QNbNodes do
      B_Matrix.SetValeur(i,j, B_Matrix.GetValeur(j,i), True);
  B_Matrix.RemoveAllZeroes();
  QT1 := Now();
  AfficherMessageErreur('Temps de construction de la matrice B: ' + DatePascalToDateHeureCondensee(QT1 - QTo));
  //B_Matrix.Lister();
end;
function TCodeDeCalcul.MakeFactorisedLMatrix(out QL: TMatriceCreuse): boolean;
var
   i, j, k   , QNbNodes: integer;
   vv        : double;
   ww        , Qik, Qjk, Q2: double;
   LowIndex  : array of Integer;
   HighIndex : array of Integer;
   QTo, QT1: TDateTime;
begin
   Result := false;
   QTo := Now();
   QNbNodes := GetMaxNode();
   AfficherMessage(Format('%s.MakeFactorisedMatrix()', [ClassName]));
   AfficherMessage('---> '+ GetResourceString(rsFIND_SUM_LIMITS));

   try
     SetLength(LowIndex, 0);
     SetLength(LowIndex, QNbNodes + 1);
     AfficherMemoryUsage();
     // rechercher les index mini et maxi des valeurs non nulles de BMatrix
     // (énorme accélération de la vitesse de calcul)
     // /!\ Ne pas utiliser QL.RecenserLowIndexes(): la matrice QL est modifiée en permanence
     //     et utilise les LowIndex de la matrice B
     for i:=1 to QNbNodes do
     begin
       if (i MOD 200 = 0) then
       begin
         AfficherMessage(Format(GetResourceString(rsLINEOFNB),[i, QNbNodes]));
         if (Assigned(self.ProcDispProgression)) then self.ProcDispProgression(GetResourceString(rsFACTORISEMATRICE), i, 0, QNbNodes);
       end;
       for j:=1 to QNbNodes do
       begin
         if (abs(B_Matrix.GetValeur(j, i)) > 0) then
         begin
           LowIndex[i] := j;
           Break;
         end;
       end;
     end;
     //*)
     AfficherMessage(GetResourceString(rsDESCENTE));
     // Durée du calcul sur Citon II 2017-09-09: 6.90 s
     for i := 1 to QNbNodes do
     begin
       if (Assigned(self.ProcDispProgression)) then
          self.ProcDispProgression(Format(GetResourceString(rsDESCENTE + ': ' + rsLINEOFNB),[i, QNbNodes]), i, 0, QNbNodes);
       if (i mod PALIERS_INDICATION = 0) then AfficherMessage(Format(GetResourceString(rsLINEOFNB),[i, QNbNodes]));
       vv := 0;
       for j := 1 to i - 1 do
       begin
         Q2 := QL.GetValeur(i,j);
         vv += Q2 * Q2;
       end;
       QL.SetValeur(i,i, Sqrt(abs(B_Matrix.GetValeur(i,i)) - vv), false);
       for j := 1 + i to QNbNodes do
       begin
         ww := 0;
         //for k := 1 to i-1 do ww += QL[i,k] * QL[j,k];
         for k := LowIndex[i] to i-1 do
         begin
           ww += QL.GetValeur(i,k) * QL.GetValeur(j,k);  // très forte accélération de la vitesse de calcul
         end;
         QL.SetValeur(j,i, (B_Matrix.GetValeur(i,j) - ww) / (QL.GetValeur(i,i) + 1e-24), false);
       end;
     end;
     //QL.CreerRepresentationMatriceEnImage(GetGHTopoDirectory() + 'L_Matrix.bmp');
     QL.RemoveAllZeroes();
     //QL.Lister();
     QT1 := Now();
     AfficherMessageErreur('Temps de factorisation de la matrice B: ' + DatePascalToDateHeureCondensee(QT1 - QTo));
     Result := True;
   except
   end;
 end;
 procedure TCodeDeCalcul.RemonterSysteme(const QL: TMatriceCreuse; const Axe: TAxeCoordonnees);
 var
   i, j, k   : integer;
   QNbNodes, QNbBranches: integer;
   vv        : double;
   ww        , QQ666, tmp: double;
   QPosPtZero: TPoint3Df;
 begin
   AfficherMessageErreur('RemonterSysteme: ');
   QNbNodes    := GetMaxNode();
   QNbBranches := GetNbBranches();
   QPosPtZero  := MakeTPoint3Df(0.00, 0.00, 0.00);
   try
     for i :=  1 to QNbNodes  do
     begin
       case Axe of
         axisX: XX[i].X := 0.00;
         axisY: XX[i].Y := 0.00;
         axisZ: XX[i].Z := 0.00;
       end;
     end;
     // second membre; triangularisation
     AfficherMessage(GetResourceString(rsTRIANGULARISATION));
     for i :=  1 to QNbNodes  do
     begin
       //if (i mod PALIERS_INDICATION = 0) then  AfficherMessage(Format(GetResourceString(rsLINEOFNB),[i, NbNodes ]));
       //if (Assigned(self.ProcDispProgression)) then self.ProcDispProgression(Format(GetResourceString(rsTRIANGULARISATION + ': ' + rsLINEOFNB),[i, NbNodes ]), i, 0, NbNodes );
       ww := 0;
       for k:= 1 to i - 1 do
         ww := ww + L_Matrix.GetValeur(i,k) * ExtractValueFromTPoint3Df(Axe, S_Vector[k]);
       QQ666 := L_Matrix.GetValeur(i,i) + 1e-24;
       case Axe of
         axisX: S_Vector[i].X  := (VecteurSecondMembre[i].X - ww) / QQ666;
         axisY: S_Vector[i].Y  := (VecteurSecondMembre[i].Y - ww) / QQ666;
         axisZ: S_Vector[i].Z  := (VecteurSecondMembre[i].Z - ww) / QQ666;
       end;
     end;
     // remontée du système; inconnues recherchées
     AfficherMessage(GetResourceString(rsREMONTEE));
     AfficherMessage('--> Remontée');
     for i := QNbNodes downto 1 do
     begin
       if (i mod PALIERS_INDICATION = 0) then AfficherMessage(Format(GetResourceString(rsLINEOFNB),[i, QNbNodes]));
       if (Assigned(self.ProcDispProgression)) then self.ProcDispProgression(Format(GetResourceString(rsREMONTEE + ': ' + rsLINEOFNB),[i, QNbNodes]), i, 0, QNbNodes);
       ww := 0;
       for k := 1 + i to QNbNodes do
         ww := ww + L_Matrix.GetValeur(k,i) * ExtractValueFromTPoint3Df(Axe, XX[k]);
       tmp := (ExtractValueFromTPoint3Df(Axe, S_Vector[i]) - ww) / (L_Matrix.GetValeur(i,i) + 1e-24);
       case Axe of
         axisX: XX[i].X := tmp;
         axisY: XX[i].Y := tmp;
         axisZ: XX[i].Z := tmp;
       end;
     end;
     XX[0] := XX[1];
     // Table des Noeuds
     QPosPtZero := self.DocuTopo.GetPositionDuPointZero();
     for i := 1 to QNbNodes do
     begin
       FTableNoeuds[i].IDNoeud := i;
       if (i mod 200 = 0) then AfficherMessage(Format(GetResourceString(rsLINEOFNB),[i, QNbNodes]));
       case Axe of
         axisX: FTableNoeuds[i].XN := (XX[i].X - XX[0].X) + QPosPtZero.X;
         axisY: FTableNoeuds[i].YN := (XX[i].Y - XX[0].Y) + QPosPtZero.Y;
         axisZ: FTableNoeuds[i].ZN := (XX[i].Z - XX[0].Z) + QPosPtZero.Z;
       end;
     end;

     // ++++++++++++++++
     // Passage par cette ligne = c'est OK
     AfficherMessage(Format(GetResourceString(rsCOORDONNEES_OK),[QNbNodes]));
   finally
     AfficherMessage(GetResourceString(rsDEL_TEMP_MATRIX));
   end;
 end;

// Lancer le calcul (nouvelle version)
function  TCodeDeCalcul.CalculComplet(const MyBDDEntites: TBDDEntites; const UseDebug: boolean): boolean;
var
  UneVisee: TUneVisee;
  Branche: TBrancheXYZ;
  EWE: String;
begin
  Result := false;
  AfficherMessage(GetResourceString(rsPURGE_TABLE_BRCH));
  if (Assigned(self.ProcDispProgression)) then self.ProcDispProgression(GetResourceString(rsPURGE_TABLE_BRCH), -1, -1, -1);
  ViderTableJonctionsBranches();
  AfficherMemoryUsage();
  // Purger la table des branches
  try
    // Branche0
    UneVisee := EmptyVisee('');
    SetLength(Branche.PointsTopo, 0);              // N
    addViseeAtBranche(Branche, UneVisee);          // N
    Branche.NoBranche    := 0;
    Branche.NomBranche   :='Point 0';
    Branche.NoeudDepart  := 0;
    Branche.NoeudArrivee := 0;
    AddBranche(Branche);

    AfficherMessage(Format('%s.CalculComplet',[self.ClassName]));
    Result := False;
    AfficherMessage('--> Jonctions');
    RecenserJonctions();
    AfficherMessage('--> Branches');
    RecenserBranches();
    AfficherMessage('--> Accroissements');
    CalculerAccroissements();  // calcul des accroissements des branches
    // élagage du réseau: suppression des branches en cul de sac
    //AfficherMessage('--> Elagage du réseau');
    //ElaguerBranches();

    AfficherMessage('--> Noeuds');
    CalculerCoordNoeuds();     // calcul matriciel des coordonnées des noeuds
    // AfficherMessage(Format('%s: %s', [{$I %FILE%}, {$I %LINE%}]));
    AfficherMessage('--> Repartition');
    RepartirEcarts();          // répartition parallèle des écarts
    //AfficherMessage('--> Calcul branches en cul de sac');
    //CalculBranchesElaguees();
    EWE := ChangeFileExt(FFichierTOP, '.pot');
    AfficherMessage('--> Contours');
    CalculContoursGaleriesExtended(MyBDDEntites, EWE, false);  // calcul des contours des galeries
    ViderTableJonctionsBranches();

    SetLength(FBMatrix_LowIndex,0);
    SetLength(FBMatrix_HighIndex,0);
    AfficherMemoryUsage();
    Result := True;

  except

  end;
end;


function TCodeDeCalcul.GetNoeud(const Serie, Station: integer): integer;
var
  i   : integer;
  Jonc: TJonctionXYZ;
begin
  Result := -1;
  for i:=0 to GetNbJonctions() - 1 do
  begin
    Jonc := GetJonction(i);
    //S:=Format(FMT_NDSER_PT,[Serie, Station]);                         // ancienne méthode
    //if (S = Jonc.IDJonction) then
    if ((Serie = Jonc.NoSer) and (Station = Jonc.NoSt)) then Exit(i);   // nouvelle méthode

  end;
end;



function TCodeDeCalcul.CalculerCoordNoeuds(): boolean;
var
  T0, T1: TDateTime;
  tm: TDateTime;
  t: integer;
  Branche : TBrancheXYZ;
  QNbBranches, QNbNodes: integer;
  WU: String;
begin
  Result := false;
  T0 := Now();

  tm := Now();
  QNbBranches := GetNbBranches();
  QNbNodes := GetMaxNode();

  InitialiseMatrices(QNbBranches, QNbNodes);

  SetLength(FTableNoeuds, 0);
  SetLength(FTableNoeuds, QNbNodes+1);
  AfficherMessage('');
  AfficherMessage(GetResourceString(rsBIND_MATRICIAL_PROBLEM));
  AfficherMessage('=================================');
  AfficherMessage(GetResourceString(rsBIND_INCIDENCE_MATRIX));
  MakeRMatrix(); // matrice d'incidence (assemblage)
  MakeWMatrix(); // matrice de pondération (diagonale)
  //MakeAVector(); // reprend les accroissements des branches
  MakeBMatrix(); // matrice de compensation
  MakeVectorSecondMembre(); // second membre

  AfficherMessage('');
  AfficherMessage(GetResourceString(rsFACTORISEMATRICE));
  // factorisation de la matrice de compensation
  MakeFactorisedLMatrix(L_Matrix);
  AfficherMessage('--> MakeFactorisedLMatrix');
  tm := Now();
  AfficherMessage(GetResourceString(rsCALC_COORDS_NOEUDS));


  //*)

  // version monocoeur

  SolveAxis(axisX);
  SolveAxis(axisY);
  SolveAxis(axisZ);
  //*)



  tm := Now() - tm;
  //WriteTableNodesInFile(ChangeFileExt(FDocumentToporobot.GetDatabaseName, '.nds'));
  // assignation des écarts
  for t:=1 to QNbBranches - 1 do //GetNbBranches() - 1 do
  begin
    Branche := GetBranche(t);
    with Branche do
    begin
      XDepart  := FTableNoeuds[NoeudDepart].XN;
      YDepart  := FTableNoeuds[NoeudDepart].YN;
      ZDepart  := FTableNoeuds[NoeudDepart].ZN;
      XArrivee := FTableNoeuds[NoeudArrivee].XN;
      YArrivee := FTableNoeuds[NoeudArrivee].YN;
      ZArrivee := FTableNoeuds[NoeudArrivee].ZN;
    end;
    PutBranche(t,Branche);
  end;
  AfficherMessage('');
  // libération mémoire
  //-- finalisation
  AfficherMessage(GetResourceString(rsFREE_TEMP_VARS));
  FinalizeMatrices();
  T1 := Now();
  AfficherMessageErreur(DatePascalToDateHeureCondensee(T1-T0));
  result := true;
end;
// répartition des écarts de fermeture entre les noeuds
procedure TCodeDeCalcul.RepartirEcarts();
var
  i, s, QNbBranches: integer;
  V, V1, V2: TUneVisee;
  Branche: TBrancheXYZ;
  EcartX : double;
  EcartY : double;
  EcartZ : double;
  EpsX, EpsY, EpsZ: double;
  Xo, Yo, Zo: double;
  XXo,YYo, ZZo: double;
  R      : double;
  DevelBranche: Double;
begin;

  AfficherMessage('');
  AfficherMessage(GetResourceString(rsREPARTIR_ECARTS));
  QNbBranches := GetNbBranches();
  AfficherMessageErreur(Format('Deformation du reseau: %d branches', [QNbBranches]));
  AfficherMessageErreur('=========================================');
  AfficherMessageErreur('Branche; X1; Y1; Z1; X2; Y2; Z2; Devel; EcartX; EcartY; EcartZ; EcartTotal; EpsX; EpsY; EpsZ; EpsTotal');

  for i:=1 to QNbBranches - 1 do
  begin
    Branche := GetBranche(i);

    EcartX := (Branche.XArrivee - Branche.XDepart) - Branche.DeltaX;
    EcartY := (Branche.YArrivee - Branche.YDepart) - Branche.DeltaY;
    EcartZ := (Branche.ZArrivee - Branche.ZDepart) - Branche.DeltaZ;
    // calcul de R
    R := 1e-10;
    for s := 0 to High(Branche.PointsTopo) do  //Branche.PointsTopo.GetNbElements() - 1do
    begin
      V := Branche.PointsTopo[s] ; //V := GetBrStation(i, s);
      R := R + Hypot3D(V.DeltaX, V.DeltaY, V.DeltaZ);
    end;
    DevelBranche := R;
    // mettre l'azimut de V[1] dans V[0]
    if (High(Branche.PointsTopo) > 0) then //if (Branche.PointsTopo.Count > 1) then
    begin
      V1 := Branche.PointsTopo[0];   //V1 := GetBrStation(i,0);
      V2 := Branche.PointsTopo[1];   //V2 := GetBrStation(i,1);
      V1.Azimut  :=  V2.Azimut;
      Branche.PointsTopo[0] := V1;   // PutBrStation(i, 0, V1);
    end;
    // calcul de répartition
    EpsX := EcartX / DevelBranche;
    EpsY := EcartY / DevelBranche;
    EpsZ := EcartZ / DevelBranche;
    Xo := Branche.XDepart;
    Yo := Branche.YDepart;
    Zo := Branche.ZDepart;
    R  := 1e-10;
    XXo := Branche.XDepart;
    YYo := Branche.YDepart;
    ZZo := Branche.ZDepart;
    for s := 0 to High(Branche.PointsTopo) do //for s := 0 to Branche.PointsTopo.GetNbElements() - 1 do
    begin
      V := Branche.PointsTopo[s]; //V :=GetBrStation(i, s);
      R:=R + Hypot3D(V.DeltaX, V.DeltaY, V.DeltaZ);
      XXo:=XXo + V.DeltaX;
      YYo:=YYo + V.DeltaY;
      ZZo:=ZZo + V.DeltaZ;

      Xo:=XXo + R * EpsX;
      Yo:=YYo + R * EpsY;
      Zo:=ZZo + R * EpsZ;

      v.DeltaX := Xo;
      v.DeltaY := Yo;
      v.DeltaZ := Zo;

      Branche.PointsTopo[s] := V; //PutBrStation(i, s, V);
    end;
    PutBranche(i, Branche);
  end;    // for i, Br
  AfficherMessageErreur('=========================================');

end;

// accroissements des branches
(*
procedure TCodeDeCalcul.MakeAVector();
var
  k, QNbBranches: integer;
  Branche: TBrancheXYZ;
begin
  AfficherMessageErreur('--> MakeAVector()');

  QNbBranches := GetNbBranches();


  for k:=1 to QNbBranches - 1 do
  begin
    Branche := GetBranche(k);
    A[k].X := Branche.DeltaX;
    A[k].Y := Branche.DeltaY;
    A[k].Z := Branche.DeltaZ;
  end;
  A[1] := MakeTPoint3Df(0.01, 0.01, 0.01);
end;
//*)

// matrice de pondération
procedure TCodeDeCalcul.MakeWMatrix();
var
  i, QNbBranches: integer;
begin
  QNbBranches := GetNbBranches();
  SetLength(W_mtx, 0);
  SetLength(W_mtx, QNbBranches + 1);
  for i:=1 to QNbBranches - 1 do W_mtx[i] := Getbranche(i).Rigidite;  // paramètre de raideur de la branche
end;
// second membre
procedure TCodeDeCalcul.MakeVectorSecondMembre();
var
  i,k    , QNbBranches, QNbNodes: integer;
  wwx, wwy, wwz, EWE: double;
  MyBranche: TBrancheXYZ;
begin;
  AfficherMessageErreur('--> MakeVectorSecondMembre()');
  QNbBranches := GetNbBranches();
  QNbNodes    := GetMaxNode();
  for i:=1 to QNbNodes do
  begin
    wwx := 0.00;
    wwy := 0.00;
    wwz := 0.00;
    for k := 1 to QNbBranches - 1 do
    begin
      EWE := R_Matrix.GetValeur(k,i) *  W_mtx[k];
      if (k = 1) then  // sécurisation branche 1
      begin
        MyBranche.DeltaX := 0.01;
        MyBranche.DeltaY := 0.01;
        MyBranche.DeltaZ := 0.01;
      end;
      MyBranche := GetBranche(k);
      wwx += EWE * MyBranche.DeltaX; //A[k].X;
      wwy += EWE * MyBranche.DeltaY; //A[k].Y;
      wwz += EWE * MyBranche.DeltaZ; //A[k].Z;
    end;
    VecteurSecondMembre[i].X := wwx;
    VecteurSecondMembre[i].Y := wwy;
    VecteurSecondMembre[i].Z := wwz;
  end;
  // on peut jeter la matrice R
  try
    R_Matrix.Finaliser();
  finally
    R_Matrix.Free;
  end;
end;

procedure TCodeDeCalcul.SolveAxis(const Axe: TAxeCoordonnees);
var
  i: Integer;
  QPosPtZero: TPoint3Df;
  QFilename: string;
begin
  AfficherMessage(GetResourceString(rsAXIS) + Chr(88 + Ord(Axe)));
  QFileName := GetGHTopoDirectory() + 'Matrice_B.mtx';
  AfficherMessageErreur(GetResourceString(rs2NDMEMBER));
  {$IFDEF SPAR_SOLVER}
     if (not FMySolver.Initialiser(GetMaxNode(), QFilename)) then exit();

     AfficherMessageErreur('** PreparerSolver()');
     PreparerSolver(Axe);
     AfficherMessageErreur('** Resoudre()');
     FMySolver.Resoudre();
     QPosPtZero := self.DocuTopo.GetPositionDuPointZero();
     for i := 0 to GetMaxNode() do
     begin
       case Axe of
         axisX: FTableNoeuds[i].XN := FMySolver.GetXValue(i) + QPosPtZero.X;
         axisY: FTableNoeuds[i].YN := FMySolver.GetXValue(i) + QPosPtZero.Y;
         axisZ: FTableNoeuds[i].ZN := FMySolver.GetXValue(i) + QPosPtZero.Z;
       end;
     end;
     FMySolver.Finaliser();
  {$ELSE}
     AfficherMessageErreur(GetResourceString(rsNODECOORDINATES));
     RemonterSysteme(L_Matrix, Axe);
  {$ENDIF SPAR_SOLVER}
end;




procedure TCodeDeCalcul.SolveAxisX();
begin
  AfficherMessageErreur('SolveAxis X');
  SolveAxis(axisX);
end;

procedure TCodeDeCalcul.SolveAxisY();
begin
  AfficherMessageErreur('SolveAxis Y');
  SolveAxis(axisY);
end;

procedure TCodeDeCalcul.SolveAxisZ();
begin
  AfficherMessageErreur('SolveAxis Z');
  SolveAxis(axisZ);
end;

function TCodeDeCalcul.ExtractValueFromTPoint3Df(const Axe: TAxeCoordonnees; const P: TPoint3Df): double;
begin
  case Axe of
    axisX: Result := P.X;
    axisY: Result := P.Y;
    axisZ: Result := P.Z;
  end;
end;

// calcul des contours de galeries (nouvelle version)
procedure TCodeDeCalcul.CalculContoursGaleriesExtended(const MyBDDEntites: TBDDEntites; const FichierTOP: string; const DoGenerateFile: boolean);
var
  VA          : TViseeAntenne;
  //Palette256  : TPalette256;
  Br          : Integer;
  EPSG        : TLabelSystemesCoordsEPSG;
  MyBranche   : TBrancheXYZ;
  function CalcViseeAntenne(const VA: TViseeAntenne;
                            out   OutEE: TEntiteEtendue): boolean;
  var
    VS      : TUneVisee;
    DX, DY, DZ, DP: double;
    EE: TEntiteEtendue;
  begin
    result := false;
    DX := 0.00; DY := 0.00; DZ := 0.00;
    if (MyBDDEntites.GetEntiteViseeFromSerSt(VA.SerieDepart, VA.PtDepart, EE)) then
    begin
      DX := EE.Une_Station_2_X;
      DY := EE.Une_Station_2_Y;
      DZ := EE.Une_Station_2_Z;
    end
    else
    begin
      exit(false);
    end;
    //*)
    VS.IDSecteur       := 0; // VA.Secteur; // TODO: Secteur à implémenter
    VS.Code            := VA.Code;
    VS.Expe            := VA.Expe;
    VS.Longueur        := VA.Longueur;
    VS.Azimut          := VA.Azimut;
    VS.Pente           := VA.Pente;
    VS.LG := 0.00;
    VS.LG := 0.00;
    VS.HZ := 0.00;
    VS.HN := 0.00;
    VS.Commentaires     := VA.Commentaires;
    VS.IDTerrainStation := VA.IDTerrainStation;
    VS.TypeVisee        := tgVISEE_RADIANTE;
    CalculerVisee(VS,
                  self.DocuTopo.GetCodeByNumero(VS.Code),
                  self.DocuTopo.GetExpeByNumero(VS.Expe),
                  DX, DY,
                  1.00,
                  DZ, DP);
    DX := EE.Une_Station_2_X + VS.DeltaX;
    DY := EE.Une_Station_2_Y + VS.DeltaY;
    DZ := EE.Une_Station_2_Z + VS.DeltaZ;

    //******************************

    // TODO: Dans GHCaveDraw, les antennes ne doivent servir que de guides et
    //       n'ont pas à être capturées
    OutEE.Entite_Serie    := -VA.SerieDepart; // -QNo;
    OutEE.Entite_Station  :=  VA.PtDepart;//0;
    // code et expé
    OutEE.eCode       := VA.Code;
    OutEE.eExpe       := VA.Expe;
    OutEE.eSecteur    := VA.Secteur;
    OutEE.eEntrance   := VA.EntranceRatt;
    OutEE.eReseau     := VA.Reseau;
    OutEE.Type_Entite := tgVISEE_RADIANTE;
    OutEE.DateLeve    := Now();
    OutEE.Drawn           := True;   // drapeau
    // données originales
    OutEE.oLongueur       := VA.Longueur;
    OutEE.oAzimut         := VA.Azimut;
    OutEE.oPente          := VA.Pente;
    OutEE.oLG             := 0.00;
    OutEE.oLD             := 0.00;
    OutEE.oHZ             := 0.00;
    OutEE.oHN             := 0.00;
    // centerline
    OutEE.Une_Station_1_X := EE.Une_Station_2_X;
    OutEE.Une_Station_1_Y := EE.Une_Station_2_Y;
    OutEE.Une_Station_1_Z := EE.Une_Station_2_Z;
    OutEE.Une_Station_2_X := DX;
    OutEE.Une_Station_2_Y := DY;
    OutEE.Une_Station_2_Z := DZ;
    // habillage
    OutEE.X1PD := EE.Une_Station_2_X;
    OutEE.Y1PD := EE.Une_Station_2_Y;
    OutEE.X1PG := EE.Une_Station_2_X;
    OutEE.Y1PG := EE.Une_Station_2_Y;
    OutEE.Z1PH := EE.Une_Station_2_Z;
    OutEE.Z1PB := EE.Une_Station_2_Z;
    OutEE.X2PD := DX;
    OutEE.Y2PD := DY;
    OutEE.X2PG := DX;
    OutEE.Y2PG := DY;
    OutEE.Z2PH := DZ;
    OutEE.Z2PB := DZ;
    // couleur par défaut
    OutEE.ColourByDepth := clGray;
    // commentaires
    OutEE.oIDLitteral   := VA.IDTerrainStation;
    OutEE.oCommentaires := VA.Commentaires;
    OutEE.IsPOI         := false;
    // passage ici = OK
    result := true;
  end;

  //****************************************************************************
  function TraiterBranche(const QBranche: TBrancheXYZ; const Br: integer): boolean;
  var
    TabVisee    : array of TStation;
    St          : Integer;
    V1          : TUneVisee;
    Expe        : TExpe;
    Code        : TCode;
    AlphaD      : Double;
    qCosAlphaD  : double;
    qSinAlphaD  : double;

    //AlphaG      : Double;
    EWE         : Integer;
    AReseau     : TReseau;
    QEntite     : TEntiteEtendue;
    Cnt         : integer;
    QEntrance   : TEntrance;
  begin
    Result := False;
    //try
      EWE := 1 + High(QBranche.PointsTopo); // QBranche.PointsTopo.GetNbElements();
      //AfficherMessage(Format('--> Branche: %d - %d points', [Br, EWE]));

      Cnt := 0;
      AReseau := self.DocuTopo.GetReseau(QBranche.NoReseau);
      SetLength(TabVisee, 0);
      SetLength(TabVisee, 2 + EWE);
      for St:= 0 to EWE - 1 do
      begin
        //AfficherMessage(Format('----> Point: %d', [St]));
        V1 := QBranche.PointsTopo[st]; //GetBrStation(Br, St);
        with TabVisee[St] do
        begin
          NumPoint  := V1.NoVisee; //  NumPoint:=V1.NoViseeSer;
          stSecteur := V1.IDSecteur;
          Expe      := self.DocuTopo.GetExpeByNumero(V1.Expe);          // WU       := V1.Expe;
          Code      := self.DocuTopo.GetCodeByNumero(V1.Code);               // WU       := V1.Code;
          stCode    := Code.IDCode;
          stExpe    := Expe.IDExpe;
          Couleur   := FPalette256.GetColorByIndex(Expe.Couleur);
          TypeVisee := V1.TypeVisee;

          Date      := GetSecuredDate(Expe.AnneeExpe, Expe.MoisExpe, Expe.JourExpe); //TODO Corriger dates

          Longueur := V1.Longueur;
          Azimut   := V1.Azimut;
          Pente    := V1.Pente;

          LD      := V1.LD;
          LG      := V1.LG;
          HZ      := V1.HZ;
          HN      := V1.HN;
          Commentaire := V1.Commentaires;
          TabVisee[St].IDTerrainStation := V1.IDTerrainStation;
          X       := V1.DeltaX;
          Y       := V1.DeltaY;
          Z       := V1.DeltaZ;
        end; // with TabVisee[St] do begin
      end;   // for St:=0 to PointsTopo.Count-1 do begin
      //AfficherMessage(Format('----> Points OK - Idx max TabVisee: %d .. %d; EWE = %d', [Low(TabVisee), high(TabVisee), EWE]));
      TabVisee[0].LD := TabVisee[1].LD;
      TabVisee[0].LG := TabVisee[1].LG;
      TabVisee[0].HZ := TabVisee[1].HZ;
      TabVisee[0].HN := TabVisee[1].HN;
      TabVisee[0].TypeVisee := TabVisee[1].TypeVisee;
      if (EWE = 1) then
      begin
        TabVisee[1]   :=  TabVisee[0];
      end
      else
      begin
        TabVisee[EWE]   :=  TabVisee[EWE - 1];
        TabVisee[EWE].X :=  TabVisee[EWE - 1].X +
                           (TabVisee[EWE - 1].X - TabVisee[EWE - 2].X);
        TabVisee[EWE].Y :=  TabVisee[EWE - 1].Y +
                           (TabVisee[EWE - 1].Y - TabVisee[EWE - 2].Y);
      end;
      // calcul contours
      for St := 1 to EWE - 1 do
      begin
        try
          if (St = 1) then
          begin
            AlphaD := CalculerAngleBissecteur(TabVisee[St].X - TabVisee[St-1].X,
                                              TabVisee[St].Y - TabVisee[St-1].Y,
                                              TabVisee[St].X - TabVisee[St-1].X,
                                              TabVisee[St].Y - TabVisee[St-1].Y);
            //AlphaG := AlphaD + PI; // Pas besoin (optimisation possible)
          end;
          with QEntite do
          begin
            sincos(AlphaD, qSinAlphaD, qCosAlphaD);
            eCode       := TabVisee[St].stCode;
            eExpe       := TabVisee[St].stExpe;
            eSecteur    := TabVisee[St].stSecteur;
            eEntrance   := QBranche.NoEntranceRatt;
            eReseau     := QBranche.NoReseau;
            Type_Entite := TabVisee[St].TypeVisee; //2
            //if (Type_Entite = tgENTRANCE) then AfficherMessageErreur(Format('%d.%d est une entree', [QBranche.NoSerie, St]));

            DateLeve    := TabVisee[St].Date;
            eEntrance       := QBranche.NoEntranceRatt;
            Entite_Serie    := QBranche.NoSerie;
            Entite_Station  := TabVisee[St].NumPoint;
            // données originales
            oLongueur       := TabVisee[St].Longueur;
            oAzimut         := TabVisee[St].Azimut;
            oPente          := TabVisee[St].Pente;
            oLG             := TabVisee[St].LG;
            oLD             := TabVisee[St].LD;
            oHZ             := TabVisee[St].HZ;
            oHN             := TabVisee[St].HN;
            // centerline
            Une_Station_1_X := TabVisee[St-1].X ;
            Une_Station_1_Y := TabVisee[St-1].Y ;
            Une_Station_1_Z := TabVisee[St-1].Z ;
            Une_Station_2_X := TabVisee[St].X ;
            Une_Station_2_Y := TabVisee[St].Y ;
            Une_Station_2_Z := TabVisee[St].Z ;
            // habillage
            X1PD := Une_Station_1_X + TabVisee[St-1].LD * qCosAlphaD; //Cos(AlphaD);
            Y1PD := Une_Station_1_Y + TabVisee[St-1].LD * qSinAlphaD; //sin(AlphaD);
            X1PG := Une_Station_1_X - TabVisee[St-1].LG * qCosAlphaD;  //Cos(AlphaG);  cos(x+pi) = -cos(x)
            Y1PG := Une_Station_1_Y - TabVisee[St-1].LG * qSinAlphaD; //Sin(AlphaG);   sin(x+pi) = -sin(x)
            Z1PH := Une_Station_1_Z + TabVisee[St-1].HZ;
            Z1PB := Une_Station_1_Z - TabVisee[St-1].HN;
            AlphaD := CalculerAngleBissecteur(TabVisee[St].X - TabVisee[St-1].X,
                                              TabVisee[St].Y - TabVisee[St-1].Y,
                                              TabVisee[St+1].X - TabVisee[St].X,
                                              TabVisee[St+1].Y - TabVisee[St].Y);
            SinCos(AlphaD, qSinAlphaD, qCosAlphaD);

            //AlphaG := AlphaD + PI;
            X2PD := Une_Station_2_X + TabVisee[St].LD * qCosAlphaD; //Cos(AlphaD);
            Y2PD := Une_Station_2_Y + TabVisee[St].LD * qSinAlphaD; //sin(AlphaD);
            X2PG := Une_Station_2_X - TabVisee[St].LG * qCosAlphaD; // Cos(AlphaG);
            Y2PG := Une_Station_2_Y - TabVisee[St].LG * qSinAlphaD; //Sin(AlphaG);
            Z2PH := Une_Station_2_Z + TabVisee[St].HZ;
            Z2PB := Une_Station_2_Z - TabVisee[St].HN;
            // couleur pour dégradé
            ColourByDepth := clBlue;
            // Points d'intérêt
            IsPOI := (Pos(KEYWORD_POI, LowerCase(TabVisee[St].Commentaire)) > 0);
            // commentaires
            oIDLitteral   := GetIDStation(QBranche.NoSerie, TabVisee[St]);
            oCommentaires := TabVisee[St].Commentaire;
          end;
          // les entrées ne sont pas des entités comme les autres et font l'objet d'une
          // liste séparée. Elles n'ont pas à figurer dans la liste des visées.
          if (QEntite.Type_Entite = tgENTRANCE) then
          begin
            //QEntrance.eNumEntree   := MyBDDEntites.GetNbEntrees();
            QEntrance.eRefSer      := QEntite.Entite_Serie;
            QEntrance.eRefSt       := QEntite.Entite_Station;
            QEntrance.eNomEntree   := QEntite.oCommentaires;
            QEntrance.eObserv      := '';
            QEntrance.eXEntree     := QEntite.Une_Station_2_X;
            QEntrance.eYEntree     := QEntite.Une_Station_2_Y;
            QEntrance.eZEntree     := QEntite.Une_Station_2_Z;
            MyBDDEntites.AddEntree(QEntrance);
          end
          else
          begin
            MyBDDEntites.AddEntiteVisee(QEntite);
          end;
          Inc(Cnt);
        except
        end;
      end;   // for Br
      // Valeur de retour: OK si égalité entre le nombre de visées initial et le nombre de visées effectivement ajouté.
      Result := (EWE = (Cnt + 1));
    //except
    //end;
  end;
  procedure TraiterViseesEnAntenne();
  var
    n, nbA : integer;
    QEntite: TEntiteEtendue;
    OldVA: TViseeAntenne;
    t, t0, t1: TDateTime;
    tt: TTimeStamp;
    HH, MM, MS, SS: word;
  begin
    // calculer les visées en antenne
    AfficherMessage(GetResourceString(rsCALCUL_ANTENNES));
    nbA := self.DocuTopo.GetNbAntennes();
    t0 := Now();
    if ( nbA > 1) then
    begin
      for n := 1 to nbA - 1 do
      begin
        OldVA := self.DocuTopo.GetViseeAntenne(n - 1);
        VA    := self.DocuTopo.GetViseeAntenne(n);
        if ((Assigned(self.ProcDispProgression)) and ((n MOD 250) = 0)) then self.ProcDispProgression(GetResourceString(rsCALCUL_ANTENNES), n, 0, nbA - 1);
        //if (CalcViseeAntenne(VA, QEntite)) then MyBDDEntites.AddEntiteAntenne(QEntite);
        if (CalcViseeAntenne(VA, QEntite)) then MyBDDEntites.AddEntiteVisee(QEntite);    // AddEntiteVisee à remplacer par AddEntiteAntenne
      end
    end else
    begin
      AfficherMessage(' --> No antenna shots');
    end;
    t1 := Now();
    DecodeTime(t1-t0, HH, MM, SS, MS);
    AfficherMessageErreur(Format('Temps de calcul des antennes: %.2d:%.2d:%.2d.%.3d', [HH, MM, SS, MS]));
  end;
  //procedure TraiterLesEntrees();
  procedure CloturerListeVisees();
  begin
    //MyBDDEntites.AddEntite(EmptyEntiteEtendue);
  end;
begin;
  AfficherMessage('');
  AfficherMessage('CalculContoursGaleriesExtended: ' + FichierTOP);
  AfficherMessage('--------------------------------');
  // démarrage de l'objet TBDDEntites
  try
    if (Not MyBDDEntites.Initialiser(self.DocuTopo.GetPositionDuPointZero()))then Exit;
    EPSG := self.DocuTopo.GetCodeEPSGSystemeCoordonnees;
    MyBDDEntites.SetCodeEPSG(EPSG.CodeEPSG);
    recopierTablesEntreesReseauxSecteursCodesExpes(MyBDDEntites);
    // construction de la liste des visées
    AfficherMessage(GetResourceString(rsSCAN_BRCHS));
    //AfficherMessage(inttostr(GetNbBranches));
    for Br:= 1 to GetNbBranches - 1 do
    begin
      MyBranche := GetBranche(Br);
      if (Not TraiterBranche(MyBranche, Br)) then AfficherMessage(Format('Erreur pour la branche %d - %s', [Br, MyBranche.NomBranche]));
    end;
    AfficherMessage(rsSCAN_BRCHS + ' OK');
    // antennes
    TraiterViseesEnAntenne();
    // visée bidon de fin  DONE: Vérifier si cette visée est nécessaire
    CloturerListeVisees();
    // recenser les dates
    MyBDDEntites.RecenserDates();
    // sauvegarde en bloc
    AfficherMessage('Calcul contours galeries OK');
    if (DoGenerateFile) then MyBDDEntites.SaveToFile(FichierTOP, '');
  finally
  end;
  // provisoire: jonctions
  EnvoyerJonctionsVersBDDEntites(MyBDDEntites);
  AfficherMessage('--------------------------------');
end;
function TCodeDeCalcul.EnvoyerJonctionsVersBDDEntites(const MyBDDEntites: TBDDEntites): boolean;
var
  n, i, ND: Integer;
  J       : TJonctionXYZ;
  //JC      : TJonctionWithCoords;
begin
  result := false;
  n := GetNbJonctions();
  if (n = 0) then Exit;
  for i := 0 to n - 1 do
  begin
    J := GetJonction(i);
    ND := GetNoeud(J.NoSer, J.NoSt);
    if (ND >= 0) then
    begin

      J.X := FTableNoeuds[i].XN;
      J.Y := FTableNoeuds[i].YN;
      J.Z := FTableNoeuds[i].ZN;
      MyBDDEntites.AddJonction(J);
    end;
  end;
  result := true;
end;

// générer un fichier de synthèse (TAB simplifié avec coordonnées)
function TCodeDeCalcul.GenererFichierSyntheseAvecCoordonnees(const MyBDDEntites: TBDDEntites; const FileName: string): boolean;
var
  WU    : String;
  fp    : TextFile;
  i, j  : Integer;

  MyEntrance: TEntrance;
  MySerie   : TObjSerie;
  MyVisee   : TUneVisee;
  MyEntite: TEntiteEtendue;
  SD, PD: Integer;
  IxdSerie: Integer;
  procedure WrtEnTeteSection(const S: string);
  begin
    WriteLn(fp, '');
    WriteLn(fp, Format('[%s]', [S]));
    WriteLn(fp, '');
  end;
  procedure WrtLn(const S: string); inline;
  begin
    WriteLn(fp, S);
  end;
  procedure WrtEntite(const S: TEntiteEtendue);
  begin
    if ((S.Entite_Serie < 0) or (S.Entite_Station < 0)) then Exit;
    WrtLn(Format('%d' + #9 + '%d' + #9 +
                 '%d' + #9 + '%d' + #9 +
                 '%.2f' + #9 + '%.2f' + #9 + '%.2f' + #9 +
                 '%.2f' + #9 + '%.2f' + #9 + '%.2f' + #9 + '%.2f' + #9 +
                 '%s' + #9 +
                 '%.2f' + #9 + '%.2f' + #9 + '%.2f' + #9 +
                 '%s',
                [S.Entite_Serie, S.Entite_Station,
                 S.eCode, S.eExpe,
                 S.oLongueur, S.oAzimut, S.oPente,
                 S.oLD, S.oLG, S.oHZ, S.oHN,
                 S.oIDLitteral,
                 S.Une_Station_2_X, S.Une_Station_2_Y, S.Une_Station_2_Z,
                 S.oCommentaires
                ]));
  end;
begin
  Result := False;
  WU := Format('%s.GenererFichierSyntheseAvecCoordonnees: %s', [ClassName, FileName]);
  AfficherMessage(WU);
  AfficherMessageErreur(WU);
  AssignFile(fp, FileName);
  try
    ReWrite(fp);
    // les entrées
    WrtEnTeteSection('Entrances');
    for i := 0 to self.DocuTopo.GetNbEntrees - 1 do
    begin
      MyEntrance := self.DocuTopo.GetEntree(i);
    end;
    // les séries
    WrtEnTeteSection('Series');

    for i := 1 to self.DocuTopo.GetNbSeries - 1 do // série 0 ignorée
    begin
      MySerie  := self.DocuTopo.GetSerie(i);
      SD       := MySerie.GetNoSerieDep();
      PD       := MySerie.GetNoPointDep();
      IxdSerie := MySerie.GetNumeroDeSerie();
      WrtLn(Format('%d' + #9 +
                   '%d' + #9 + '%d' + #9 +
                   '%d' + #9 + '%d' + #9 +
                   '%s',
                   [MySerie.GetNumeroDeSerie() * (-1),
                    SD, PD,
                    MySerie.GetNoSerieArr, MySerie.GetNoPointArr,
                    MySerie.GetNomSerie
                   ]));
      // première visée
      MyBDDEntites.GetEntiteViseeFromSerSt(MySerie.GetNoSerieDep(), MySerie.GetNoPointDep(), MyEntite);
      WrtEntite(MyEntite);
      for j := 1 to MySerie.GetNbVisees() - 1 do
      begin
        MyBDDEntites.GetEntiteViseeFromSerSt(IxdSerie, j, MyEntite);
        WrtEntite(MyEntite);
      end;
      WrtLn('');
    end;
    AfficherMessage('--- Generation OK');
    Result := True;
  finally
    CloseFile(fp);
  end;
end;

{$IFDEF SPAR_SOLVER}
function TCodeDeCalcul.PreparerSolver(const Axe: TAxeCoordonnees): boolean;
var
  QNbNodes, QNbBranches, i, j, k: Integer;
  ww : double;
begin
  result := false;
  AfficherMessage(Format('%s.PreparerSolver()', [ClassName]));
  QNbNodes     := GetMaxNode();
  QNbBranches  := GetNbBranches();

  //if (Axe = axisX) then
  begin
    // rechercher les index mini et maxi des valeurs non nulles de RMatri
    SetLength(FBMatrix_LowIndex,0);
    SetLength(FBMatrix_HighIndex,0);
    SetLength(FBMatrix_LowIndex  , QNbNodes + 1);
    SetLength(FBMatrix_HighIndex , QNbNodes + 1);
    AfficherMessage('---> '+ GetResourceString(rsFIND_SUM_LIMITS));
    for i := 1 to QNbNodes do
    begin
      if (i MOD 200 = 0) then  AfficherMessage(Format(GetResourceString(rsLINEOFNB),[i, QNbNodes]));
      for j := 1 to QNbBranches - 1 do
      begin
        if (Abs(R_Matrix[j, i]) > 0) then
        begin
          FBMatrix_LowIndex[i] := j;
          Break;
        end;
      end;
    end;
    for i := 1 to QNbNodes do
    begin
      if (i MOD 200 = 0) then AfficherMessage(Format(GetResourceString(rsLINEOFNB),[i, QNbNodes]));
      for j := QNbBranches - 1 downto 1 do
      begin
        if (Abs(R_Matrix[j,i]) > 0) then
        begin
          FBMatrix_HighIndex[i]:=j;
          Break;
        end;
      end;
    end;
    for i := 1 to QNbNodes do
    begin
      //if (Assigned(self.ProcDispProgression)) then self.ProcDispProgression(Format('Matrice de compensation: ' + rsLINEOFNB, [i, QNbNodes]), i, 0, QNbNodes);
      for j:=1 to QNbNodes do
      begin
        ww := 0.00;
        for k := FBMatrix_LowIndex[i] to FBMatrix_HighIndex[i] do
        begin
          ww := ww + R_Matrix[k,i] * W_mtx[k] * R_Matrix[k,j];
        end;
        if (not IsZero(ww)) then  FMySolver.AddValueAtMatrixA(i, j, ww);
      end;
    end;
    // on sauvegarde la matrice
    //FMySolver.SaveMatrixAToFile();

  end;

  // second membre
  for i := 1 to QNbNodes do FMySolver.AddValueAtVectorB(i, ExtractValueFromTPoint3Df(Axe, VecteurSecondMembre[i]));
end;
{$ENDIF SPAR_SOLVER}

procedure TCodeDeCalcul.recopierTablesEntreesReseauxSecteursCodesExpes(const MyBDDEntites: TBDDEntites);
var
  i           : Integer;
  MyEntree    : TEntrance;
  MyReseau    : TReseau;
  MySecteur   : TSecteur;
  MyCode      : TCode;
  MyExpe      : TExpe;
begin
  AfficherMessage('recopierTablesEntreesReseauxSecteursCodesExpes');
  // lister les entrées
  for i := 0 to self.DocuTopo.GetNbEntrees() - 1   do MyBDDEntites.AddEntree(self.DocuTopo.GetEntree(i));
  // lister les réseaux
  for i := 0 to self.DocuTopo.GetNbReseaux() - 1   do MyBDDEntites.AddReseau(self.DocuTopo.GetReseau(i));
  // lister les secteurs
  for i := 0 to self.DocuTopo.GetNbSecteurs() - 1  do MyBDDEntites.AddSecteur(self.DocuTopo.GetSecteur(i));
  // lister les codes
  for i := 0 to self.DocuTopo.GetNbCodes() - 1     do MyBDDEntites.AddCode(self.DocuTopo.GetCode(i));
  // lister les expés
  for i := 0 to self.DocuTopo.GetNbExpes() - 1     do MyBDDEntites.AddExpe(self.DocuTopo.GetExpe(i));
end;

end.
