unit unitProfilTopo;

{$INCLUDE CompilationParameters.inc}

interface

uses
  {$INCLUDE SelectionLangues.inc} // insère les unités en fonction de la langue
  StructuresDonnees,
  Common,
  UnitListesSimplesWithGeneriques,
  Classes, SysUtils, Graphics;
// profil de terrain
type

{ TProfilTopo }

 TProfilTopo = class(TListeSimple<TPoint3DfOrderedByP>)
  private
    FProfilName   : string;
    FProfilColor  : TColor;
    FExtremite1   : TPoint2Df;
    FExtremite2   : TPoint2Df;
    FTableauGaleriesRecoupees: array of TEntiteEtendue; // pour création de transects
  public
    function  Initialiser(const P1, P2: TPoint2Df): boolean;
    procedure Finaliser();
    property  ProfilName  : string read FProfilName  write FProfilName;
    property  ProfilColor : TColor read FProfilColor write FProfilColor;
    function  GetExtremite1(): TPoint2Df; inline;
    function  GetExtremite2(): TPoint2Df; inline;
    procedure SetExtremite1(const P: TPoint2Df); inline;
    procedure SetExtremite2(const P: TPoint2Df); inline;

    procedure SortByAbscisses();
    function  GetAltitudeProfil(const P: double): double;
    procedure Epurage();

    procedure AddGalerieRecoupee(const P: TEntiteEtendue);
    function  GetNbGaleriesRecoupees(): integer;
    function  GetGalerieRecoupee(const Idx: integer): TEntiteEtendue;
end;

// profil de terrain
type

{ TListeProfilsTopo }

 TListeProfilsTopo = class(TListeSimple<TProfilTopo>)
  private
  public
    procedure RemoveProfil(const Idx: integer);
    procedure ClearListeProfils();
end;

implementation

{ TProfilTopo }
function SortAbscissesCriteria(Item1, Item2: Pointer): Integer;
var
  E1, E2: ^TPoint3DfOrderedByP;
begin
  E1:=Item1;
  E2:=Item2;
  if (E1.P < E2.P) then
    Result := -1
  else if (E1.P = E2.P) then
    Result :=  0
  else
    Result :=  1;
end;

procedure TListeProfilsTopo.RemoveProfil(const Idx: integer);
var
  P: TProfilTopo;
begin
  P := self.GetElement(Idx);
  try
    P.Finaliser();
  finally
    self.RemoveElement(Idx);
  end;
end;

// liste des profils topo
procedure TListeProfilsTopo.ClearListeProfils();
var
  i, Nb: Integer;
  QProfilTopo: TProfilTopo;
begin
  Nb := self.GetNbElements();
  if (Nb > 0) then
  begin
    for i := 0 to Nb - 1 do
    begin
      QProfilTopo := self.GetElement(i);
      try      // nettoyage des enfants de QProfilTopo;
        QProfilTopo.Finaliser();
      finally  // et libération
        QProfilTopo.Free;
      end;
    end;
  end;
  self.ClearListe();
end;

(* TProfilTopo *)
function TProfilTopo.Initialiser(const P1, P2: TPoint2Df): boolean;
begin
  result := false;
  self.ClearListe();
  SetExtremite1(P1);
  SetExtremite2(P2);
  SetLength(FTableauGaleriesRecoupees, 0);
  result := True;
end;

procedure TProfilTopo.Finaliser();
begin
  self.ClearListe();
  SetLength(FTableauGaleriesRecoupees, 0);
end;
procedure TProfilTopo.SortByAbscisses();
begin
  self.Sort(SortAbscissesCriteria);
end;

procedure TProfilTopo.Epurage();
var
  P0, P1: TPoint3DfOrderedByP;
  Nb, i: Integer;
  d: Double;
begin
  Nb := self.GetNbElements();
  if (Nb = 0) then Exit;
  for i := GetNbElements() - 1 downto 1 do
  begin
    P1 := GetElement(i);
    P0 := GetElement(i-1);
    d := P1.P - P0.P;
    if (d < 0.01) then RemoveElement(i);
  end;
end;
procedure TProfilTopo.AddGalerieRecoupee(const P: TEntiteEtendue);
var
  n: Integer;
begin
  n := Length(FTableauGaleriesRecoupees);
  Setlength(FTableauGaleriesRecoupees, n + 1);
  FTableauGaleriesRecoupees[High(FTableauGaleriesRecoupees)] := P;
end;
function TProfilTopo.GetNbGaleriesRecoupees(): integer;
begin
  Result := Length(FTableauGaleriesRecoupees);
end;
function TProfilTopo.GetGalerieRecoupee(const Idx: integer): TEntiteEtendue;
begin
  Result := FTableauGaleriesRecoupees[Idx];
end;



function  TProfilTopo.GetExtremite1(): TPoint2Df; inline;
begin
  Result := FExtremite1;
end;

function  TProfilTopo.GetExtremite2(): TPoint2Df; inline;
begin
  Result := FExtremite2;
end;

procedure TProfilTopo.SetExtremite1(const P: TPoint2Df); inline;
begin
  FExtremite1 := P;
end;

procedure TProfilTopo.SetExtremite2(const P: TPoint2Df); inline;
begin
  FExtremite2 := P;
end;


function  TProfilTopo.GetAltitudeProfil(const P: double): double;
var
  Nb, i: Integer;
  P0, P1: TPoint3DfOrderedByP;
  dP, dZ: Double;
begin
  Result := 0.00;
  Nb := self.GetNbElements();
  if (Nb < 2) then exit(0.00);
  for i := 0 to Nb-2 do
  begin
    P0 := GetElement(i);
    P1 := GetElement(i+1);
    if (IsInRange(P, P0.P, P1.P)) then
    begin
      dP := P1.P - P0.P;
      dZ := P1.Z - P0.Z;
      Result := P0.Z + (P - P0.P) * dZ/dP;
      Exit;
    end;
  end;
end;





end.

