unit unitX3D;

interface
{$INCLUDE CompilationParameters.inc}
uses
  StructuresDonnees,
  Common,
  Graphics,
  Classes, SysUtils;

type

{ TSceneX3D }

 TSceneX3D = class
  strict private
    FOutput: TextFile;
    FFichierX3D: TStringDirectoryFilename;
    FDoXML : boolean;
    procedure BeginScene(const NameScene: string);
    procedure EndScene();
    procedure WriteFooter();
    procedure WriteHeader();
    procedure WriteLine(const S: string);
  public
    function  Initialiser(const FileName: TStringDirectoryFilename; const DoXML: boolean; const NameScene: string): boolean;
    procedure Finaliser();

    procedure BeginShape(const NameShape: string);
    procedure DescribeShapeMaterial(const NomShape: string; const Color: TColorRGBA);
    procedure EndShape(const NameShape: string);

    procedure BeginIndexedFaceSetSection(const DoColorPerVertex, DoSolid: boolean);
    procedure BeginIndexFaceSetList();
    procedure BeginCoordinatesList();
    procedure EndCoordinatesList();
    procedure EndIndexedFaceSetSection();
    procedure EndIndexFaceSetList();
    procedure Write3DFace(const T: TMNTTriangleABC);
    procedure WriteVertex(const X, Y, Z: double); overload;
    procedure WriteVertex(const V: TVertex); overload; inline;


    procedure TraceCube(const NomCube: string; const CBG, CHD: TPoint3Df);
    procedure TraceMorceauTube(const E: TEntiteEtendue; const CouleurTube: TColorRGBA);
end;


implementation
const
  BALISE_SCENE                          = 'Scene';
  BALISE_SHAPE                          = 'Shape';
  BALISE_APPEARANCE                     = 'Appearance';
  BALISE_INDEXED_FACE_SET               = 'IndexedFaceSet';
  SHAPE_NAME_MNT                        = 'MNT';

{ TSceneX3D }

function TSceneX3D.Initialiser(const FileName: TStringDirectoryFilename; const DoXML: boolean; const NameScene: string): boolean;
begin
  result := false;
  FFichierX3D := FileName;
  AfficherMessage(format('%s.Initialiser(): %s', [ClassName, FFichierX3D]));
  FDoXML := DoXML;
  AssignFile(FOutput, FFichierX3D);
  try
    ReWrite(FOutput);
    WriteHeader();
      BeginScene(NameScene);
    result := true;
  except
  end;
end;

procedure TSceneX3D.Finaliser();
begin
  try
    EndScene();
    WriteFooter();
  finally
    Closefile(FOutput);
  end;
end;
//******************************************************************************
procedure TSceneX3D.WriteLine(const S: string); inline;
begin
  WriteLn(FOutput, S);
end;
procedure TSceneX3D.WriteHeader();
begin
  if (FDoXML) then
  begin
    WriteLine(FormatterVersionEncodageXML('1.0', 'UTF-8'));
    WriteLine('<!DOCTYPE X3D PUBLIC "ISO//Web3D//DTD X3D 3.2//EN"');
    WriteLine('"http://www.web3d.org/specifications/x3d-3.2.dtd">');
    WriteLine('<X3D profile="Interchange" version="3.2"');
    WriteLine('   xmlns:xsd="http://www.w3.org/2001/XMLSchema-instance"');
    WriteLine('   xsd:noNamespaceSchemaLocation="http://www.web3d.org/specifications/x3d-3.2.xsd">');
  end else
  begin
    WriteLine('#X3D V3.2 utf8');
    WriteLine('PROFILE Interchange');
    WriteLine('META "generator" "view3dscene, http://castle-engine.sourceforge.net/view3dscene.php"');
    WriteLine(Format('META "source" "%s"', [ExtractFileName(FFichierX3D)]));
  end;
end;
procedure TSceneX3D.WriteFooter();
begin
  if (FDoXML) then WriteLine('</X3D>') else pass;
end;
//******************************************************************************
procedure TSceneX3D.BeginScene(const NameScene: string);
begin
  if (FDoXML) then
  begin
    WriteLine(Format('  <%s>', [BALISE_SCENE]));
    WriteLine(Format('    <NavigationInfo type=''"%s" "%s" "%s" "%s"''/>', ['EXAMINE', 'WALK', 'FLY', 'ANY']));
  end else
  begin
    WriteLine('NavigationInfo {');
    WriteLine('  avatarSize [');
    WriteLine(format('    %.5f, %.5f, %.5f,', [0.25, 1.60, 0.75]));
    WriteLine('  ]');
    WriteLine('  type [');
    WriteLine('    "EXAMINE",');
    WriteLine('    "WALK",');
    WriteLine('    "FLY",');
    WriteLine('    "ANY",');
    WriteLine('  ]');
    WriteLine('}');

    pass; // aucun marqueur de scène
  end;
end;
procedure TSceneX3D.EndScene();
begin
  if (FDoXML) then WriteLine(Format('  </%s>', [BALISE_SCENE]))
              else pass;  // aucun marqueur de scène
end;
procedure TSceneX3D.BeginShape(const NameShape: string);
begin
  if (FDoXML) then WriteLine(Format('    <%s>', [BALISE_SHAPE]))
              else WriteLine('Shape {');
end;
procedure TSceneX3D.DescribeShapeMaterial(const NomShape: string; const Color: TColorRGBA);
var
  qr, qg, qb, qt: Extended;
begin
  qr := Color.R / 255.0;
  qg := Color.G / 255.0;
  qb := Color.B / 255.0;
  qt := 1.0 - Color.A / 255.0;
  if (FDoXML) then
  begin
    WriteLine(Format('     <%s>', [BALISE_APPEARANCE]));
    WriteLine(Format('        <Material emissiveColor=''%.3f, %.3f, %.3f'' transparency=''%.2f''/>', [qr, qg, qb, qt]));            // emissiveColor='0 0 1'/>
    WriteLine(Format('     </%s>', [BALISE_APPEARANCE]));
  end else
  begin
    WriteLine(#9 + 'appearance Appearance {');
		WriteLine(#9 + #9 + 'material Material {');
		WriteLine(#9 + #9 + '}');
	  WriteLine(#9 + '}');
  end;
end;

procedure TSceneX3D.EndShape(const NameShape: string);
begin
  if (FDoXML) then WriteLine(Format('    </%s>', [BALISE_SHAPE]))
              else WriteLine('}');
end;
procedure TSceneX3D.BeginIndexedFaceSetSection(const DoColorPerVertex, DoSolid: boolean);
var
  B1, B2: String;
begin
  if (FDoXML) then
  begin
    B1 := BoolToStr(DoColorPerVertex, 'true', 'false');
    B2 := BoolToStr(DoSolid         , 'true', 'false');
    // crease angle = lissage (compris entre 0 et pi)
    Write(FOutput, Format('      <%s creaseAngle=''%.3f'' colorPerVertex=''%s'' solid=''%s'' ', [BALISE_INDEXED_FACE_SET, 3.14, B1, B2]));
  end else
  begin
    pass;
  end;
end;

procedure TSceneX3D.EndIndexedFaceSetSection();
begin
  if (FDoXML) then WriteLine(Format('      </%s>', [BALISE_INDEXED_FACE_SET]))
              else WriteLine(#9 + '}');
end;

procedure TSceneX3D.BeginIndexFaceSetList();
begin
  if (FDoXML) then
  begin
    Write(FOutput, ' coordIndex=''');
    Write(FOutput, #13#10);
  end else
  begin
    WriteLine(#9 + 'geometry IndexedFaceSet {');
    WriteLine(#9 + #9 + 'solid FALSE');
    WriteLine(#9 + #9 + 'coordIndex [');
  end;
end;
procedure TSceneX3D.EndIndexFaceSetList();
begin
  if (FDoXML) then WriteLine('''>')
              else WriteLine(#9 + #9 + ']');
end;
procedure TSceneX3D.BeginCoordinatesList();
begin
  if (FDoXML) then
  begin
    WriteLine('    <Coordinate point=''');
  end else
  begin
    WriteLine(#9 + #9 + 'colorPerVertex FALSE'); //+ BoolToStr(DoColorPerVertex, 'TRUE', 'FALSE')
    WriteLine(#9 + #9 + 'coord Coordinate {');
		WriteLine(#9 + #9 + #9 + 'point [');
  end;
end;
procedure TSceneX3D.EndCoordinatesList();
begin
  if (FDoXML) then
  begin
    WriteLine('''/>');
  end else
  begin
    WriteLine(#9 + #9 + #9 +']');
    WriteLine(#9 + #9 + '}');
  end;
end;

procedure TSceneX3D.Write3DFace(const T: TMNTTriangleABC);
begin

  if (FDoXML) then
  begin
    WriteLine(Format('%d %d %d %d ', [T.PointA, T.PointB, T.PointC, -1]));
  end else
  begin // 818, 958, 817, -1,
    WriteLine(Format(#9 + #9 + #9 + '%d, %d, %d, %d, ', [T.PointA, T.PointB, T.PointC, -1]));
  end;

end;
procedure TSceneX3D.WriteVertex(const X, Y, Z: double); overload;
begin
  if (FDoXML) then WriteLine(Format('%.3f %.3f %.3f ', [X, Y, Z]))
              else WriteLine(Format(#9 + #9 + #9 + '%.3f, %.3f, %.3f, ', [X, Y, Z]));
end;
procedure TSceneX3D.WriteVertex(const V: TVertex); overload; inline;
begin
  WriteVertex(V.X, V.Y, V.Z);
end;

procedure TSceneX3D.TraceCube(const NomCube: string; const CBG, CHD: TPoint3Df);
begin
  BeginShape(NomCube);
    DescribeShapeMaterial(NomCube, GetColorRGBA(clBlue, 255));
    WriteLine(Format('      <IndexedLineSet DEF=''%s'' coordIndex=''%s''>', ['ILS', '0 1 2 3 0 4 5 6 7 4 -1 1 5 -1 2 6 -1 3 7 -1']));
    WriteLine(Format('        <Coordinate DEF=''%s'' point=''', ['CoordsCorners']));
      WriteVertex(CBG.X, CBG.Y, CBG.Z);
      WriteVertex(CHD.X, CBG.Y, CBG.Z);
      WriteVertex(CHD.X, CHD.Y, CBG.Z);
      WriteVertex(CBG.X, CHD.Y, CBG.Z);
      WriteVertex(CBG.X, CBG.Y, CHD.Z);
      WriteVertex(CHD.X, CBG.Y, CHD.Z);
      WriteVertex(CHD.X, CHD.Y, CHD.Z);
      WriteVertex(CBG.X, CHD.Y, CHD.Z);
    WriteLine('      ''/>');
    WriteLine('      </IndexedLineSet>');
  EndShape(NomCube);
end;


//******************************************************************************
procedure TSceneX3D.TraceMorceauTube(const E: TEntiteEtendue; const CouleurTube: TColorRGBA);
var
  EWE: String;
  q, r, s: Integer;
begin
  if (E.Drawn AND (E.Type_Entite in [tgDEFAULT, tgFOSSILE, tgVADOSE, tgENNOYABLE, tgSIPHON, tgTUNNEL, tgMINE])) then
  begin
    EWE := Format('Tube%d-%d', [E.Entite_Serie, E.Entite_Station]);
    BeginShape(EWE);
      DescribeShapeMaterial(SHAPE_NAME_MNT, CouleurTube);
      BeginIndexedFaceSetSection(false, false);
        BeginIndexFaceSetList();
          WriteLine('0 1 2 3 4 5 0 -1');
          WriteLine('6 11 10 9 8 7 6 -1');
          WriteLine('0 1 2 3 4 5 0 -1');
          for q := 0 to 5 do
          begin
            r := q mod 6;
            s := (q + 1) mod 6;
            WriteLine(Format('%d %d %d %d -1', [r, r + 6, s + 6, s]));
          end;
        EndIndexFaceSetList();
        BeginCoordinatesList();
          WriteVertex(E.Une_Station_1_X, E.Une_Station_1_Y, E.Z1PB);
          WriteVertex(E.X1PD, E.Y1PD, 0.5 * (E.Z1PB + E.Une_Station_1_Z));
          WriteVertex(E.X1PD, E.Y1PD, 0.5 * (E.Z1PH + E.Une_Station_1_Z));
          WriteVertex(E.Une_Station_1_X, E.Une_Station_1_Y, E.Z1PH);
          WriteVertex(E.X1PG, E.Y1PG, 0.5 * (E.Z1PH + E.Une_Station_1_Z));
          WriteVertex(E.X1PG, E.Y1PG, 0.5 * (E.Z1PB + E.Une_Station_1_Z));

          WriteVertex(E.Une_Station_2_X, E.Une_Station_2_Y, E.Z2PB);
          WriteVertex(E.X2PD, E.Y2PD, 0.5 * (E.Z2PB + E.Une_Station_2_Z));
          WriteVertex(E.X2PD, E.Y2PD, 0.5 * (E.Z2PH + E.Une_Station_2_Z));
          WriteVertex(E.Une_Station_2_X, E.Une_Station_2_Y, E.Z2PH);
          WriteVertex(E.X2PG, E.Y2PG, 0.5 * (E.Z2PH + E.Une_Station_2_Z));
          WriteVertex(E.X2PG, E.Y2PG, 0.5 * (E.Z2PB + E.Une_Station_2_Z));
        EndCoordinatesList();
      EndIndexedFaceSetSection();
    EndShape(EWE);
  end;
end;

end.

