unit UnitEntitesExtended;

// Table des entités étendue
// 06/11/2013: Fonctions de tri fixées
// 27/11/2013: Recensement des dates (nb de dates inférieur ou égal au nb de séances)
// 13/04/2013: Export SVG
// 14/04/2014: Fixation de bugs
// 04/08/2014: Export KML et OSM: Correction des pbs liés aux entrées
// 20/12/2014: Export des antennes au format GPX
// 06/02/2015: Le mode ftFLAGS du MétaFiltre est désactivé (fonctionne mais jamais utilisé en pratique)
// 06/02/2015: Remaniment des fonctions de recherche
// 12/02/2015: Ajout d'un tri ascendant sur le membre TagDouble des entités
// 19/06/2015: Export OSM: Entrées OK
// 24/06/2015: Export GCD: Les ID antennes contiennent le point de rattachement
// 12/10/2015: Export PostScript
// 22/12/2015: Génération de cartes dynamiques Leaflet
// 29/02/2015: Les fonctions de recherche balayent la table des entrées puis celle des entités
// 20/03/2016: Les listes passent aux génériques
// 27/04/2016: Ajout du filtre ENTREE_RATTACHEMENT au MétaFiltre
// 20/05/2016: La liste des entités passe aux génériques
// 11/08/2016: Prise en compte du paramètre USE_GENERIC_TYPES
//             pour compilation FPC 2.6 sous ARM Raspberry
// 03/08/2017: Modification de GetEntiteFromSerSt'): Très forte augmentation de la vitesse de recherche
// 02/01/2018: Les fonctions de dessin des diagrammes vont dans TCdrRoseDiagramExt et TCdrDepthDiagramExt
// 13/04/2018: Unification de TPostScriptCanvas et TSVGCanvas
{$INCLUDE CompilationParameters.inc}
interface

uses
  {$INCLUDE SelectionLangues.inc} // insère les unités en fonction de la langue
  StructuresDonnees,
  UnitStructuresDonneesEntites,
  Common,
  {$IFDEF USE_GENERIC_TYPES}
  UnitListesSimplesWithGeneriques,
  {$ELSE}
  UnitListesSimplesWithoutGeneriques,
  {$ENDIF}
  UnitClassPalette,
  Classes, SysUtils, Graphics, math,
  ConvertisseurJPC,
  //UnitDXFDrawing,
  SVGCanvasUnit,
  PostScriptCanvasUnit,
  UnitSilhouetteGalerie,
  UnitDXFDrawing;
type
{ TBDDEntites }
  {$IFDEF USE_GENERIC_TYPES}
  TBDDEntites = class
  {$ELSE}
  TBDDEntites = class(TList)
  {$ENDIF}
  private
    // code EPSG
    FCodeEPSG     : integer;
    // tables des entités
    FTableDesVisees: TTableEntites; // visées
    FTableAntennes : TTableEntites; // antennes
    // tables
    FTableEntrees : TTableEntrees;
    FTableSecteurs: TTableSecteurs;
    FTableReseaux : TTableReseaux;
    FTableExpes   : TTableExpes;
    FTableCodes   : TTableCodes;
    FTableJonctions   : TTableJonctionsXYZ; //TTableJonctionsWithCoords;

    FPalette256   : TPalette256;
    // barre de progression
    FProcDispProgression: TProcDisplayProgression;
    // altitude du point zéro
    FPositionDuPointZero: TPoint3Df;
    // encombrement du reseau
    FCoinBasGauche : TPoint3Df;
    FCoinHautDroit : TPoint3Df;
    // encombrement de la zone métafiltrée
    FMetafilteredCoinBasGauche : TPoint3Df;
    FMetafilteredCoinHautDroit : TPoint3Df;
    // ID des stations extremes
    FIDStationXMini: string;
    FIDStationYMini: string;
    FIDStationZMini: string;
    FIDStationXMaxi: string;
    FIDStationYMaxi: string;
    FIDStationZMaxi: string;
    // histogramme des profondeurs et des directions
    FClassRepartDepthDiagram: TVecteurDouble;
    FClassRepartRoseDiagram : TVecteurDouble;
    // tableaux pour statistiques
    FSpeleometrieParReseaux,
    FSpeleometrieParSecteurs,
    FSpeleometrieParExpes,
    FSpeleometrieParCodes,
    FSpeleometrieParDates: TTableauVentilationSpeleometrie;
    FRecapitulation : TVentilationSpeleometrie;
    // couleurs mini et maxi
    FCouleurZMin    : TColor;
    FCouleurZMax    : TColor;
    // tableau des dates
    FListeDesDates : array of TDateTime;
    FNbDates       : integer;
    // nombre de visées retenues par le MétaFiltre
    FNbreViseesRetenues: integer;

    // Derniers résultats de GetEntiteFromSerSt
    FLastEntiteFound : TEntiteEtendue;
    FLastSerFound    : TNumeroSerie;
    FLastPtFound     : integer;

    {$IFDEF USE_GENERIC_TYPES}
    {$ELSE}
    procedure AddElement(const E: TEntiteEtendue);
    procedure ClearListe;
    function GetElement(const Idx: integer): TEntiteEtendue;
    function GetNbElements: integer;
    procedure PutElement(const Idx: integer; const E: TEntiteEtendue);
    function RemoveElement(const Idx: integer): boolean;
    {$ENDIF}


    procedure ViderTablesSimples();
    function CalcSpeleometrieReseauOuSecteur(const QMode: integer; const IdxReseauOuSecteur: integer; out TR: TSpeleometrieReseauOuSecteur): boolean;

  public
    constructor Create;
    function  Initialiser(const PositionPointZero: TPoint3Df): boolean;
    procedure SetProcDisplayProgression(const ProcProgression: TProcDisplayProgression);
    procedure ResetTables();
    procedure ViderLesViseesEtAntennes();
    procedure Finaliser();
    // Code EPSG
    function  GetCodeEPSG(): integer;
    procedure SetCodeEPSG(const C: integer);

    // Accès aux visées
    procedure AddEntiteVisee(const E: TEntiteEtendue);
    function  GetEntiteVisee(const Idx: integer): TEntiteEtendue;
    procedure PutEntiteVisee(const Idx: integer; const E: TEntiteEtendue);
    procedure SetEntiteViseeVisible(const Idx: integer; const QVisible: boolean);
    function GetNbEntitesVisees(): integer;
    // Accès aux antennes
    procedure AddEntiteAntenne(const E: TEntiteEtendue);
    function  GetEntiteAntenne(const Idx: integer): TEntiteEtendue;
    function  GetNbEntitesAntennes: integer;
    procedure PutEntiteAntenne(const Idx: integer; const E: TEntiteEtendue);
    procedure SetEntiteAntenneVisible(const Idx: integer; const QVisible: boolean);
    // couleurs en fonction des réseaux
    function  GetCouleurEntiteRGBAByReseau(const E: TEntiteEtendue): TColorRGBA;
    function  GetCouleurEntiteRGBByReseau(const E: TEntiteEtendue): TColor;
    function GetCouleurRGBAByIdxReseau(const Idx: integer): TColorRGBA;
    function GetCouleurRGBByIdxReseau(const Idx: integer): TColor;
    // couleurs en fonction des secteurs
    function  GetCouleurEntiteRGBABySecteur(const E: TEntiteEtendue): TColorRGBA;
    function  GetCouleurEntiteRGBBySecteur(const E: TEntiteEtendue): TColor;
    function GetCouleurRGBAByIdxSecteur(const Idx: integer): TColorRGBA;
    function GetCouleurRGBByIdxSecteur(const Idx: integer): TColor;
    // couleurs en fonction des expés
    function  GetCouleurEntiteRGBAByExpe(const E: TEntiteEtendue): TColorRGBA;
    function  GetCouleurEntiteRGBByExpe(const E: TEntiteEtendue): TColor;
    function  GetCouleurRGBExpeByIDX(const IDX: integer): TColor;
    function  GetCouleurRGBAExpeByIdx(const IDX: integer): TColorRGBA;
    // gestion des entrées
    procedure AddEntree(const AEntree: TEntrance);
    function  GetEntree(const NumEntree: integer): TEntrance;
    function  GetNbEntrees(): integer;
    // gestion des réseaux
    procedure AddReseau(const AReseau: TReseau);
    function  GetReseau(const NumReseau: integer): TReseau;
    function  GetNbReseaux(): integer;
    // gestion des secteurs
    procedure AddSecteur(const ASecteur: TSecteur);
    function  GetSecteur(const NumSecteur: integer): TSecteur;
    function  GetNbSecteurs(): integer;
    // gestion des expés
    procedure AddExpe(const LaExpe: TExpe);
    function  GetExpe(const NumExpe: integer): TExpe;
    function  GetExpeByIndex(const Idx: integer): TExpe;
    function  GetNbExpes(): integer;
    // gestion des codes
    procedure AddCode(const LeCode: TCode);
    function  GetCode(const NoCode: integer): TCode;
    function  GetCodeByIndex(const Idx: integer): TCode;
    function  GetNbCodes(): integer;
    // jonctions
    procedure AddJonction(const J: TJonctionXYZ);
    function GetJonction(const N: integer): TJonctionXYZ;
    function GetNbJonctions(): integer;


    // utilitaires de tri
    procedure SortByDepth(); inline;
    procedure SortBySerSts(); inline;
    procedure SortByTagDouble(); inline;
    // métafiltre
    function MetaFiltre(const Filtre: string; out DevelVisees: double): integer;
    // encombrement du reseau
    procedure SetMinMax(const DoMetafilteredOnly: boolean);
    // encombrement de la cavité
    function GetCoinBasGauche(): TPoint3Df; inline;
    function GetCoinHautDroit(): TPoint3Df; inline;
    // encombrement de la zone métafiltrée
    function GetMetafilteredCoinBasGauche(): TPoint3Df; inline;
    function GetMetafilteredCoinHautDroit(): TPoint3Df; inline;
    function GetIDStationXMaxi(): string; inline;
    function GetIDStationXMini(): string; inline;
    function GetIDStationYMaxi(): string; inline;
    function GetIDStationYMini(): string; inline;
    function GetIDStationZMaxi(): string; inline;
    function GetIDStationZMini(): string; inline;
    // fonctions de recherche
    function GetEntiteViseeFromXY(const AntennesAussi: boolean; const X, Y: double; out EE: TEntiteEtendue; out InternalIdxEntite: integer): boolean;
    function GetEntiteViseeFromCle(const DoMatchExact: boolean; const Cle: string; out EE: TEntiteEtendue): boolean;
    function GetEntiteViseeFromSerSt(const Ser, St: Integer; out EE: TEntiteEtendue): boolean;
    function FindStationByCle(const DoMatchExact: boolean; const Cle: string; out E: TEntiteEtendue): Boolean;
    function FindEntranceByCle(const DoMatchExact: boolean; const Cle: string; out E: TEntrance): boolean;


    // statistiques
    procedure CalculSpeleometrie();
    // définir l'altitude du point zéro de la cavité
    procedure SetPositionPointZero(const QPositionPointZero: TPoint3Df);
    function  GetPositionPointZero(): TPoint3Df;
    function  GetDeltaXYZFromPositionSt0(const ST: TEntiteEtendue): TPoint3Df;
    function GetSpeleometrieParReseauxByIdx(const Idx: integer): TVentilationSpeleometrie;
    function GetSpeleometrieParSecteurByIdx(const Idx: integer): TVentilationSpeleometrie;
    function GetSpeleometrieParExpeByIdx(const Idx: integer): TVentilationSpeleometrie;
    function GetSpeleometrieParCodeByIdx(const Idx: integer): TVentilationSpeleometrie;
    function GetSpeleometrieParDateByIdx(const Idx: integer): TVentilationSpeleometrie;
    // recensement des dates
    function RecenserDates(): integer;
    function GetUneDate(const Idx: integer): TDateTime;
    function GetNbDates(): integer;
    // sauvegarde/restauration (TODO: voir si obsolète)
    function  SaveToFile(const Filename: TStringDirectoryFilename; const Filtres: string): boolean; deprecated;
    function  LoadFromFile(const FileName: TStringDirectoryFilename): boolean; deprecated;
    // utilitaire d'export vers logiciels de cartographie
    // GPX only: Export des visées rayonnantes
    procedure ExportForCarto(const FileName: TStringDirectoryFilename;
                             const Convertisseur: TConversionSysteme;
                             const QSystemeGeogEPSG: integer;
                             const ModeOutputFormatGIS: TOutputFormatGIS;
                             const DoMakeSilhouette: boolean;
                             const Filtres: string;
                             const PrefixStations: string;
                             const CouleurDefaut: TColor;
                             const UseColorGroupes: boolean);
    // export GHCavedraw
    procedure ExportForGHCaveDraw(const MyFile: TStringDirectoryFilename; const QMyFiltre: string);
    // couleur en fonction de l'altitude
    procedure CalcCouleursByDepth(const CouleurZMin, CouleurZMax: TColor);
    // couleur en fonction du mode de travail
    function  GetColorViseeFromModeRepresentation(const M: TModeRepresentationGaleries; const E: TEntiteEtendue): TColor;
    // diagrammes des directions et des altitudes
    function  ParseRoseDiagram(const NbPetales: integer; const ViseeMini: double): boolean;
    function  ParseDepthHistogramme(const NbPetales: integer; const ViseeMini: double; const Normalize: boolean): Boolean;
    // export OpenOffice;org
    procedure ExportVue2DSVG(const Filename: TStringDirectoryFilename; const DoXHTML: boolean; const FM: TModeRepresentationGaleries; const FE: TElementsDrawn; const FillOpacity: byte);
    procedure ExportVue2DPostScript(const PSFileName: TStringDirectoryFilename; const FM: TModeRepresentationGaleries; const FE: TElementsDrawn);
    // export AutoCAD (obsolète mais constamment demandé)
    procedure ExportVue2DAcadDXF(const PSFileName: TStringDirectoryFilename; const FM: TModeRepresentationGaleries; const FE: TElementsDrawn); deprecated;
    // export des antennes comme nuage de points
    procedure ExporterAntennesNuagePoints(const QFilename: TStringDirectoryFilename);
    // spéléométrie simplifiée pour réseaux ou secteurs
    // Mode: 1 = réseau; 2 = secteurs
    function CalcSpeleometrieParReseauxOuSecteurs(const QMode: integer; out TR: TArraySpeleometrieReseauOuSecteur): boolean;
    function GetNbEntitesMetafiltrees(): integer;

    // calculer une visée puis l'ajouter
    // ceci afin de pouvoir visualiser immédiatement une visée sans avoir à recalculer le réseau
    function GenererEntiteDepuisUneVisee(const QStation: TToporobotStation;
                                         const QV: TUneVisee; out EE: TEntiteEtendue): boolean;

    // extraction des data précalculées pour les diagrammes
    function GetClassesRepartDepthDiagram(): TVecteurDouble; inline;
    function GetClassesRepartRoseDiagram(): TVecteurDouble; inline;
end;


implementation
uses
  unitExportGIS;
const
  LOW_INDEX = 0;
  INVDEGREES = 180.0 / PI;


// Trier les visées par ZOrder
function SortEntiteZOrder(Item1, Item2: Pointer): Integer;
var
  E1, E2: ^TEntiteEtendue;
  bidon1: Extended;
  bidon2: Extended;
begin
  E1:=Item1;
  E2:=Item2;
  bidon1 := 0.50 * (E1.Une_Station_1_Z + E1.Une_Station_2_Z);
  bidon2 := 0.50 * (E2.Une_Station_1_Z + E2.Une_Station_2_Z);
  if      (bidon1 < bidon2) then Result := -1
  else if (bidon1 = bidon2) then Result :=  0
  else                           Result :=  1;
end;
// Trier les visées par TagDouble (e.g. distance)
function SortEntiteByTagDouble(Item1, Item2: Pointer): Integer;
var
  E1, E2: ^TEntiteEtendue;
  bidon1: Extended;
  bidon2: Extended;
begin
  E1 := Item1;
  E2 := Item2;
  bidon1 := E1.TagDouble;
  bidon2 := E2.TagDouble;
  if      (bidon1 < bidon2) then Result := -1
  else if (bidon1 = bidon2) then Result :=  0
  else                           Result :=  1;
end;
// TODO: Valider
function SortEntiteSerSt(Item1, Item2: Pointer): Integer;
const
  QMULT_SERIE = 100000;
var
  E1, E2: ^TEntiteEtendue;
  P1, P2: int64;
begin
  E1:=Item1;
  E2:=Item2;
  P1 := QMULT_SERIE * E1.Entite_Serie + E1.Entite_Station;
  P2 := QMULT_SERIE * E2.Entite_Serie + E2.Entite_Station;
  if      (P1 < P2) then Result := -1
  else if (P1 = P2) then Result :=  0
  else                   Result :=  1;
end;
//******************************************************************************
{ TBDDEntites }
{$IFDEF USE_GENERIC_TYPES}
{$ELSE}
procedure TBDDEntites.ClearListe;
var
  i: Integer;
begin
  try
    if (self.Count > 0) then for i:= 0 to self.Count - 1 do Dispose(self.Items[i]);
  finally
    self.Clear;
  end;
end;

function TBDDEntites.GetNbElements: integer;
begin
  Result := self.Count;
end;

procedure TBDDEntites.AddElement(const E: TEntiteEtendue);
var
  pE: ^TEntiteEtendue;
begin
 New(pE);
 pE^ := E;
 self.Add(pE);
end;

function TBDDEntites.GetElement(const Idx: integer): TEntiteEtendue;
var
  pE: ^TEntiteEtendue;
begin
  pE:= self.Items[Idx];
  Result := pE^;
end;



procedure TBDDEntites.PutElement(const Idx: integer; const E: TEntiteEtendue);
var
  pE: ^TEntiteEtendue;
begin
  pE := self.Items[Idx];
  pE^:= E;
  self.Items[Idx] := pE;
end;

function TBDDEntites.RemoveElement(const Idx: integer): boolean;
begin
  Result := False;
  try
    Dispose(self.Items[Idx]);
    self.Delete(Idx);
    Result := True;
  except
  end;
end;

{$ENDIF}

constructor TBDDEntites.Create;
begin
  inherited Create;
  // créer la palette
  FPalette256 := TPalette256.Create;
  FPalette256.GenerateTOPOROBOTPalette;
  FNbreViseesRetenues := 0;
  // créer les tables
  FTableDesVisees      := TTableEntites.Create; // les visées
  FTableAntennes       := TTableEntites.Create; // les antennes

  FTableEntrees        := TTableEntrees.Create;
  FTableReseaux        := TTableReseaux.Create;
  FTableSecteurs       := TTableSecteurs.Create;
  FTableCodes          := TTableCodes.Create;
  FTableExpes          := TTableExpes.Create;
  FTableJonctions      := TTableJonctionsXYZ.Create;
end;


function TBDDEntites.Initialiser(const PositionPointZero: TPoint3Df): boolean;
begin
  AfficherMessage(Format('%s.Initialise()', [ClassName]));
  FLastSerFound := -1;
  FLastPtFound  :=  0;
  // position du point zéro
  SetPositionPointZero(PositionPointZero);


  //----------------
  Result := false;
  try
    ResetTables();
    Result := True;
  finally
  end;
end;



procedure TBDDEntites.SetProcDisplayProgression(const ProcProgression: TProcDisplayProgression);
begin
  FProcDispProgression := ProcProgression;
end;

procedure TBDDEntites.ResetTables;
begin
  ViderLesViseesEtAntennes();
  ViderTablesSimples();
end;
procedure TBDDEntites.ViderTablesSimples;
begin
  try
    SetLength(FListeDesDates, 0);
    FNbDates := 0;
    ViderLesViseesEtAntennes();
    FTableEntrees.ClearListe();
    FTableReseaux.ClearListe();
    FTableSecteurs.ClearListe();
    FTableCodes.ClearListe();
    FTableExpes.ClearListe();
    FTableJonctions.ClearListe();

    AfficherMessage('Vidage des tables simples OK');
  except
    AfficherMessage('Echec purge des tables simples');
  end;
end;

procedure TBDDEntites.ViderLesViseesEtAntennes();
begin
  FTableDesVisees.ClearListe();
  FTableAntennes.ClearListe();
end;

procedure TBDDEntites.Finaliser;
begin
  AfficherMessage(Format('%s.Finaliser', [ClassName]));
  try
    try
      ResetTables();
    except
      ;
    end;
  finally
    FPalette256.Free;
    FTableDesVisees.Free;
    FTableAntennes.Free;
    FTableEntrees.Free;    // FTableEntrees        := TTableEntrees.Create;
    FTableReseaux.Free;    // FTableReseaux        := TTableReseaux.Create;
    FTableSecteurs.Free;   // FTableSecteurs       := TTableSecteurs.Create;
    FTableExpes.Free;      // FTableCodes          := TTableCodes.Create;
    FTableCodes.Free;      // FTableExpes          := TTableExpes.Create;
    FTableJonctions.Free;
  end;
end;

function TBDDEntites.GetCodeEPSG: integer;
begin
  Result := FCodeEPSG;
end;

procedure TBDDEntites.SetCodeEPSG(const C: integer);
begin
  FCodeEPSG := C;
end;

// ajout, modif, suppression, lecture d'éléments
// les visées
procedure TBDDEntites.AddEntiteVisee(const E: TEntiteEtendue);
begin
  FTableDesVisees.AddElement(E);
end;

function TBDDEntites.GetEntiteVisee(const Idx: integer): TEntiteEtendue;
begin
  Result := FTableDesVisees.GetElement(Idx);
end;

procedure TBDDEntites.PutEntiteVisee(const Idx: integer; const E: TEntiteEtendue);
begin
  FTableDesVisees.PutElement(Idx, E);
end;

procedure TBDDEntites.SetEntiteViseeVisible(const Idx: integer; const QVisible: boolean);
var
  WU: TEntiteEtendue;
begin
  WU := GetEntiteVisee(Idx);
  WU.Drawn := QVisible;
  PutEntiteVisee(Idx, WU);
end;

function TBDDEntites.GetNbEntitesVisees: integer;
begin
  Result := FTableDesVisees.GetNbElements();
end;

// les antennes
// ajout, modif, suppression, lecture d'éléments
procedure TBDDEntites.AddEntiteAntenne(const E: TEntiteEtendue);
begin
  FTableAntennes.AddElement(E);
end;

function TBDDEntites.GetEntiteAntenne(const Idx: integer): TEntiteEtendue;
begin
  Result := FTableAntennes.GetElement(Idx);
end;

procedure TBDDEntites.PutEntiteAntenne(const Idx: integer; const E: TEntiteEtendue);
begin
  FTableAntennes.PutElement(Idx, E);
end;

procedure TBDDEntites.SetEntiteAntenneVisible(const Idx: integer; const QVisible: boolean);
var
  WU: TEntiteEtendue;
begin
  WU := GetEntiteAntenne(Idx);
  WU.Drawn := QVisible;
  PutEntiteAntenne(Idx, WU);
end;

function TBDDEntites.GetNbEntitesAntennes: integer;
begin
  Result := FTableAntennes.GetNbElements();
end;
//*************************************
// les entrées

procedure TBDDEntites.AddEntree(const AEntree: TEntrance);
begin
  FTableEntrees.AddElement(AEntree);
end;

function TBDDEntites.GetEntree(const NumEntree: integer): TEntrance;
begin
  Result := FTableEntrees.GetElement(NumEntree);
end;

function TBDDEntites.GetNbEntrees: integer;
begin
  Result := FTableEntrees.GetNbElements;
end;

procedure TBDDEntites.AddReseau(const AReseau: TReseau);
begin
  FTableReseaux.AddElement(AReseau);
end;

function TBDDEntites.GetReseau(const NumReseau: integer): TReseau;
begin
  try
    Result := FTableReseaux.GetElement(NumReseau);
  except
    Result := FTableReseaux.GetElement(0);
  end;
end;


function TBDDEntites.GetNbReseaux: integer;
begin
  Result := FTableReseaux.GetNbElements;
end;

procedure TBDDEntites.AddSecteur(const ASecteur: TSecteur);
begin
  FTableSecteurs.AddElement(ASecteur);
end;

function TBDDEntites.GetSecteur(const NumSecteur: integer): TSecteur;
begin
  try
    Result := FTableSecteurs.GetElement(NumSecteur);
  except
    Result := FTableSecteurs.GetElement(0);
  end;
end;


function TBDDEntites.GetNbSecteurs: integer;
begin
  Result := FTableSecteurs.GetNbElements;
end;

procedure TBDDEntites.AddExpe(const LaExpe: TExpe);
begin
  FTableExpes.AddElement(LaExpe);
end;

function TBDDEntites.GetExpe(const NumExpe: integer): TExpe;
begin
  Result := FTableExpes.GetElement(NumExpe);
end;

function TBDDEntites.GetExpeByIndex(const Idx: integer): TExpe;
begin
  Result := FTableExpes.GetElementByIndex(Idx);
end;

function TBDDEntites.GetNbExpes: integer;
begin
  Result := FTableExpes.GetNbElements;
end;


procedure TBDDEntites.AddCode(const LeCode: TCode);
begin
  FTableCodes.AddElement(LeCode);
end;

function TBDDEntites.GetCode(const NoCode: integer): TCode;
begin
  Result := FTableCodes.GetElement(NoCode);
end;

function TBDDEntites.GetCodeByIndex(const Idx: integer): TCode;
begin
  Result := FTableCodes.GetElementByIndex(Idx);
end;

function TBDDEntites.GetNbCodes: integer;
begin
  Result := FTableCodes.GetNbElements;
end;
// liste des jonctions
procedure TBDDEntites.AddJonction(const J: TJonctionXYZ);
begin
  FTableJonctions.AddElement(J);

end;

function TBDDEntites.GetJonction(const N: integer): TJonctionXYZ;
begin
  try
    Result := FTableJonctions.GetElement(N);
  except
    Result := FTableJonctions.GetElement(0);
  end;
end;

function TBDDEntites.GetNbJonctions: integer;
begin
  Result := FTableJonctions.GetNbElements;
end;


//******************************************************************************
function TBDDEntites.GetCouleurEntiteRGBAByReseau(const E: TEntiteEtendue): TColorRGBA;
var
  WU : TColor;
  EWE: TReseau;
begin
  try
    EWE  := self.GetReseau(E.eReseau);
    WU   := EWE.ColorReseau;
  except
    WU   := clWhite;
  end;
  Result := GetColorRGBA(WU, $FF);
end;

function TBDDEntites.GetCouleurEntiteRGBByReseau(const E: TEntiteEtendue): TColor;
var
  EWE: TReseau;
begin
  try
    EWE    := self.GetReseau(E.eReseau);
    Result := EWE.ColorReseau;
  except
    Result := clWhite;
  end;
end;
function TBDDEntites.GetCouleurRGBAByIdxReseau(const Idx: integer): TColorRGBA;
var
  WU : TColor;
  EWE: TReseau;
begin
  try
    EWE  := self.GetReseau(Idx);
    WU   := EWE.ColorReseau;
  except
    WU   := clWhite;
  end;
  Result := GetColorRGBA(WU, $FF);
end;
function TBDDEntites.GetCouleurRGBByIdxReseau(const Idx: integer): TColor;
var
  EWE: TReseau;
begin
  try
    EWE    := self.GetReseau(Idx);
    Result := EWE.ColorReseau;
  except
    Result := clWhite;
  end;
end;

function TBDDEntites.GetCouleurEntiteRGBABySecteur(const E: TEntiteEtendue): TColorRGBA;
var
  EWE: TSecteur;
  WU : TColor;
begin
  try
    EWE := self.GetSecteur(E.eSecteur);
    WU  := EWE.CouleurSecteur;
  except
    WU  := clWhite;
  end;
  Result := GetColorRGBA(WU, $FF);
end;

function TBDDEntites.GetCouleurEntiteRGBBySecteur(const E: TEntiteEtendue): TColor;
var
  EWE: TSecteur;
begin
  try
    EWE     := self.GetSecteur(E.eSecteur);
    Result  := EWE.CouleurSecteur;
  except
    Result := clWhite;
  end;
end;
function TBDDEntites.GetCouleurRGBAByIdxSecteur(const Idx: integer): TColorRGBA;
var
  EWE: TSecteur;
  WU : TColor;
begin
  try
    EWE := self.GetSecteur(Idx);
    WU  := EWE.CouleurSecteur;
  except
    WU  := clWhite;
  end;
  Result := GetColorRGBA(WU, $FF);
end;

function TBDDEntites.GetCouleurRGBByIdxSecteur(const Idx: integer): TColor;
var
  EWE: TSecteur;
begin
  try
    EWE     := self.GetSecteur(Idx);
    Result  := EWE.CouleurSecteur;
  except
    Result := clWhite;
  end;
end;

function TBDDEntites.GetCouleurEntiteRGBByExpe(const E: TEntiteEtendue): TColor;
var
  EWE: TExpe;
begin
  try
    EWE := self.GetExpeByIndex(E.eExpe);
    Result  := FPalette256.GetColorByIndex(EWE.Couleur);
  except
    Result := clWhite;
  end;
end;

function TBDDEntites.GetCouleurEntiteRGBAByExpe(const E: TEntiteEtendue): TColorRGBA;
var
  EWE: TExpe;
  WU : TColor;
begin
  try
    EWE := self.GetExpeByIndex(E.eExpe);
    WU  := FPalette256.GetColorByIndex(EWE.Couleur);
  except
    WU := clWhite;
  end;
  Result := GetColorRGBA(WU, $FF);
end;
function TBDDEntites.GetCouleurRGBAExpeByIdx(const IDX: integer): TColorRGBA;
var
  EWE: TExpe;
  WU : TColor;
begin
  try
    EWE := self.GetExpeByIndex(IDX);
    WU  := FPalette256.GetColorByIndex(Idx);
  except
    WU := clWhite;
  end;
  Result := GetColorRGBA(WU, $FF);
end;
function TBDDEntites.GetCouleurRGBExpeByIDX(const IDX: integer): TColor;
var
  EWE: TExpe;
begin
  try
    EWE := self.GetExpeByIndex(IDX);
    Result  := FPalette256.GetColorByIndex(Idx);
  except
    Result := clWhite;
  end;
end;
//******************************************************************************
// Encombrement du réseau
// définir mini et maxi
// DONE: Membres XMini, XMaxi, YMini, YMaxi,ZMini, ZMaxi : double; localisés ici:
// font double emploi avec les membres privés FCoinBasGauche et FCoinHautDroit
procedure TBDDEntites.SetMinMax(const DoMetafilteredOnly: boolean);
const
  GM = 1e20;
var
  i, NbEntrances: integer;
  XMini, XMaxi, YMini, YMaxi, ZMini, ZMaxi: double;
  EWE: TEntiteEtendue;
  procedure ZDar(const E666: TEntiteEtendue);
  begin
    if (E666.Une_Station_1_X > XMaxi) then
    begin
      XMaxi := E666.Une_Station_1_X;
      if (E666.Drawn) then FMetafilteredCoinHautDroit.X := XMaxi;
      FIDStationXMaxi := E666.oIDLitteral;
    end;
    if (E666.Une_Station_1_X < XMini) then
    begin
      XMini := E666.Une_Station_1_X;
      if (E666.Drawn) then FMetafilteredCoinBasGauche.X := XMini;
      FIDStationXMini := E666.oIDLitteral;
    end;
    if (E666.Une_Station_1_Y > YMaxi) then
    begin
      YMaxi := E666.Une_Station_1_Y;
      if (E666.Drawn) then FMetafilteredCoinHautDroit.Y := YMaxi;
      FIDStationYMaxi := E666.oIDLitteral;
    end;
    if (E666.Une_Station_1_Y < YMini) then
    begin
      YMini := E666.Une_Station_1_Y;
      if (E666.Drawn) then FMetafilteredCoinBasGauche.Y := YMini;
      FIDStationYMini := E666.oIDLitteral;
    end;
    if (E666.Une_Station_1_Z > ZMaxi) then
    begin
      ZMaxi := E666.Une_Station_1_Z;
      if (E666.Drawn) then FMetafilteredCoinHautDroit.Z := ZMaxi;
      FIDStationZMaxi := E666.oIDLitteral;
    end;
    if (E666.Une_Station_1_Z < ZMini) then
    begin
      ZMini := E666.Une_Station_1_Z;
      if (E666.Drawn) then FMetafilteredCoinBasGauche.Z := ZMini;
      FIDStationZMini := E666.oIDLitteral;
    end;
    if (E666.Une_Station_2_X > XMaxi) then
    begin
      XMaxi := E666.Une_Station_2_X;
      if (E666.Drawn) then FMetafilteredCoinHautDroit.X := XMaxi;
      FIDStationXMaxi := E666.oIDLitteral;
    end;
    if (E666.Une_Station_2_X < XMini) then
    begin
      XMini := E666.Une_Station_2_X;
      if (E666.Drawn) then FMetafilteredCoinBasGauche.X := XMini;
      FIDStationXMini := E666.oIDLitteral;
    end;
    if (E666.Une_Station_2_Y > YMaxi) then
    begin
      YMaxi := E666.Une_Station_2_Y;
      if (E666.Drawn) then FMetafilteredCoinHautDroit.Y := YMaxi;
      FIDStationYMaxi := E666.oIDLitteral;
    end;
    if (E666.Une_Station_2_Y < YMini) then
    begin
      YMini := E666.Une_Station_2_Y;
      if (E666.Drawn) then FMetafilteredCoinBasGauche.Y := YMini;
      FIDStationYMini := E666.oIDLitteral;
    end;
    if (E666.Une_Station_2_Z > ZMaxi) then
    begin
      ZMaxi := E666.Une_Station_2_Z;
      if (E666.Drawn) then FMetafilteredCoinHautDroit.Z := ZMaxi;
      FIDStationZMaxi := E666.oIDLitteral;
    end;
    if (E666.Une_Station_2_Z < ZMini) then
    begin
      ZMini := E666.Une_Station_2_Z;
      if (E666.Drawn) then FMetafilteredCoinBasGauche.Z := ZMini;
      FIDStationZMini := E666.oIDLitteral;
    end;
  end;
begin

  AfficherMessage(ClassName + '.SetMinMax');
  //SetPositionPointZero;
  with self do
  begin
    XMini :=  GM;
    YMini :=  GM;
    ZMini :=  GM;
    XMaxi := -GM;
    YMaxi := -GM;
    ZMaxi := -GM;
    if (DoMetafilteredOnly) then
    begin
      for i := LOW_INDEX to GetNbEntitesVisees() - 1 do
      begin
        EWE := GetEntiteVisee(i);
        if (EWE.Drawn) then ZDar(EWE);
      end;
    end
    else
    begin
      for i := LOW_INDEX to GetNbEntitesVisees() - 1 do
      begin
        EWE := GetEntiteVisee(i);
        ZDar(EWE);
      end;
    end;
    // TODO: Les entrées
    NbEntrances := GetNbEntrees();
    for i := 0 to NbEntrances - 1 do
    begin
      ;
    end;
    // affectation des membres privés suivants
    FCoinBasGauche := MakeTPoint3Df(XMini, YMini, ZMini);
    FCoinHautDroit := MakeTPoint3Df(XMaxi, YMaxi, ZMaxi);
    AfficherMessage(Format('-- Reseau complet : Mini: %.0f, %.0f, %.0f', [FCoinBasGauche.X, FCoinBasGauche.Y, FCoinBasGauche.Z]));
    AfficherMessage(Format('-- Reseau complet : Maxi: %.0f, %.0f, %.0f', [FCoinHautDroit.X, FCoinHautDroit.Y, FCoinHautDroit.Z]));
    AfficherMessage(Format('-- Zone filtree   : Mini: %.0f, %.0f, %.0f', [FMetafilteredCoinBasGauche.X, FMetafilteredCoinBasGauche.Y, FMetafilteredCoinBasGauche.Z]));
    AfficherMessage(Format('-- Zone filtree   : Maxi: %.0f, %.0f, %.0f', [FMetafilteredCoinHautDroit.X, FMetafilteredCoinHautDroit.Y, FMetafilteredCoinHautDroit.Z]));
  end;
end;

// Coordonnees extremes
function TBDDEntites.GetCoinBasGauche: TPoint3Df;
begin
  Result := FCoinBasGauche;
end;

function TBDDEntites.GetCoinHautDroit: TPoint3Df;
begin
  Result := FCoinHautDroit;
end;

function TBDDEntites.GetMetafilteredCoinBasGauche: TPoint3Df;
begin
  Result := FMetafilteredCoinBasGauche;
end;

function TBDDEntites.GetMetafilteredCoinHautDroit: TPoint3Df;
begin
  Result := FMetafilteredCoinHautDroit;
end;

// ID des stations extremes
function TBDDEntites.GetIDStationXMini: string;
begin
  Result := FIDStationXMini;
end;

function TBDDEntites.GetIDStationYMini: string;
begin
  Result := FIDStationYMini;
end;
function TBDDEntites.GetIDStationZMini: string;
begin
  Result := FIDStationZMini;
end;

function TBDDEntites.GetIDStationXMaxi: string;
begin
  Result := FIDStationXMaxi;
end;
function TBDDEntites.GetIDStationYMaxi: string;
begin
  Result := FIDStationYMaxi;
end;
function TBDDEntites.GetIDStationZMaxi: string;
begin
  Result := FIDStationZMaxi;
end;

//******************************************************************************

procedure TBDDEntites.SortByDepth;
begin
  //AfficherMessage('--- SortByDepth()');
  FTableDesVisees.Sort(SortEntiteZOrder);
end;
procedure TBDDEntites.SortBySerSts;
begin
  //AfficherMessage('--- SortBySerSts()');
  FTableDesVisees.Sort(SortEntiteSerSt);
end;
procedure TBDDEntites.SortByTagDouble;
begin
  AfficherMessageErreur('--- SortByTagDouble()');
  FTableDesVisees.Sort(SortEntiteByTagDouble);
end;

/// fonctions de recherche
// recherche d'une entité près d'un point (x, y)
// Retourne: VRAI si entité trouvée, FAUX sinon
// Transmet: Entité trouvée
//           Index interne de l'entité
function TBDDEntites.GetEntiteViseeFromXY(const AntennesAussi: boolean;
                                          const X, Y: double;
                                          out   EE: TEntiteEtendue;
                                          out   InternalIdxEntite: integer) : boolean;
var
  dMax: Extended;
  i: Integer;
  E: TEntiteEtendue;
  d: Double;
begin
  Result := false;
  InternalIdxEntite := -1;
  dMax := 1.00e10;
  for i := LOW_INDEX to self.GetNbEntitesVisees() - 1 do
  begin
    E := self.GetEntiteVisee(i);
    // on zappe les antennes
    if (E.Type_Entite = tgVISEE_RADIANTE) then Continue;
    // on zappe les entitées métafiltrées
    if (not E.Drawn) then Continue;
    d := Hypot2D(E.Une_Station_2_X - X, E.Une_Station_2_Y - Y);
    if (d < dMax) then
    begin
      dMax := d;
      EE := E;
      Result := True;
      InternalIdxEntite := i;
    end;
  end;
end;

function TBDDEntites.GetEntiteViseeFromCle(const DoMatchExact: boolean; const Cle: string; out EE: TEntiteEtendue): boolean;
var
  WU: String;
  i, Nb: Integer;
  S: String;
  QCond: Boolean;
  Entr: TEntrance;
begin
  Result := false;
  WU := UpperCase(Trim(Cle));
  // d'abord les entrées
  Nb := GetNbEntrees();
  if (Nb > 0) then
  begin
    for i := 0 to Nb - 1 do
    begin
      Entr := GetEntree(i);
      if (DoMatchExact) then
      begin
        S := UpperCase(Trim(Entr.eIDTerrain));
        QCond := (S = WU);
      end
      else
      begin
        S  := UpperCase(Trim(Entr.eIDTerrain) + '|' + Entr.eNomEntree + '|' + Trim(Entr.eObserv));
        QCond := (Pos(WU, S) > 0);
      end;

      if (QCond) then
      begin
        EE.Type_Entite     := tgENTRANCE;
        EE.Entite_Serie    := Entr.eRefSer;
        EE.Entite_Station  := Entr.eRefSt;
        EE.oIDLitteral     := Format('%s - %d.%d - Entree: %s', [Entr.eIDTerrain, Entr.eRefSer, Entr.eRefSt, Entr.eNomEntree]);
        EE.oCommentaires   := Entr.eObserv;
        EE.Une_Station_2_X := Entr.eXEntree;
        EE.Une_Station_2_Y := Entr.eYEntree;
        EE.Une_Station_2_Z := Entr.eZEntree;
        Result := True;
        Exit;
      end;
    end;
  end;

  // enuite, la table des entités
  for i := LOW_INDEX to self.GetNbEntitesVisees() - 1 do
  begin
    EE := GetEntiteVisee(i);
    if (EE.Type_Entite = tgVISEE_RADIANTE) then Continue;
    if (DoMatchExact) then
    begin
      S := UpperCase(Trim(EE.oIDLitteral));
      QCond := (S = WU);
    end
    else
    begin
      S  := UpperCase(Trim(EE.oIDLitteral) + '|' + Trim(EE.oCommentaires));
      QCond := (Pos(WU, S) > 0);
    end;
    if (QCond) then
    begin
      Result := True;
      Exit;
    end;
  end;
end;
// Cette fonction est coûteuse: on vérifie si le précédent appel de GetEntiteFromSerSt()
// a été effectué avec le même couple (Ser, St)
// Si oui, on retourne le résultat de cette recherche, préalablement sauvegardé
// Si non, on effectue la recherche et on sauvegarde son résultat
// Comme les entités ne sont pas forcément triées selon leur ID,
// seul un balayage est possible
function TBDDEntites.GetEntiteViseeFromSerSt(const Ser, St: Integer; out EE: TEntiteEtendue): boolean;
var
  i, Nb: Integer;
  Entr: TEntrance;
  procedure QSetLastSuccessfullRequest();
  begin
    FLastEntiteFound := EE;
    FLastSerFound := Ser;
    FLastPtFound  := St;
  end;
begin
  Result := false;

  if ((FLastSerFound = Ser) and (FLastPtFound = St)) then
  begin
    EE := FLastEntiteFound;
    exit(true);
  end;
  Nb := self.GetNbEntrees();
  if (Nb > 0) then
  begin
    for i := 0 to Nb - 1 do
    begin
      Entr := GetEntree(i);
      if ((Entr.eRefSer = Ser) and (Entr.eRefSt = St)) then
      begin
        EE.Type_Entite := tgENTRANCE;
        EE.Entite_Serie := Ser;
        EE.Entite_Station := St;
        EE.oIDLitteral     := Entr.eIDTerrain + ' (Entree)';
        EE.oCommentaires   := Entr.eObserv;
        EE.Une_Station_2_X := Entr.eXEntree;
        EE.Une_Station_2_Y := Entr.eYEntree;
        EE.Une_Station_2_Z := Entr.eZEntree;
        QSetLastSuccessfullRequest(); // on sauvegarde le résultat de cette recherche
        Exit(True);
      end;
    end;
  end;
  Nb := self.GetNbEntitesVisees();
  for i := LOW_INDEX to Nb - 1 do
  begin
    EE := self.GetEntiteVisee(i);
    if (EE.Type_Entite  = tgVISEE_RADIANTE) then Continue;
    if (EE.Entite_Serie = Ser) and (EE.Entite_Station = St) then
    begin
      QSetLastSuccessfullRequest();  // on sauvegarde le résultat de cette recherche
      Exit(True);
    end;
  end;
end;
// recherche d'une entité à partir d'une clé
// Si la clé est de la forme 101.33, c'est à priori un couple série-station; on utilise GetEntiteFromSerSt()
// Si échec de cette première recherche, on utilise GetEntiteFromCle()
// Retourne: VRAI si entité trouvée, FAUX sinon
// Transmet: Entité trouvée
function TBDDEntites.FindStationByCle(const DoMatchExact: boolean; const Cle: string; out E: TEntiteEtendue): Boolean;
var
  S: String;
  WU: Integer;
  S1, S2: String;
  N: Integer;
  qSer, qSt: Integer;
begin
  Result := False;
  S := UpperCase(Trim(Cle));
  AfficherMessage(Format('%s.FindStationByCle(%s)', [self.ClassName, Cle]));
  // analyse de la clé
  // la clé contient un point --> recherche par série/station
  WU := Pos('.', S);
  if (WU > 0) then
  begin
    AfficherMessage('--> Recherche directe station.');
    N := Length(S);
    S1 := Trim(Copy(S, 0, WU - 1));
    S2 := Trim(Copy(S, WU + 1, N));
    qSer := StrToIntDef(S1, 0);
    qSt  := StrToIntDef(S2, 0);
    if (GetEntiteViseeFromSerSt(qSer, qSt, E)) then
    begin
      Result := True;
      Exit;
    end;
  end;
  // station non trouvee ? Tentative de Recherche par clé
  Result := GetEntiteViseeFromCle(DoMatchExact, S, E);
end;
// recherche d'une entrée par clé ou couple série/pt.
// Si DoMatchExact est armé, la recherche ne se fait que sur eIDTerrain
function TBDDEntites.FindEntranceByCle(const DoMatchExact: boolean; const Cle: string; out E: TEntrance): boolean;
var
  i, Nb: Integer;
  EWE: TEntrance;
  QKey, QAT: String;
  QCond: Boolean;
begin
  Result := false;
  QKey := LowerCase(trim(Cle));
  Nb := self.GetNbEntrees();
  if (Nb = 0) then exit;
  for i := 0 to Nb - 1 do
  begin
    EWE := GetEntree(i);
    if (DoMatchExact) then
    begin
      QAT := LowerCase(Trim(EWE.eIDTerrain)); // seul le champ eIDTerrain est évalué
      QCond := (QKey = QAT);
    end
    else
    begin
      QAT := Lowercase(Trim(EWE.eIDTerrain + '|' + trim(EWE.eNomEntree) + '|' + trim(EWE.eObserv)));
      QCond := (Pos(QKey, QAT) > 0);
    end;
    if (QCond) then
    begin
      E := EWE;
      Exit; // a tester
    end;
  end;
end;


//******************************************************************************
// Statistiques
// calculs statistiques
// semble OK
procedure TBDDEntites.CalculSpeleometrie;
var
  i,j : integer;
  EWE: TEntiteEtendue;
  WU: TExpe;
  X69: TDateTime;
  QAT: TCode;
  function  GetEmptyVentilation: TVentilationSpeleometrie;
  begin
    with Result do
    begin
      Fossiles   := 0.00;
      Vadoses    := 0.00;
      Ennoyables := 0.00;
      Siphons    := 0.00;
      Tunnels    := 0.00;
      Filons     := 0.00;
      Speciaux   := 0.00;
    end;
  end;
  function GetLigneRecapitulative(const QV: TVentilationSpeleometrie; const EE: TEntiteEtendue): TVentilationSpeleometrie;
  var
    dv: Double;
  begin
    Result := QV;
    dv := Hypot3D(EE.Une_Station_2_X - EE.Une_Station_1_X,
                  EE.Une_Station_2_Y - EE.Une_Station_1_Y,
                  EE.Une_Station_2_Z - EE.Une_Station_1_Z);
    case EE.Type_Entite of
      tgENTRANCE   : ; // entrées
      tgVISEE_RADIANTE: ; // visées en antenne
      tgDEFAULT,
      tgFOSSILE    : Result.Fossiles    := Result.Fossiles   + dv;
      tgVADOSE     : Result.Vadoses     := Result.Vadoses    + dv;
      tgENNOYABLE  : Result.Ennoyables  := Result.Ennoyables + dv;
      tgSIPHON     : Result.Siphons     := Result.Siphons    + dv;
      tgFIXPOINT   : ; // points fixes
      tgSURFACE    : Result.Speciaux    := Result.Speciaux   + dv;
      tgTUNNEL     : Result.Filons      := Result.Filons     + dv;
      tgMINE       : Result.Tunnels     := Result.Tunnels    + dv;
    else
                     Result.Fossiles    := Result.Fossiles   + dv;
    end;
  end;
  procedure  AjouterLongueurVisee(var QTableauRecap: TTableauVentilationSpeleometrie; const QE: TEntiteEtendue; const QIdx: integer);
  begin
    QTableauRecap[QIdx] := GetLigneRecapitulative(QTableauRecap[QIdx], QE);
  end;
  procedure PurgerTableau(var QTableauRecap: TTableauVentilationSpeleometrie; const QNb: integer);
  var
    n: Integer;
  begin
    SetLength(QTableauRecap, QNb);
    for n := 0 to QNb - 1 do QTableauRecap[n] := GetEmptyVentilation;
  end;
begin
  AfficherMessage('TdlgSpeleometrie.CalculSpeleometrie()');
  AfficherMessage('');
  // init tableaux
  PurgerTableau(FSpeleometrieParReseaux   , GetNbReseaux);
  PurgerTableau(FSpeleometrieParSecteurs  , GetNbSecteurs);
  PurgerTableau(FSpeleometrieParExpes     , GetNbExpes);
  PurgerTableau(FSpeleometrieParCodes     , GetNbCodes);
  PurgerTableau(FSpeleometrieParDates     , GetNbDates);
  AfficherMessage(format('  Speleometrie par expes (%d expes)',[GetNbExpes]));
  //-----------------------
  try
    for i := LOW_INDEX to GetNbEntitesVisees() - 1 do
    begin
      EWE := GetEntiteVisee(i);
      for j:=0 to GetNbExpes - 1 do
      begin
        WU := GetExpe(j);
        if (EWE.eExpe = WU.IDExpe) then AjouterLongueurVisee(FSpeleometrieParExpes, EWE, j);
      end;
    end;
    AfficherMessage(' ---> Expes OK');
  except
    AfficherMessage(' ---> Expes KO');
  end;
  AfficherMessage(format('  Speleometrie par dates (%d dates)',[GetNbDates]));
  //-----------------------
  try
    for i:= LOW_INDEX to GetNbEntitesVisees() - 1 do
    begin
      EWE := GetEntiteVisee(i);
      //if ((i mod 100) = 0) then AfficherMessage(Format('%d - %s', [i, DateToStr(EWE.DateLeve)]));
      for j:=0 to GetNbDates - 1 do
      begin
        // troncature des dates et comparaison
        if (Trunc(EWE.DateLeve) = Trunc(FListeDesDates[j])) then AjouterLongueurVisee(FSpeleometrieParDates, EWE, j);
      end;
    end;
    AfficherMessage(' ---> Dates OK');
  except
    AfficherMessage(' ---> Dates KO');
  end;


  AfficherMessage(format('  Speleometrie par codes (%d codes)',[GetNbCodes]));
  //-----------------------
  try
    for i:= LOW_INDEX to GetNbEntitesVisees() - 1 do
    begin
      EWE := GetEntiteVisee(i);
      for j:=0 to GetNbCodes - 1 do
      begin
        QAT := GetCode(j);
        if (EWE.eCode = QAT.IDCode) then
          AjouterLongueurVisee(FSpeleometrieParCodes, EWE, j);
      end;
    end;
    AfficherMessage(' ---> Codes OK');
  except
    AfficherMessage(' ---> Codes KO');
  end;
  // calcul de la spéléométrie par réseaux
  AfficherMessage(format('  Speleometrie par reseaux (%d reseaux)',[GetNbReseaux]));
  try
    for i := LOW_INDEX to GetNbEntitesVisees() - 1 do
    begin
      EWE := GetEntiteVisee(i);
      for j := 0 to GetNbReseaux -1 do
      begin
        if (EWE.eReseau = j) then AjouterLongueurVisee(FSpeleometrieParReseaux, EWE, j);
      end;
    end;
    AfficherMessage(' ---> Reseaux OK');
  except
    AfficherMessage(' ---> Reseaux KO');
  end;
  // calcul de la spéléométrie par secteurs
  AfficherMessage(format('  Speleometrie par secteurs (%d secteurs)',[GetNbSecteurs]));
  //SetLength(FSpeleometrieParDates,FNbSeances);
  try
    for i := LOW_INDEX to GetNbEntitesVisees() - 1 do
    begin
      EWE := GetEntiteVisee(i);
      for j := 0 to GetNbSecteurs -1 do
      begin
        if (EWE.eSecteur = j) then AjouterLongueurVisee(FSpeleometrieParSecteurs, EWE, j);
      end;
    end;
    AfficherMessage(' ---> Secteurs OK');
  except
    AfficherMessage(' ---> Secteurs OK');
  end;
  // calcul de la spéléométrie totale
  AfficherMessage('Speleometrie totale');
  //SetLength(FSpeleometrieParDates,FNbSeances);
  try
    FRecapitulation := GetEmptyVentilation;
    for i := LOW_INDEX to GetNbEntitesVisees() - 1 do
    begin
      EWE := GetEntiteVisee(i);
      FRecapitulation := GetLigneRecapitulative(FRecapitulation, EWE);
    end;
    AfficherMessage(' ---> Total OK');
  except
  end;
  //*)
end;
// attraper les spéléométries par index
function TBDDEntites.GetSpeleometrieParExpeByIdx(const Idx: integer): TVentilationSpeleometrie;
begin
  Result := FSpeleometrieParExpes[Idx];
end;

function TBDDEntites.GetSpeleometrieParCodeByIdx(const Idx: integer): TVentilationSpeleometrie;
begin
   Result := FSpeleometrieParCodes[Idx];
end;

function TBDDEntites.GetSpeleometrieParDateByIdx(const Idx: integer): TVentilationSpeleometrie;
begin
  Result := FSpeleometrieParDates[Idx];
end;

function TBDDEntites.GetSpeleometrieParReseauxByIdx(const Idx: integer): TVentilationSpeleometrie;
begin
  Result := FSpeleometrieParReseaux[Idx];
end;
function TBDDEntites.GetSpeleometrieParSecteurByIdx(const Idx: integer): TVentilationSpeleometrie;
begin
  Result := FSpeleometrieParSecteurs[Idx];
end;
//******************************************************************************
// Exportations

// semble OK - A vérifier - Couleurs à implémenter.
// GPX: Semble OK
// OSM: Semble OK
// KML: Semble OK
// GPX only: Export des visées rayonnantes OK
// Leaflet: OK pour les silhouettes - TODO: Implementer les entrées
procedure TBDDEntites.ExportForCarto(const FileName: TStringDirectoryFilename;
                                     const Convertisseur: TConversionSysteme;
                                     const QSystemeGeogEPSG: integer;
                                     const ModeOutputFormatGIS: TOutputFormatGIS;
                                     const DoMakeSilhouette: boolean;
                                     const Filtres: string;
                                     const PrefixStations: string;
                                     const CouleurDefaut: TColor;
                                     const UseColorGroupes: boolean);

var
  i: integer;
  //MyEntite: TEntite;
  E1, E2: TEntiteEtendue;
  OldSerie: LongInt;
  MySilhouette: TSilhouetteGalerie;
  PtsParois, PtsParois2: TPtsSectionTransversale;
  PolygoneSilhouette: Boolean;
  foo: String;
  NbreSilhouettes: Integer;
  FConvertisseurEPSG: TConversionSysteme;
  //_WU: TColorGaleries;
  MyColorEntite: TColor;
  EWE: TLabelSystemesCoordsEPSG;
  QNbEntites, QNBEntrees: Integer;
  MyEntrance: TEntrance;
  OSMNode0: Int64;
  MyExportGIS: TExportGIS;
  QDevelViseesVisibles: double;


  //----------------------------------------------------------------------------
  //----------------------------------------------------------------------------
  // dessiner le polygone de silhouette
  procedure DessinerPolygoneSilhouette(const QSilhouette: TSilhouetteGalerie;
                                       const TagString: string;
                                       const TagNum: Int64);
  const
    DELTA_Z_SURFACE = 10.00;
  var
    QC: TColorRGBA;
    ii : integer;
    PP: TPoint2Df;
    P1, P2: TProjUV;
    Nb: LongInt;
    EE: TPtsSectionTransversale;
    QStart: Integer;
    QStartWay: Integer;
    fred_mignon: String;  // le Fred Mignon est en string :lol:
    Wizz: string;
  begin
     // construction du polygone
     if (not QSilhouette.BuildPolygoneParois) then Exit;
     case ModeOutputFormatGIS of
      gisKML:
        begin // KML
          Nb := QSilhouette.GetNbPointsPolygoneParois;
          MyExportGIS.WriteLigne('      <Placemark>');
          //MyExportGIS.WriteLigne(Format('       <!-- Polygone: %s (%d sommets) -->',['MyPolygone', Nb]));
          MyExportGIS.WriteLigne(Format('         <name>%s</name>', [TagString]));
          if (UseColorGroupes) then    // utilise les couleurs de groupe de la topographie
          begin
            QC := QSilhouette.GetFillColor;
            fred_mignon := STYLE_GALERIE_PAR_DEFAUT +
                           format('%.2X%.2X%.2X%.2X', [QC.R, QC.G, QC.B, QC.A]);
            MyExportGIS.WriteLigne(Format('         <styleUrl>#%s</styleUrl>', [fred_mignon]));
          end
          else                        // sinon, utilise la couleur spécifiée
          begin
            MyExportGIS.WriteLigne(Format('         <styleUrl>#%s</styleUrl>', [STYLE_GALERIE_PAR_DEFAUT]));
          end;
          MyExportGIS.WriteLigne('         <Polygon>');
          MyExportGIS.WriteLigne(Format('           <tesselate>%d</tesselate>', [1]));
          MyExportGIS.WriteLigne('           <outerBoundaryIs>');
          MyExportGIS.WriteLigne('             <LinearRing>');
          MyExportGIS.WriteLigne('               <coordinates>');
          for ii := 0 to Nb - 1 do
          begin
            PP := QSilhouette.GetPointPolygoneParois(ii);
            P1.U := PP.X;
            P1.V := PP.Y;
            P2 := FConvertisseurEPSG.ConversionSyst1ToSyst2EPSG(QSystemeGeogEPSG, CODE_EPSG_WGS84, P1);
            MyExportGIS.WriteLigne(Format('             %s,%s,%s', [
                               ensureConvertLongOrLatToXML(P2.V),
                               ensureConvertLongOrLatToXML(P2.U),
                               ensureConvertAltitudeToXML(DELTA_Z_SURFACE)]));   //Lon, Lat
          end;
          MyExportGIS.WriteLigne('               </coordinates>');
          MyExportGIS.WriteLigne('             </LinearRing>');
          MyExportGIS.WriteLigne('           </outerBoundaryIs>');
          MyExportGIS.WriteLigne('         </Polygon>');
  	      MyExportGIS.WriteLigne('      </Placemark>');
        end;
      gisGPX:
        begin // GPX
          Nb := QSilhouette.GetNbPointsPolygoneParois;
          //MyExportGIS.WriteLigne(Format('<!-- Polygone: %s (%d sommets) -->',['MyPolygone', Nb]));
          MyExportGIS.WriteLigne('    <trk> <trkseg>');
          for ii := 0 to Nb - 1 do
          begin
            PP := QSilhouette.GetPointPolygoneParois(ii);
            P1.U := PP.X;
            P1.V := PP.Y;
            P2 := FConvertisseurEPSG.ConversionSyst1ToSyst2EPSG(QSystemeGeogEPSG, CODE_EPSG_WGS84, P1);
            MyExportGIS.WriteLigne(Format('        <trkpt lat="%s" lon="%s"></trkpt>', [
                                          ensureConvertLongOrLatToXML(P2.U),
                                          ensureConvertLongOrLatToXML(P2.V)]))    //QLat, QLon
          end;
          MyExportGIS.WriteLigne('    </trkseg> </trk>');
        end;

      gisLEAFLET: // LeafLet
        begin
          //********************************************************************
          Nb := QSilhouette.GetNbPointsPolygoneParois;

          MyExportGIS.WriteLigne(Format('       //Polygone: %s (%d sommets) ',[TagString, Nb]));
          if (UseColorGroupes) then QC := QSilhouette.GetFillColor   // utilise les couleurs de groupe de la topographie
                               else QC := GetColorRGBA(clRed, 192);

          MyExportGIS.WriteLigne('      L.polygon([');
          for ii := 0 to Nb - 1 do
          begin
            PP := QSilhouette.GetPointPolygoneParois(ii);
            P1.U := PP.X;
            P1.V := PP.Y;
            if (ii < (Nb-1)) then Wizz := ',' else Wizz := '';
            P2 := FConvertisseurEPSG.ConversionSyst1ToSyst2EPSG(QSystemeGeogEPSG, CODE_EPSG_WGS84, P1);
            MyExportGIS.WriteLigne(Format('             [%s, %s] %s // %s', [
                        ensureConvertLongOrLatToXML(P2.U),
                        ensureConvertLongOrLatToXML(P2.V),
                        Wizz,
                        ensureConvertAltitudeToXML(DELTA_Z_SURFACE)]));   //Lon, Lat
          end;
          MyExportGIS.WriteLigne('       ], {');
          MyExportGIS.WriteLigne(Format('          color: ''%s'', ', [ColorToHTMLColor(RGBToColor(QC.R, QC.G, QC.B))]));
          MyExportGIS.WriteLigne(Format('          weight: %d, ', [1]));
          MyExportGIS.WriteLigne(Format('          opacity: %.2f, ', [1.00]));
          MyExportGIS.WriteLigne(Format('          fillcolor: ''%s'', ', [ColorToHTMLColor(RGBToColor(QC.R, QC.G, QC.B))]));
          MyExportGIS.WriteLigne(Format('          fillopacity: %.2f', [1.00])); //[QScrap.Opacite / 256.0]));
          MyExportGIS.WriteLigne('        }');
          // on utilise TagString, auquel le nom du groupe a été passé
          MyExportGIS.WriteLigne(Format('      ).addTo(map).bindPopup("%s");', [QSilhouette.GetNomSilhouette()]));


          //********************************************************************
        end;
    end;
  end;
  // export des entrées
  procedure DessinerMarkerEntrance(const QEntrance: TEntrance;
                                   const NoBasePt : Int64);
  begin
    MyExportGIS.WriteAEntrance(MyEntrance, OSMNode0 + i);
  end;
begin
  // affectation du convertisseur
  FConvertisseurEPSG := Convertisseur;
  NbreSilhouettes := 0;
  foo := ChooseString(Ord(ModeOutputFormatGIS), ['--', 'Google Earth', 'GPX standard', 'OpenStreetMap', 'Leaflet', 'Carto Exploreur', 'Memory Map']);
  //CouleurDefaut := clRed;
  // controle paramètres
  AfficherMessage(Format('%s.ExportForCarto(%s)',[self.ClassName,FileName]));
  AfficherMessage(Format('--> Filtres: %s',[Filtres]));
  //AfficherMessage(Format('--> Mode: %d; %s',[IIF(DoMakeSilhouette, 'Silhouette des parois', 'Polygonale seule')]));
  AfficherMessage(Format('--> Systeme de coordonnees: EPSG:%d',[QSystemeGeogEPSG]));

  EWE := FConvertisseurEPSG.GetEPSGSystemeFromCodeEPSG(QSystemeGeogEPSG);
  AfficherMessage(Format('--> Systeme coordonnees: %d = %s',[EWE.CodeEPSG, EWE.NomEPSG]));
  AfficherMessage(Format('--> Export vers: %s', [foo]));
  // Un coup de MétaFiltre pour commencer
  self.MetaFiltre(Filtres, QDevelViseesVisibles);
  SetMinMax(False);
  // Tri par séries et stations (peut-être inutile)
  SortBySerSts();
  // ouverture fichier
  try
    try
      MyExportGIS := TExportGIS.Create;
      MyExportGIS.Initialiser(ModeOutputFormatGIS,
                              Self,
                              FConvertisseurEPSG,
                              QSystemeGeogEPSG,
                              UseColorGroupes,
                              CouleurDefaut,
                              1200, 900,
                              FileName);
      MyExportGIS.WriteHeader(false, -1.0, -1.0);
      if (DoMakeSilhouette) then
      begin
        AfficherMessage('Mode silhouettes ON');
        MyExportGIS.WriteCommentaire('Mode silhouettes ON');
        MySilhouette := TSilhouetteGalerie.Create;
        // premier polygone
        E1 := self.GetEntiteVisee(1);
        AfficherMessage('00');
        MyColorEntite := GetCouleurEntiteRGBByExpe(E1);
        MySilhouette.ClearSilhouette(MyColorEntite, MyColorEntite);
        AfficherMessage('01');
        PtsParois.ParoiDroiteX := E1.X1PD;
        PtsParois.ParoiDroiteY := E1.Y1PD;
        PtsParois.ParoiGaucheX := E1.X1PG;
        PtsParois.ParoiGaucheY := E1.Y1PG;
        PtsParois.CenterlineX  := E1.Une_Station_1_X;
        PtsParois.CenterlineY  := E1.Une_Station_1_Y;
        PtsParois.CenterlineZ  := E1.Une_Station_1_Z;

        MySilhouette.AddPtsParois(PtsParois);
        AfficherMessage(format('%d entités',[self.GetNbEntitesVisees]));
        QNbEntites := self.GetNbEntitesVisees();
        for i:= 2 to QNbEntites - 1 do
        begin
          if (Assigned(FProcDispProgression)) then FProcDispProgression(Format('Entite %d / %d', [i, QNbEntites]), i, 1, QNbEntites);
          E1 := GetEntiteVisee(i);
          E2 := GetEntiteVisee(i-1);
          PtsParois.ParoiDroiteX := E1.X2PD;
          PtsParois.ParoiDroiteY := E1.Y2PD;
          PtsParois.ParoiGaucheX := E1.X2PG;
          PtsParois.ParoiGaucheY := E1.Y2PG;
          PtsParois.CenterlineX  := E1.Une_Station_2_X;
          PtsParois.CenterlineY  := E1.Une_Station_2_Y;
          PtsParois.CenterlineZ  := E1.Une_Station_2_Z;

          if (IsSameSerie(E1, E2)) then
          begin
            MySilhouette.AddPtsParois(PtsParois);
            // donner à la silhouette le nom de la série
            MySilhouette.SetNomSilhouette(Format('Serie %d', [E1.Entite_Serie]));
          end else
          begin
            // on construit et trace le polygone
            DessinerPolygoneSilhouette(MySilhouette, Format('Galerie%.5d',[NbreSilhouettes]), NbreSilhouettes);
            NbreSilhouettes := 1 + NbreSilhouettes;
            // on purge la silhouette courante pour en accueillir la suivante
            MyColorEntite := GetCouleurEntiteRGBByExpe(E1);
            MySilhouette.ClearSilhouette(MyColorEntite, MyColorEntite);
            PtsParois2.ParoiDroiteX := E1.X1PD;
            PtsParois2.ParoiDroiteY := E1.Y1PD;
            PtsParois2.ParoiGaucheX := E1.X1PG;
            PtsParois2.ParoiGaucheY := E1.Y1PG;
            PtsParois2.CenterlineX   := E1.Une_Station_1_X;
            PtsParois2.CenterlineY   := E1.Une_Station_1_Y;
            PtsParois2.CenterlineZ   := E1.Une_Station_1_Z;

            MySilhouette.AddPtsParois(PtsParois2);
          end;
        end;  // for
        AfficherMessage('Mode silhouettes DONE');
        MyExportGIS.WriteCommentaire('Mode silhouettes DONE');
        // dernier polygone
        DessinerPolygoneSilhouette(MySilhouette, Format('Galerie%.5d',[NbreSilhouettes]), NbreSilhouettes);
        NbreSilhouettes := 1 + NbreSilhouettes;
        MySilhouette.ClearSilhouette(clWhite, clWhite);
        MySilhouette.Free;
      end   // if (silhouette)
      else  // if (silhouette)
      begin
        AfficherMessage('Mode silhouettes OFF');
        MyExportGIS.WriteCommentaire('Mode silhouettes OFF');
        E1 := GetEntiteVisee(1);
        MyExportGIS.WriteConvertedPoint(tpmENTETE_POLY, E1);
        MyExportGIS.WriteConvertedPoint(tpmSTART_POLY, E1);
        QNbEntites := self.GetNbEntitesVisees();
        for i:= 2 to QNbEntites - 1 do
        begin
          if (Assigned(FProcDispProgression)) then FProcDispProgression(Format('Entite %d / %d', [i, QNbEntites]), i, 1, QNbEntites);
          E1 := GetEntiteVisee(i);
          if (E1.Type_Entite = tgENTRANCE) then Continue;
          if (E1.Type_Entite = tgVISEE_RADIANTE) then Continue;
          E2 := GetEntiteVisee(i-1);
          if (IsSameSerie(E1, E2)) then
          begin
            MyExportGIS.WriteConvertedPoint(tpmPOINT_POLY, E1);
          end else
          begin
            MyExportGIS.WriteConvertedPoint(tpmEND_POLY, E1); // pour GPX uniquement: balise de cloture
            MyExportGIS.WriteConvertedPoint(tpmENTETE_POLY, E1);
            MyExportGIS.WriteConvertedPoint(tpmSTART_POLY, E1);
          end;
        end;
        MyExportGIS.WriteConvertedPoint(tpmEND_POLY, E1); // pour GPX uniquement: balise de cloture
        // antennes
        (*
        for i:= 1 to QNbEntites - 1 do
        begin
          if (Assigned(FProcDispProgression)) then FProcDispProgression(Format('Entite %d / %d', [i, QNbEntites]), i, 1, QNbEntites);
          E1 := GetEntite(i);
          if (E1.Type_Entite = tgENTRANCE) then Continue;
          if (E1.Type_Entite = tgVISEE_RADIANTE) then MyExportGIS.WriteSegmentShot(E1);
        end;
        //*)
        MyExportGIS.WriteCommentaire('Centerlines DONE');
      end;  // if (silhouette)
      // cloture du document
      // exporter les entrées

      QNBEntrees := self.GetNbEntrees();
      OSMNode0   := QNbEntites + QNBEntrees + 100;
      AfficherMessageErreur(Format('OSM: Export %d entrées', [QNBEntrees]));
      if (QNBEntrees > 0) then
      begin
        AfficherMessage(Format('-- Export %d entrances', [QNBEntrees]));
        MyExportGIS.WriteCommentaire(Format('**** %d entrances ***', [QNBEntrees]));
        for i := 0 to QNBEntrees - 1 do
        begin
          MyEntrance := self.GetEntree(i);
          AfficherMessageErreur(Format('%d: %s', [i, MyEntrance.eNomEntree]));

          DessinerMarkerEntrance(MyEntrance, OSMNode0 + i);
        end;
        MyExportGIS.WriteCommentaire(Format('**** End of entrances ***', [QNBEntrees]));

      end;
      // départ de la numérotation des noeuds OSM
      (*
      for i := 0 to QNB - 1 do
      begin
        MyEntrance := self.GetEntree(i);
        MyExportGIS.WriteOSMEntrance(MyEntrance, OSMNode0 + i);
      end;
       end;
      //*)
      // -----------------------------------------------------------------------
      AfficherMessageErreur('OSM: Export entrées OK');

      MyExportGIS.WriteFooter();
      except
    end;
  finally
    MyExportGIS.Finaliser();
    MyExportGIS.Free;
  end;
end;
// semble OK. A valider
// SetMinMax inutile
procedure TBDDEntites.ExportForGHCaveDraw(const MyFile: TStringDirectoryFilename; const QMyFiltre: string);
var
  F: TextFile;
  i: integer;
  QID: Int64;
  OldFilter: string;
  E: TEntiteEtendue;
  Nb: Integer;
  NbViseesAntenne: Integer;
  QDevelViseesVisibles: double;
  procedure WrtLn(const S: string);
  begin
    WriteLn(F, S);
  end;
begin
  AfficherMessage(Format('%s.ExportForGHCaveDraw(%s) - Filtres: %s',[ClassName, MyFile, QMyFiltre]));
  AssignFile(F, MyFile);
  try
    // sauvegarder le filtre
    OldFilter := QMyFiltre;
    // export vers GHCaveDraw = on envoie tout
    MetaFiltre('', QDevelViseesVisibles);
    // **********************************
    ReWrite(F);
    // en-tête
    // points de base des éléments du dessin
    WrtLn('# Base points');
    WrtLn('');
    WrtLn('basepoints');
    Nb := self.GetNbEntitesVisees();
    NbViseesAntenne := 0;
    for i := LOW_INDEX to Nb - 1 do
    begin
      E  := GetEntiteVisee(i);
      //if (Not E.Drawn) then Continue;
      // NOTA: Les visées en antennes ne sont pas prises en compte par
      //       les fonctions de recherche et d'indexation de GHCaveDraw
      QID := MakeGHCaveDrawIDPtCenterline(E, NbViseesAntenne);
      WrtLn(Format(FMT_BASEPOINTS,
                  [QID  ,
                   E.oIDLitteral,
                   E.Type_Entite,
                   GetCouleurEntiteRGBByExpe(E), //E.ColorEntite,
                   E.Une_Station_1_X, E.Une_Station_1_Y, E.Une_Station_1_Z,
                   E.Une_Station_2_X, E.Une_Station_2_Y, E.Une_Station_2_Z,
                   E.X2PG, E.Y2PG,  E.Z1PB,
                   E.X2PD, E.Y2PD,  E.Z1PH
                   ]));
    end;
    WrtLn('endbasepoints');
  finally
    // restauration du filtre précédent
    MetaFiltre(OldFilter, QDevelViseesVisibles);
    CloseFile(F);
  end;
end;



// Validé OK
procedure TBDDEntites.CalcCouleursByDepth(const CouleurZMin, CouleurZMax: TColor);
var
  i, Nb: Integer;
  RStart, GStart, BStart : byte;
  RStop,  GStop,  BStop  : byte;

  RDiff, GDiff, BDiff    : Integer; // doit être un entier signé

  DeltaZ: Extended;
  EWE: TEntiteEtendue;
  qz: Double;
  qc: Extended;
  qr: Extended;
  CC: TColor;
  qMax666: Extended;
  qMin666: Extended;
  WU0, WU1: LongInt;
begin
  FCouleurZMin := CouleurZMin;
  FCouleurZMax := CouleurZMax;

  // couleur de départ et pas
  RedGreenBlue(ColorToRGB(FCouleurZMin),  RStart, GStart, BStart);
  RedGreenBlue(ColorToRGB(FCouleurZMax),  RStop,  GStop,  BStop);

  RDiff := RStop - RStart;
  GDiff := GStop - GStart;
  BDiff := BStop - BStart;

  DeltaZ := FCoinHautDroit.Z - FCoinBasGauche.Z;
  Nb := GetNbEntitesVisees();
  qMax666 :=  1E-20;
  qMin666 := -qMax666;

  for i:= LOW_INDEX to Nb - 1 do
  begin
    EWE := GetEntiteVisee(i);
    // cote moyenne
    qz   := 0.50 * (EWE.Une_Station_1_Z + EWE.Une_Station_2_Z); // altitude absolue
    qc   := qz - FCoinBasGauche.Z;                              // cote par rapport au point bas
    qr   := qc / DeltaZ;                                        // rapport
    if (qr > qMax666) then qMax666:= qr;
    if (qr < qMin666) then qMin666:= qr;


    CC   := RGBToColor(Trunc(RStart + qr * RDiff),
                       Trunc(GStart + qr * GDiff),
                       Trunc(BStart + qr * BDiff));
    EWE.ColourByDepth := CC;
    PutEntiteVisee(i, EWE);
  end;
  AfficherMessage(Format('CalcCouleursByDepth: %.3f, %.3f', [qMin666, qMax666]));
end;



function TBDDEntites.GetColorViseeFromModeRepresentation(const M: TModeRepresentationGaleries; const E: TEntiteEtendue): TColor;
begin
  case M of
    rgRESEAUX : Result := GetCouleurEntiteRGBByReseau(E);
    rgSECTEURS: Result := GetCouleurEntiteRGBBySecteur(E);
    rgSEANCES : Result := GetCouleurEntiteRGBByExpe(E);
    rgGRAY    : Result := clSilver;
    rgDEPTH   : Result := E.ColourByDepth; // DONE: Dégradé à implémenter
  otherwise
    Result := clSilver;
  end;
end;


//******************************************************************************
// MétaFiltre
//******************************************************************************
//Signature: function TBDDEntites.MetaFiltre(const Filtre: string): integer;
{$INCLUDE MetaFiltre.inc} // Ne pas modifier ce fichier
//******************************************************************************
function TBDDEntites.RecenserDates: integer;
var
  i: integer;
  Year, Month, Day: word;
  EWE: TEntiteEtendue;
  LS: TStringList;
begin
  FNbDates := 0;
  AfficherMessage(Format('%s.RecenserDates ()', [ClassName]));
  Result := 0;
  LS := TStringList.Create;
  try
    try
      LS.Sorted := True;
      LS.Duplicates := dupIgnore;
      LS.Clear;
      for i := LOW_INDEX to GetNbEntitesVisees() - 1 do
      begin
        EWE := self.GetEntiteVisee(i);
        SafeDecodeDate(EWE.DateLeve, Year, Month, Day);
        LS.Add(Format('%.4d-%.2d-%.2d', [Year, Month, Day]));
      end;
      // remplir le tableau des dates
      SetLength(FListeDesDates, 0);
      SetLength(FListeDesDates, LS.Count);
      for i := 0 to LS.Count - 1 do
      begin
        Year  := StrToIntDef(Copy(LS.Strings[i], 1, 4), 1900);
        Month := StrToIntDef(Copy(LS.Strings[i], 6, 2), 1);
        Day   := StrToIntDef(Copy(LS.Strings[i], 9, 2), 1);
        FListeDesDates[i] := GetSecuredDate(Year, Month, Day);
        //AfficherMessage(Format('   Date %d = %s', [i, DateToStr(FListeDesDates[i])]));
      end;
      Result   := LS.Count;
      FNbDates := LS.Count;
      // destruction de la liste
      LS.Clear;
    except
    end;
  finally
    LS.Free;
  end;
  AfficherMessage(Format('%s.RecenserDates () OK', [ClassName]));
end;

function TBDDEntites.GetUneDate(const Idx: integer): TDateTime;
begin
  try
    Result := FListeDesDates[Idx];
  except
    Result := TDateTime(0.00);
  end;
end;

function TBDDEntites.GetNbDates: integer;
begin
  Result := FNbDates;
end;

//******************************************************************************
// stats et diagrammes
// histogramme des profondeurs : calcul
function TBDDEntites.ParseDepthHistogramme(const NbPetales: integer; const ViseeMini: double; const Normalize: boolean): Boolean;
var
  i, j: integer;
  Interval: double;
  dx, dy, dz: double;
  LPT, LP: double;
  Z1, Z2, ZMoy: double;
  ClasseMax: double;
  SinP: double; // sinus de la pente
  MyEntite: TEntiteEtendue;
begin
  Result := false;
  AfficherMessage(Format('%s.ParseDepthHistogramme: Nb = %d; Mini=%.2f', [ClassName, NbPetales, ViseeMini]));
  try
    // pente maxi
    SinP := 0.50; // angle de 30°
    // intervalle
    Interval := (FCoinHautDroit.Z - FCoinBasGauche.Z) / NbPetales;
    // table des classes
    SetLength(FClassRepartDepthDiagram, 0);
    SetLength(FClassRepartDepthDiagram, NbPetales);
    // initialiser cette table
    for i := 0 to NbPetales - 1 do FClassRepartDepthDiagram[i] := 0.00;
    // boucle de calcul
    LPT := 0.00;
    for i := LOW_INDEX to GetNbEntitesVisees() - 1 do
    begin
      // calcul longueur projetee
      MyEntite := GetEntiteVisee(i);
      with MyEntite do
      begin
        if (not Drawn)                then Continue; // visées non dessinées
        //if (Type_Entite = tgENTRANCE) then Continue;  // entrées
        // /!\ Utiliser IsViseetInNaturalCave et non IsViseetInCaveOrTunnel
        if (not IsViseetInNaturalCave(MyEntite)) then Continue; // galeries non naturelles
        dx := Une_Station_2_X - Une_Station_1_X;
        dy := Une_Station_2_Y - Une_Station_1_Y;
        dz := Une_Station_2_Z - Une_Station_1_Z;
        ZMoy := (Une_Station_2_Z + Une_Station_1_Z) / 2.0;
        LP := Hypot3D(dx, dy, dz);
      end;
      // test: passer à entitée suivante si LP<ViseeMini
      if (LP < ViseeMini) then CONTINUE;
      if (not IsInRange(dz / LP, -SinP, SinP)) then Continue;
      // cumul des longueurs projetées
      LPT := LPT + LP;
      // répartition
      for j := 0 to NbPetales - 1 do
      begin
        Z1 := FCoinBasGauche.Z + Interval * j;
        Z2 := FCoinBasGauche.Z + Interval * (j + 1);
        if (IsInRange(ZMoy, Z1, Z2)) then FClassRepartDepthDiagram[j] := FClassRepartDepthDiagram[j] + LP;
      end;
    end; // for i ..
    // Longueur totale nulle -> sortie de la fonction
    if (LPT = 0) then Exit;
    if (Normalize) then
    begin
      // mise en pourcentage
      ClasseMax := 0.0;
      for j := 0 to NbPetales - 1 do
        if (FClassRepartDepthDiagram[j] > ClasseMax) then
          ClasseMax := FClassRepartDepthDiagram[j];
      for j := 0 to NbPetales - 1 do FClassRepartDepthDiagram[j] := FClassRepartDepthDiagram[j] / ClasseMax;
    end;
    Result := True;
  except
  end;
end;


// histogramme des directions : calcul
function TBDDEntites.ParseRoseDiagram(const NbPetales: integer; const ViseeMini: double): boolean;
var
  i, j: integer;
  LP: double;
  dx: double;
  dy: double;
  dz: double;
  Ang: double;
  LPT: double;
  s: string;
  ClasseMax: double;
  b: boolean;
  Interval: double;
  MyEntite: TEntiteEtendue;
begin
  //AfficherMessage(Format('%s.ParseDiagram: Nb = %d; Mini=%.2f', [ClassName, NbPetales, ViseeMini]));
  Result := false;
  try
    Interval := PI / NbPetales;
    // dimensionner table des classes
    SetLength(FClassRepartRoseDiagram, 0);
    SetLength(FClassRepartRoseDiagram, NbPetales);
    // initialiser cette table
    for i := 0 to NbPetales - 1 do FClassRepartRoseDiagram[i] := 0.00;
    // boucle de calcul
    LPT := 0.00;
    for i := LOW_INDEX to GetNbEntitesVisees() - 1 do
    begin
      MyEntite := GetEntiteVisee(i);
      // calcul longueur projetee
      with MyEntite do
      begin
        if (not Drawn) then Continue;
        if (Type_Entite = tgENTRANCE) then Continue;
        //if (Type_Entite = tgTYPE_ENTITES_VISEES_ANTENNE) then Continue;
        // /!\ Utiliser IsViseetInNaturalCave et non IsViseetInCaveOrTunnel
        if (not IsViseetInNaturalCave(MyEntite)) then Continue;
        dx := Une_Station_2_X - Une_Station_1_X;
        dy := Une_Station_2_Y - Une_Station_1_Y;
        dz := Une_Station_2_Z - Une_Station_1_Z;
        LP := Sqrt(Sqr(dx) + Sqr(dy));
      end;
      // test: passer à entitée suivante si LP<ViseeMini
      if (LP < ViseeMini) then Continue;
      // cumul des longueurs projetées
      LPT := LPT + LP;
      // calcul de l'angle
      // pour le calcul d'une rosette symétrique
      // l'angle retourné est dans la plage [0; PI]
      //Ang:=(0.50*PI)-ArcSin(dy/LP);
      Ang := ArcTan2(dx, dy + 1e-05);
      if (Ang < 0) then Ang := Ang + 2 * PI;
      // répartition
      for j := 0 to NbPetales - 1 do
      begin
        if ((Ang >= j * Interval) and (Ang < (j + 1) * Interval)) then
        begin
          FClassRepartRoseDiagram[j] := FClassRepartRoseDiagram[j] + LP;
        end;
        if ((Ang >= (PI + j * Interval)) and (Ang < (PI + (j + 1) * Interval))) then
        begin
          FClassRepartRoseDiagram[j] := FClassRepartRoseDiagram[j] + LP;
        end;
      end;
    end; // for i ..
    // Longueur totale nulle -> sortie de la fonction
    if (LPT = 0) then Exit;
     // mise en pourcentage
    ClasseMax := 0;
    for j := 0 to NbPetales - 1 do
      if (FClassRepartRoseDiagram[j] > ClasseMax) then ClasseMax := FClassRepartRoseDiagram[j];
    for j := 0 to NbPetales - 1 do FClassRepartRoseDiagram[j] := FClassRepartRoseDiagram[j] / ClasseMax;
    Result := True;
  except
  end;// with self ...
end;




procedure TBDDEntites.ExportVue2DSVG(const Filename: TStringDirectoryFilename;
                                     const DoXHTML: boolean;
                                     const FM: TModeRepresentationGaleries;
                                     const FE: TElementsDrawn;
                                     const FillOpacity: byte);
const
  NOM_STYLE_ENTRANCE            = 'entrance';
  NOM_STYLE_TEXTE_ENTRANCE      = 'entrance-caption';
  NOM_STYLE_CENTERLINE          = 'centerline';
  NOM_STYLE_VISEE_RAYONNANTE    = 'antenna';
  NOM_STYLE_CROSS_SECTION       = 'cross-section';
  NOM_STYLE_PAROIS_DEFAUT       = 'parois-default';

  NOM_STYLE_SILHOUETTE_DEFAUT   = 'SilhouetteDefaut';
  NOM_STYLE_CADRE               = 'CadreDessin';
  FMT_STYLE_RESEAU  = 'Reseau%d';
  FMT_STYLE_SECTEUR = 'Secteur%d';
  FMT_STYLE_SEANCE  = 'Seance%d';
var
  SVGCanvas: TSVGCanvas;
  miou: String;
  chat: String;
  Idx666: Integer;
  MyReseau: TReseau;
  QQQQ: String;
  MySecteur: TSecteur;
  MyExpe: TExpe;

  procedure ListerStylesReseaux;
  var
    i, Nb: Integer;
    EWE  : TReseau;
    NS   : String;
    WU: TColor;
  begin
    Nb := GetNbReseaux;
    for i := 0 to Nb - 1 do
    begin
      EWE := GetReseau(i);
      NS  := Format(FMT_STYLE_RESEAU, [i]);
      WU  := GetCouleurRGBByIdxReseau(i);
      SVGCanvas.WriteStyleSolidPolygone(NS, WU, WU, FillOpacity, 0.0, EWE.NomReseau);
    end;
  end;
  procedure ListerStylesSecteurs;
  var
    i, Nb: Integer;
    EWE  : TSecteur;
    NS   : String;
    WU   : TColor;
  begin
    Nb := GetNbSecteurs;
    for i := 0 to Nb - 1 do
    begin
      EWE := GetSecteur(i);
      NS  := Format(FMT_STYLE_SECTEUR, [i]);
      WU  := GetCouleurRGBByIdxSecteur(i);
      SVGCanvas.WriteStyleSolidPolygone(NS, WU, WU, FillOpacity, 0.0, EWE.NomSecteur);
    end;
  end;
  procedure ListerStylesSeances;
  var
    i, Nb: Integer;
    EWE  : TExpe;
    NS   : String;
    WU: TColor;
  begin
    Nb := GetNbExpes;
    for i := 0 to Nb - 1 do
    begin
      EWE := GetExpe(i);
      NS  := Format(FMT_STYLE_SEANCE, [EWE.IDExpe]);
      WU :=  GetCouleurRGBExpeByIDX(EWE.Couleur);
      SVGCanvas.WriteStyleSolidPolygone(NS, WU, WU, FillOpacity, 0.0, EWE.Commentaire);
    end;
  end;
  procedure DessinerCenterlines(const QMode: TModeTOptionsCenterLine);
  var
    i, Nb: Integer;
    V: TEntiteEtendue;
  begin
    AfficherMessage('--> Dessin centerlines');
    Nb := GetNbEntitesVisees();
    for i := LOW_INDEX to Nb - 1 do
    begin
      V := GetEntiteVisee(i);
      if (V.Type_Entite = tgENTRANCE) then Continue;
      if (V.Type_Entite = tgVISEE_RADIANTE) then Continue;
      if (mclCENTERLINE     in QMode) then SVGCanvas.DrawLine(NOM_STYLE_CENTERLINE   , V.Une_Station_1_X, V.Une_Station_1_Y, V.Une_Station_2_X, V.Une_Station_2_Y);
      if (mclCROSS_SECTIONS in QMode) then SVGCanvas.DrawLine(NOM_STYLE_CROSS_SECTION, V.X2PG, V.Y2PG, V.X2PD, V.Y2PD);
    end;
  end;
  procedure DessinerAntennes();
  var
    i, Nb: Integer;
    V: TEntiteEtendue;
  begin
    AfficherMessage('--> Dessin antennes');
    Nb := GetNbEntitesVisees();
    for i := LOW_INDEX to Nb - 1 do
    begin
      V := GetEntiteVisee(i);
      if (V.Type_Entite = tgVISEE_RADIANTE) then
      begin
        SVGCanvas.DrawLine(NOM_STYLE_VISEE_RAYONNANTE   , V.Une_Station_1_X, V.Une_Station_1_Y, V.Une_Station_2_X, V.Une_Station_2_Y);
      end;
    end;
  end;

  procedure DessinerSilhouette(const AvecParois: boolean; const QIdxReseau, QIdxsecteur, QIdxExpe: integer);
  var
    Quad: TArrayPoints2Df;
    i, Nb: Integer;
    V: TEntiteEtendue;
    QQC: TColor;
    QQR: TReseau;
    QQS: TSecteur;
    QQE: TExpe;
    QStyle: String;
  begin
    AfficherMessage(Format('--> Dessin Silhouette: %d - %d - %d' , [QIdxReseau, QIdxsecteur, QIdxExpe]));
    Nb := GetNbEntitesVisees();
    for i := 2 to Nb - 1 do
    begin
      V := GetEntiteVisee(i);
      if (V.Type_Entite =tgENTRANCE) then Continue;
      case FM of
        rgRESEAUX:
          begin
            QQR := GetReseau(V.eReseau);
            if (V.eReseau <> QIdxReseau) then Continue;
            QStyle := Format(FMT_STYLE_RESEAU, [V.eReseau]);
          end;
        rgSECTEURS:
          begin
            QQS := GetSecteur(V.eSecteur);
            if (V.eSecteur <> QIdxsecteur) then Continue;
            QStyle := Format(FMT_STYLE_SECTEUR, [V.eSecteur]);
          end;
        rgSEANCES:
          begin
            QQE := GetExpeByIndex(V.eExpe);
            if (QQE.IDExpe <> QIdxExpe) then Continue;
            QStyle := Format(FMT_STYLE_SEANCE, [QQE.IDExpe]);
          end;
        rgGRAY:
          begin
            QStyle := NOM_STYLE_SILHOUETTE_DEFAUT;
          end;
        rgDEPTH:
          begin
            QQC := GetColorViseeFromModeRepresentation(FM, V);
          end;
      end;
      SetLength(Quad, 4);
      Quad[0] := MakeTPoint2Df(V.X1PD, V.Y1PD);
      Quad[1] := MakeTPoint2Df(V.X2PD, V.Y2PD);
      Quad[2] := MakeTPoint2Df(V.X2PG, V.Y2PG);
      Quad[3] := MakeTPoint2Df(V.X1PG, V.Y1PG);
      if (FM = rgDEPTH) then SVGCanvas.DrawPolygonSansStyle(QQC, QQC, 0.00, Quad, True)
                        else SVGCanvas.DrawPolygon(QStyle, Quad);
      if (AvecParois) then
      begin
        //SVGCanvas.DrawPolygonSansStyle(NOM_STYLE_CENTERLINE   , V.Une_Station_1_X, V.Une_Station_1_Y, V.Une_Station_2_X, V.Une_Station_2_Y);
      end;
    end;
  end;
  procedure DessinerEntrances(const B: boolean);
  var
    NbE, i: Integer;
    MyEntrance: TEntrance;
  begin
    NbE := GetNbEntrees();
    if (NbE = 0) then Exit;
    for i := 0 to NbE - 1 do
    begin
      MyEntrance := GetEntree(i);
      SVGCanvas.DrawCircle(NOM_STYLE_ENTRANCE, MyEntrance.eXEntree, MyEntrance.eYEntree, 5.00);
      SVGCanvas.DrawText(NOM_STYLE_TEXTE_ENTRANCE, 1, MyEntrance.eXEntree, MyEntrance.eYEntree, 0.00, MyEntrance.eNomEntree);
    end;
  end;

begin
  miou := ChooseString(Ord(FM), ['rgRESEAUX', 'rgSECTEURS', 'rgSEANCES', 'rgGRAY', 'rgDEPTH']);
  chat := '';
  if (edPolygonals     in FE) then chat += 'edPolygonals, ';
  if (edStations       in FE) then chat += 'edStations, ';
  if (edIDStations     in FE) then chat += 'edIDStations, ';
  if (edAltitudes      in FE) then chat += 'edAltitudes, ';
  if (edCotes          in FE) then chat += 'edCotation, ';
  if (edWalls          in FE) then chat += 'edWalls, ';
  if (edCrossSections  in FE) then chat += 'edCrossSections, ';
  if (edFillGalerie    in FE) then chat += 'edFillGalerie, ';
  if (edQuadrilles     in FE) then chat += 'edQuadrilles, ';
  if (edENTRANCES      in FE) then chat += 'edENTRANCES, ';
  if (edANNOTATIONS    in FE) then chat += 'edANNOTATIONS, ';
  if (edANTENNES       in FE) then chat += 'edANTENNES, ';
  if (edJONCTIONS      in FE) then chat += 'edJONCTIONS, ';



  AfficherMessage(Format('%s.ExportVue2DSVG(%s): %s', [ClassName, Filename, BoolToStr(DoXHTML, 'XHTML', 'SVG')]));
  AfficherMessage(Format('-- Mode représentation: %s', [miou]));
  AfficherMessage(Format('-- Elements dessinés  : %s', [chat]));
  SVGCanvas := TSVGCanvas.Create;
  try
    // en-tetes
    SVGCanvas.Commentaire := '';
    SVGCanvas.Scale       := 1.00;
    // maxi et mini
    //SVGCanvas.SetDrawingBounds(FCoinBasGauche.X, FCoinBasGauche.Y, FCoinHautDroit.X, FCoinHautDroit.Y);
    if (Not SVGCanvas.InitializeDocument(Filename,
                                         DoXHTML,
                                         'Vue 3D by ' + ApplicationName,
                                         Format('%s: Vue en plan', [ExtractFileName(FileName)]),
                                         FCoinBasGauche.X, FCoinBasGauche.Y, FCoinHautDroit.X, FCoinHautDroit.Y)) then Exit;
    // patterns
    SVGCanvas.BeginPatternsSection;
      // les patterns ici
    SVGCanvas.EndPatternsSection;
    // section de styles
    SVGCanvas.BeginStylesSection;
      // styles de cube englobant
      SVGCanvas.WriteStyleSolidLine(NOM_STYLE_CADRE, clred, 0.1, 'bounding_box_cube');
      // entrées
      SVGCanvas.WriteStyleSolidPolygone(NOM_STYLE_ENTRANCE, clBlack, clBlue, 128, 0.00, 'Entrances');
      SVGCanvas.WriteStyleText(NOM_STYLE_TEXTE_ENTRANCE, 'Arial', [], clBlue, 6.0, 'Entrances labels');





      // centerlines, station et sections
      SVGCanvas.WriteStyleSolidLine(NOM_STYLE_CENTERLINE    , clBlack, 0.1, 'centerlines');
      SVGCanvas.WriteStyleSolidLine(NOM_STYLE_CROSS_SECTION , clGray , 0.05, 'cross_sections');
      SVGCanvas.WriteStyleSolidLine(NOM_STYLE_PAROIS_DEFAUT , clBlack, 0.4, 'default walls');
      // visées en antenne
      SVGCanvas.WriteStyleSolidLine(NOM_STYLE_VISEE_RAYONNANTE, clGray, 0.01, 'antenna_shots');
      // style par défaut: gris
      SVGCanvas.WriteStyleSolidPolygone(NOM_STYLE_SILHOUETTE_DEFAUT, clGray, clGray, FillOpacity, 0.00, 'Default Silhouettes');

      // les styles de galeries ici
      case FM of
        rgRESEAUX  : ListerStylesReseaux;
        rgSECTEURS : ListerStylesSecteurs;
        rgSEANCES  : ListerStylesSeances;
      end;
    SVGCanvas.EndStylesSection;
    // dessin
    SVGCanvas.BeginDrawingSection();
      if (edPolygonals     in FE) then DessinerCenterLines([mclCENTERLINE]);
      if (edCrossSections  in FE) then DessinerCenterlines([mclCROSS_SECTIONS]);
      if (edANTENNES       in FE) then DessinerAntennes();
      case FM of
        rgRESEAUX:
          begin
            for Idx666 := 0 to GetNbReseaux - 1 do
            begin
              MyReseau := GetReseau(Idx666);
              QQQQ     := Format('Reseau%d', [Idx666]);
              SVGCanvas.BeginGroupe(QQQQ, MyReseau.NomReseau, 0.0, 0.0);
                if (edFillGalerie    in FE) then DessinerSilhouette(False, Idx666, -1, -1);
              SVGCanvas.EndGroupe(QQQQ);
            end;
          end;
        rgSECTEURS:
          begin
            for Idx666 := 0 to GetNbSecteurs - 1 do
            begin
              MySecteur := GetSecteur(Idx666);
              QQQQ     := Format('Secteur%d', [Idx666]);
              SVGCanvas.BeginGroupe(QQQQ, MySecteur.NomSecteur, 0.0, 0.0);
                if (edFillGalerie    in FE) then DessinerSilhouette(False, -1, Idx666, -1);
              SVGCanvas.EndGroupe(QQQQ);
            end;
          end;
        rgSEANCES:
          begin
            for Idx666 := 0 to GetNbExpes - 1 do
            begin
              MyExpe   := GetExpe(Idx666);
              QQQQ     := Format('Expe%d', [MyExpe.IDExpe]);
              SVGCanvas.BeginGroupe(QQQQ, MyExpe.Commentaire, 0.0, 0.0);
                if (edFillGalerie    in FE) then DessinerSilhouette(False, -1, -1, MyExpe.IDExpe);
              SVGCanvas.EndGroupe(QQQQ);
            end;
          end;
        rgDEPTH:
          begin
            DessinerSilhouette(False, -1, -1, -1);
          end;
      end;
      // entrées
      if (edENTRANCES in FE) then DessinerEntrances(True);
    SVGCanvas.EndDrawingSection();
    // cloture
    SVGCanvas.FinalizeDocument;
  finally
    SVGCanvas.Free;
  end;
end;

procedure TBDDEntites.SetPositionPointZero(const QPositionPointZero: TPoint3Df);

begin
  FPositionDuPointZero := QPositionPointZero;

  AfficherMessage(Format('%s.SetAltitudePointZero(%.2f, %.2f, %.2f)', [ClassName, FPositionDuPointZero.X, FPositionDuPointZero.Y, FPositionDuPointZero.Z]));
end;

function TBDDEntites.GetPositionPointZero: TPoint3Df;
begin
  Result := FPositionDuPointZero;
end;

function TBDDEntites.GetDeltaXYZFromPositionSt0(const ST: TEntiteEtendue): TPoint3Df;
begin
  Result.X := ST.Une_Station_2_X - FPositionDuPointZero.X;
  Result.Y := ST.Une_Station_2_Y - FPositionDuPointZero.Y;
  Result.Z := ST.Une_Station_2_Z - FPositionDuPointZero.Z;
end;

//==============================================================================
// exporter les antennes comme nuages de points
procedure TBDDEntites.ExporterAntennesNuagePoints(const QFilename: TStringDirectoryFilename);
var
  i, Nb: Integer;
  E: TEntiteEtendue;
  fp : TextFile;
begin
  AfficherMessage(Format('%s.ExporterAntennesNuagePoints: %s', [ClassName, QFilename]));
  Nb := self.GetNbEntitesVisees();
  AssignFile(fp, QFilename);
  try
    Rewrite(fp);
    for i := 1 to Nb -1 do
    begin
      E := self.GetEntiteVisee(i);
      if (E.Type_Entite = tgVISEE_RADIANTE) then
      begin
        WriteLn(fp, Format('%.3f, %.3f, %.3f',
                           [E.Une_Station_2_X,
                            E.Une_Station_2_Y,
                            E.Une_Station_2_Z
                           ]));
      end;
    end;
  finally
    CloseFile(fp);
  end;
end;
//******************************************************************************
//******************************************************************************
// sauvegarde dans un fichier de texte
// Statut: Opérationnel - A valider définitivement (notamment les filtres)
function TBDDEntites.SaveToFile(const Filename: TStringDirectoryFilename; const Filtres: string): boolean; deprecated;
var
  fp: TextFile;
  i, N: integer;
  QDevelViseesVisibles: double;
  procedure WrtLn(const L: string); inline;
  begin
    WriteLn(fp, L);
  end;
  procedure WrtLnNbElts(const QN: integer; const CS: string);
  begin
    WriteLn(fp, Format('%d', [QN]) + #9 + CS);
  end;
  procedure WrtEntree(const QE: TEntrance; const QIdx: integer);
  var
    WU: String;
  begin
    WU := Format('%d', [QIdx]) + #9 +
          QE.eNomEntree + #9 +
          Format('%d', [QE.eRefSer]) + #9 +
          Format('%d', [QE.eRefSt]) + #9 +
          Format('%.2f', [QE.eXEntree]) + #9 +
          Format('%.2f', [QE.eYEntree]) + #9 +
          Format('%.2f', [QE.eZEntree]) + #9 +
          QE.eObserv;
    WriteLn(fp, WU);
  end;
  procedure WrtReseau(const QE: TReseau; const QIdx: integer);
  var
    WU: String;
  begin
    WU := Format('%d', [QIdx]) + #9 +
          Format('%d', [QE.TypeReseau]) + #9 +
          Format('%d', [QE.ColorReseau]) + #9 +
          QE.NomReseau + #9 +
          QE.ObsReseau;
    WriteLn(fp, WU);
  end;
  procedure WrtSecteur(const QE: TSecteur; const QIdx: integer);
  var
    WU: String;
  begin
    WU := Format('%d', [QIdx]) + #9 +
          Format('%d', [QE.CouleurSecteur]) + #9 +
          QE.NomSecteur;
    WriteLn(fp, WU);
  end;
  procedure WrtCode(const QE: TCode);
  var
    WU: String;
  begin
    WU := Format('%d', [QE.IDCode]) + #9 +
          Format('%.2f', [QE.GradAz]) + #9 +
          Format('%.2f', [QE.GradInc]) + #9 +
          Format('%.3f', [QE.PsiL]) + #9 +
          Format('%.3f', [QE.PsiAz]) + #9 +
          Format('%.3f', [QE.PsiP]) + #9 +
          Format('%.4f', [QE.FactLong]) + #9 +
          Format('%.4f', [1.00]) + #9 +
          Format('%.2f', [QE.AngLimite]) + #9 +
          Format('%.d',  [0]) + #9 +                      //Format('%.d',  [QE.ReservedInt]) + #9 +
          QE.Commentaire;
          // ne pas écrire les paramètres des fonctions de correction, inutiles ici
    WriteLn(fp, WU);
  end;
  procedure WrtExpe(const QE: TExpe);
  var
    WU: String;
  begin
    WU := Format('%d', [QE.IDExpe]) + #9 +
          Format('%d', [QE.JourExpe]) + #9 +
          Format('%d', [QE.MoisExpe]) + #9 +
          Format('%d', [QE.AnneeExpe]) + #9 +
          Format('%d', [QE.ModeDecl]) + #9 +
          Format('%.2f', [QE.Declinaison]) + #9 +
          Format('%.2f', [0.00]) + #9 + //  QE.Inclinaison
          Format('%d', [QE.Couleur]) + #9 +
          QE.Operateur + #9 +
          QE.ClubSpeleo + #9 +
          QE.Commentaire;
    WriteLn(fp, WU);
  end;
  procedure WrtEntite(const QE: TEntiteEtendue);
  var
    AAAA, MM, JJ: word;
    WU: String;
  begin
    if (Not QE.Drawn) then Exit;
    DecodeDate(QE.DateLeve, AAAA, MM, JJ);
    WU := // serie et point
          Format('%d', [QE.Entite_Serie]) + #9 +
          Format('%d', [QE.Entite_Station]) + #9 +
          // type de visée
          Format('%d', [Ord(QE.Type_Entite)]) + #9 +
          // réseau et secteur
          Format('%d', [QE.eReseau]) + #9 +
          Format('%d', [QE.eSecteur]) + #9 +
          // codes, expés
          Format('%d', [QE.eCode]) + #9 +
          Format('%d', [QE.eExpe]) + #9 +
          // date du levé
          Format('%d', [AAAA]) + #9 +
          Format('%d', [MM]) + #9 +
          Format('%d', [JJ]) + #9 +
          // dessiné
          Format('%d', [IIF(QE.Drawn, 1, 0)]) + #9 +
          // valeurs initiales
          Format('%.2f', [QE.oLongueur]) + #9 +
          Format('%.2f', [QE.oAzimut]) + #9 +
          Format('%.2f', [QE.oPente]) + #9 +
          Format('%.2f', [QE.oLG]) + #9 +
          Format('%.2f', [QE.oLD]) + #9 +
          Format('%.2f', [QE.oHZ]) + #9 +
          Format('%.2f', [QE.oHN]) + #9 +
          // valeurs calculées: centerline
          Format('%.2f', [QE.Une_Station_1_X]) + #9 +
          Format('%.2f', [QE.Une_Station_1_Y]) + #9 +
          Format('%.2f', [QE.Une_Station_1_Z]) + #9 +
          Format('%.2f', [QE.Une_Station_2_X]) + #9 +
          Format('%.2f', [QE.Une_Station_2_Y]) + #9 +
          Format('%.2f', [QE.Une_Station_2_Z]) + #9 +
          // valeurs calculées: silhouette
          Format('%.2f', [QE.X1PD]) + #9 +
          Format('%.2f', [QE.Y1PD]) + #9 +
          Format('%.2f', [QE.X1PG]) + #9 +
          Format('%.2f', [QE.Y1PG]) + #9 +

          Format('%.2f', [QE.X2PD]) + #9 +
          Format('%.2f', [QE.Y2PD]) + #9 +
          Format('%.2f', [QE.X2PG]) + #9 +
          Format('%.2f', [QE.Y2PG]) + #9 +

          Format('%.2f', [QE.Z1PH]) + #9 +
          Format('%.2f', [QE.Z1PB]) + #9 +
          Format('%.2f', [QE.Z2PH]) + #9 +
          Format('%.2f', [QE.Z2PB]) + #9 +
          QE.oIDLitteral + #9 +
          QE.oCommentaires;
    WriteLn(fp, WU);
  end;
begin
  Result := False;
  AfficherMessage(Format('%s.SaveToFile(%s)', [ClassName, Filename]));
  // filtres
  N := MetaFiltre(Filtres, QDevelViseesVisibles);
  AfficherMessage(Format('-- Filtres: %s - %d entites retenues', [Filtres, N]));
  AssignFile(fp, Filename);
  try
    ReWrite(fp);
    // entête
    WrtLn(Format('%s created %s by GHTopo', [Filename, DateTimeToStr(Now)]));
    WrtLn('*** DO NOT MODIFY ***');
    WrtLn(Format('Filters: %s', [Filtres]));
    N := GetNbEntrees();
    WrtLnNbElts(N, 'entrances');
    for i := 0 to N - 1 do WrtEntree(GetEntree(i), i);
    N := GetNbReseaux();
    WrtLnNbElts(N, 'reseaux');
    for i := 0 to N - 1 do WrtReseau(GetReseau(i), i);
    N := GetNbSecteurs();
    WrtLnNbElts(N, 'secteurs');
    for i := 0 to N - 1 do WrtSecteur(GetSecteur(i), i);
    N := GetNbCodes();
    WrtLnNbElts(N, 'codes');
    for i := 0 to N - 1 do WrtCode(GetCode(i));
    N := GetNbExpes();
    WrtLnNbElts(N, 'expes');
    for i := 0 to N - 1 do WrtExpe(GetExpe(i));
    N := GetNbEntitesVisees();
    WrtLnNbElts(N, 'shots');
    for i := LOW_INDEX to N - 1 do WrtEntite(GetEntiteVisee(i));
  finally
    CloseFile(fp);
  end;
end;
// Statut: Opérationnel - A valider définitivement
function TBDDEntites.LoadFromFile(const FileName: TStringDirectoryFilename): boolean; deprecated;
var
  fp: TextFile;
  NbLignesLues: Integer;
  MyLigne: string;
  i, Nb: Integer;
  QEntree: TEntrance;
  QReseau: TReseau;
  QSecteur: TSecteur;
  QCode: TCode;
  QExpe: TExpe;
  QEntite: TEntiteEtendue;
  function LireLigne: string;
  begin
    ReadLn(fp, Result);
    Inc(NbLignesLues);
  end;
  function LitEtDecomposeLigne(): TGHStringArray;
  begin
    Result := Split(LireLigne, #9);
  end;
  function GetNbElementsDansSection: integer;
  var
    EWE: TGHStringArray;
  begin
    EWE     := LitEtDecomposeLigne;
    Result  := StrToIntDef(EWE[0], -1);
  end;

  procedure LitEtAjouteEntree;
  var
    QT666: TGHStringArray;
    E    : TEntrance;
  begin
    try
      QT666 := LitEtDecomposeLigne;
      //E.eNumEntree := StrToIntDef(QT666[0], -1);
      E.eNomEntree := Trim(QT666[1]);
      E.eRefSer    := StrToIntDef(QT666[2], -1);
      E.eRefSt     := StrToIntDef(QT666[3], -1);
      E.eXEntree   := ConvertirEnNombreReel(QT666[4], -1.00);
      E.eYEntree   := ConvertirEnNombreReel(QT666[5], -1.00);
      E.eZEntree   := ConvertirEnNombreReel(QT666[6], -1.00);
      E.eObserv    := Trim(QT666[7]);
      //if (E.eNumEntree >= 0) then AddEntree(E);
    except
    end;
  end;
  procedure LitEtAjouteReseau;
  var
    QT666: TGHStringArray;
    E    : TReseau;
  begin
    try
      QT666 := LitEtDecomposeLigne();
      E.TypeReseau  := StrToIntDef(QT666[1], -1);
      E.ColorReseau := TColor(StrToIntDef(QT666[2], 0));
      E.NomReseau   := Trim(QT666[3]);
      E.ObsReseau   := Trim(QT666[4]);
      AddReseau(E);
    except
    end;
  end;
  procedure LitEtAjouteSecteur;
  var
    QT666: TGHStringArray;
    E    : TSecteur;
  begin
    try
      QT666 := LitEtDecomposeLigne();
      E.CouleurSecteur := TColor(StrToIntDef(QT666[1], 0));
      E.NomSecteur     := Trim(QT666[2]);
    except
    end;
  end;
  procedure LitEtAjouteCode;
  var
    QT666: TGHStringArray;
    E    : TCode;
  begin
    try
      QT666 := LitEtDecomposeLigne;
      E.IDCode         := StrToIntDef(QT666[0], -1);
      E.GradAz         := ConvertirEnNombreReel(QT666[1], 360);
      E.GradInc        := ConvertirEnNombreReel(QT666[2], 360);
      E.PsiL           := ConvertirEnNombreReel(QT666[3], 1.00);
      E.PsiAz          := ConvertirEnNombreReel(QT666[4], 1.00);
      E.PsiAz          := ConvertirEnNombreReel(QT666[5], 1.00);
      // TODO: Support de FactLong a valider
      E.FactLong       := 1.00; // ConvertirEnNombreReel(QT666[6], 1.00);
      E.AngLimite      := ConvertirEnNombreReel(QT666[7], 1.00);
      //E.ReservedInt    := StrToIntDef(QT666[8], -1);
      E.Commentaire    := Trim(QT666[9]);
      // ne pas laisser de variables non initialisées
      E.ParamsFuncCorrAz  := MakeTParamFoncCorrectionAngulaire(0.00, 0.00, 0.00);
      E.ParamsFuncCorrInc := MakeTParamFoncCorrectionAngulaire(0.00, 0.00, 0.00);
      if (E.IDCode >= 0) then AddCode(E);
    except
    end;
  end;
  procedure LitEtAjouteExpe;
  var
    QT666: TGHStringArray;
    E    : TExpe;
  begin
    try
      QT666            := LitEtDecomposeLigne;
      E.IDExpe         := StrToIntDef(QT666[0], -1);
      E.JourExpe       := StrToIntDef(QT666[1], 1);
      E.MoisExpe       := StrToIntDef(QT666[2], 1);
      E.AnneeExpe      := GetAnneeISO(StrToIntDef(QT666[3], 0));
      E.ModeDecl       := StrToIntDef(QT666[4], 0);
      E.Declinaison    := ConvertirEnNombreReel(QT666[5], 1.00);
      //E.Inclinaison    := ConvertirEnNombreReel(QT666[6], 1.00);
      E.Couleur        := StrToIntDef(QT666[7], 0);
      E.Operateur    := Trim(QT666[8]);
      E.ClubSpeleo   := Trim(QT666[9]);
      E.Commentaire    := Trim(QT666[10]);
      if (E.IDExpe >= 0) then AddExpe(E);
    except
    end;
  end;
  procedure LitEtAjouteEntite;
  var
    QT666: TGHStringArray;
    E    : TEntiteEtendue;
    AAAA, MM, JJ: word;
    QAT: Integer;
    QIDX: Integer;
  begin
    try
      QT666            := LitEtDecomposeLigne;
      // série et station
      QIDX := 0;
      E.Entite_Serie         := StrToIntDef(QT666[QIDX + 0], -1);   //0
      E.Entite_Station       := StrToIntDef(QT666[QIDX + 1], 0);    //1
      // type de visée
      QIDX := 2;
      E.Type_Entite          := TTypeDeVisee(StrToIntDef(QT666[QIDX], 0)); //2
      // réseau et secteur // codes, expés
      QIDX := 3;
      E.eReseau              := StrToIntDef(QT666[QIDX + 0], 0);   //3
      E.eEntrance            := 0;
      E.eSecteur             := StrToIntDef(QT666[QIDX + 1], 0);   //4
      E.eCode                := StrToIntDef(QT666[QIDX + 2], 0);   //5
      E.eExpe                := StrToIntDef(QT666[QIDX + 3], 0);   //6
      // date du levé
      QIDX := 7;
      AAAA := GetAnneeISO(StrToIntDef(QT666[QIDX + 0], 1));        //7
      MM   := StrToIntDef(QT666[QIDX + 1], 1);                     //8
      JJ   := StrToIntDef(QT666[QIDX + 2], 1899);                  //9
      E.DateLeve := GetSecuredDate(AAAA, MM, JJ);//EncodeDate(AAAA, MM, JJ);
      // dessiné
      QIDX := 10;
      QAT      := StrToIntDef(QT666[QIDX], 1);                     //10
      E.Drawn  := (QAT = 1);
      // valeurs initiales
      QIDX := 11;
      E.oLongueur   := ConvertirEnNombreReel(QT666[QIDX + 0], 0.00);       //11
      E.oAzimut     := ConvertirEnNombreReel(QT666[QIDX + 1], 0.00);       //12
      E.oPente      := ConvertirEnNombreReel(QT666[QIDX + 2], 0.00);       //13
      E.oLG         := ConvertirEnNombreReel(QT666[QIDX + 3], 0.00);       //14
      E.oLD         := ConvertirEnNombreReel(QT666[QIDX + 4], 0.00);       //15
      E.oHZ         := ConvertirEnNombreReel(QT666[QIDX + 5], 0.00);       //16
      E.oHN         := ConvertirEnNombreReel(QT666[QIDX + 6], 0.00);       //17
      QIDX := 18;
      // valeurs calculées: centerline
      E.Une_Station_1_X := ConvertirEnNombreReel(QT666[QIDX + 0], 0.00);   //18
      E.Une_Station_1_Y:= ConvertirEnNombreReel(QT666[QIDX + 1], 0.00);    //19
      E.Une_Station_1_Z:= ConvertirEnNombreReel(QT666[QIDX + 2], 0.00);    //20
      E.Une_Station_2_X:= ConvertirEnNombreReel(QT666[QIDX + 3], 0.00);    //21
      E.Une_Station_2_Y:= ConvertirEnNombreReel(QT666[QIDX + 4], 0.00);    //22
      E.Une_Station_2_Z:= ConvertirEnNombreReel(QT666[QIDX + 5], 0.00);    //23
      // valeurs calculées: silhouette
      QIDX := 24;
      E.X1PD := ConvertirEnNombreReel(QT666[QIDX + 0], 0.00);  //24
      E.Y1PD := ConvertirEnNombreReel(QT666[QIDX + 1], 0.00);  //25
      E.X1PG := ConvertirEnNombreReel(QT666[QIDX + 2], 0.00);  //26
      E.Y1PG := ConvertirEnNombreReel(QT666[QIDX + 3], 0.00);  //27

      E.X2PD := ConvertirEnNombreReel(QT666[QIDX + 4], 0.00); //28
      E.Y2PD := ConvertirEnNombreReel(QT666[QIDX + 5], 0.00); //29
      E.X2PG := ConvertirEnNombreReel(QT666[QIDX + 6], 0.00); //30
      E.Y2PG := ConvertirEnNombreReel(QT666[QIDX + 7], 0.00); //31

      E.Z1PH := ConvertirEnNombreReel(QT666[QIDX + 8], 0.00); //32
      E.Z1PB := ConvertirEnNombreReel(QT666[QIDX + 9], 0.00); //33
      E.Z2PH := ConvertirEnNombreReel(QT666[QIDX + 10], 0.00); //34
      E.Z2PB := ConvertirEnNombreReel(QT666[QIDX + 11], 0.00); //35
      // littéral et commentaires
      QIDX   := 36;
      E.oIDLitteral   := Trim(QT666[QIDX + 0]);
      E.oCommentaires := Trim(QT666[QIDX + 1]);
      // code d'erreur égal à 0 ici:
      // Supérieur à 0 => appartient à une série ordinaire
      // Inférieur à 0 => est une visée en antenne
      if (E.Entite_Serie <> 0) then AddEntiteVisee(E);
    except
    end;
  end;

begin
  Result := False;
  AfficherMessage(Format('%s.LoadFromFile(%s)', [ClassName, Filename]));
  // vider les tables
  ViderTablesSimples();
  ViderLesViseesEtAntennes();
  AfficherMessage('-- Purge des tables OK');
  // ouverture du fichier
  NbLignesLues := 0;
  AssignFile(fp, FileName);
  try
    Reset(fp);
    // zapping des premières lignes
    MyLigne := LireLigne; // C:\Documents and Settings\jean-pierre.cassou\Bureau\toto666.pot created 07/11/2013 10:31:06 by GHTopo
    MyLigne := LireLigne; // *** DO NOT MODIFY ***
    MyLigne := LireLigne; // Filters: reseau=2
    // Lecture des entrées
    Nb      := GetNbElementsDansSection;
    if (Nb > 0) then
    begin
      for i := 0 to Nb - 1 do LitEtAjouteEntree;
      AfficherMessage(Format('-- [%d] Lecture des %d entrees', [NbLignesLues, Nb]));
      for i := 0 to GetNbEntrees - 1 do
      begin
        QEntree := GetEntree(i);
        AfficherMessage(Format('--> %d: %d - %s - %d/%d - %.0f, %.0f, %.0f', [i, i, QEntree.eNomEntree, QEntree.eRefSer, QEntree.eRefSt, QEntree.eXEntree, QEntree.eYEntree, QEntree.eZEntree]));
      end;
    end;
    // Lecture des réseaux
    Nb      := GetNbElementsDansSection;
    if (Nb > 0) then
    begin
      for i := 0 to Nb - 1 do LitEtAjouteReseau;
      AfficherMessage(Format('-- [%d] Lecture des %d reseaux', [NbLignesLues, Nb]));
      for i := 0 to GetNbReseaux - 1 do
      begin
        QReseau := GetReseau(i);
        AfficherMessage(Format('--> %d: %d - %d - %X - %s (%s)', [i, i, QReseau.TypeReseau, QReseau.ColorReseau, QReseau.NomReseau, QReseau.ObsReseau]));
      end;
    end;
    // Lecture des secteurs
    Nb      := GetNbElementsDansSection;
    if (Nb > 0) then
    begin
      for i := 0 to Nb - 1 do LitEtAjouteSecteur;
      AfficherMessage(Format('-- [%d] Lecture des %d secteurs', [NbLignesLues, Nb]));
      for i := 0 to GetNbSecteurs - 1 do
      begin
        QSecteur := GetSecteur(i);
        AfficherMessage(Format('--> %d: %d - %X - %s', [i, i, QSecteur.CouleurSecteur, QSecteur.NomSecteur]));
      end;
    end;
    // Lecture des codes
    Nb      := GetNbElementsDansSection;
    if (Nb > 0) then
    begin
      for i := 0 to Nb - 1 do LitEtAjouteCode;
      AfficherMessage(Format('-- [%d] Lecture des %d codes', [NbLignesLues, Nb]));
      for i := 0 to GetNbCodes - 1 do
      begin
        QCode := GetCode(i);
        AfficherMessage(Format('--> %d: %d - %.0f, %.0f - %s', [i, QCode.IDCode, QCode.GradAz, QCode.GradInc, QCode.Commentaire]));
      end;
    end;
    // Lecture des expés
    Nb      := GetNbElementsDansSection;
    if (Nb > 0) then
    begin
      for i := 0 to Nb - 1 do LitEtAjouteExpe;
      AfficherMessage(Format('-- [%d] Lecture des %d Expes', [NbLignesLues, Nb]));
      for i := 0 to GetNbExpes - 1 do
      begin
        QExpe := GetExpe(i);
        AfficherMessage(Format('--> %d: %d - %.2d/%.2d/%.4d - %s', [i, QExpe.IDExpe, QExpe.JourExpe, QExpe.MoisExpe, QExpe.AnneeExpe, QExpe.Commentaire]));
      end;
    end;
    // Lecture des visées
    Nb      := GetNbElementsDansSection;
    if (Nb > 0) then
    begin
      for i := 0 to Nb - 1 do LitEtAjouteEntite;
      AfficherMessage(Format('-- [%d] Lecture des %d entites', [NbLignesLues, Nb]));
      for i := LOW_INDEX to GetNbEntitesVisees() - 1 do QEntite := GetEntiteVisee(i);
      //  AfficherMessage(Format('--> %d: %d.%d - %.2f, %.2f, %.2f - %s', [i, QEntite.Entite_Serie, QEntite.Entite_Station, QEntite.oLongueur, QEntite.oAzimut, QEntite.oPente, QEntite.oIDLitteral]));
      // préparation des vues
      SetMinMax(False);
      SortByDepth;
    end;
    // OK s'il y a des entités
    Result := (GetNbEntitesVisees() > 0);
  finally
    CloseFile(fp);
  end;
end;

function TBDDEntites.CalcSpeleometrieParReseauxOuSecteurs(const QMode: integer; out TR: TArraySpeleometrieReseauOuSecteur): boolean;
var
  i, Nb: Integer;
  EWE: TSpeleometrieReseauOuSecteur;
begin
  Result := false;
  case QMode of
    0: Nb := self.GetNbReseaux();
    1: Nb := self.GetNbSecteurs();
  end;
  AfficherMessage(Format('%s.CalcSpeleometriePourReseauOuSecteur: (%s) Nb = %d',
                         [ClassName,
                          IIF((QMode = 1), 'RESEAUX', 'SECTEURS'),
                          Nb
                          ]));

  SetLength(TR, 0);
  SetLength(TR, Nb);
  for i := 0 to Nb - 1 do
  begin
    CalcSpeleometrieReseauOuSecteur(QMode, i, EWE);
    TR[i] := EWE;
  end;
end;

function TBDDEntites.GetNbEntitesMetafiltrees: integer;
begin
  Result := FNbreViseesRetenues;
end;

function TBDDEntites.GenererEntiteDepuisUneVisee(const QStation: TToporobotStation;
                                                 const QV: TUneVisee; out EE: TEntiteEtendue): boolean;
var
  CC: TCode;
  EX: TExpe;
  EWE: TEntiteEtendue;
  VQ: TUneVisee;
  dX, dY, dZ, dP: double;
begin
  Result := false;
  try
    VQ := QV;
    // rechercher la station d'accrochage
    AfficherMessageErreur(Format('GenererEntiteDepuisUneVisee(): Depart: %d.%d - C=%d, E=%d - L=%.3f, A=%.3f, P=%.3f',
                                 [QStation.eSerie, QStation.ePoint,
                                  QV.Code, QV.Expe,
                                  QV.Longueur, QV.Azimut, QV.Pente]));
    // extraction du point de départ (coordonnées)
    //if (not
    GetEntiteViseeFromSerSt(QStation.eSerie, QStation.ePoint, EWE); // then Exit;

    AfficherMessageErreur(Format('%d.%d - Type: %d - %.3f, %.3f, %.3f', [EWE.Entite_Serie, EWE.Entite_Station, QV.TypeVisee, EWE.Une_Station_2_X, EWE.Une_Station_2_Y, EWE.Une_Station_2_Z]));
    CC := GetCodeByIndex(QV.Code);
    EX := GetExpeByIndex(QV.Expe);
    // préparation du résultat
    EE.Entite_Serie    := QStation.eSerie;
    EE.Entite_Station  := QStation.ePoint;
    EE.DateLeve        := Now();
    EE.Drawn           := true;
    EE.eCode           := QV.Code;
    EE.eExpe           := QV.Expe;
    EE.oIDLitteral     := QV.IDTerrainStation;
    EE.eSecteur        := QV.IDSecteur;

    // tout ceci sera recalculé à la prochaine compilation
    EE.eReseau         := 0;
    EE.eEntrance       := 0;
    EE.oLongueur       := QV.Longueur;
    EE.oAzimut         := QV.Azimut;
    EE.oPente          := QV.Pente;
    EE.oLD             := QV.LD;
    EE.oLG             := QV.LG;
    EE.oHZ             := QV.HZ;
    EE.oHN             := QV.HN;


    // calcul coordonnées
    CalculerVisee(VQ, CC, EX, dX, dY, 1.0, dZ, dP);

    EE.Une_Station_1_X := EWE.Une_Station_2_X;
    EE.Une_Station_1_Y := EWE.Une_Station_2_Y;
    EE.Une_Station_1_Z := EWE.Une_Station_2_Z;

    EE.Une_Station_2_X := EE.Une_Station_1_X + dX;
    EE.Une_Station_2_Y := EE.Une_Station_1_Y + dY;
    EE.Une_Station_2_Z := EE.Une_Station_1_Z + dZ;
    EE.Type_Entite     := QV.TypeVisee;
    Result := True;
  except

  end;
end;

// retourne le tableau de stats (dev, encombrement) pour un réseau ou un secteur donné
function TBDDEntites.CalcSpeleometrieReseauOuSecteur(const QMode: integer; const IdxReseauOuSecteur: integer; out TR: TSpeleometrieReseauOuSecteur): boolean;
var
  MyReseau: TReseau;
  MySecteur: TSecteur;
  WU: String;
  i, Nb: Integer;
  EE: TEntiteEtendue;
  IsNatural: Boolean;
  EWE: Boolean;
  procedure CheckC1(const QE: TEntiteEtendue);
  begin
    if (QE.Une_Station_1_X < TR.CoordMini.X) then TR.CoordMini.X := QE.Une_Station_1_X;
    if (QE.Une_Station_1_Y < TR.CoordMini.Y) then TR.CoordMini.Y := QE.Une_Station_1_Y;
    if (QE.Une_Station_1_Z < TR.CoordMini.Z) then TR.CoordMini.Z := QE.Une_Station_1_Z;

    if (QE.Une_Station_2_X < TR.CoordMini.X) then TR.CoordMini.X := QE.Une_Station_2_X;
    if (QE.Une_Station_2_Y < TR.CoordMini.Y) then TR.CoordMini.Y := QE.Une_Station_2_Y;
    if (QE.Une_Station_2_Z < TR.CoordMini.Z) then TR.CoordMini.Z := QE.Une_Station_2_Z;

  end;
  procedure CheckC2(const QE: TEntiteEtendue);
  begin
    if (QE.Une_Station_1_X > TR.CoordMaxi.X) then TR.CoordMaxi.X := QE.Une_Station_1_X;
    if (QE.Une_Station_1_Y > TR.CoordMaxi.Y) then TR.CoordMaxi.Y := QE.Une_Station_1_Y;
    if (QE.Une_Station_1_Z > TR.CoordMaxi.Z) then TR.CoordMaxi.Z := QE.Une_Station_1_Z;

    if (QE.Une_Station_2_X > TR.CoordMaxi.X) then TR.CoordMaxi.X := QE.Une_Station_2_X;
    if (QE.Une_Station_2_Y > TR.CoordMaxi.Y) then TR.CoordMaxi.Y := QE.Une_Station_2_Y;
    if (QE.Une_Station_2_Z > TR.CoordMaxi.Z) then TR.CoordMaxi.Z := QE.Une_Station_2_Z;
  end;

begin
  Result := false;
  try
    case QMode of
      0: MyReseau  := self.GetReseau(IdxReseauOuSecteur);
      1: MySecteur := self.GetSecteur(IdxReseauOuSecteur);
    end;
    case QMode of
      0: WU := MyReseau.NomReseau;
      1: WU := MySecteur.NomSecteur;
    end;
    AfficherMessage(Format('%s.CalcSpeleometriePourReseauOuSecteur: (%s) %d - %s',
                           [ClassName,
                           IIF((QMode = 1), 'RESEAUX', 'SECTEURS'),
                           IdxReseauOuSecteur,
                           WU
                           ]));
    Nb := self.GetNbEntitesVisees();
    TR.IDReseauSecteur  := IdxReseauOuSecteur;
    TR.Developpement    := 0.00;
    TR.NomReseauSecteur := WU;
    TR.NbVisees         := 0;
    TR.CoordMini        := MakeTPoint3Df( INFINI,  INFINI,  INFINI);
    TR.CoordMaxi        := MakeTPoint3Df(-INFINI, -INFINI, -INFINI);
    TR.Etendue          := MakeTPoint3Df(0.00, 0.00, 0.00);
    for i := 0 to Nb - 1 do
    begin
      EE := self.GetEntiteVisee(i);
      if (EE.Type_Entite = tgSURFACE)   then continue;
      if (EE.Type_Entite = tgFIXPOINT)  then continue;
      if (EE.Type_Entite = tgENTRANCE)  then continue;
      if (EE.Type_Entite = tgVISEE_RADIANTE) then continue;

      IsNatural := (EE.Type_Entite = tgDEFAULT) OR
                   (EE.Type_Entite = tgFOSSILE) OR
                   (EE.Type_Entite = tgVADOSE) OR
                   (EE.Type_Entite = tgENNOYABLE) OR
                   (EE.Type_Entite = tgSIPHON);
      case QMode of
        0: EWE := IsNatural and (EE.eReseau  = IdxReseauOuSecteur);
        1: EWE := IsNatural and (EE.eSecteur = IdxReseauOuSecteur);
        2: EWE := IsNatural and (EE.eEntrance = IdxReseauOuSecteur);
      end;
      if (EWE) then
      begin
        TR.Developpement := TR.Developpement + Hypot3D(EE.Une_Station_2_X - EE.Une_Station_1_X,
                                                       EE.Une_Station_2_Y - EE.Une_Station_1_Y,
                                                       EE.Une_Station_2_Z - EE.Une_Station_1_Z);
        CheckC1(EE);
        CheckC2(EE);
        TR.NbVisees  += 1;
        TR.Etendue   := MakeTPoint3Df(TR.CoordMaxi.X - TR.CoordMini.X,
                                      TR.CoordMaxi.Y - TR.CoordMini.Y,
                                      TR.CoordMaxi.Z - TR.CoordMini.Z);
      end
    end;
    TR.DonneesValides := (TR.Developpement > 0.0001);
    Result := True;
  except

  end;
end;

// Export en PostScript
procedure TBDDEntites.ExportVue2DPostScript(const PSFileName: TStringDirectoryFilename; const FM: TModeRepresentationGaleries; const FE: TElementsDrawn);
const
  LAYER_CENTERLINES = 'centerlines';
  LAYER_SECTIONS    = 'sections';
  LAYER_WALLS       = 'walls';
var
  MyPSCanvas: TPostScriptCanvas;
  MyEntite: TEntiteEtendue;
  i, Nb: Integer;
  EWE: TColor;
  PP: array[0..3] of TPoint2Df;
  MyPen: TPenPSProperties;
begin
  AfficherMessage(Format('%s.ExportVue2DPostScript: %s', [ClassName, PSFileName]));
  MyPSCanvas := TPostScriptCanvas.Create;
  try
    MyPSCanvas.InitializeDocument(PSFileName, false,
                                  'Vue en plan', 'Vue en plan en PostScript',
                                  self.FCoinBasGauche.X, self.FCoinBasGauche.Y,
                                  self.FCoinHautDroit.X, self.FCoinHautDroit.Y);
      // centerlines
      MyPSCanvas.BeginLayer(LAYER_CENTERLINES);
      Nb := self.GetNbEntitesVisees();
      for i := 0 to Nb - 1 do
      begin
        MyEntite := GetEntiteVisee(i);

        if (MyEntite.Type_Entite in [tgDEFAULT, tgFOSSILE, tgVADOSE, tgENNOYABLE, tgSIPHON]) then
        begin
          MyPen.Color := GetColorViseeFromModeRepresentation(FM, MyEntite);
          MyPen.fWidth:= 0.50;
          MyPSCanvas.SetPen(MyPen);
          MyPSCanvas.DrawLine('',
                              MyEntite.Une_Station_1_X, MyEntite.Une_Station_1_Y,
                              MyEntite.Une_Station_2_X, MyEntite.Une_Station_2_Y);
        end;
      end;
      // antennes
      if (edANTENNES in FE) then
      begin
        MyPen.Color := clGray;
        MyPen.fWidth:= 0.00;
        for i := 0 to Nb - 1 do
        begin
          MyEntite := GetEntiteVisee(i);
          if (MyEntite.Type_Entite = tgVISEE_RADIANTE) then
          begin
            MyPSCanvas.SetPen(MyPen);
            MyPSCanvas.DrawLine('',
                                MyEntite.Une_Station_1_X, MyEntite.Une_Station_1_Y,
                                MyEntite.Une_Station_2_X, MyEntite.Une_Station_2_Y);
          end;
        end;
      end;
      MyPSCanvas.EndLayer(LAYER_CENTERLINES);
      // sections
      MyPSCanvas.BeginLayer(LAYER_SECTIONS);
      MyPen.Color := clGray;
      MyPen.fWidth:= 0.00;
      for i := 0 to Nb - 1 do
      begin
        MyEntite := GetEntiteVisee(i);
        if (MyEntite.Type_Entite in [tgDEFAULT, tgFOSSILE, tgVADOSE, tgENNOYABLE, tgSIPHON]) then
        begin
          MyPSCanvas.SetPen(MyPen);
          MyPSCanvas.DrawLine('',
                              MyEntite.X1PG, MyEntite.Y1PG,
                              MyEntite.X1PD, MyEntite.Y1PD);
        end;
      end;
      MyPSCanvas.EndLayer(LAYER_SECTIONS);
      // contours
      MyPSCanvas.BeginLayer(LAYER_WALLS);
      if (edWalls in FE) then
      begin
        for i := 0 to Nb - 1 do
        begin
          MyEntite := GetEntiteVisee(i);
          if (MyEntite.Type_Entite in [tgDEFAULT, tgFOSSILE, tgVADOSE, tgENNOYABLE, tgSIPHON]) then
          begin
            MyPen.Color := GetColorViseeFromModeRepresentation(FM, MyEntite);
            MyPen.fWidth:= 0.00;
            PP[0] := MakeTPoint2Df(MyEntite.X1PG, MyEntite.Y1PG);
            PP[1] := MakeTPoint2Df(MyEntite.X1PD, MyEntite.Y1PD);
            PP[2] := MakeTPoint2Df(MyEntite.X2PD, MyEntite.Y2PD);
            PP[3] := MakeTPoint2Df(MyEntite.X2PG, MyEntite.Y2PG);
            MyPSCanvas.DrawLine('', PP[1].X, PP[1].Y, PP[2].X, PP[2].Y);
            MyPSCanvas.DrawLine('', PP[0].X, PP[0].Y, PP[3].X, PP[3].Y);
          end;
        end;
      end;
      MyPSCanvas.EndLayer(LAYER_WALLS);
    MyPSCanvas.FinalizeDocument();
  finally
    MyPSCanvas.Free;
  end;
end;


procedure TBDDEntites.ExportVue2DAcadDXF(const PSFileName: TStringDirectoryFilename; const FM: TModeRepresentationGaleries; const FE: TElementsDrawn);
var
  DXFOut: TDXFOutput;
  n, i, NbreSilhouettes: Integer;
  EWE, E1, E2: TEntiteEtendue;
  QAT: TEntrance;
  DoMakeSilhouette, DoMakePolyCenterlines: boolean;
  MySilhouette: TSilhouetteGalerie;
  MyColorEntite: TColor;
  PtsParois, PtsParois2: TPtsSectionTransversale;
begin
  AfficherMessage(Format('%s.ExportVue2DAcadDXF: %s - %s', [ClassName, GetDescTModeRepresentationGaleriesType(FM), PSFileName]));
  DXFOut := TDXFOutput.Create;
  try
    if (DXFOut.Initialiser(PSFileName)) then
    begin
      DXFOut.SetLimitesDessin(FCoinBasGauche, FCoinHautDroit);
      DXFOut.writeHeader();
        // stations

        n := self.GetNbEntitesVisees();
        for i := 0 to n - 1 do
        begin
          EWE := GetEntiteVisee(i);
          DXFOut.DXFDrawStation(EWE);
        end;
        //*)
        // entrées
        n := GetNbEntrees();
        for i := 0 to n - 1 do
        begin
          QAT := GetEntree(i);
          DXFOut.DXFDrawEntrance(QAT);
        end;
        // silhouettes
        DoMakeSilhouette := True;
        if (DoMakeSilhouette) then
        begin
          SortBySerSts;
          AfficherMessage('Mode silhouettes');
          MySilhouette := TSilhouetteGalerie.Create;

          // premier polygone
          E1 := self.GetEntiteVisee(1);
          AfficherMessage('00');
          MyColorEntite := clRed;
          MySilhouette.ClearSilhouette(MyColorEntite, MyColorEntite);
          AfficherMessage('01');
          PtsParois.ParoiDroiteX := E1.X1PD;
          PtsParois.ParoiDroiteY := E1.Y1PD;
          PtsParois.ParoiGaucheX := E1.X1PG;
          PtsParois.ParoiGaucheY := E1.Y1PG;
          PtsParois.CenterlineX  := E1.Une_Station_1_X;
          PtsParois.CenterlineY  := E1.Une_Station_1_Y;
          PtsParois.CenterlineZ  := E1.Une_Station_1_Z;

          MySilhouette.AddPtsParois(PtsParois);
          AfficherMessage(format('%d entités',[self.GetNbEntitesVisees()]));
          // TODO: Vérifier si les index doivent partir de LOW_INDEX ou de 2
          //for i:= 2 to self.GetNbEntites - 1 do
          for i:= 2 to self.GetNbEntitesVisees() - 1 do
          begin
            //if (Assigned(FProcDispProgression)) then FProcDispProgression(i, 2, self.GetNbEntites, Format('Entite %d / %d', [i, self.GetNbEntites]));
            E1 := GetEntiteVisee(i);
            E2 := GetEntiteVisee(i-1);

            PtsParois.ParoiDroiteX := E1.X2PD;
            PtsParois.ParoiDroiteY := E1.Y2PD;
            PtsParois.ParoiGaucheX := E1.X2PG;
            PtsParois.ParoiGaucheY := E1.Y2PG;
            PtsParois.CenterlineX  := E1.Une_Station_2_X;
            PtsParois.CenterlineY  := E1.Une_Station_2_Y;
            PtsParois.CenterlineZ  := E1.Une_Station_2_Z;

            if (IsSameSerie(E1, E2)) then
            begin
              MySilhouette.AddPtsParois(PtsParois);
              // donner à la silhouette le nom de la série
              MySilhouette.SetNomSilhouette(Format('Serie %d', [E1.Entite_Serie]));
            end else
            begin
              // on construit et trace le polygone
              DXFOut.DXFDrawSilhouettes(MySilhouette);
              DXFOut.DXFDrawCenterlineSilhouettes(MySilhouette);
              NbreSilhouettes := 1 + NbreSilhouettes;
              // on purge la silhouette courante pour en accueillir la suivante
              MyColorEntite := clRed;
              MySilhouette.ClearSilhouette(MyColorEntite, MyColorEntite);
              PtsParois2.ParoiDroiteX := E1.X1PD;
              PtsParois2.ParoiDroiteY := E1.Y1PD;
              PtsParois2.ParoiGaucheX := E1.X1PG;
              PtsParois2.ParoiGaucheY := E1.Y1PG;
              PtsParois2.CenterlineX   := E1.Une_Station_1_X;
              PtsParois2.CenterlineY   := E1.Une_Station_1_Y;
              PtsParois2.CenterlineZ   := E1.Une_Station_1_Z;
              MySilhouette.AddPtsParois(PtsParois2);
            end;
          end;  // for
          AfficherMessage('Mode silhouettes terminé');
          // dernier polygone
          DXFOut.DXFDrawSilhouettes(MySilhouette);
          DXFOut.DXFDrawCenterlineSilhouettes(MySilhouette);

          NbreSilhouettes := 1 + NbreSilhouettes;
          MySilhouette.ClearSilhouette(clWhite, clWhite);
          MySilhouette.Free;
        end;   // if (silhouette)
      DXFOut.writeFooter();
    end;
    DXFOut.Finaliser();
  finally
    DXFOut.Free;
  end;
end;


function TBDDEntites.GetClassesRepartDepthDiagram(): TVecteurDouble;
begin
  Result := FClassRepartDepthDiagram;
end;
function TBDDEntites.GetClassesRepartRoseDiagram(): TVecteurDouble;
begin
  Result := FClassRepartRoseDiagram;
end;


end.

