unit UnitTSAGeoMag;
{$mode delphi}
// Activate it ONLY if you uses GHTopo source code (for AfficherMessage() function)
{$DEFINE IN_GHTOPO}

// Magnetic Declination with WMM model:
////////////////////////////////////////////////////////////////////////////
//
//GeoMag.java - originally geomag.c
//Ported to Java 1.0.2 by Tim Walker
//tim.walker@worldnet.att.net
//tim@acusat.com
//Ported to FreePascal / Lazarus / Delphi by JP CASSOU from Java version
//jean.pierre.cassou@numericable.fr
//jean-pierre.cassou@developpement-durable.gouv.fr
//
//Original source geomag.c available at
//http://www.ngdc.noaa.gov/seg/potfld/DoDWMM.html
//
//NOTE: original comments from the geomag.c source file are in ALL CAPS
//Tim's added comments for the port to Java are not
//
// PUBLIC DOMAIN NOTICE --   License Statement from the NOAA
////////////////////////////////////////////////////////////////////////////
(* *	This is a class to generate the magnetic declination,
 *	magnetic field strength and inclination for any point
 *	on the earth.  The true bearing = magnetic bearing + declination.
 * 	This class is adapted from an Applet from the NOAA National Data Center
 *	at <a href ="http://www.ngdc.noaa.gov/seg/segd.shtml"> http://www.ngdc.noaa.gov/seg/segd.shtml.</a>
 * 	None of the calculations
 *	were changed.  This class requires an input file named WMM-95.DAT, which
 *	must be in the same directory that the application is run from.
 *	NOTE: The original version of
 *	the applet calculated the declination as a function of the date.
 *	This class assumes a date of 1 January 1999, using the 1995 geomagnetic
 *	data.  The WMM-95.DAT file will probably be updated by NOAA in Dec 1999.
 *	This version assumes the calculation is wanted for an altitude of sea level.</p><p>
 *
 *	Using the correct date, the declination is accurate to about 0.5 degrees.</p><p>
 *
 *  This is the LANL D-3 version of the GeoMagnetic calculator from
 *	the NOAA National Data Center at http://www.ngdc.noaa.gov/seg/segd.shtml.</p><p>
 *
 *	Adapted by John St. Ledger, Los Alamos National Laboratory
 *	June 25, 1999</p><p>
 *
 *	Verified by comparison to the source code downloaded from the web.
 *	The declination calculated by that source is exactly the same as this class.
 *	The applet on the web calculates a declination several minutes
 *	different than this class.  Suspect this is the difference between
 *	the DOD model used in this class and the IGRF model of the magnetic field.
 *	Does not return the grid variation near the poles.  These lines commented out,
 *	and can easily be restored.</p><p>
 *
 *	Version 2 Comments:  The world magnetic model is updated every 5 years.
 *	The data for 2000 uses the same algorithm to calculate the magnetic
 *	field variables.  The only change is in the spherical harmonic coefficients
 *	in the input file.  The input file has been renamed to WMM.COF.  Once again,
 *	the date was fixed.  This time to January 1, 2001.  Also, a deprecated
 *	constructor for StreamTokenizer was replaced, and the error messages in the catch
 *	clause were changed.  Methods to get the field strength and inclination
 *	were added.</p><p>
 *
 *	Found out some interesting information about the altitude. The altitude entered
 *      for the calculations is the height above the WGS84 spheroid, not height MSL. Using
 *      MSL height means that the altitude could be in error by as much as 200 meters.
 *      This should not be significant for our applications.</p>
 *
 *     <ul>References:
 *
 *       <li>JOHN M. QUINN, DAVID J. KERRIDGE AND DAVID R. BARRACLOUGH,
 *            WORLD MAGNETIC CHARTS FOR 1985 - SPHERICAL HARMONIC
 *            MODELS OF THE GEOMAGNETIC FIELD AND ITS SECULAR
 *            VARIATION, GEOPHYS. J. R. ASTR. SOC. (1986) 87,
 *            PP 1143-1157</li>
 *
 *       <li>DEFENSE MAPPING AGENCY TECHNICAL REPORT, TR 8350.2:
 *            DEPARTMENT OF DEFENSE WORLD GEODETIC SYSTEM 1984,
 *            SEPT. 30 (1987)</li>
 *
 *       <li>JOSEPH C. CAIN, ET AL.; A PROPOSED MODEL FOR THE
 *            INTERNATIONAL GEOMAGNETIC REFERENCE FIELD - 1965,
 *            J. GEOMAG. AND GEOELECT. VOL. 19, NO. 4, PP 335-355
 *            (1967) (SEE APPENDIX)</li>
 *
 *       <li>ALFRED J. ZMUDA, WORLD MAGNETIC SURVEY 1957-1969,
 *            INTERNATIONAL ASSOCIATION OF GEOMAGNETISM AND
 *            AERONOMY (IAGA) BULLETIN #28, PP 186-188 (1971)</li>
 *
 *       <li>JOHN M. QUINN, RACHEL J. COLEMAN, MICHAEL R. PECK, AND
 *            STEPHEN E. LAUBER; THE JOINT US/UK 1990 EPOCH
 *            WORLD MAGNETIC MODEL, TECHNICAL REPORT NO. 304,
 *            NAVAL OCEANOGRAPHIC OFFICE (1991)</li>
 *
 *       <li>JOHN M. QUINN, RACHEL J. COLEMAN, DONALD L. SHIEL, AND
 *            JOHN M. NIGRO; THE JOINT US/UK 1995 EPOCH WORLD
 *            MAGNETIC MODEL, TECHNICAL REPORT NO. 314, NAVAL
 *            OCEANOGRAPHIC OFFICE (1995)</li></ul>
 *     <p>WMM-2000 is a National Imagery and Mapping Agency (NIMA) standard
 *     product. It is covered under NIMA Military Specification:
 *     MIL-W-89500 (1993).
 *
 *                     EMAIL:   quinn@ghtmail.cr.usgs.gov<br>
 *)

// 19/08/2014: Validé.


interface

uses
  Classes, SysUtils,
  Math,
  {$IFDEF IN_GHTOPO}
  Common, // for AfficherMessage() function
  {$ENDIF}
  dateutils;

type TDataLineGeomag = record
  YearStart: word;
  YearEnd  : word;
  m    : integer;
  n    : integer;
  gnm  : double;
  hnm  : double;
  dgnm : double;
  dhnm : double;
end;
type TMatrix13    = array[0 .. 12] of array[0 .. 12] of double;
type TVecteur13   = array[0 .. 12] of double;
type TVecteur169  = array[0 .. 168] of double;


type

{ TSAGeoMag }

 TSAGeoMag = class(TList)
  private
    //--------------------------------
    fYearMini: word;
    fYearMaxi: word;

    //--------------------------------
    alt: double; //	Geodetic altitude in km = 0;
    glat, glon : double; // 	Geodetic latitude and longitude in deg.  An input.
    decimaltime: double; //     = 2012.5 Time in decimal years.  An input.
    OutDec        : double; //     Geomagnetic declination in deg.
    OutDip        : double; //     Geomagnetic inclination in deg ; Down is positive, up is negative.
    ti         : double; //     Geomagnetic total intensity, in nano Teslas.
    gv         : double; //     Geomagnetic grid variation, referenced to grid North.  Not calculated or output in version 5.0.
    maxdeg     : integer; //     = 12 The maximum number of degrees of the spherical harmonic model.
    maxord     : integer; //    The maximum order of spherical harmonic model.
    defaultdate: double; //     = 2012.5
    defaultAltitude: double; // = 0.00 default altitude
    c          : TMatrix13;  // The gauss coefficients of main geomagnetic model (nt).
    cd         : TMatrix13;  // The gauss coefficients of secular geomagnetic model (nt/yr).
    tc         : TMatrix13;  // The time adjusted geomagnetic gauss coefficients (nt).
    dp         : TMatrix13;  // The theta derivative of p(n,m) (unnormalized).
    snorm      : TVecteur169;// The Schmidt normalization factors.
    sp, cp     : TVecteur13; // The sine and cosine of (m*spherical coord. longitude).
    fn, fm     : TVecteur13;
    pp         : TVecteur13; // The associated Legendre polynomials for m=1 (unnormalized).
    k          : TMatrix13;

    (*
     * The variables otime (old time), oalt (old altitude),
     * olat (old latitude), olon (old longitude), are used to
     * store the values used from the previous calculation to
     * save on calculation time if some inputs don't change.
    //*)
    otime, oalt, olat, olon: double;
    // The date in years, for the start of the valid time of the fit coefficients
    epoch : double;
    bx,by,bz,bh              : double;
    re,a2,b2,c2,a4,b4,c4     : double;
    r,d,ca,sa,ct, st         : double;
    //--------------------------------
    procedure calcGeoMag(const fLat, fLon, qyear, qaltitude: double);

    function  LoadVoidTMatrix13(): TMatrix13;
    function  LoadVoidTVecteur13(): TVecteur13;
    function  LoadVoidTVecteur169(): TVecteur169;
    procedure ViderListeData();
    procedure AddDataGeomag(const D: TDataLineGeomag);
    procedure AddDataGeoMagByValues(const StartYear: word; const qn, qm: integer; const qgnm, qhnm, qdgnm, qdhnm: double);
    function  GetDataGeomag(const Idx: integer): TDataLineGeomag;
    function  GetNbDataGeomag(): integer;
    procedure ReconstruireDataGeomag();

    procedure initModel();
    procedure initMachinSchmidt();


  public
    function Initialiser(const FichierCOF: string): boolean;
    procedure Finaliser();
    function GetDescriptionDeclimag(): string;
    function CalculerDeclinaison(const lon, lat, alt: double; const qdate: TDateTime): double;
    function getDeclination(const dlat, dlong: double; const year: Integer; const altitude: double): double;
end;

implementation

{ TSAGeoMag }
(*
*	Returns the declination from the Department of
*	Defense geomagnetic model and data, in degrees.  The
*	magnetic heading + declination = true heading.
* 	(True heading + variation = magnetic heading.)
*
* @param	dlat		Latitude in decimal degrees.
* @param 	dlong		Longitude in decimal degrees.
* @param	year		The date as a decimial year.
* @param	altitude	The altitude in kilometers.
*
* @return  The declination in degrees.
*)
function TSAGeoMag.getDeclination(const dlat, dlong: double; const year: Integer; const altitude: double): double;
begin
  calcGeoMag( dlat, dlong, year, altitude);
  result := OutDec;
end;

function TSAGeoMag.LoadVoidTMatrix13: TMatrix13;
var
  i, j: Integer;
begin
  for i := 0 to 12 do
    for j := 0 to 12 do
      Result[i][j] := 0.00;
end;

function TSAGeoMag.LoadVoidTVecteur13: TVecteur13;
var
  i: Integer;
begin
  for i := 0 to 12 do Result[i] := 0.00;
end;

function TSAGeoMag.LoadVoidTVecteur169: TVecteur169;
var
  i: Integer;
begin
  for i := 0 to 168 do Result[i] := 0.00;
end;

procedure TSAGeoMag.ViderListeData;
var
  i, Nb: Integer;
begin
  Nb := self.Count;
  if (Nb > 0) then
  begin
    for i := 0 to Nb - 1 do Dispose(self.Items[i]);
  end;
  self.Clear;
end;

procedure TSAGeoMag.AddDataGeomag(const D: TDataLineGeomag);
var
  P: ^TDataLineGeomag;
begin
  New(P);
  P^ := D;
  self.Add(P);
end;

procedure TSAGeoMag.AddDataGeoMagByValues(const StartYear: word; const qn, qm: integer; const qgnm, qhnm, qdgnm, qdhnm: double);
var
  D : TDataLineGeomag;
begin
  D.YearStart := StartYear;
  D.YearEnd   := StartYear + 5;
  D.n    := qn;
  D.m    := qm;
  D.gnm  := qgnm;
  D.hnm  := qhnm;
  D.dgnm := qdgnm;
  D.dhnm := qdhnm;
  self.AddDataGeomag(D);
end;

function TSAGeoMag.GetDataGeomag(const Idx: integer): TDataLineGeomag;
var
  P: ^TDataLineGeomag;
begin
  P := self.Items[Idx];
  Result := P^;
end;

function TSAGeoMag.GetNbDataGeomag: integer;
begin
  Result := self.Count;
end;


procedure TSAGeoMag.ReconstruireDataGeomag;
var
  Nb: Integer;
  i: Integer;
  D: TDataLineGeomag;
begin
  ViderListeData();
  // WMM.cof 2010..2015
  AddDataGeoMagByValues(2010, 1, 0, -29496.60, 0.00, 11.60, 0.00);
  AddDataGeoMagByValues(2010, 1, 1, -1586.30, 4944.40, 16.50, -25.90);
  AddDataGeoMagByValues(2010, 2, 0, -2396.60, 0.00, -12.10, 0.00);
  AddDataGeoMagByValues(2010, 2, 1, 3026.10, -2707.70, -4.40, -22.50);
  AddDataGeoMagByValues(2010, 2, 2, 1668.60, -576.10, 1.90, -11.80);
  AddDataGeoMagByValues(2010, 3, 0, 1340.10, 0.00, 0.40, 0.00);
  AddDataGeoMagByValues(2010, 3, 1, -2326.20, -160.20, -4.10, 7.30);
  AddDataGeoMagByValues(2010, 3, 2, 1231.90, 251.90, -2.90, -3.90);
  AddDataGeoMagByValues(2010, 3, 3, 634.00, -536.60, -7.70, -2.60);
  AddDataGeoMagByValues(2010, 4, 0, 912.60, 0.00, -1.80, 0.00);
  AddDataGeoMagByValues(2010, 4, 1, 808.90, 286.40, 2.30, 1.10);
  AddDataGeoMagByValues(2010, 4, 2, 166.70, -211.20, -8.70, 2.70);
  AddDataGeoMagByValues(2010, 4, 3, -357.10, 164.30, 4.60, 3.90);
  AddDataGeoMagByValues(2010, 4, 4, 89.40, -309.10, -2.10, -0.80);
  AddDataGeoMagByValues(2010, 5, 0, -230.90, 0.00, -1.00, 0.00);
  AddDataGeoMagByValues(2010, 5, 1, 357.20, 44.60, 0.60, 0.40);
  AddDataGeoMagByValues(2010, 5, 2, 200.30, 188.90, -1.80, 1.80);
  AddDataGeoMagByValues(2010, 5, 3, -141.10, -118.20, -1.00, 1.20);
  AddDataGeoMagByValues(2010, 5, 4, -163.00, 0.00, 0.90, 4.00);
  AddDataGeoMagByValues(2010, 5, 5, -7.80, 100.90, 1.00, -0.60);
  AddDataGeoMagByValues(2010, 6, 0, 72.80, 0.00, -0.20, 0.00);
  AddDataGeoMagByValues(2010, 6, 1, 68.60, -20.80, -0.20, -0.20);
  AddDataGeoMagByValues(2010, 6, 2, 76.00, 44.10, -0.10, -2.10);
  AddDataGeoMagByValues(2010, 6, 3, -141.40, 61.50, 2.00, -0.40);
  AddDataGeoMagByValues(2010, 6, 4, -22.80, -66.30, -1.70, -0.60);
  AddDataGeoMagByValues(2010, 6, 5, 13.20, 3.10, -0.30, 0.50);
  AddDataGeoMagByValues(2010, 6, 6, -77.90, 55.00, 1.70, 0.90);
  AddDataGeoMagByValues(2010, 7, 0, 80.50, 0.00, 0.10, 0.00);
  AddDataGeoMagByValues(2010, 7, 1, -75.10, -57.90, -0.10, 0.70);
  AddDataGeoMagByValues(2010, 7, 2, -4.70, -21.10, -0.60, 0.30);
  AddDataGeoMagByValues(2010, 7, 3, 45.30, 6.50, 1.30, -0.10);
  AddDataGeoMagByValues(2010, 7, 4, 13.90, 24.90, 0.40, -0.10);
  AddDataGeoMagByValues(2010, 7, 5, 10.40, 7.00, 0.30, -0.80);
  AddDataGeoMagByValues(2010, 7, 6, 1.70, -27.70, -0.70, -0.30);
  AddDataGeoMagByValues(2010, 7, 7, 4.90, -3.30, 0.60, 0.30);
  AddDataGeoMagByValues(2010, 8, 0, 24.40, 0.00, -0.10, 0.00);
  AddDataGeoMagByValues(2010, 8, 1, 8.10, 11.00, 0.10, -0.10);
  AddDataGeoMagByValues(2010, 8, 2, -14.50, -20.00, -0.60, 0.20);
  AddDataGeoMagByValues(2010, 8, 3, -5.60, 11.90, 0.20, 0.40);
  AddDataGeoMagByValues(2010, 8, 4, -19.30, -17.40, -0.20, 0.40);
  AddDataGeoMagByValues(2010, 8, 5, 11.50, 16.70, 0.30, 0.10);
  AddDataGeoMagByValues(2010, 8, 6, 10.90, 7.00, 0.30, -0.10);
  AddDataGeoMagByValues(2010, 8, 7, -14.10, -10.80, -0.60, 0.40);
  AddDataGeoMagByValues(2010, 8, 8, -3.70, 1.70, 0.20, 0.30);
  AddDataGeoMagByValues(2010, 9, 0, 5.40, 0.00, 0.00, 0.00);
  AddDataGeoMagByValues(2010, 9, 1, 9.40, -20.50, -0.10, 0.00);
  AddDataGeoMagByValues(2010, 9, 2, 3.40, 11.50, 0.00, -0.20);
  AddDataGeoMagByValues(2010, 9, 3, -5.20, 12.80, 0.30, 0.00);
  AddDataGeoMagByValues(2010, 9, 4, 3.10, -7.20, -0.40, -0.10);
  AddDataGeoMagByValues(2010, 9, 5, -12.40, -7.40, -0.30, 0.10);
  AddDataGeoMagByValues(2010, 9, 6, -0.70, 8.00, 0.10, 0.00);
  AddDataGeoMagByValues(2010, 9, 7, 8.40, 2.10, -0.10, -0.20);
  AddDataGeoMagByValues(2010, 9, 8, -8.50, -6.10, -0.40, 0.30);
  AddDataGeoMagByValues(2010, 9, 9, -10.10, 7.00, -0.20, 0.20);
  AddDataGeoMagByValues(2010, 10, 0, -2.00, 0.00, 0.00, 0.00);
  AddDataGeoMagByValues(2010, 10, 1, -6.30, 2.80, 0.00, 0.10);
  AddDataGeoMagByValues(2010, 10, 2, 0.90, -0.10, -0.10, -0.10);
  AddDataGeoMagByValues(2010, 10, 3, -1.10, 4.70, 0.20, 0.00);
  AddDataGeoMagByValues(2010, 10, 4, -0.20, 4.40, 0.00, -0.10);
  AddDataGeoMagByValues(2010, 10, 5, 2.50, -7.20, -0.10, -0.10);
  AddDataGeoMagByValues(2010, 10, 6, -0.30, -1.00, -0.20, 0.00);
  AddDataGeoMagByValues(2010, 10, 7, 2.20, -3.90, 0.00, -0.10);
  AddDataGeoMagByValues(2010, 10, 8, 3.10, -2.00, -0.10, -0.20);
  AddDataGeoMagByValues(2010, 10, 9, -1.00, -2.00, -0.20, 0.00);
  AddDataGeoMagByValues(2010, 10, 10, -2.80, -8.30, -0.20, -0.10);
  AddDataGeoMagByValues(2010, 11, 0, 3.00, 0.00, 0.00, 0.00);
  AddDataGeoMagByValues(2010, 11, 1, -1.50, 0.20, 0.00, 0.00);
  AddDataGeoMagByValues(2010, 11, 2, -2.10, 1.70, 0.00, 0.10);
  AddDataGeoMagByValues(2010, 11, 3, 1.70, -0.60, 0.10, 0.00);
  AddDataGeoMagByValues(2010, 11, 4, -0.50, -1.80, 0.00, 0.10);
  AddDataGeoMagByValues(2010, 11, 5, 0.50, 0.90, 0.00, 0.00);
  AddDataGeoMagByValues(2010, 11, 6, -0.80, -0.40, 0.00, 0.10);
  AddDataGeoMagByValues(2010, 11, 7, 0.40, -2.50, 0.00, 0.00);
  AddDataGeoMagByValues(2010, 11, 8, 1.80, -1.30, 0.00, -0.10);
  AddDataGeoMagByValues(2010, 11, 9, 0.10, -2.10, 0.00, -0.10);
  AddDataGeoMagByValues(2010, 11, 10, 0.70, -1.90, -0.10, 0.00);
  AddDataGeoMagByValues(2010, 11, 11, 3.80, -1.80, 0.00, -0.10);
  AddDataGeoMagByValues(2010, 12, 0, -2.20, 0.00, 0.00, 0.00);
  AddDataGeoMagByValues(2010, 12, 1, -0.20, -0.90, 0.00, 0.00);
  AddDataGeoMagByValues(2010, 12, 2, 0.30, 0.30, 0.10, 0.00);
  AddDataGeoMagByValues(2010, 12, 3, 1.00, 2.10, 0.10, 0.00);
  AddDataGeoMagByValues(2010, 12, 4, -0.60, -2.50, -0.10, 0.00);
  AddDataGeoMagByValues(2010, 12, 5, 0.90, 0.50, 0.00, 0.00);
  AddDataGeoMagByValues(2010, 12, 6, -0.10, 0.60, 0.00, 0.10);
  AddDataGeoMagByValues(2010, 12, 7, 0.50, 0.00, 0.00, 0.00);
  AddDataGeoMagByValues(2010, 12, 8, -0.40, 0.10, 0.00, 0.00);
  AddDataGeoMagByValues(2010, 12, 9, -0.40, 0.30, 0.00, 0.00);
  AddDataGeoMagByValues(2010, 12, 10, 0.20, -0.90, 0.00, 0.00);
  AddDataGeoMagByValues(2010, 12, 11, -0.80, -0.20, -0.10, 0.00);
  AddDataGeoMagByValues(2010, 12, 12, 0.00, 0.90, 0.10, 0.00);
  //(IGRF 2015-2020)
(*      2015.0            WMM-2015        12/15/2014
  1  0  -29438.5       0.0       10.7        0.0
  1  1   -1501.1    4796.2       17.9      -26.8
  2  0   -2445.3       0.0       -8.6        0.0
  2  1    3012.5   -2845.6       -3.3      -27.1
  2  2    1676.6    -642.0        2.4      -13.3
  3  0    1351.1       0.0        3.1        0.0
  3  1   -2352.3    -115.3       -6.2        8.4
  3  2    1225.6     245.0       -0.4       -0.4
  3  3     581.9    -538.3      -10.4        2.3
  4  0     907.2       0.0       -0.4        0.0
  4  1     813.7     283.4        0.8       -0.6
  4  2     120.3    -188.6       -9.2        5.3
  4  3    -335.0     180.9        4.0        3.0
  4  4      70.3    -329.5       -4.2       -5.3
  5  0    -232.6       0.0       -0.2        0.0
  5  1     360.1      47.4        0.1        0.4
  5  2     192.4     196.9       -1.4        1.6
  5  3    -141.0    -119.4        0.0       -1.1
  5  4    -157.4      16.1        1.3        3.3
  5  5       4.3     100.1        3.8        0.1
  6  0      69.5       0.0       -0.5        0.0
  6  1      67.4     -20.7       -0.2        0.0
  6  2      72.8      33.2       -0.6       -2.2
  6  3    -129.8      58.8        2.4       -0.7
  6  4     -29.0     -66.5       -1.1        0.1
  6  5      13.2       7.3        0.3        1.0
  6  6     -70.9      62.5        1.5        1.3
  7  0      81.6       0.0        0.2        0.0
  7  1     -76.1     -54.1       -0.2        0.7
  7  2      -6.8     -19.4       -0.4        0.5
  7  3      51.9       5.6        1.3       -0.2
  7  4      15.0      24.4        0.2       -0.1
  7  5       9.3       3.3       -0.4       -0.7
  7  6      -2.8     -27.5       -0.9        0.1
  7  7       6.7      -2.3        0.3        0.1
  8  0      24.0       0.0        0.0        0.0
  8  1       8.6      10.2        0.1       -0.3
  8  2     -16.9     -18.1       -0.5        0.3
  8  3      -3.2      13.2        0.5        0.3
  8  4     -20.6     -14.6       -0.2        0.6
  8  5      13.3      16.2        0.4       -0.1
  8  6      11.7       5.7        0.2       -0.2
  8  7     -16.0      -9.1       -0.4        0.3
  8  8      -2.0       2.2        0.3        0.0
  9  0       5.4       0.0        0.0        0.0
  9  1       8.8     -21.6       -0.1       -0.2
  9  2       3.1      10.8       -0.1       -0.1
  9  3      -3.1      11.7        0.4       -0.2
  9  4       0.6      -6.8       -0.5        0.1
  9  5     -13.3      -6.9       -0.2        0.1
  9  6      -0.1       7.8        0.1        0.0
  9  7       8.7       1.0        0.0       -0.2
  9  8      -9.1      -3.9       -0.2        0.4
  9  9     -10.5       8.5       -0.1        0.3
 10  0      -1.9       0.0        0.0        0.0
 10  1      -6.5       3.3        0.0        0.1
 10  2       0.2      -0.3       -0.1       -0.1
 10  3       0.6       4.6        0.3        0.0
 10  4      -0.6       4.4       -0.1        0.0
 10  5       1.7      -7.9       -0.1       -0.2
 10  6      -0.7      -0.6       -0.1        0.1
 10  7       2.1      -4.1        0.0       -0.1
 10  8       2.3      -2.8       -0.2       -0.2
 10  9      -1.8      -1.1       -0.1        0.1
 10 10      -3.6      -8.7       -0.2       -0.1
 11  0       3.1       0.0        0.0        0.0
 11  1      -1.5      -0.1        0.0        0.0
 11  2      -2.3       2.1       -0.1        0.1
 11  3       2.1      -0.7        0.1        0.0
 11  4      -0.9      -1.1        0.0        0.1
 11  5       0.6       0.7        0.0        0.0
 11  6      -0.7      -0.2        0.0        0.0
 11  7       0.2      -2.1        0.0        0.1
 11  8       1.7      -1.5        0.0        0.0
 11  9      -0.2      -2.5        0.0       -0.1
 11 10       0.4      -2.0       -0.1        0.0
 11 11       3.5      -2.3       -0.1       -0.1
 12  0      -2.0       0.0        0.1        0.0
 12  1      -0.3      -1.0        0.0        0.0
 12  2       0.4       0.5        0.0        0.0
 12  3       1.3       1.8        0.1       -0.1
 12  4      -0.9      -2.2       -0.1        0.0
 12  5       0.9       0.3        0.0        0.0
 12  6       0.1       0.7        0.1        0.0
 12  7       0.5      -0.1        0.0        0.0
 12  8      -0.4       0.3        0.0        0.0
 12  9      -0.4       0.2        0.0        0.0
 12 10       0.2      -0.9        0.0        0.0
 12 11      -0.9      -0.2        0.0        0.0
 12 12       0.0       0.7        0.0        0.0
999999999999999999999999999999999999999999999999
999999999999999999999999999999999999999999999999
//*)
  Nb := GetNbDataGeomag;
  {$IFDEF IN_GHTOPO}
  AfficherMessage(Format('%s.ReconstruireDataGeomag: %d data', [ClassName, Nb]));
  for i := 0 to Nb - 1 do
  begin
    D := GetDataGeomag(i);
    AfficherMessage(Format(' %d: %d, %d, %.1f, %.1f, %.1f, %.1f' , [i, D.n, D.m, D.gnm, D.hnm, D.dgnm, D.dhnm]));
  end;
  {$ENDIF}
end;

procedure TSAGeoMag.initModel;
var
  a: Extended;
  b: Extended;
  i, Nb: Integer;
  EWE: TDataLineGeomag;
begin
  {$IFDEF IN_GHTOPO}  AfficherMessage(Format('%s.InitModel', [ClassName])); {$ENDIF}
  glat := 0;
  glon := 0;

  // INITIALIZE CONSTANTS
  sp[0] := 0.0;
  cp[0]    := 1.00;
  snorm[0] := 1.00;
  pp[0]    := 1.0;
  dp[0][0] := 0.0;
  a := 6378.137;         // Semi-major axis of WGS-84 ellipsoid, in km.
  b := 6356.7523142;     // Semi-minor axis of WGS-84 ellipsoid, in km.
  re := 6371.2;   // Mean radius of IAU-66 ellipsoid, in km.
  a2 := a * a;
  b2 := b * b;
  c2 := a2 - b2;
  a4 := a2 * a2;
  b4 := b2 * b2;
  c4 := a4 - b4;
   // READ WORLD MAGNETIC MODEL SPHERICAL HARMONIC COEFFICIENTS
  c[0][0]  := 0.0;
  cd[0][0] := 0.0;
  epoch    := 2010.0;
  defaultDate := epoch + 2.5;
  //loop to get data from file
  Nb := GetNbDataGeomag;
  for i := 0 to Nb - 1 do
  begin
    EWE := GetDataGeomag(i);
    if (EWE.m <= EWE.n) then
    begin
      c[EWE.m][EWE.n]  := EWE.gnm;
      cd[EWE.m][EWE.n] := EWE.dgnm;

      if (EWE.m <> 0) then
      begin
        c[EWE.n][EWE.m-1]  := EWE.hnm;
        cd[EWE.n][EWE.m-1] := EWE.dhnm;
      end;
    end;
  end;
  initMachinSchmidt;
end;

procedure TSAGeoMag.initMachinSchmidt;
var
  m, n, j: Integer;
  D1, D2 : Integer;
  flnmj  : Extended;
begin
  {$IFDEF IN_GHTOPO} AfficherMessage(Format('%s.initMachinSchmidt', [ClassName])); {$ENDIF}
  // CONVERT SCHMIDT NORMALIZED GAUSS COEFFICIENTS TO UNNORMALIZED
  snorm[0] := 1.0;
  for n := 1 to maxord do     //for (int n = 1; n <= maxord; n++){
  begin
    snorm[n] := snorm[n - 1] * (2 * n - 1) / n;
    j  := 2;
    m  := 0;
    D1 := 1;
    D2 := n - m + D1;
    // for(int m = 0,D1 = 1,D2 = (n - m + D1) / D1; D2 > 0; D2--, m += D1){
    for D2 := n - m + D1 downto 1 do
    begin
      k[m][n] := (((n - 1) * (n - 1))-(m * m))
                 /
                 ((2 * n-1)*(2*n-3));
      if (m > 0) then
      begin
        flnmj := ((n - m + 1) * j) / (n + m);
        snorm[n + m * 13] := snorm[n + (m -1 ) * 13] * sqrt(flnmj);
        j := 1;
        c[n][m-1]  := snorm[n + m * 13] * c[n][m-1];
        cd[n][m-1] := snorm[n + m * 13] * cd[n][m-1];
      end;
      c[m][n]  := snorm[n + m * 13] * c[m][n];
      cd[m][n] := snorm[n + m * 13] * cd[m][n];
      // instructions supplémentaires de la 3e zone du for
      m := m + D1;
    end;	//for(m...)
    fn[n] := n+1;
    fm[n] := n;
  end; //}	//for(n...)
  k[1][1] := 0.0;
  otime := -1000.0;
  oalt  := -1000.0;
  olat  := -1000.0;
  olon  := -1000.0;

end;

function TSAGeoMag.GetDescriptionDeclimag: string;
begin
  Result := Format('WMM-2010 with WGS84 gravity model - From %d to %d', [fYearMini, fYearMaxi]);
end;

// *     NOTE:  THIS VERSION OF GEOMAG USES THE WMM-2010 GEOMAGNETIC
//    *            MODEL REFERENCED TO THE WGS-84 GRAVITY MODEL ELLIPSOID</p>
//    *
//    * @param	fLat			The latitude in decimal degrees.
//    * @param	fLon			The longitude in decimal degrees.
//    * @param	year			The date as a decimal year.
//    * @param	altitude		The altitude in kilometers.
procedure TSAGeoMag.calcGeoMag(const fLat, fLon, qyear, qaltitude: double);
const
  dtr = pi/180.0;
var
  dt: double;

  rlon, rlat: Extended;
  srlon, srlat: ValReal;
  crlon, crlat: ValReal;
  srlat2, crlat2: Extended;
  q, q1, q2: ValReal;
  r2: Extended;
  aor: Extended;
  ar, br, bt, bp, bpp: Extended;
  D3, D4: Integer;
  n, m: Integer;
  par: Extended;
  temp1, temp2: Extended;
  parp: Extended;
begin
  OutDec := -666.00;
  OutDip := -777.00;
  glat :=  fLat;
  glon :=  fLon;
  alt  :=  qaltitude;
  decimaltime := qyear;  // The date in decimal years for calculating the magnetic field components.
  dt   := decimaltime - epoch;
  rlon := glon * dtr;
  rlat := glat * dtr;
  sincos(rlon, srlon, crlon);
  sincos(rlat, srlat, crlat);
  srlat2 := srlat * srlat;
  crlat2 := crlat * crlat;
  sp[1] := srlon;
  cp[1] := crlon;

  // CONVERT FROM GEODETIC COORDS. TO SPHERICAL COORDS.
  //if (alt != oalt || glat != olat){
  if ((alt <> oalt) OR (glat <> olat)) then
  begin
     q  := sqrt(a2 - c2 * srlat2);
     q1 := alt * q;
     q2 := ((q1 + a2) / (q1 + b2)) * ((q1 + a2) / (q1 + b2));
     ct := srlat / sqrt(q2 * crlat2 + srlat2);
     st := sqrt(1.0 - (ct * ct));
     r2 := ((alt*alt) + 2.0 * q1 + (a4 - c4 * srlat2) / (q * q));
     r := sqrt(r2);
     d := sqrt(a2 * crlat2 + b2 * srlat2);
     ca := (alt + d) / r;
     sa := c2 * crlat * srlat / (r * d);
  end;
  if (glon <> olon) then
  begin
    for m := 2 to maxord do
    begin
      sp[m] := sp[1] * cp[m-1] + cp[1] * sp[m-1];
      cp[m] := cp[1] * cp[m-1] - sp[1] * sp[m-1];
    end;
  end;
  aor := re / r;
  ar  := aor * aor;
  br  := 0.00;
  bt  := 0.00;
  bp  := 0.00;
  bpp := 0.00;

  //for(int n = 1; n <= maxord; n++){
  for n := 1 to maxord do
  begin
    ar := ar * aor;
    D3 := 1;
    D4 := (n + m + D3);
    m := 0;
    //for (int m = 0,D3 = 1,D4 = (n + m + D3) / D3; D4 > 0; D4--,m += D3)
    for D4 := (n + m + D3) downto 1 do
    begin
      //COMPUTE UNNORMALIZED ASSOCIATED LEGENDRE POLYNOMIALS
      //AND DERIVATIVES VIA RECURSION RELATIONS
      if ((alt <> oalt) or (glat <> olat)) then
      begin
        if (n = m) then
        begin
          snorm[n + m * 13] := st * snorm[n - 1 + (m - 1) * 13];
          dp[m][n]          := st * dp[m-1][n-1]+ ct* snorm[n - 1 + (m - 1) * 13];
        end;
        if ((n = 1) and (m = 0)) then
        begin
          snorm[n + m * 13] := ct * snorm[n - 1 + m * 13];
          dp[m][n]          := ct * dp[m][n - 1] - st * snorm[n - 1 + m * 13];
        end;
        if ((n > 1) and (n <> m)) then
        begin
          if (m > (n - 2)) then
          begin
             snorm[n - 2 + m * 13] := 0.0;
             dp[m][n - 2]          := 0.0;
          end;
          snorm[n + m * 13] := ct * snorm[n - 1 + m * 13] - k[m][n] * snorm[n - 2 + m * 13];
          dp[m][n]          := ct * dp[m][n - 1] - st * snorm[n - 1 + m * 13] - k[m][n] * dp[m][n - 2];
        end;
      end; // if ((alt <> oalt) or (glat <> olat)) then

      //TIME ADJUST THE GAUSS COEFFICIENTS

      if (decimaltime <> otime) then
      begin
        tc[m][n] := c[m][n] + dt * cd[m][n];
        if (m <> 0) then tc[n][m - 1] := c[n][m - 1]+ dt * cd[n][m - 1];
      end;
      //ACCUMULATE TERMS OF THE SPHERICAL HARMONIC EXPANSIONS
      par := ar * snorm[ n + m * 13];
      if (m = 0) then
      begin
        temp1 := tc[m][n] * cp[m];
        temp2 := tc[m][n] * sp[m];
      end
      else
      begin
        temp1 := tc[m][n] * cp[m] + tc[n][m - 1] * sp[m];
        temp2 := tc[m][n] * sp[m] - tc[n][m - 1] * cp[m];
      end;

      bt := bt - ar * temp1 * dp[m][n];
      bp += (fm[m] * temp2 * par);
      br += (fn[n] * temp1 * par);

      //SPECIAL CASE:  NORTH/SOUTH GEOGRAPHIC POLES

      if ((st = 0) and (m = 1)) then
      begin
         if (n = 1) then pp[n] := pp[n - 1]
                    else pp[n] := ct * pp[n - 1] - k[m][n] * pp[n - 2];
         parp := ar * pp[n];
         bpp += (fm[m] * temp2 * parp);
      end;
      // instructions supplémentaires de la 3e zone du for
      m += D3;
    end; // for(m...)

  end;	 //for(n...)
  if (st = 0) then bp := bpp
                     else bp /= st;

  //ROTATE MAGNETIC VECTOR COMPONENTS FROM SPHERICAL TO
  //GEODETIC COORDINATES
  // bx must be the east-west field component
  // by must be the north-south field component
  // bz must be the vertical field component.
  bx := -bt * ca - br * sa;
  by := bp;
  bz := bt * sa - br * ca;

  //COMPUTE DECLINATION (DEC), INCLINATION (DIP) AND
  //TOTAL INTENSITY (TI)

  bh := sqrt((bx * bx)+(by * by));
  ti := sqrt((bh * bh)+(bz * bz));
  //	Calculate the declination.
  OutDec := arctan2(by, bx) / dtr;
  OutDip := arctan2(bz, bh) / dtr;
  //AfficherMessage(format('-- Dec = %.6f, bx = %.5f, by = %.5f, bz = %.5f, bh = %.5f', [OutDec, bx, by, bz, bh]));

  //	This is the variation for grid navigation.
  //	Not used at this time.  See St. Ledger for explanation.
  //COMPUTE MAGNETIC GRID VARIATION IF THE CURRENT
  //GEODETIC POSITION IS IN THE ARCTIC OR ANTARCTIC
  //(I.E. GLAT > +55 DEGREES OR GLAT < -55 DEGREES)
  // Grid North is referenced to the 0 Meridian of a polar
  // stereographic projection.
  otime := time;
  oalt  := alt;
  olat  := glat;
  olon  := glon;
end;
//------------------------------------------------------------------------------
// calcul de déclimag avec interpolation
function TSAGeoMag.CalculerDeclinaison(const lon, lat, alt: double; const qDate: TDateTime): double;
var
  ///qd    : TDateTime;
  YYYY  : integer;
  NbD   : Word;
  rr1: Extended;
  dd1, dd2, ddd: Double;
begin
  YYYY  := YearOf(qdate);
  // calcul du rapport d'interpolation rr1 = Nb_jours_ecoules / Nb_jours_annee
  rr1   := DayOfTheYear(qDate) / DaysInAYear(YYYY);
  // valeurs d'interpolation dd1 et dd2 pour le 1er janvier des années N et N+1
  calcGeoMag(lat, lon, YYYY, alt / 1000);
  dd1    := OutDec;
  calcGeoMag(lat, lon, YYYY + 1, alt / 1000);
  dd2 := OutDec;
  // variations entre année N et N+1
  ddd := dd2 - dd1;
  Result := dd1 + ddd * rr1;
end;
//------------------------------------------------------------------------------
function TSAGeoMag.Initialiser(const FichierCOF: string): boolean;
var
  EWE: Double;
  i: Integer;
begin
  Result := False;
  fYearMini := 2000;
  fYearMaxi := 2015;
  {$IFDEF IN_GHTOPO}AfficherMessage(Format('%s.Initialise: %s', [self.ClassName, FichierCOF])); {$ENDIF}
  alt     := 0.00; //	Geodetic altitude in km = 0;
  glat    := 0.00;
  glon    := 0.00;
  decimaltime  := 2012.5;
  OutDec          := 0.00; //     Geomagnetic declination in deg.
  OutDip          := 0.00; //     Geomagnetic inclination in deg ; Down is positive, up is negative.
  ti           := 0.00; //     Geomagnetic total intensity, in nano Teslas.
  gv           := 0.00; //     Geomagnetic grid variation, referenced to grid North.  Not calculated or output in version 5.0.
  maxdeg       := 12;   //     = 12 The maximum number of degrees of the spherical harmonic model.
  maxord       := maxdeg; //    The maximum order of spherical harmonic model.
  defaultdate  := 2012.5;
  defaultAltitude := 0.00;
  c          := LoadVoidTMatrix13;  // The gauss coefficients of main geomagnetic model (nt).
  cd         := LoadVoidTMatrix13;  // The gauss coefficients of secular geomagnetic model (nt/yr).
  tc         := LoadVoidTMatrix13;  // The time adjusted geomagnetic gauss coefficients (nt).
  dp         := LoadVoidTMatrix13;  // The theta derivative of p(n,m) (unnormalized).
  snorm      := LoadVoidTVecteur169;// The Schmidt normalization factors.
  sp         := LoadVoidTVecteur13; // The sine and cosine of (m*spherical coord. longitude).
  fn         := LoadVoidTVecteur13;
  cp         := LoadVoidTVecteur13;
  fm         := LoadVoidTVecteur13;
  pp         := LoadVoidTVecteur13; // The associated Legendre polynomials for m=1 (unnormalized).
  k          := LoadVoidTMatrix13;
  try
    ViderListeData();
    ReconstruireDataGeomag();
    initModel();
    for i := 2000 to 2015 do
    begin
      EWE := getDeclination(42.800, 0.02, i, 0.4);
      {$IFDEF IN_GHTOPO}AfficherMessage(Format('%d: %.5f', [i, EWE])); {$ENDIF}
    end;
    Result := True;
  except
  end;
end;
procedure TSAGeoMag.Finaliser;
begin
  ViderListeData;
end;

end.


public class TSAGeoMag
    private void initModel()
    {
        glat = 0;
        glon = 0;
        //bOutDated = false;
        //String strModel = new String();
        //String strFile = new String("WMM.COF");
        //		String strFile = new String("wmm-95.dat");

        // INITIALIZE CONSTANTS
        maxord = maxdeg;
        sp[0] = 0.0;
        cp[0] = snorm[0] = pp[0] = 1.0;
        dp[0][0] = 0.0;
        /**
         *	Semi-major axis of WGS-84 ellipsoid, in km.
         */
        double a = 6378.137;
        /**
         *	Semi-minor axis of WGS-84 ellipsoid, in km.
         */
        double b = 6356.7523142;
        /**
         *	Mean radius of IAU-66 ellipsoid, in km.
         */
        re = 6371.2;
        a2 = a * a;
        b2 = b * b;
        c2 = a2 - b2;
        a4 = a2 * a2;
        b4 = b2 * b2;
        c4 = a4 - b4;
        String EWE = TCallDialogs.getGHTopoDirectory() + File.separatorChar + "WMM.COF";
        if (FileExists(EWE)) TCallDialogs.showMessage("", EWE);
        try
        {
            //open data file and parse values
            //InputStream is;
            Reader is;
            InputStream input = getClass().getResourceAsStream(EWE);
            if(input == null) throw new FileNotFoundException(EWE + " not found");
            is = new InputStreamReader(input);
            StreamTokenizer str = new StreamTokenizer(is);


            // READ WORLD MAGNETIC MODEL SPHERICAL HARMONIC COEFFICIENTS
            c[0][0] = 0.0;
            cd[0][0] = 0.0;
            str.nextToken();
            epoch = str.nval;
            defaultDate = epoch + 2.5;
            GeneralFunctions.afficherMessage("TSAGeoMag Epoch is: " + epoch);
            GeneralFunctions.afficherMessage("TSAGeoMag default date is: " + defaultDate);
            str.nextToken();
            //strModel = str.sval;
            str.nextToken();

            //loop to get data from file
            while(true){
                str.nextToken();
                if(str.nval == 9999)	// end of file
                    break;

                int n = (int)str.nval;
                str.nextToken();
                int m = (int)str.nval;
                str.nextToken();
                double gnm = str.nval;
                str.nextToken();
                double hnm = str.nval;
                str.nextToken();
                double dgnm = str.nval;
                str.nextToken();
                double dhnm = str.nval;

                if (m <= n)
                {
                    c[m][n] = gnm;
                    cd[m][n] = dgnm;

                    if (m != 0)
                    {
                        c[n][m-1] = hnm;
                        cd[n][m-1] = dhnm;
                    }
                }

            }	//while(true)

            is.close();
        }	//try
        // version 2, catch FileNotFound and IO exceptions separately,
        // rather than catching all exceptions.
        // Version 5.4 add logger support, and comment out System.out.println
        catch(FileNotFoundException e)
        {
            String msg = sprintf("\nNOTICE      NOTICE      NOTICE      \n" +
            		"WMMCOF file not found in TSAGeoMag.InitModel()\n" +
            		"The input file %s was not found in the same\n" +
            		"directory as the application.\n" +
            		"The magnetic field components are set to internal values.\n",
                        EWE);
            GeneralFunctions.afficherMessage(msg);
            setCoeff();
        }

        catch(IOException e)
        {
            String msg = sprintf("\nNOTICE      NOTICE      NOTICE      \n" +
			"Problem reading the WMMCOF file in TSAGeoMag.InitModel()\n" +
			"The input file %s was found, but there was a problem \n" +
			"reading the data.\n" +
			"The magnetic field components are set to internal values.",
                        EWE);
            GeneralFunctions.afficherMessage(msg);
            setCoeff();
        }
        // CONVERT SCHMIDT NORMALIZED GAUSS COEFFICIENTS TO UNNORMALIZED
        snorm[0] = 1.0;
        for (int n = 1; n <= maxord; n++){

            snorm[n] = snorm[n - 1] * (2 * n - 1) / n;
            int j = 2;

            for(int m = 0,D1 = 1,D2 = (n - m + D1) / D1; D2 > 0; D2--, m += D1){
                k[m][n] = (double) (((n - 1) * (n - 1))-(m * m))/(double) ((2 * n-1)*(2*n-3));
                if(m > 0){
                    double flnmj = ((n - m + 1) * j) / (double) (n + m);
                    snorm[n + m * 13] = snorm[n + (m -1 ) * 13] * Math.sqrt(flnmj);
                    j = 1;
                    c[n][m-1] = snorm[n + m * 13] * c[n][m-1];
                    cd[n][m-1] = snorm[n + m * 13] * cd[n][m-1];
                }
                c[m][n] = snorm[n + m * 13] * c[m][n];
                cd[m][n] = snorm[n + m * 13] * cd[m][n];
            }	//for(m...)

            fn[n] = (n+1);
            fm[n] = n;

        }	//for(n...)

        k[1][1] = 0.0;

        otime = oalt = olat = olon = -1000.0;


    }

    /**     <p><b>PURPOSE:</b>  THIS ROUTINE COMPUTES THE DECLINATION (DEC),
     *               INCLINATION (DIP), TOTAL INTENSITY (TI) AND
     *               GRID VARIATION (GV - POLAR REGIONS ONLY, REFERENCED
     *               TO GRID NORTH OF POLAR STEREOGRAPHIC PROJECTION) OF
     *               THE EARTH'S MAGNETIC FIELD IN GEODETIC COORDINATES
     *               FROM THE COEFFICIENTS OF THE CURRENT OFFICIAL
     *               DEPARTMENT OF DEFENSE (DOD) SPHERICAL HARMONIC WORLD
     *               MAGNETIC MODEL (WMM-2010).  THE WMM SERIES OF MODELS IS
     *               UPDATED EVERY 5 YEARS ON JANUARY 1'ST OF THOSE YEARS
     *               WHICH ARE DIVISIBLE BY 5 (I.E. 1980, 1985, 1990 ETC.)
     *               BY THE NAVAL OCEANOGRAPHIC OFFICE IN COOPERATION
     *               WITH THE BRITISH GEOLOGICAL SURVEY (BGS).  THE MODEL
     *               IS BASED ON GEOMAGNETIC SURVEY MEASUREMENTS FROM
     *               AIRCRAFT, SATELLITE AND GEOMAGNETIC OBSERVATORIES.</p><p>
     *
     *
     *
     *     <b>ACCURACY:</b>  IN OCEAN AREAS AT THE EARTH'S SURFACE OVER THE
     *                ENTIRE 5 YEAR LIFE OF A DEGREE AND ORDER 12
     *                SPHERICAL HARMONIC MODEL SUCH AS WMM-95, THE ESTIMATED
     *                RMS ERRORS FOR THE VARIOUS MAGENTIC COMPONENTS ARE:</p>
     *<ul>
     *                DEC  -   0.5 Degrees<br>
     *                DIP  -   0.5 Degrees<br>
     *                TI   - 280.0 nanoTeslas (nT)<br>
     *                GV   -   0.5 Degrees<br></ul>
     *
     *                <p>OTHER MAGNETIC COMPONENTS THAT CAN BE DERIVED FROM
     *                THESE FOUR BY SIMPLE TRIGONOMETRIC RELATIONS WILL
     *                HAVE THE FOLLOWING APPROXIMATE ERRORS OVER OCEAN AREAS:</p>
     *<ul>
     *                X    - 140 nT (North)<br>
     *                Y    - 140 nT (East)<br>
     *                Z    - 200 nT (Vertical)  Positive is down<br>
     *                H    - 200 nT (Horizontal)<br></ul>
     *
     *                <p>OVER LAND THE RMS ERRORS ARE EXPECTED TO BE SOMEWHAT
     *                HIGHER, ALTHOUGH THE RMS ERRORS FOR DEC, DIP AND GV
     *                ARE STILL ESTIMATED TO BE LESS THAN 0.5 DEGREE, FOR
     *                THE ENTIRE 5-YEAR LIFE OF THE MODEL AT THE EARTH's
     *                SURFACE.  THE OTHER COMPONENT ERRORS OVER LAND ARE
     *                MORE DIFFICULT TO ESTIMATE AND SO ARE NOT GIVEN.</p><p>
     *
     *                THE ACCURACY AT ANY GIVEN TIME OF ALL FOUR
     *                GEOMAGNETIC PARAMETERS DEPENDS ON THE GEOMAGNETIC
     *                LATITUDE.  THE ERRORS ARE LEAST AT THE EQUATOR AND
     *                GREATEST AT THE MAGNETIC POLES.</p><p>
     *
     *                IT IS VERY IMPORTANT TO NOTE THAT A DEGREE AND
     *                ORDER 12 MODEL, SUCH AS WMM-2010 DESCRIBES ONLY
     *                THE LONG WAVELENGTH SPATIAL MAGNETIC FLUCTUATIONS
     *                DUE TO EARTH'S CORE.  NOT INCLUDED IN THE WMM SERIES
     *                MODELS ARE INTERMEDIATE AND SHORT WAVELENGTH
     *                SPATIAL FLUCTUATIONS OF THE GEOMAGNETIC FIELD
     *                WHICH ORIGINATE IN THE EARTH'S MANTLE AND CRUST.
     *                CONSEQUENTLY, ISOLATED ANGULAR ERRORS AT VARIOUS
     *                POSITIONS ON THE SURFACE (PRIMARILY OVER LAND, IN
     *                CONTINENTAL MARGINS AND OVER OCEANIC SEAMOUNTS,
     *                RIDGES AND TRENCHES) OF SEVERAL DEGREES MAY BE
     *                EXPECTED. ALSO NOT INCLUDED IN THE MODEL ARE
     *                NONSECULAR TEMPORAL FLUCTUATIONS OF THE GEOMAGNETIC
     *                FIELD OF MAGNETOSPHERIC AND IONOSPHERIC ORIGIN.
     *                DURING MAGNETIC STORMS, TEMPORAL FLUCTUATIONS CAN
     *                CAUSE SUBSTANTIAL DEVIATIONS OF THE GEOMAGNETIC
     *                FIELD FROM MODEL VALUES.  IN ARCTIC AND ANTARCTIC
     *                REGIONS, AS WELL AS IN EQUATORIAL REGIONS, DEVIATIONS
     *                FROM MODEL VALUES ARE BOTH FREQUENT AND PERSISTENT.</p><p>
     *
     *                IF THE REQUIRED DECLINATION ACCURACY IS MORE
     *                STRINGENT THAN THE WMM SERIES OF MODELS PROVIDE, THEN
     *                THE USER IS ADVISED TO REQUEST SPECIAL (REGIONAL OR
     *                LOCAL) SURVEYS BE PERFORMED AND MODELS PREPARED BY
     *                THE USGS, WHICH OPERATES THE US GEOMAGNETIC
     *                OBSERVATORIES.  REQUESTS OF THIS NATURE SHOULD
     *                BE MADE THROUGH NIMA AT THE ADDRESS ABOVE.</p><p>
     *
     *
     *
     *     NOTE:  THIS VERSION OF GEOMAG USES THE WMM-2010 GEOMAGNETIC
     *            MODEL REFERENCED TO THE WGS-84 GRAVITY MODEL ELLIPSOID</p>
     *
     * @param	fLat			The latitude in decimal degrees.
     * @param	fLon			The longitude in decimal degrees.
     * @param	year			The date as a decimal year.
     * @param	altitude		The altitude in kilometers.
     */
    private void calcGeoMag(double fLat, double fLon, double year, double altitude)
    {

        glat =  fLat;
        glon =  fLon;
        alt = altitude;
        /**
         *	The date in decimal years for calculating the magnetic field components.
         */
        time = year;

        double dt = time - epoch;
        //if (otime < 0.0 && (dt < 0.0 || dt > 5.0))
        //		if(bCurrent){
        //			if (dt < 0.0 || dt > 5.0)
        //				bOutDated = true;
        //			else
        //				bOutDated = false;
        //		}

        double pi = Math.PI;
        double dtr = (pi/180.0);
        double rlon = glon * dtr;
        double rlat = glat * dtr;
        double srlon = Math.sin(rlon);
        double srlat = Math.sin(rlat);
        double crlon = Math.cos(rlon);
        double crlat = Math.cos(rlat);
        double srlat2 = srlat * srlat;
        double crlat2 = crlat * crlat;
        sp[1] = srlon;
        cp[1] = crlon;

        // CONVERT FROM GEODETIC COORDS. TO SPHERICAL COORDS.
        if (alt != oalt || glat != olat){
            double q = Math.sqrt(a2 - c2 * srlat2);
            double q1 = alt * q;
            double q2 = ((q1 + a2) / (q1 + b2)) * ((q1 + a2) / (q1 + b2));
            ct = srlat / Math.sqrt(q2 * crlat2 + srlat2);
            st = Math.sqrt(1.0 - (ct * ct));
            double r2 = ((alt*alt) + 2.0 * q1 + (a4 - c4 * srlat2) / (q * q));
            r = Math.sqrt(r2);
            d = Math.sqrt(a2 * crlat2 + b2 * srlat2);
            ca = (alt + d) / r;
            sa = c2 * crlat * srlat / (r * d);
        }
        if (glon != olon){
            for (int m = 2; m <= maxord; m++){
                sp[m] = sp[1] * cp[m-1] + cp[1] * sp[m-1];
                cp[m] = cp[1] * cp[m-1] - sp[1] * sp[m-1];
            }
        }
        double aor = re / r;
        double ar = aor * aor;
        double br = 0, bt = 0, bp = 0, bpp = 0;

        for(int n = 1; n <= maxord; n++){
            ar = ar * aor;
            for (int m = 0,D3 = 1,D4 = (n + m + D3) / D3; D4 > 0; D4--,m += D3){

                //COMPUTE UNNORMALIZED ASSOCIATED LEGENDRE POLYNOMIALS
                //AND DERIVATIVES VIA RECURSION RELATIONS
                if(alt != oalt || glat != olat){
                    if(n == m){
                        snorm[n + m * 13] = st * snorm[n - 1 + (m - 1) * 13];
                        dp[m][n] = st * dp[m-1][n-1]+ ct* snorm[n - 1 + (m - 1) * 13];
                    }
                    if(n == 1 && m == 0){
                        snorm[n + m * 13] = ct * snorm[n - 1 + m * 13];
                        dp[m][n] = ct * dp[m][n - 1] - st * snorm[n - 1 + m * 13];
                    }
                    if(n > 1 && n != m){
                        if(m > n - 2)
                            snorm[n - 2 + m * 13] = 0.0;
                        if(m > n - 2)
                            dp[m][n - 2] = 0.0;
                        snorm[n + m * 13] = ct * snorm[n - 1 + m * 13] - k[m][n] * snorm[n - 2 + m * 13];
                        dp[m][n] = ct * dp[m][n - 1] - st * snorm[n - 1 + m * 13] - k[m][n] * dp[m][n - 2];
                    }
                }

                //TIME ADJUST THE GAUSS COEFFICIENTS

                if(time != otime){
                    tc[m][n] = c[m][n] + dt * cd[m][n];

                    if(m != 0)
                        tc[n][m - 1] = c[n][m - 1]+ dt * cd[n][m - 1];
                }

                //ACCUMULATE TERMS OF THE SPHERICAL HARMONIC EXPANSIONS
                double temp1, temp2;
                double par = ar * snorm[ n + m * 13];
                if(m == 0){
                    temp1 = tc[m][n] * cp[m];
                    temp2 = tc[m][n] * sp[m];
                }
                else{
                    temp1 = tc[m][n] * cp[m] + tc[n][m - 1] * sp[m];
                    temp2 = tc[m][n] * sp[m] - tc[n][m - 1] * cp[m];
                }

                bt = bt - ar * temp1 * dp[m][n];
                bp += (fm[m] * temp2 * par);
                br += (fn[n] * temp1 * par);

                //SPECIAL CASE:  NORTH/SOUTH GEOGRAPHIC POLES

                if(st == 0.0 && m == 1){
                    if(n == 1)
                        pp[n] = pp[n - 1];
                    else
                        pp[n] = ct * pp[n - 1] - k[m][n] * pp[n - 2];
                    double parp = ar * pp[n];
                    bpp += (fm[m] * temp2 * parp);
                }

            }	//for(m...)

        }	//for(n...)


        if(st == 0.0)
            bp = bpp;
        else
            bp /= st;

        //ROTATE MAGNETIC VECTOR COMPONENTS FROM SPHERICAL TO
        //GEODETIC COORDINATES
        // bx must be the east-west field component
        // by must be the north-south field component
        // bz must be the vertical field component.
        bx = -bt * ca - br * sa;
        by = bp;
        bz = bt * sa - br * ca;

        //COMPUTE DECLINATION (DEC), INCLINATION (DIP) AND
        //TOTAL INTENSITY (TI)

        bh = Math.sqrt((bx * bx)+(by * by));
        ti = Math.sqrt((bh * bh)+(bz * bz));
        //	Calculate the declination.
        dec = (Math.atan2(by, bx) / dtr);
        GeneralFunctions.afficherMessage("Dec is: " + dec );
        dip = (Math.atan2(bz, bh) / dtr);

        //	This is the variation for grid navigation.
        //	Not used at this time.  See St. Ledger for explanation.
        //COMPUTE MAGNETIC GRID VARIATION IF THE CURRENT
        //GEODETIC POSITION IS IN THE ARCTIC OR ANTARCTIC
        //(I.E. GLAT > +55 DEGREES OR GLAT < -55 DEGREES)
        // Grid North is referenced to the 0 Meridian of a polar
        // stereographic projection.

        //OTHERWISE, SET MAGNETIC GRID VARIATION TO -999.0
        /*
         gv = -999.0;
         if (Math.abs(glat) >= 55.){
         if (glat > 0.0 && glon >= 0.0)
         gv = dec-glon;
         if (glat > 0.0 && glon < 0.0)
         gv = dec + Math.abs(glon);
         if (glat < 0.0 && glon >= 0.0)
         gv = dec+glon;
         if (glat < 0.0 && glon < 0.0)
         gv = dec - Math.abs(glon);
         if (gv > +180.0)
         gv -= 360.0;
         if (gv < -180.0)
         gv += 360.0;
         }
         */
        otime = time;
        oalt = alt;
        olat = glat;
        olon = glon;

    }
    /**
     *	Returns the declination from the Department of
     *	Defense geomagnetic model and data, in degrees.  The
     *	magnetic heading + declination = true heading. The date and
     *  altitude are the defaults, of half way through the valid
     *  5 year period, and 0 elevation.
     * 	(True heading + variation = magnetic heading.)
     *
     * @param	dlat	Latitude in decimal degrees.
     * @param 	dlong	Longitude in decimal degrees.
     *
     * @return  The declination in degrees.
     */
    public double getDeclination( double dlat, double dlong )
    {
            calcGeoMag( dlat, dlong, defaultDate, defaultAltitude );
            return  dec;
    }
    /**
     *	Returns the declination from the Department of
     *	Defense geomagnetic model and data, in degrees.  The
     *	magnetic heading + declination = true heading.
     * 	(True heading + variation = magnetic heading.)
     *
     * @param	dlat		Latitude in decimal degrees.
     * @param 	dlong		Longitude in decimal degrees.
     * @param	year		The date as a decimial year.
     * @param	altitude	The altitude in kilometers.
     *
     * @return  The declination in degrees.
     */
    public double getDeclination( double dlat, double dlong, int year, double altitude )
    {
            afficherMessageFmt("getDeclination: lat = %.8f, lon = %.8f, alt = %.0f m, year=%d", dlat, dlong, altitude * 1000, year);
            calcGeoMag( dlat, dlong, year, altitude);
            return  dec;
    }
    /**
     *	Returns the magnetic field intensity from the
     *	Department of Defense geomagnetic model and data
     *	in nano Tesla. The date and
     *  altitude are the defaults, of half way through the valid
     *  5 year period, and 0 elevation.
     *
     * @param	dlat	Latitude in decimal degrees.
     * @param 	dlong	Longitude in decimal degrees.
     *
     * return  Magnetic field strength in nano Tesla.
     */
    public double getIntensity( double dlat, double dlong )
    {
            calcGeoMag( dlat, dlong, defaultDate, defaultAltitude );
            return  ti;
    }

    /**
     *	Returns the magnetic field intensity from the
     *	Department of Defense geomagnetic model and data
     *	in nano Tesla.
     *
     * @param	dlat	 	Latitude in decimal degrees.
     * @param 	dlong		Longitude in decimal degrees.
     * @param	year		Date of the calculation in decimal years.
     * @param	altitude	Altitude of the calculation in kilometers.
     *
     * return  Magnetic field strength in nano Tesla.
     */
    public double getIntensity( double dlat, double dlong, double year, double altitude )
    {
            calcGeoMag( dlat, dlong, year, altitude );
            return  ti;
    }
    /**
     *	Returns the horizontal magnetic field intensity from the
     *	Department of Defense geomagnetic model and data
     *	in nano Tesla. The date and
     *  altitude are the defaults, of half way through the valid
     *  5 year period, and 0 elevation.
     *
     * @param	dlat	Latitude in decimal degrees.
     * @param 	dlong	Longitude in decimal degrees.
     *
     * return  The horizontal magnetic field strength in nano Tesla.
     */
    public double getHorizontalIntensity( double dlat, double dlong )
    {
            calcGeoMag( dlat, dlong, defaultDate, defaultAltitude );
            return  bh;
    }

    /**
     *	Returns the horizontal magnetic field intensity from the
     *	Department of Defense geomagnetic model and data
     *	in nano Tesla.
     *
     * @param	dlat	 	Latitude in decimal degrees.
     * @param 	dlong		Longitude in decimal degrees.
     * @param	year		Date of the calculation in decimal years.
     * @param	altitude	Altitude of the calculation in kilometers.
     *
     * return  The horizontal magnetic field strength in nano Tesla.
     */
    public double getHorizontalIntensity( double dlat, double dlong, double year, double altitude )
    {
            calcGeoMag( dlat, dlong, year, altitude );
            return  bh;
    }
    /**
     *	Returns the vertical magnetic field intensity from the
     *	Department of Defense geomagnetic model and data
     *	in nano Tesla. The date and
     *  altitude are the defaults, of half way through the valid
     *  5 year period, and 0 elevation.
     *
     * @param	dlat	Latitude in decimal degrees.
     * @param 	dlong	Longitude in decimal degrees.
     *
     * @return  The vertical magnetic field strength in nano Tesla.
     */
    public double getVerticalIntensity( double dlat, double dlong )
    {
            calcGeoMag( dlat, dlong, defaultDate, defaultAltitude );
            return  bz;
    }

    /**
     *	Returns the vertical magnetic field intensity from the
     *	Department of Defense geomagnetic model and data
     *	in nano Tesla.
     *
     * @param	dlat	 	Latitude in decimal degrees.
     * @param 	dlong		Longitude in decimal degrees.
     * @param	year		Date of the calculation in decimal years.
     * @param	altitude	Altitude of the calculation in kilometers.
     *
     * @return  The vertical magnetic field strength in nano Tesla.
     */
    public double getVerticalIntensity( double dlat, double dlong, double year, double altitude )
    {
            calcGeoMag( dlat, dlong, year, altitude );
            return  bz;
    }
    /**
     *	Returns the northerly magnetic field intensity from the
     *	Department of Defense geomagnetic model and data
     *	in nano Tesla. The date and
     *  altitude are the defaults, of half way through the valid
     *  5 year period, and 0 elevation.
     *
     * @param	dlat	Latitude in decimal degrees.
     * @param 	dlong	Longitude in decimal degrees.
     *
     * return  The northerly component of the magnetic field strength in nano Tesla.
     */
    public double getNorthIntensity( double dlat, double dlong )
    {
            calcGeoMag( dlat, dlong, defaultDate, defaultAltitude );
            return  by;
    }

    /**
     *	Returns the northerly magnetic field intensity from the
     *	Department of Defense geomagnetic model and data
     *	in nano Tesla.
     *
     * @param	dlat	 	Latitude in decimal degrees.
     * @param 	dlong		Longitude in decimal degrees.
     * @param	year		Date of the calculation in decimal years.
     * @param	altitude	Altitude of the calculation in kilometers.
     *
     * return  The northerly component of the magnetic field strength in nano Tesla.
     */
    public double getNorthIntensity( double dlat, double dlong, double year, double altitude )
    {
            calcGeoMag( dlat, dlong, year, altitude );
            return  by;
    }
    /**
     *	Returns the easterly magnetic field intensity from the
     *	Department of Defense geomagnetic model and data
     *	in nano Tesla. The date and
     *  altitude are the defaults, of half way through the valid
     *  5 year period, and 0 elevation.
     *
     * @param	dlat		Latitude in decimal degrees.
     * @param 	dlong	Longitude in decimal degrees.
     *
     * return  The easterly component of the magnetic field strength in nano Tesla.
     */
    public double getEastIntensity( double dlat, double dlong )
    {
            calcGeoMag( dlat, dlong, defaultDate, defaultAltitude );
            return  bx;
    }

    /**
     *	Returns the easterly magnetic field intensity from the
     *	Department of Defense geomagnetic model and data
     *	in nano Tesla.
     *
     * @param	dlat	 	Latitude in decimal degrees.
     * @param 	dlong		Longitude in decimal degrees.
     * @param	year		Date of the calculation in decimal years.
     * @param	altitude	Altitude of the calculation in kilometers.
     *
     * return  The easterly component of the magnetic field strength in nano Tesla.
     */
    public double getEastIntensity( double dlat, double dlong, double year, double altitude )
    {
            calcGeoMag( dlat, dlong, year, altitude );
            return  bx;
    }
    /**
     *	Returns the magnetic field dip angle from the
     *	Department of Defense geomagnetic model and data,
     *	in degrees.  The date and
     *  altitude are the defaults, of half way through the valid
     *  5 year period, and 0 elevation.
     *
     * @param	dlat	Latitude in decimal degrees.
     * @param 	dlong	Longitude in decimal degrees.
     *
     * return  The magnetic field dip angle, in degrees.
     */
    public double getDipAngle( double dlat, double dlong )
    {
            calcGeoMag( dlat, dlong, defaultDate, defaultAltitude );
            return  dip;
    }

    /**
     *	Returns the magnetic field dip angle from the
     *	Department of Defense geomagnetic model and data,
     *	in degrees.
     *
     * @param	dlat			Latitude in decimal degrees.
     * @param 	dlong		Longitude in decimal degrees.
     * @param	year			Date of the calculation in decimal years.
     * @param	altitude		Altitude of the calculation in kilometers.
     *
     * return  The magnetic field dip angle, in degrees.
     */
    public double getDipAngle( double dlat, double dlong, double year, double altitude )
    {
            calcGeoMag( dlat, dlong, year, altitude );
            return  dip;
    }

    /**	This method sets the input data to the internal fit coefficents.
     *	If there is an exception reading the input file WMM.COF, these values
     *	are used.
     *
     *  NOTE:  This method is not tested by the JUnit test, unless the WMM.COF file
     *         is missing.
     */
    private void setCoeff()
    {
        c[0][0] = 0.0;
        cd[0][0] = 0.0;

        epoch = Double.parseDouble(input[0].trim().split("[\\s]+")[0]);
        defaultDate = epoch + 2.5;

        String [] tokens;

        //loop to get data from internal values
        for(int i=1; i<input.length; i++)
        {
            tokens = input[i].trim().split("[\\s]+");

            int n = Integer.parseInt(tokens[0]);
            int m = Integer.parseInt(tokens[1]);
            double gnm = Double.parseDouble(tokens[2]);
            double hnm = Double.parseDouble(tokens[3]);
            double dgnm = Double.parseDouble(tokens[4]);
            double dhnm = Double.parseDouble(tokens[5]);

            if (m <= n)
            {
                c[m][n] = gnm;
                cd[m][n] = dgnm;

                if (m != 0)
                {
                    c[n][m-1] = hnm;
                    cd[n][m-1] = dhnm;
                }
            }
        }
    }

    /**<p>
     *   Given a Gregorian Calendar object, this returns the decimal year
     *   value for the calendar, accurate to the day of the input calendar.
     *   The hours, minutes, and seconds of the date are ignored.</p><p>
     *
     *   If the input Gregorian Calendar is new GregorianCalendar(2012, 6, 1), all of
     *   the first of July is counted, and this returns 2012.5. (183 days out of 366)</p><p>
     *
     *   If the input Gregorian Calendar is new GregorianCalendar(2010, 0, 0), the first
     *   of January is not counted, and this returns 2010.0</p><p>
     *
     * @param cal  Has the date (year, month, and day of the month)
     * @return double	   The date in decimal years
     */
    public double decimalYear(GregorianCalendar cal)
    {
	int year = cal.get(Calendar.YEAR);
	double daysInYear;
        daysInYear = (cal.isLeapYear(year)) ? 366.0 : 365.0;
	return year + (cal.get(Calendar.DAY_OF_YEAR))/daysInYear;
    }
}


