unit ToporobotClasses2012;
//****************************************************
// Projet     : GHTopo
// Modules    : Classe pour gestion des données TOPOROBOT
// Licence    : General Public License - WITHOUT WARRANTY
// Auteur     : JP CASSOU
// OS         : Windows NT - Linux
// Langage    : Free-Pascal 3.0.x sous Lazarus
//-----------------------------------------------------
// Eléments du cahier des charges:
// - Les données sont stockées sous format texte
// - Gestion de la base entièrement en mémoire sous forme d'une arborescence
// - Les listes simples sont des TList dont les membres sont des 'record'.
// - Les séries sont des objets autonomes stockés dans une liste de type TList
//   L'accès aux membres de l'objet se fait uniquement par getters et setters
// - La création de nouveaux items se fait par des fonctions
//   de la forme AddObjet(const MonObjet)
// - Le processus d'édition d'un objet est le suivant:
//   1. Extraction d'un objet (expé, code, entrée, série, antenne, réseau) depuis sa liste
//      via une fonction de la forme GetObjet(const Idx: integer): TMonObjet
//   2. Utilisation de l'objet et modification de ses propriétés
//   3. Mise à jour de l'objet dans sa liste
//      via une fonction de la forme PutObjet(const Idx: integer; const Obj: TMonObjet);
// - Support du format Toporobot TAB en lecture-écriture
// - Support du format XTB ( = format TAB avec sections additionnelles)
// - Lecture des fichiers Text (format considéré comme obsolète et peu robuste.
//                              L'écriture dans ce format ne figure pas dans le CdC)
// - Evolutivité vers un nouveau format de type XML
// - Exportation vers d'autres logiciels
//   * Visual Topo (indispensable) - La méthode TOPOROBOT est incompatible avec la notation alphanumérique
//   * Therion (en cours de validation)
//-----------------------------------------------------
// Module     : ToporobotClasses:
//            : Structure des données d'un dossier Toporobot
// Compatibilité intégrale avec le standard Toporobot 1994.
// 22/08/2012 : Refonte du noyau de GHTopo (ToporobotClasses)
//              en séparant le SGBD et le code de calcul
// 24/08/2012 : Opérationnel
// 14/02/2013 : Réacculturation au code de GHTopo. Petites corrections
//              Préparation de fonctions de recherche plein texte dans les listes
// 03/05/2013 : Réorganisation du fichier de paramètres de compilation
// 13/06/2013 : Quelques modifications dont ExtractFileNameOnly
// 14/06/2013 : Implantation du nouveau format GHTopo XML
// 03/10/2013 : DONE: Implémentation du support des lasermètres-clinomètres type Stanley TLM330
// 04/10/2013 : Conversions UTF8 <> ANSI fixed
// 23/10/2013 : Fonctions de recherche dans les tables
// 24/10/2013 : Support des secteurs, utilisés au niveau visée
// 25/10/2013 : Découplage des listes simples de l'unité ToporobotClasses - Vont dans UnitListesSimples.pas
// 10/11/2013 : Export en CSV au niveau des tables - DONE: Squelette des fonctions OK + En-tête séries/
// 11/11/2013 : Check des données: Longueurs OK, Azimuts OK, Pentes: Clinomètres std et lasermètre TLM330: OK (95% des configurations).
// 15/11/2013 : DONE: Ajout des secteurs dans le format XML
// **** CHANGEMENT DANS LE FORMAT XTB
// 14/02/2014 : Ajout du support de la colonne 'Secteur' de la section AntennaShots
// 30/06/2014 : Harmonisation avec la version Java. L'ajout de nouveaux éléments préparamétrés se fait ici.
// 29/07/2014 : Stockage des composantes des couleurs dans le format XML pour éviter les complications de conversions
// 20/11/2014 : Ajout de la fonction HasStationForAntennaShot()
// 08/12/2014 : Export au format TOPOROBOT Text (pour PocketTopo)
// 16/12/2014 : Fixation d'un bug dans LoadFileText (fonction deprecated)
// 03/02/2015 : Suppression de stations et de leurs visées rayonnantes associées
// 13/02/2015 : Fixation d'un bug dans les chargeurs de fichier (entrée 0 absente = doit être créée)
// 31/03/2015 : Correction dans SaveToTab (section -9)
// 12/05/2015 : Changement de numéro de série met à jour les visées rayonnantes
// 13/05/2015 : Correction dans l'export VTopo
// 18/05/2015 : corrections dans l'export/import Toporobot Text
// 21/05/2015 : Nettoyage des antennes: Suppression des antennes accrochées à un point de la forme nnnn.0
// 29/05/2015 : Correction dans le chargement/sauvegrd XML: Pb de secteurs
// 04/06/2015 : On passe aux génériques pour certaines listes ^^
// 16/07/2015 : Fonctions de correction d'erreur systématique des instruments: OK avec les Grottes du Roy
// 22/02/2016 : Fonction FindPtTopoByCle(): Recherche par occurrence exacte ou texte contenant la clé
// 24/04/2016 : Suppression de PutSerie(); inutile: un TObjSerie est un pointeur, toute modif validée dans un TObjSerie est répercutée dans la base.
// 20/07/2016 : Les étiquettes de terrain et alias de stations topo sont gérés dans une table à part entière
// 11/08/2016 : Prise en compte du paramètre USE_GENERIC_TYPES
//              pour compilation FPC 2.6 sous ARM Raspberry
// 25/12/2016 : Tables FTableStylesAnnotations et FTableAnnotations pour les annotations
// 04/08/2017 : Support des notes de version
// 20/09/2018 : Section nouvelle: Filtres personnalisés
//
//-----------------------------------------------------
//                     Lecture
//                     |  Ajout
//                     |  |  Modif
//                     |  |  |  Suppr
//                     |  |  |  |  Count
//                     |  |  |  |  |  Tri
//                     |  |  |  |  |  |  Recherche
// Table entrées       X  X  X  X  X  0  X
// Table réseaux       X  X  X  X  X  0  X
// Table secteurs      X  X  X  X  X  0  X
// Table expés         X  X  X  X  X  X  X
// Table codes         X  X  X  X  X  X  X
// Table antennes      X  X  X  X  X  0  0
// Table séries        X  X  X  X  X  X  X
//        |-Stations   X  X  X  X  X  0  0
// Table Antennes      X  X  X  X  X  0  0
//
// LoadFromTab()       OK
// LoadFichierText     OK
// SaveToFile()        OK
// LoadFromXML         OK
// SaveToXML           OK
// Export THERION      OK   /!\ A valider, tests OK à raison de la lecture par Loch
// Export Visual Topo: OK   /!\ Visual Topo est très peu tolérant.

// Export Therion: Etat au 05/03/2014:
// - Toporabot OK
// - Charentais OK
// - Saint Marcel OK
// - Shuanghe OK
// TODO: Gestion des projections géographiques
// 10/06/2014: Correction dans le format XML (les éléments 0 ne sont pas exportés et sont reconstruits avant lecture)
// 07/10/2016 : Mise en place d'un callback pour affichage de la progression (refonte de l'interface)
// 13/10/2016 : Fixation d'un crash lié à l'absence de la série numéro 1.
//              Remède: La série est recréée automatiquement avec avertissement.
// 27/07/2017 : Unification des getters et setters pour séries, codes, expés
//              les numéros de série, d'expé, de code, de réseau et de secteur deviennent des types scalaires (entiers)

{$INCLUDE CompilationParameters.inc}
interface

uses
  {$INCLUDE SelectionLangues.inc} // insère les unités en fonction de la langue
  StructuresDonnees,
  {$IFDEF USE_GENERIC_TYPES}
  UnitListesSimplesWithGeneriques,
  {$ELSE}
  UnitListesSimplesWithoutGeneriques,
  {$ENDIF}
  unitobjetserie,
  Common,
  FileUtil, dateutils,
  UnitClassPalette,
  math,
  Classes,
  SysUtils,
  Graphics,
  UnitTSAGeoMag,
  DOM, XMLWrite, XMLRead;
// REFONTE DU CONTENEUR TTOPOROBOTSTRUCTURE
//------------------------------------------------------------------------------
type

{ TToporobotStructure2012 }

 TToporobotStructure2012 = class
    //constructor Create(AOwner: TObject);
    //destructor Free;
    constructor Create;
  private
    // callback pour suivi de progression
    FProcDispProgression: TProcDisplayProgression;
    // général
    FDatabaseName             : TStringDirectoryFilename;
    FNameSpace                : string; // espace de noms       ; ex: TP19
    FNomEtude                 : string; // nom du complexe      ; ex: Puts dets Tachous
    FCommentaireEtude         : string;
    FSystemeDeCoordonneesCodeEPSG : TLabelSystemesCoordsEPSG;

    // coordonnées et station de référence
    FPositionPointZero: TPoint3Df;
    FRefSer        : integer;
    FRefPt         : integer;
    // listes simples
    FTableEntrees : TTableEntrees;   //TODO: Ajouter un champ 'ID littéral'

    FTableSecteurs: TTableSecteurs;
    FTableReseaux : TTableReseaux;
    FTableExpes   : TTableExpes;
    FTableCodes   : TTableCodes;
    FTableViseesAntenne       : TTableViseesAntenne;
    FTableStylesAnnotations   : TTableStylesAnnotations;
    FTableAnnotations         : TTableAnnotations;
    // filtres personnalisés
    FTableFiltresPersonnalises: TTableFiltresPersonnalises;
    // Points d'intérêt
    FTablePointsOfInterest    : TListePointsOfInterest;

    // liste des IDs terrain
    FTableIDsTerrain   : TTableIDTerrain;

    // liste des séries
    FListeDesSeries : TList;//TList;
    // dernier texte cherché
    FLastFindTextWhat      : string;
    // dernières occurrences trouvées
    FLastFoundIdxEntrances : integer;
    FLastFoundIdxReseaux   : integer;
    FLastFoundIdxSecteurs  : integer;
    FLastFoundIdxExpes     : integer;
    FLastFoundIdxCodes     : integer;
    FLastFoundIdxSeries    : integer;

    // série et station courants
    FCurrentNumeroSerie     : TNumeroSerie;
    FCurrentIndexStation   : integer;
    // pour mises en surbrillance
    FCurrentNumeroExpe     : integer;
    FCurrentNumeroCode     : integer;
    FCurrentNumeroReseau   : integer;
    FCurrentNumeroSecteur  : integer;

    function VerifieSerie1(): boolean;
  public

    procedure Finaliser();
    procedure SetCurrentNumeroSerie(const S: TNumeroSerie);
    function  GetCurrentNumeroSerie(): TNumeroSerie;
    procedure SetCurrentIndexPointOfSerie(const S: integer);
    procedure SetCurrentIndexSeriePoint(const S, P: integer);
    function  GetCurrentIndexPointOfSerie(): integer;
    procedure SetCurrentNumeroCode(const C: TNumeroCode);
    procedure SetCurrentNumeroExpe(const C: TNumeroExpe);
    procedure SetCurrentNumeroReseau(const C: integer);
    procedure SetCurrentNumeroSecteur(const C: integer);
    function  GetCurrentNumeroCode(): TNumeroCode;
    function  GetCurrentNumeroExpe(): TNumeroExpe;
    function  GetCurrentNumeroReseau(): TNumeroReseau;
    function  GetCurrentNumeroSecteur(): TNumeroSecteur;

    procedure ReInitialiser(const DoCreateItemsZero: boolean);
    procedure SetProcDisplayProgression(const ProcProgression: TProcDisplayProgression);
    function  GetProcDisplayProgression(): TProcDisplayProgression;
    procedure ViderTablesSimples();
    procedure ViderListeViseesAntennes();
    // s'agit-il d'un document GHTopo étendu ?
    function IsGHTopoCodesInstrumentsExtended(): boolean;
    // chargement document topo
    function  LoadFichierTab(const FichierTAB: TStringDirectoryFilename): integer;
    function  LoadFromXML(const FichierXML: TStringDirectoryFilename): integer;
    function  LoadFichierText(const FichierText: TStringDirectoryFilename): integer; deprecated;
    // LoadFromPocketTopoTXT() a été supprimé: non utilisé; PocketTopo exporte en Toporobot Text

    // sauvegarde document
    procedure SaveToFile(const FichierTAB: TStringDirectoryFilename;
                         const ModeSaveTAB: TModeSaveTAB;
                         const TextFileFormat: TTextFileFormat);
    procedure SaveToXML(const FichierXML: TStringDirectoryFilename);
    // section General
    procedure SetNameSpace(const S: string);
    procedure SetNomEtude(const S: string);
    procedure SetCommentairesEtude(const S: string);
    procedure SetDatabaseName(const S: string);
    function  GetNameSpace(): string;
    function  GetNomEtude(): string;
    function  GetCommentairesEtude(): string;
    procedure SetRefSeriePoint(const Ser, Pt: integer);
    // systèmes de coordonnées
    procedure SetCodeEPSGSystemeCoordonnees(const CodeEPSG: TLabelSystemesCoordsEPSG); overload;
    procedure SetCodeEPSGSystemeCoordonnees(const qCodeEPSG: integer; const qNomEPSG: string); overload;
    function  GetCodeEPSGSystemeCoordonnees(): TLabelSystemesCoordsEPSG;


    // gestion des entrées
    procedure AddEntree(const AEntree: TEntrance);
    function  GetEntree(const NumEntree: integer): TEntrance;
    procedure PutEntree(const NoEntree: integer; const AEntree: TEntrance);
    function  RemoveEntree(const Idx: integer): boolean; inline;
    function  GetNbEntrees(): integer;
//    function  getNextValidIdxEntree(): integer;
    function  CreateNouveauEntree(): boolean;

    // gestion des réseaux
    procedure AddReseau(const AReseau: TReseau); inline;
    function  GetReseau(const NumReseau: integer): TReseau; inline;
    procedure PutReseau(const NoReseau: integer; const AReseau: TReseau); inline;
    function  RemoveReseau(const Idx: integer): boolean; inline;
    function  GetNbReseaux(): integer; inline;
    function  CreateNouveauReseau(): boolean;
    function  GetLastReseau(): TReseau;
    function  HasNumeroReseau(const Idx: integer): boolean;
    // styles d'annotations: c'est un simple tableau à 10 éléments

    function MakeTStyleAnnotation(const QStyleName  : string;
                                  const QFontName   : string;
                                  const QFontColor  : TColor;
                                  const QBackColor    : TColor;
                                  const QFontStyle    : TFontStyles;
                                  const QHauteurTexte : double;
                                  const QPosition     : byte;
                                  const QAngleRot     : integer): TTexteAttributs;

    procedure InitTableStylesAnnotations();
    function  GetStyleAnnotation(const Idx: integer): TTexteAttributs;
    procedure PutStyleAnnotation(const Idx: integer; const QT: TTexteAttributs);
    function  GetNbStylesAnnotations(): integer;


    // annotations
    procedure AddAnnotation(const Annotation: TAnnotation); inline;
    function  GetAnnotation(const NumAnnotation: integer): TAnnotation; inline;
    procedure PutAnnotation(const NoAnnotation: integer; const AAnnotation: TAnnotation); inline;
    function  RemoveAnnotation(const Idx: integer): boolean; inline;
    function  GetNbAnnotations(): integer; inline;
    function  GetLastAnnotation(): TAnnotation;

    // filtres personnalisés
    procedure AddFiltrePerso(const FiltrePerso: TFiltrePersonnalise); inline;
    function  GetFiltrePerso(const NumFiltrePerso: integer): TFiltrePersonnalise; inline;
    procedure PutFiltrePerso(const NoFiltrePerso: integer; const AFiltrePerso: TFiltrePersonnalise); inline;
    function  RemoveFiltrePerso(const Idx: integer): boolean; inline;
    function  GetNbFiltresPersos(): integer; inline;
    function  GetLastFiltrePerso(): TFiltrePersonnalise;
    procedure ViderListeFiltres();
    // points d'intérêt (POI)
    procedure AddPointOfInterest(const APOI: TPointOfInterest); inline;
    function  GetNbPointsOfInterests(): integer; inline;
    function  GetPointOfInterest(const Idx: integer): TPointOfInterest; inline;
    procedure PutPointOfInterest(const Idx: integer; const APOI: TPointOfInterest); inline;
    function  RemovePointOfInterest(const Idx: integer): boolean; inline;
    procedure RecenserPointsOfInterest();

    // gestion des secteurs
    procedure AddSecteur(const ASecteur: TSecteur); inline;
    function  GetSecteur(const NumSecteur: integer): TSecteur; inline;
    procedure PutSecteur(const NoSecteur: integer; const ASecteur: TSecteur); inline;
    function  RemoveSecteur(const Idx: integer): boolean; inline;
    function  GetNbSecteurs(): integer; inline;
    function  CreateNouveauSecteur()  : boolean;
    function  GetLastSecteur(): TSecteur;
    function  HasNumeroSecteur(const Idx: integer): boolean;

    // gestion des expés
    procedure AddExpe(const LaExpe: TExpe); inline;
    function  GetExpe(const Idx: integer): TExpe; inline;
    procedure PutExpe(const Idx: integer; const LaExpe: TExpe); inline;
    function  RemoveExpe(const Idx: integer): boolean; inline;
    function  GetExpeByNumero(const NumeroExpe: TNumeroExpe): TExpe; inline;
    function  GetNbExpes(): integer; inline;
    function  CreateNouveauExpe()  : boolean;
    function  GetLastExpe(): TExpe;

    // gestion des codes
    procedure AddCode(const LeCode: TCode); inline;
    function  RemoveCode(const Idx: integer): boolean; inline;
    procedure PutCode(const Idx: integer; const LeCode: TCode); inline;
    function  GetCode(const Idx: integer): TCode; inline;
    function  GetCodeByNumero(const NumeroCode: TNumeroCode): TCode; inline;
    function  GetNbCodes(): integer;
    function  CreateNouveauCode()  : boolean;
    function  GetLastCode(): TCode;

    // gestion des antennes
    procedure AddViseeAntenne(const VA: TViseeAntenne); inline;
    function  GetViseeAntenne(const No: integer): TViseeAntenne; inline;
    procedure PutViseeAntenne(const No: integer; const VA: TViseeAntenne); inline;
    function  RemoveViseeAntenne(const No: integer): boolean; inline;
    function  GetNbAntennes(): integer; inline;
    function  NettoyerAntennes(const longMax: double; const qIdxSerie, qIdxStation: integer): integer;
    function  ExtractViseesAntennesOfStation(const NoSerie: TNumeroSerie;
                                             const NoStation: Integer;
                                             out ArrAntennesFound: TArrayOfTViseeAntenne): integer;
    {$IFDEF SUPPORT_NOTES_VERSION}
    // gestion des notes de version (commun avec GHCaveDraw)
    procedure AddNoteVersion(const VA: TReleaseNotes); inline;
    function  GetNoteVersion(const No: integer): TReleaseNotes; inline;
    procedure PutNoteVersion(const No: integer; const VA: TReleaseNotes); inline;
    function  GetNbNotesVersion(): integer; inline;
    function  RemoveNoteVersion(const Idx: integer): boolean; inline;
    {$ENDIF SUPPORT_NOTES_VERSION}

    // vérifier si la visée en antenne possède une station de rattachement
    // utilise l'IDTerrain, la série et le point
    // /!\ Modifie la visée passée en paramètre
    function  HasStationForAntennaShot(var VA: TViseeAntenne): boolean;
    // OuvrirAntennesDepuisPocketTopo() supprimé: inutilisé
    // gestion des séries
    procedure ClearListeSeries();
    //function  CreateNewSerie(const SD, PD: integer; const QNomSerie: string): boolean;
    function  CreateNewSerie(const SD, PD: integer;
                             const QNumSerie: TNumeroSerie;
                             const QNomSerie: string): boolean;

    function  GetNbSeries(): integer; inline;
    procedure AddSerie(const S: TObjSerie);
    // /!\ Retourne un POINTEUR sur un objet TObjSerie
    function  GetSerie(const Idx: integer): TObjSerie;
    function  GetSerieByNumeroSerie(const Idx: TNumeroSerie; out SR: TObjSerie; out InternalIdxSerie: integer): boolean;
    function  FindIndexesSeriesRattachedAtSerie(const Idx: integer; out LS: TListOfIntegers): boolean;

    //procedure PutSerie(const Idx: integer; const S: TObjSerie);
    //procedure PutLastSerie(const S: TObjSerie);
    function  RemoveSerie(const Idx: integer): boolean;
    function  GetLastSerie(): TObjSerie;
    function  ExistsSerie(const QIndexSerie: TNumeroSerie): boolean;
    // prochains index valides
    function  getMaxIdxCode(): TNumeroCode;
    function  getMaxIdxExpe(): TNumeroExpe;
    function  getMaxIdxSerie() : TNumeroSerie;
    function  getMaxIdxSecteur(): integer;
    function  getMaxIdxReseau(): integer;



    // recherche
    function  FindIdxEntranceByText(const S: string): integer;
    function  FindIdxReseauByText(const S: string): integer;
    function  FindIdxSecteurByText(const S: string): integer;
    function  FindIdxCodeByText(const S: string): integer;
    function  FindIdxExpeByText(const S: string): integer;
    function  FindIdxSerieByText(const S: string): integer;
    // procédures de tri
    procedure SortExpes();
    procedure SortCodes();
    procedure SortSeries();
    function  SortAntennesByLengths(): integer;
    function  SortAntennesBySerieStation(): integer;

                                                // coordonnées par défaut
    function  GetDatabaseName: TStringDirectoryFilename;
    procedure SetDefaultCoords(const X, Y, Z: double); overload;
    procedure SetDefaultCoords(const P0: TPoint3Df); overload;
    function  GetPositionDuPointZero(): TPoint3Df;


    // fonctions de recherche
    function FindPtTopoByCle(const DoMatchExact: boolean;
                             var Cle: string;
                             var Sr, Pt: integer): boolean;
                             //out ListeStationsFound: TArrayStationsMatchFound): boolean;
    function HasPtTopoBySerSt(const qSr, qPt: integer): boolean;
    // aller à un index d'enregistrement depuis son numéro (pouvant être différent de l'index)
    function GetIdxEntreeByNumero(const N: TNumeroEntrance): integer; inline; // simple wrapper
    function GetIdxReseauByNumero(const N: TNumeroReseau): integer; inline; // simple wrapper
    function GetIdxSecteurByNumero(const N: TNumeroSecteur): integer; inline; // simple wrapper

    function GetIdxCodeByNumero(const N: TNumeroCode): integer;
    function GetIdxExpeByNumero(const N: TNumeroExpe): integer;

    function GetIdxSerieByNumero(const qIdx: TNumeroSerie): integer;

    function GetRefPoint(): integer;
    function GetRefSerie(): integer;
    // infos générales sur les données de la cavité
    function GetInfosCavite(const LS: TStrings): integer;
    // export vers autres logiciels
    procedure ExportVisualTopo(const FichierTRO: TStringDirectoryFilename; const DoExportViseesRadiantes: boolean);  // Visual Topo
    procedure ExportVersTopodroid(const FichierCSV: TStringDirectoryFilename); // Android Abris


    procedure ExporterVersToporobotTEXT(const QFileName: TStringDirectoryFilename;
                                        const LongueurMaxAntenne: double;
                                        const DoExportAntennaShots: boolean); deprecated;   // Toporobot TEXT
    // génération d'un dossier complet Therion
    procedure GenererThConfig(const THC, FTH: string);
    procedure GenererDossierTherion(const DossierTherion: TStringDirectoryFilename);
    procedure ExportVersTherion(const FichierTH: string);  // Therion .th file
    // check des valeurs
    function  CheckAzimutByIndex(const Idx: TNumeroCode; const QAz: double; out MsgErr: string): boolean;
    function  CheckPenteByIndex(const Idx: TNumeroCode; const QP: double; out MsgErr: string): boolean;
    // test d'existence des éléments
//    function  ExistsIdxEntrance(const Idx: TNumEntrance): boolean;
    function  ExistsIdxCode(const Idx: integer): boolean; inline;
    function  ExistsIdxExpe(const Idx: integer): boolean; inline;
    function  ExistsIdxReseau(const Idx: integer): boolean; inline;
    function  ExistsIdxEntrance(const Idx: integer): boolean; inline;

    function  GetLastFindText(): string; inline;
    // export de listes en CSV
    procedure ExportListeEntreesCSV(const FileName: TStringDirectoryFilename);
    procedure ExportListeReseauxCSV(const FileName: TStringDirectoryFilename);
    procedure ExportListeSecteursCSV(const FileName: TStringDirectoryFilename);
    procedure ExportListeCodesCSV(const FileName: TStringDirectoryFilename);
    procedure ExportListeExpesCSV(const FileName: TStringDirectoryFilename);
    procedure ExportListeEntetesSeries(const FileName: TStringDirectoryFilename);
    procedure ExportListeAnnotationsCSV(const FileName: TStringDirectoryFilename);
    procedure ExportSeriesCSV(const QFileName: TStringDirectoryFilename);
    // check des données
    procedure CheckerLesDonneesTopo();
    procedure CheckerLesEntrees();
    procedure CheckerLesSeries();
    procedure CheckerLesAntennes();

    function CheckSerie(const MySerie: TObjSerie; const QIdx: integer): integer; // retourne: -1 si série vide; numéro de visée si erreur, 0 si OK
    function CheckVisee(const V: TUneVisee; out MsgErr: string): TErrorVisee;
    // recalcul des déclinaisons magnétiques
    procedure RecalculerDeclinaisonsMagnetiquesExpes();

    // supprimme toutes les visées rayonnantes d'une station
    function RemoveViseesRayonnantesAssocieesAStation(const qSer, qSt: integer): integer;
    // renommer les points d'origine des visées rayonnantes associées à une station
    function DeplacerOrigineViseesRayonnantes(const OldSerie, OldPoint, NewSerie, NewPoint: integer): integer;
    function ReAttribuerViseesRayonnantesDeUneSerie(const SerieCourante: TObjSerie; const OldNumSerie: integer): integer;
    function UpdateSerPtExtremsSeriesRattachees(const OldNumSerie,NewNumSerie: integer): integer;

    // gestion des IDs terrain
    procedure RecenserLesIDsTerrain();
    function  FindIdxIDTerrainByText(const Str: string): integer;
    procedure AddLabelTerrain(const ALabelTerrain: TEtiquetteTerrain);
    function  GetLabelTerrain(const NumLabelTerrain: integer): TEtiquetteTerrain;
    function  GetNbLabelTerrain: integer;
    procedure PutLabelTerrain(const NoLabelTerrain: integer; const ALabelTerrain: TEtiquetteTerrain);
    function  RemoveLabelTerrain(const Idx: integer): boolean;

    // ajout de données depuis fichiers CSV
    procedure AddEntrancesFromFile(const FileName: TStringDirectoryFilename; const DoReplaceExisting: boolean);
    procedure AddReseauxFromFile(const FileName: TStringDirectoryFilename; const DoReplaceExisting: boolean);
    procedure AddSecteursFromFile(const FileName: TStringDirectoryFilename; const DoReplaceExisting: boolean);
    procedure AddCodesFromFile(const FileName: TStringDirectoryFilename; const DoReplaceExisting: boolean);
    procedure AddExpesFromFile(const FileName: TStringDirectoryFilename; const DoReplaceExisting: boolean);
    procedure AddAntennesFromFile(const FileName: TStringDirectoryFilename; const DoReplaceExisting: boolean);


    // réattribution des codes, réseaux et expés pour les visées en antennes
    function ReattribuerCodesExpesViseesAntennes(): boolean;
    // fonctions de remplacement d'index
    procedure ReplaceIndexSeriesInEntrances(const OldIndex, NewIndex: TNumeroSerie);
    procedure ReplaceIndexInStations(const OldIndex, NewIndex: integer; const ModeRemplacement: TModeRemplacementStations);
    procedure ReplaceIndexInAntennes(const OldIndex, NewIndex: integer;  const ModeRemplacement: TModeRemplacementAntennes);
    procedure ReplaceIdxCodes(const OldIndex, NewIndex: TNumeroCode);
    procedure ReplaceIdxExpes(const OldIndex, NewIndex: TNumeroExpe);
    procedure ReplaceIdxSeries(const OldIndex, NewIndex: TNumeroSerie);
end;
//------------------------------------------------------------------
procedure RegenererEbaucheXTB(const NomEbauche: string; const DefaultCodeEPSG: integer);

implementation
uses
  // pour la déclinaison magnétique
  {$IFDEF MSWINDOWS}
    UnitWrapperDeclimag;
  {$ENDIF}
  {$IFDEF LINUX}
    UnitWrapperDeclimagLinux;
  {$ENDIF}
  {$IFDEF MAC_OS_X_VERSION_MIN_REQUIRED}
    UnitWrapperDeclimagLinux;
  {$ENDIF}

  //********************
  // Inclusion des constantes GtxKey
  {$INCLUDE GTX_Keys.inc}
const FMTSTS = 's%d_%d';

// Régénération d'un stub topo
procedure RegenererEbaucheXTB(const NomEbauche: string; const DefaultCodeEPSG: integer);
const
  QFMT_SERIE    = '%d' + TAB + '%d' + TAB +
                  '%d' + TAB + '%d' + TAB + '%d' + TAB + '%d' + TAB +
                  '%d' + TAB +
                  '%d' + TAB + '%d' + TAB +
                  '%s' + TAB + '%s' + TAB +
                  '%d' + TAB + '%.2f';
  QLINE_STATION = '%d'+TAB+'%d'+TAB+
                 '%d'+TAB+'%d'+TAB+
                 '%.2f'+TAB+'%.2f'+TAB+'%.2f'+TAB+
                 '%.2f'+TAB+'%.2f'+TAB+'%.2f'+TAB+'%.2f'+TAB+
                 '%s' + TAB+'%s'+TAB+'%d'+TAB+'%d';

var
  fp: TextFile;
  procedure WriteLigne(const L: string);
  begin
    WriteLn(fp, L);
  end;
  procedure WriteVisee(const QIdxSerie, QNoSt, QCode, QExpe: integer;
                       const QLong, QAz, QPente, QLG, QLD, QHZ, QHN: double;
                       const QCommentaire, QIDTerrain: string;
                       const QTypeVisee, QIDSecteur: integer);
  begin
    WriteLigne(Format(QLINE_STATION,
                                  [QIdxSerie, QNoSt,
                                   QCode, QExpe,
                                   QLong, QAz, QPente,
                                   QLG, QLD, QHZ, QHN,
                                   QCommentaire,
                                   QIDTerrain,
                                   QTypeVisee,
                                   QIDSecteur
                                  ]));

  end;
begin
  AfficherMessage(Format('RegenererEbaucheXTB: %s', [NomEbauche]));
  AssignFile(fp, NomEbauche);
  try
    ReWrite(fp);
    WriteLigne('-20' + #9 + '12/11/2014' + #9 + '10:30:00' + #9 + 'CaveStub.xtb');
    WriteLigne('-19' + #9 + 'Ebauche de document - Document stub');
    WriteLigne('-15' + #9 + format('%d', [DefaultCodeEPSG]));
    WriteLigne('-6'  + #9 + '1' + #9 + 'Entree principale');
    WriteLigne('-5'  + #9 + '1' + #9 + '0.00' + #9 + '0.00' + #9 + '0.00' + #9 + '1' + #9 + '0');
    WriteLigne('-2'  + #9 + '1' + #9 + '12' + #9 + '11' + #9 + '2014' + #9 + 'Equipe1' + #9 + 'Equipe1' + #9 + '0' + #9 + '0.00' + #9 + '0' + #9 + '39');
    WriteLigne('-1'  + #9 + '1' + #9 + '360.00' + #9 + '360.00' + #9 + '0.01' + #9 + '1.00' + #9 + '1.00' + #9 + '0.00' + #9 + '0.00' + #9 + 'Code01' + #9 + '0');
    //1	-1	1	0	1	2	2	0	0	Serie principale	A modifier	0	0.01
    WriteLigne(Format(QFMT_SERIE, [1, -1,
                                   1, 0, 1, 2,
                                   2,
                                   0, 0,
                                   'Serie principale', '', 0, 0.01]));

    WriteVisee(1, 0, 1, 1, 0.01, 135.00, -45.00, 0.02, 0.02, 0.02, 0.02, 'Point 1.0', '1.0', 0, 0);
    WriteVisee(1, 1, 1, 1, 0.05, 135.00, -45.00, 0.02, 0.02, 0.02, 0.02, '', '1.1', 0, 0);
    WriteVisee(1, 2, 1, 1, 0.05, 135.00, -45.00, 0.02, 0.02, 0.02, 0.02, '', '1.2', 0, 0);

  finally
    CloseFile(fp);
  end;
end;

// Routines internes non incorporées dans les objets
function SortCodesCriteria(Item1, Item2: Pointer): Integer;
var
  E1, E2: ^TCode;
begin
  E1:=Item1;
  E2:=Item2;
  if (E1.IDCode<E2.IDCode) then
    Result:=-1
  else if (E1.IDCode=E2.IDCode) then
    Result:=0
  else
    Result:=1;
end;
// Trier les expés par ordre croissant
function SortExpesCriteria(Item1, Item2: Pointer): Integer;
var
  E1, E2: ^TExpe;
begin
  E1:=Item1;
  E2:=Item2;
  if (E1.IDExpe < E2.IDExpe) then
    Result:=-1
  else if (E1.IDExpe=E2.IDExpe) then
    Result:=0
  else
    Result:=1;
end;
// Trier les séries par ordre croissant
function SortSerieCriteria(Item1, Item2: Pointer): Integer;
var
  E1, E2: ^TObjSerie;
begin
  E1:=Item1;
  E2:=Item2;
  if (E1.GetNumeroDeSerie() < E2.GetNumeroDeSerie()) then
    Result:=-1
  else if (E1.GetNumeroDeSerie() = E2.GetNumeroDeSerie()) then
    Result:=0
  else
    Result:=1;
end;

// Trier les antennes par ordre de longueur croissante
function SortViseeAntenneCriteriaLength(Item1, Item2: Pointer): Integer;
var
  E1, E2: ^TViseeAntenne;
begin
  E1:=Item1;
  E2:=Item2;
  if (E1.Longueur < E2.Longueur) then
    Result := -1
  else if (E1.Longueur = E2.Longueur) then
    Result := 0
  else
    Result := 1;
end;
function SortViseeAntenneCriteriaSerSt(Item1, Item2: Pointer): Integer;
const
  Q_MULT_SER = 100000;
var
  E1, E2: ^TViseeAntenne;
  QSerStE1, QSerStE2: Int64;
begin
  E1:=Item1;
  E2:=Item2;

  QSerStE1 := Q_MULT_SER * Integer(E1.SerieDepart) + E1.PtDepart;
  QSerStE2 := Q_MULT_SER * Integer(E2.SerieDepart) + E2.PtDepart;


  if (QSerStE1 < QSerStE2) then
    Result := -1
  else if (QSerStE1 = QSerStE2) then
    Result := 0
  else
    Result := 1;
end;

// TToporobotStructure2012
constructor TToporobotStructure2012.Create;//(AOwner: TObject);
begin
  inherited Create;
  // current
  FCurrentNumeroSerie   := 0;
  FCurrentIndexStation  := 0;
  FCurrentNumeroExpe    := 0;
  FCurrentNumeroCode    := 0;
  FCurrentNumeroReseau  := 0;
  FCurrentNumeroSecteur := 0;
  // point zéro
  FPositionPointZero := MakeTPoint3Df(0.00, 0.00, 0.00);
  AfficherMessage(Format('%s.Create', [self.ClassName]));
  FListeDesSeries := TList.Create;
  FListeDesSeries.Clear;
  InitTableStylesAnnotations();
  // tables simples
  FTableEntrees := TTableEntrees.Create;
  FTableReseaux := TTableReseaux.Create;
  FTableSecteurs:= TTableSecteurs.Create;
  FTableExpes   := TTableExpes.Create;
  FTableCodes   := TTableCodes.Create;
  FTableViseesAntenne := TTableViseesAntenne.Create;
  FTableIDsTerrain    := TTableIDTerrain.Create;
  FTableAnnotations   := TTableAnnotations.Create;
  FTableFiltresPersonnalises := TTableFiltresPersonnalises.Create;
  FTablePointsOfInterest     := TListePointsOfInterest.Create;
  {$IFDEF SUPPORT_NOTES_VERSION}
  FTableNotesVersion  := TTableNotesVersion.Create;
  {$ENDIF SUPPORT_NOTES_VERSION}
  AfficherMessage('-- Tables created');
end;



procedure TToporobotStructure2012.ReInitialiser(const DoCreateItemsZero: boolean);
var
  UneEntree       : TEntrance;
  UnReseau        : TReseau;
  UnSecteur       : TSecteur;
  UneExpe         : TExpe;
  UnCode          : TCode;
  UneSerie        : TObjSerie;
  UneViseeAntenne : TViseeAntenne;
begin
  ClearListeSeries;
  ViderTablesSimples;
  // numéros d'items courants
  FCurrentNumeroSerie    := 0;
  FCurrentNumeroExpe     := 0;
  FCurrentNumeroCode     := 0;
  FCurrentNumeroReseau   := 0;
  FCurrentNumeroSecteur  := 0;

  // dernières occurrences trouvées
  FLastFindTextWhat      := '';
  FLastFoundIdxEntrances := -1;
  FLastFoundIdxReseaux   := -1;
  FLastFoundIdxSecteurs  := -1;
  FLastFoundIdxExpes     := -1;
  FLastFoundIdxCodes     := -1;
  FLastFoundIdxSeries    := -1;
  if (Not DoCreateItemsZero) then Exit;
  // Général
  SetNomEtude('SansNom');
  SetNameSpace('Namespace00');
  SetCodeEPSGSystemeCoordonnees(DEFAULT_SYSTEME_COORDONNEES_CODE_EPSG, DEFAULT_SYSTEME_COORDONNEES_NOM);
  SetCommentairesEtude('');
  // Coordonnées du point zéro = celles de l'entrée principale
  SetDefaultCoords(0.00, 0.00, 0.00);
  if (self.GetNbEntrees() > 0) then
  begin
    UneEntree := GetEntree(0);
    SetDefaultCoords(UneEntree.eXEntree, UneEntree.eYEntree, UneEntree.eZEntree);
  end;
  // Réseau 0
  UnReseau := MakeTReseau(COULEUR_RESEAU_0, 0, rsMAIN_NETWORK, '');
  AddReseau(UnReseau);
  // Secteur 0
  UnSecteur := MakeTSecteur(clSilver, 'Undefined');
  AddSecteur(UnSecteur);
  // Expé 0
  with UneExpe do
  begin
    IDExpe:=0;
    JourExpe      := 01;
    MoisExpe      := 01;
    AnneeExpe     := 00;
    Operateur     := '';             // spéléomètre
    ClubSpeleo    := '';             // spéléographe
    ModeDecl      := 0;   // déclinaison auto ?
    Declinaison   := 0.00; // déclinaison
    //Inclinaison   := 0.00; // correction clino x10
    Couleur       := 0;  // couleur
    Commentaire   := rsFORFIXEDPOINTS;            // commentaire
  end;
  AddExpe(UneExpe);
  // Code 0
  with UnCode do
  begin
   IDCode          := 0;
   GradAz          := 400.00;
   GradInc         := 400.00;
   PsiL            := 0.00;
   PsiAz           := 0.00;
   PsiP            := 0.00;
   FactLong        := 1.00;
   AngLimite       := 0.00;
   ErreurTourillon := 0.00;
   //ReservedInt     := 0;
   Commentaire     := rsFORFIXEDPOINTS;
   ParamsFuncCorrAz  := MakeTParamFoncCorrectionAngulaire(0.00, 0.00, 0.00);
   ParamsFuncCorrInc := MakeTParamFoncCorrectionAngulaire(0.00, 0.00, 0.00);
  end;
  AddCode(UnCode);
  // Antennes
  with UneViseeAntenne do
  begin
    IDViseeAntenne   := 0;
    Secteur          := 0;
    Reseau           := 0;
    SerieDepart      := 1;
    PtDepart         := 0;
    Code             := 1;
    Expe             := 1;
    IDTerrainStation := '';
    Longueur         := 0.001;
    Azimut           := 0.00;
    Pente            := 0.00;
    Commentaires     := '';
  end;
  AddViseeAntenne(UneViseeAntenne);
  // série
  //Serie 0
  UneSerie := TObjSerie.Create;
  with UneSerie do begin
    ClearStations;
    SetNumeroReseau(0);
    SetNomObsSerie(rsSERIE_INITIALISATION, '');
    SetNumeroSerie(0);
    SetSeriePtExtremites(0,0,1,0);
    SetChanceObstacle(0,0);
    SetRaideur(1.00);
    AddVisee(EmptyVisee('Point 1/0'));
  end;
  AddSerie(UneSerie);
end;


procedure TToporobotStructure2012.ViderTablesSimples();
var i: integer;
begin
  AfficherMessage('');
  AfficherMessage('Vidage des tables simples');
  try
    FTableReseaux.ClearListe();
    FTableSecteurs.ClearListe();
    FTableCodes.ClearListe();
    FTableEntrees.ClearListe();
    FTableExpes.ClearListe();
    FTableViseesAntenne.ClearListe();
    FTableIDsTerrain.ClearListe();
    FTableAnnotations.ClearListe();
    FTableFiltresPersonnalises.ClearListe();
    FTablePointsOfInterest.ClearListe();
    {$IFDEF SUPPORT_NOTES_VERSION} FTableNotesVersion.ClearListe(); {$ENDIF SUPPORT_NOTES_VERSION}
    AfficherMessage('Vidage des tables simples OK');
  except
    AfficherMessage('Echec purge des tables simples');
  end;
end;
procedure TToporobotStructure2012.ViderListeViseesAntennes();
begin
  FTableViseesAntenne.ClearListe();
end;

// A invoquer avant libération
// SIGSEGV inexpliqué si implanté dans le destructeur
procedure TToporobotStructure2012.Finaliser();
begin
  try
    ViderTablesSimples();
    ClearListeSeries();
    AfficherMessage(' --- Series table cleared');
  finally

    FListeDesSeries.Free;
    // libération des tables simples
    FTableEntrees.Free;
    FTableReseaux.Free;
    FTableSecteurs.Free;
    FTableExpes.Free;
    FTableCodes.Free;
    FTableViseesAntenne.Free;
    FTableIDsTerrain.Free;
    FTableAnnotations.Free;
    FTableFiltresPersonnalises.Free;
    FTablePointsOfInterest.Free;
    {$IFDEF SUPPORT_NOTES_VERSION} FTableNotesVersion.Free; {$ENDIF SUPPORT_NOTES_VERSION}
  end;
end;


function TToporobotStructure2012.GetPositionDuPointZero(): TPoint3Df;
begin
  Result := FPositionPointZero;
end;







procedure TToporobotStructure2012.SetDefaultCoords(const X, Y, Z: double); overload;
begin
  FPositionPointZero := MakeTPoint3Df(X, Y, Z);
end;

procedure TToporobotStructure2012.SetDefaultCoords(const P0: TPoint3Df); overload;
begin
  FPositionPointZero := P0;
end;

function TToporobotStructure2012.GetRefSerie(): integer;
begin
  Result := FRefSer;
end;

function TToporobotStructure2012.GetRefPoint(): integer;
begin
  Result:= FRefPt;
end;

procedure TToporobotStructure2012.SetRefSeriePoint(const Ser, Pt: integer);
begin
  FRefSer := Ser;
  FRefPt  := Pt;
end;

procedure TToporobotStructure2012.SetNomEtude(const S: string);
begin
  FNomEtude := S;
end;
procedure TToporobotStructure2012.SetNameSpace(const S: string);
begin
  FNameSpace := S;
end;

procedure TToporobotStructure2012.SetProcDisplayProgression(const ProcProgression: TProcDisplayProgression);
begin
  FProcDispProgression := ProcProgression;
end;
function TToporobotStructure2012.GetProcDisplayProgression(
  ): TProcDisplayProgression;
begin
  Result := FProcDispProgression;
end;

procedure TToporobotStructure2012.SetCommentairesEtude(const S: string);
begin
  FCommentaireEtude := S;
end;


// la station topo courante
procedure TToporobotStructure2012.SetCurrentIndexPointOfSerie(const S: integer);
var
  SR: TObjSerie;
  QInternalIdxSerie: integer;
begin

  if (S = -1) then // -1 = on attrape la dernière station de la série courante
  begin
    GetSerieByNumeroSerie(FCurrentNumeroSerie, SR, QInternalIdxSerie);
    FCurrentIndexStation := SR.GetNbVisees() - 1;
  end
  else
  begin
    FCurrentIndexStation := S;
  end;
  AfficherMessageErreur(format('SetCurrentIndexPointOfSerie(): %d - %d', [S, FCurrentIndexStation]));
end;

procedure TToporobotStructure2012.SetCurrentNumeroSerie(const S: TNumeroSerie);
begin
  FCurrentNumeroSerie := S;
  AfficherMessageErreur(format('SetCurrentIndexSerie(): %d', [S, FCurrentNumeroSerie]));
end;

procedure TToporobotStructure2012.SetCurrentIndexSeriePoint(const S, P: integer);
begin
  AfficherMessageErreur(Format('SetCurrentIndexSeriePoint: %d.%d', [S, P]));
  self.SetCurrentNumeroSerie(S);
  self.SetCurrentIndexPointOfSerie(P);
end;

procedure TToporobotStructure2012.SetCurrentNumeroCode(const C: TNumeroCode);
begin
  FCurrentNumeroCode := C;
end;

procedure TToporobotStructure2012.SetCurrentNumeroExpe(const C: TNumeroExpe);
begin
  FCurrentNumeroExpe := C;
end;

procedure TToporobotStructure2012.SetCurrentNumeroReseau(const C: integer);
begin
  FCurrentNumeroReseau := C;
end;

procedure TToporobotStructure2012.SetCurrentNumeroSecteur(const C: integer);
begin
  FCurrentNumeroSecteur := C;
end;




function TToporobotStructure2012.GetCurrentIndexPointOfSerie(): integer;
begin
  Result := FCurrentIndexStation;
end;

function TToporobotStructure2012.GetCurrentNumeroCode(): TNumeroCode;
begin
  Result := FCurrentNumeroCode;
end;

function TToporobotStructure2012.GetCurrentNumeroExpe(): TNumeroExpe;
begin
  Result := FCurrentNumeroExpe;
end;

function TToporobotStructure2012.GetCurrentNumeroReseau(): TNumeroReseau;
begin
  Result := FCurrentNumeroReseau;
end;

function TToporobotStructure2012.GetCurrentNumeroSecteur(): TNumeroSecteur;
begin
  Result := FCurrentNumeroSecteur;
end;

function TToporobotStructure2012.GetCurrentNumeroSerie(): TNumeroSerie;
begin
  Result := FCurrentNumeroSerie;
end;

//******************************************************************************
function TToporobotStructure2012.GetNomEtude(): string;
begin
  Result := FNomEtude;
end;


function TToporobotStructure2012.GetCommentairesEtude(): string;
begin
  Result := FCommentaireEtude;
end;
procedure TToporobotStructure2012.SetDatabaseName(const S: string);
begin
  FDatabaseName:=S;
end;

function TToporobotStructure2012.GetNameSpace(): string;
begin
  result := FNameSpace;
end;

function TToporobotStructure2012.GetDatabaseName: TStringDirectoryFilename;
begin
  Result := FDatabaseName;
end;

(* TToporobotStructure2012 *)

//**************************
// système de coordonnées EPSG
function TToporobotStructure2012.GetCodeEPSGSystemeCoordonnees(): TLabelSystemesCoordsEPSG;
begin
  try
    Result := FSystemeDeCoordonneesCodeEPSG;
    AfficherMessage(format('%s.GetCodeEPSGSystemeCoordonnees: %d - %s', [ClassName, FSystemeDeCoordonneesCodeEPSG.CodeEPSG, FSystemeDeCoordonneesCodeEPSG.NomEPSG]));

	except
    on E: Exception do AfficherMessage('In GetCodeEPSGSystemeCoordonnees():' +  E.Message);
	end;
end;

procedure TToporobotStructure2012.SetCodeEPSGSystemeCoordonnees(const CodeEPSG: TLabelSystemesCoordsEPSG); overload;
begin
  FSystemeDeCoordonneesCodeEPSG := CodeEPSG;
end;

procedure TToporobotStructure2012.SetCodeEPSGSystemeCoordonnees(const qCodeEPSG: integer; const qNomEPSG: string); overload;
begin
  FSystemeDeCoordonneesCodeEPSG.CodeEPSG := qCodeEPSG;
  FSystemeDeCoordonneesCodeEPSG.NomEPSG  := qNomEPSG;
end;

//**************************

procedure TToporobotStructure2012.ClearListeSeries();
var
  ii: Integer;
begin
  AfficherMessage(Format('-- %s.ClearListeSeries()',[self.ClassName]));
  if (FListeDesSeries.Count > 0) then
  begin
    try
      for ii := 0 to FListeDesSeries.Count - 1 do
      begin
        // cette construction n'est pas valide:
        // et elle est inutile: un objet est un POINTEUR
        // OO := TObjSerie(FListeDesSeries.Items[ii]);
        // OO.Free;
        // La simple libération par Dispose suffit
        Dispose(FListeDesSeries.Items[ii]);
      end;
    except
    end;
  end;
  //*)

  FListeDesSeries.Clear;
  AfficherMessage(Format('-- %s.ClearListeSeries() OK',[self.ClassName]));
  AfficherMessage('');
end;
function TToporobotStructure2012.CreateNewSerie(const SD, PD: integer;
                                                const QNumSerie: TNumeroSerie;
                                                const QNomSerie: string): boolean;
var
  nb: Integer;
  SR: TObjSerie;
  NSR: TObjSerie;
begin
  Result := False;
  try
    NSR := TObjSerie.Create;
    NSR.ClearStations;
    nb := GetNbSeries;
    SR := GetSerie(nb - 1);
    if      (ExistsSerie(QNumSerie)) then NSR.SetNumeroSerie(1 + getMaxIdxSerie())
    else if (QNumSerie < 0) then  NSR.SetNumeroSerie(1 + getMaxIdxSerie())
    else     NSR.SetNumeroSerie(QNumSerie);

    NSR.SetChanceObstacle(0, 0);
    NSR.SetNomObsSerie(QNomSerie, '');
    NSR.SetRaideur(1.00);
    NSR.SetSeriePtExtremites(SD, PD, NSR.GetNumeroDeSerie(), 0);
    NSR.SetNumeroReseau(SR.GetNumeroReseau());
    NSR.SetCouleur(SR.GetCouleur);
    NSR.AddVisee(0, 0,
                1, 1,
                tgDEFAULT,
                0.001, 0.00, 0.00,
                0.00, 0.00, 0.00, 0.00,
                '', '');
    self.AddSerie(NSR);
    result := True;
  except
  end;
end;
(*old
function TToporobotStructure2012.CreateNewSerie(const SD, PD: integer; const QNomSerie: string): boolean;
var
  nb: Integer;
  SR: TObjSerie;
  NSR: TObjSerie;
begin
  Result := False;
  try
    NSR := TObjSerie.Create;
    NSR.ClearStations;
    nb := GetNbSeries;
    SR := GetSerie(nb - 1);

    NSR.SetNumeroSerie(1 + getMaxIdxSerie());
    NSR.SetChanceObstacle(0, 0);
    NSR.SetNomObsSerie(QNomSerie, '');
    NSR.SetRaideur(1.00);
    NSR.SetSeriePtExtremites(SD, PD, NSR.GetNumeroDeSerie(), 0);
    NSR.SetNumeroReseau(SR.GetNumeroReseau());
    NSR.SetCouleur(SR.GetCouleur);
    NSR.AddVisee(0, 0,
                1, 1,
                tgDEFAULT,
                0.001, 0.00, 0.00,
                0.00, 0.00, 0.00, 0.00,
                '', '');
    self.AddSerie(NSR);
    result := True;
  except
  end;
end;
//*)


function TToporobotStructure2012.GetNbSeries(): integer;
begin
  Result := FListeDesSeries.Count;
end;




procedure TToporobotStructure2012.AddSerie(const S: TObjSerie);
var
  pS: ^TObjSerie;
begin
  New(pS);
  pS^ := S;
  FListeDesSeries.Add(pS);
end;

// /!\ Retourne un POINTEUR sur un objet TObjSerie
// i.e: Toutes les modifications apportées à l'objet retourné par cette fonction
//      sont validées
// Rappel: En Pascal, un TObject EST un pointeur
function TToporobotStructure2012.GetSerie(const Idx: integer): TObjSerie;
var
  pS: ^TObjSerie;
begin
  pS := FListeDesSeries.Items[Idx];
  Result := pS^;
end;
//*)
function TToporobotStructure2012.GetSerieByNumeroSerie(const Idx: TNumeroSerie; out SR: TObjSerie; out InternalIdxSerie: integer): boolean;
var
  i, Nb: Integer;
begin
  Result := false;
  Nb := GetNbSeries;
  //AfficherMessageErreur(Format('%s.GetSerieByIdxSerie: %d series', [ClassName, Nb]));
  if (Nb = 0) then Exit;
  for i := 0 to Nb - 1 do
  begin
    SR := GetSerie(i);
    InternalIdxSerie := i;
    if (Idx = SR.GetNumeroDeSerie()) then Exit(True);
  end;
end;


function TToporobotStructure2012.ExistsSerie(const QIndexSerie: TNumeroSerie): boolean;
var
  n, i: Integer;
  SR: TObjSerie;
begin
  Result := false;
  //AfficherMessageErreur(Format('%s.ExistsSerie(%d) %d', [ClassName, QIndexSerie, GetNbSeries()]));
  n := GetNbSeries();
  if (n = 0) then Exit;
  for i := 0 to n - 1 do
  begin
    try
      SR := GetSerie(i);
      if (QIndexSerie = SR.GetNumeroDeSerie()) then Exit(True);
    except
      AfficherMessageErreur(Format('%s.ExistsSerie(%d): Invalid serie: %d', [ClassName, QIndexSerie, i]));
    end;
  end;
end;

function TToporobotStructure2012.RemoveSerie(const Idx: integer): boolean;
var
  MySerie: TObjSerie;
  n, v: Integer;
begin
  Result := False;
  n := GetNbSeries();
  AfficherMessageErreur(format('%s.RemoveSerie(%d de %d):', [ClassName, Idx, n]));

  try
    try
      MySerie := self.GetSerie(Idx);
      if (MySerie.GetNumeroDeSerie() < 25000) then
      begin
        AfficherMessageErreur(IIF(Assigned(MySerie), 'Objet OK', 'Objet non alloué'));
        AfficherMessageErreur('Vidage stations');
        MySerie.ClearStations();
        MySerie.Free;
      end;
      AfficherMessageErreur('Destruction item liste interne');
      Dispose(FListeDesSeries.Items[Idx]);
      Result := True;
    except
      AfficherMessageErreur(format('%s.RemoveSerie(%d): ****', [ClassName, Idx]));
    end;
  finally
    AfficherMessageErreur('Série détruite');
    FListeDesSeries.Delete(Idx);
  end;
  //*)
end;

function TToporobotStructure2012.GetLastSerie(): TObjSerie;
begin
  Result := GetSerie(FListeDesSeries.Count - 1);
end;








// fonctions de recherche
function TToporobotStructure2012.FindIdxEntranceByText(const S: string): integer;
var
  i, N, P : integer;
  EWE: TEntrance;
begin
  Result := -1;
  N := GetNbEntrees - 1;
  P := IIF((FLastFoundIdxEntrances = -1), 0, FLastFoundIdxEntrances);
  for i := P to N do
  begin
    EWE := GetEntree(i);
    if (Pos(LowerCase(S), LowerCase(EWE.eNomEntree)) > 0) then
    begin
      // mémorise la clé pour recherche de l'occurrence suivante
      FLastFindTextWhat   := S;
      // mémorise l'indice de parcours
      FLastFoundIdxEntrances := i + 1;
      // on boucle depuis le début de la liste
      if (FLastFoundIdxEntrances > N) then FLastFoundIdxEntrances := -1;
      Result := i;
      Exit;
    end;
  end;
end;
function TToporobotStructure2012.FindIdxReseauByText(const S: string): integer;
var
  i, N, P : integer;
  EWE: TReseau;
begin
  Result := -1;
  N := GetNbReseaux - 1;
  P := IIF((FLastFoundIdxReseaux = -1), 0, FLastFoundIdxReseaux);
  for i := P to N do
  begin
    EWE := GetReseau(i);
    if (Pos(LowerCase(S), LowerCase(EWE.NomReseau)) > 0) then
    begin
      // mémorise la clé pour recherche de l'occurrence suivante
      FLastFindTextWhat := S;
      // mémorise l'indice de parcours
      FLastFoundIdxReseaux := i + 1;
      // on boucle depuis le début de la liste
      if (FLastFoundIdxReseaux > N) then FLastFoundIdxReseaux := -1;
      Result := i;
      Exit;
    end;
  end;
end;

function TToporobotStructure2012.FindIdxSecteurByText(const S: string): integer;
var
  i, N, P : integer;
  EWE: TSecteur;
begin
  Result := -1;
  N := GetNbSecteurs - 1;
  P := IIF((FLastFoundIdxSecteurs = -1), 0, FLastFoundIdxSecteurs);
  for i := P to N do
  begin
    EWE := GetSecteur(i);
    if (Pos(LowerCase(S), LowerCase(EWE.NomSecteur)) > 0) then
    begin
      // mémorise la clé pour recherche de l'occurrence suivante
      FLastFindTextWhat   := S;
      // mémorise l'indice de parcours
      FLastFoundIdxSecteurs := i + 1;
      // on boucle depuis le début de la liste
      if (FLastFoundIdxSecteurs > N) then FLastFoundIdxSecteurs := -1;
      Result := i;
      Exit;
    end;
  end;
end;


function TToporobotStructure2012.FindIdxExpeByText(const S: string): integer;
var
  i, N, P : integer;
  EWE     : TExpe;
  zdar    : String;
  Q       : String;
begin
  Result := -1;
  N := GetNbExpes - 1;
  P := IIF((FLastFoundIdxExpes = -1), 0, FLastFoundIdxExpes);
  for i := P to N do
  begin
    EWE := GetExpe(i);
    // DONE: Recherche sur date, équipe et commentaire à implémenter
    Q := DateToStr(GetSecuredDate(EWE.AnneeExpe, EWE.MoisExpe, EWE.JourExpe));
    zdar := Format('%s | %s | %s | %s',
                  [Q,
                   EWE.Operateur,
                   EWE.ClubSpeleo,
                   EWE.Commentaire
                  ]);
    AfficherMessage(zdar);

    if (Pos(LowerCase(S), LowerCase(zdar)) > 0) then
    begin
      FLastFindTextWhat   := S;               // mémorise la clé pour recherche de l'occurrence suivante
      FLastFoundIdxExpes := i + 1;            // mémorise l'indice de parcours
      // on boucle depuis le début de la liste
      if (FLastFoundIdxExpes > N) then FLastFoundIdxExpes := -1;
      Exit(i);
    end;
  end;
end;
function TToporobotStructure2012.FindIdxCodeByText(const S: string): integer;
var
  i, N, P : integer;
  EWE     : TCode;
begin
  Result := -1;
  N := GetNbSeries - 1;
  P := IIF((FLastFoundIdxCodes = -1), 0, FLastFoundIdxCodes);
  for i := P to N do
  begin
    EWE := GetCode(i);
    if (Pos(LowerCase(S), LowerCase(EWE.Commentaire)) > 0) then
    begin
      // mémorise la clé pour recherche de l'occurrence suivante
      FLastFindTextWhat   := S;
      // mémorise l'indice de parcours
      FLastFoundIdxCodes := i + 1;
      // on boucle depuis le début de la liste
      if (FLastFoundIdxCodes > N) then FLastFoundIdxCodes := -1;
      Result := i;
      Exit;
    end;
  end;
end;
function TToporobotStructure2012.FindIdxSerieByText(const S: string): integer;
var
  EWE : TObjSerie;
  i, N, P: Integer;
begin
  Result := -1;
  N := GetNbSeries - 1;
  P := IIF((FLastFoundIdxSeries = -1), 0, FLastFoundIdxSeries);
  for i := P to N do
  begin
    EWE := GetSerie(i);
    if (Pos(LowerCase(S), LowerCase(EWE.GetNomSerie)) > 0) then
    begin
      // mémorise la clé pour recherche de l'occurrence suivante
      FLastFindTextWhat   := S;
      // mémorise l'indice de parcours
      FLastFoundIdxSeries := i + 1;
      // on boucle depuis le début de la liste
      if (FLastFoundIdxSeries > N) then FLastFoundIdxSeries := -1;
      Result := i;
      Exit;
    end;
  end;
end;
// fonctions de tri
procedure TToporobotStructure2012.SortExpes();
begin
  AfficherMessage(Format('%s.SortExpes: Tri des seances',[ClassName]));
  FTableExpes.Sort(SortExpesCriteria);
end;

procedure TToporobotStructure2012.SortCodes();
begin
  AfficherMessage(Format('%s.SortCodes: Tri des codes',[ClassName]));
  FTableCodes.Sort(SortCodesCriteria);
end;

procedure TToporobotStructure2012.SortSeries();
begin
  AfficherMessage(Format('%s.SortSeries: Tri des series',[ClassName]));
  FListeDesSeries.Sort(SortSerieCriteria);
end;

function  TToporobotStructure2012.SortAntennesByLengths(): integer;
begin
  Result := -1;
  try
    AfficherMessage(Format('%s.SortAntennesByLengths: Tri des antennes',[ClassName]));
    FTableViseesAntenne.Sort(SortViseeAntenneCriteriaLength());
    result := FTableViseesAntenne.GetNbElements();
  except
  end;
end;

function  TToporobotStructure2012.SortAntennesBySerieStation(): integer;
begin
  Result := -1;
  try
    AfficherMessage(Format('%s.SortAntennesBySerieStation: Tri des antennes',[ClassName]));
    FTableViseesAntenne.Sort(SortViseeAntenneCriteriaSerSt());
    result := FTableViseesAntenne.GetNbElements();
  except
  end;
end;

// charger depuis fichier Tab ou XTB
// /!\ Ne pas faire les conversions en UTF8 ici:
// Le format interne de GHTopo est ANSI pour compatibilité avec Windows.
function TToporobotStructure2012.LoadFichierTab(const FichierTAB: TStringDirectoryFilename): integer;
var
  EPSGDefault : TLabelSystemesCoordsEPSG;
  MyFiltre    : TFiltrePersonnalise;
  pTAB                : TextFile;
  // compteurs
  NoLigneTAB, i       : integer;
  Prefix1, Prefix2    : Integer;
  WU : string;
  // date
  yyy: Word;
  mmm: Word;
  ddd: Word;
  // variables de lignes
  LigneTab: String;
  PrmsLn, QArrStr, QArrStrg: TGHStringArray;
  ErrMsg, QWU, QStr: String;
  // liste provisoire pour lecture de la section -6
  ProvListeEntrees: TStringList;
  // série
  UneSerie: TObjSerie;
  qVisee: TUneVisee;
  qNbVisees: Integer;

  // items de tables simples
  UneEntree: TEntrance;

  UnReseau : TReseau;
  UnSecteur: TSecteur;
  UneExpe  : TExpe;
  UnCode   : TCode;
  UneViseeAntenne: TViseeAntenne;
  UneNoteVersion : TReleaseNotes;
  NbErrorsInLoading: Integer;

  Ex: Integer;
  qNoSerie: Integer;
  qNoStation: Integer;
  blaireau, QN: Integer;
  UneAnnotation: TAnnotation;
  function LireLigne: string;
  var
    Lign: string;
  begin
    ReadLn(pTab, Lign);
    Inc(NoLigneTAB);
    Result:= Lign;// PurgerAccents(Lign);
  end;
  procedure WriteWarning(const Msg: string);
  begin
    AfficherMessageErreur(Msg);
  end;
begin
  Result := -1;
  AfficherMemoryUsage();
  AfficherMessage(Format('%s.LoadFichierTAB(%s)',[ClassName, FichierTAB]));
  // exemple d'affichage à l'exécution du numero de la ligne (utile pour déboguer)
  //AfficherMessage(  {$INCLUDE %FILE%} );
  //AfficherMessage(  {$INCLUDE %LINE%} );
  // nom étude par défaut
  SetNomEtude('Etude001');
  SetNameSpace('Namespace00');
  SetCommentairesEtude('');

  EPSGDefault.CodeEPSG := DEFAULT_SYSTEME_COORDONNEES_CODE_EPSG;
  EPSGDefault.NomEPSG  := DEFAULT_SYSTEME_COORDONNEES_NOM;
  SetCodeEPSGSystemeCoordonnees(EPSGDefault);
  //SetDatabaseName(ExtractFileNameWoExt(FichierTAB));
  ReInitialiser(True);
  NoLigneTAB := 0;
  NbErrorsInLoading := 0;
  AssignFile(pTAB, FichierTAB);
  ProvListeEntrees := TStringList.Create;
  // création de la première série
  //UneSerie := TObjSerie.Create;
  try
    ProvListeEntrees.Clear;
    ReSet(pTAB);
    while Not Eof(pTAB) do
    begin
      try // traitement local d'exceptions dans la lecture des lignes
        // réinitialisation du message d'erreur par défaut
        ErrMsg:= rsRD_TAB_MSG_ERR;
        Prefix1:=0;
        LigneTab := LireLigne;
        if (Trim(LigneTab)='')    then Prefix1 := -100  // lignes vides
        else if (LigneTab[1]='#') then Prefix1 := -1000 // commentaires sur une ligne
        else if LigneTab[1]='{' then    // commentaires sur plusieurs lignes
        begin
          AfficherMessage(Format(rsRD_TAB_D_MULTI_OBS,[NoLigneTAB]));
          while Not (LigneTab[1]='}') do LigneTab := LireLigne; /// PurgerAccents(LireLigne);
          AfficherMessage(Format(rsRD_TAB_F_MULTI_OBS,[NoLigneTAB]));
        end
        else begin
          //LigneTab := PurgerAccents(LireLigne);
          PrmsLn   := split(LigneTab, TAB);
          Prefix1  := StrToIntDef(PrmsLn[0], -110);
        end;
        // routage selon le préfixe
        case Prefix1 of
          -9999: Break; // Arrêt forcé du traitement (utiliser avec précaution)
          -1000: AfficherMessage(Format(rsRD_TAB_LN_OBS,[NoLigneTab]));   // commentaire
           -900: ; // balise de pause
           -110: ; // Ligne invalide; ignorée
           -100: ; // ignorer les lignes vides
            -20: AfficherMessage(Format(rsRD_TAB_LASTSAVES,[PrmsLn[1], PrmsLn[2]])); // horodatage
            -19: begin
                   self.SetNomEtude(Trim(PrmsLn[1])); // Nouvelle section: Nom de l'étude
                   self.SetCommentairesEtude(Trim(PrmsLn[2]));
                 end;
            -18: begin
                   self.SetNameSpace(Trim(PrmsLn[1])); // Nouvelle section: Espace de noms (pour les synthèses)
                   self.SetCommentairesEtude(Trim(PrmsLn[2]));
                 end;
            {$IFDEF SUPPORT_NOTES_VERSION}
            -17: begin // notes de version
                   //-18 2017-08-02	CASSOU JP	Correction série 210
                   UneNoteVersion.Date        := DateTimeSQLToDateTimePascal(Trim(PrmsLn[1]));
                   UneNoteVersion.Author      := Trim(PrmsLn[2]);
                   UneNoteVersion.Description := Trim(PrmsLn[3]);
                   self.AddNoteVersion(UneNoteVersion);
                 end;
            {$ENDIF SUPPORT_NOTES_VERSION}
            -16: begin // Nouvelle section: Filtres personnalisés
                   //#### #-16	IDFiltre	NomFiltre	CouleurFiltre_R	CouleurFiltre_G	CouleurFiltre_B	FiltreExpr	Description
                   //-16	0	Aquatopo	0	128	0	SERIE=227
                   MyFiltre.CouleurFiltre :=  RGBToColor(StrToIntDef(PrmsLn[2], 0),
                                                         StrToIntDef(PrmsLn[3], 0),
                                                         StrToIntDef(PrmsLn[4], 0));
                   MyFiltre.NomFiltre     := Trim(PrmsLn[5]);
                   MyFiltre.Expression    := Trim(PrmsLn[6]);
                   MyFiltre.Description   := Trim(PrmsLn[7]);
                   AddFiltrePerso(MyFiltre);

                 end;
            -15: begin // Nouvelle section: Système de coordonnées :
                   //-19	Grottes de Villesèque
                   //-18	Reprise de la topographie de la grotte de Villesèque
                   // # Système de coordonnées
                   // #1	Code EPSG	Code GHTopo	Libellé du système de coordonnées
                   // -15	31563	LT3	Lambert 3 France
                   //try
                     //AfficherMessageErreur(PrmsLn[1]);

                     EPSGDefault.CodeEPSG := StrToIntDef(Trim(PrmsLn[1]), DEFAULT_SYSTEME_COORDONNEES_CODE_EPSG);
                     EPSGDefault.NomEPSG  := '';

                     SetCodeEPSGSystemeCoordonnees(EPSGDefault);

                   //except
                   //  FDoSetDefaultCoordSyst := True;
                   //end;
                   //*)
                 end;


            -11: begin // Nouvelle section: secteurs
                   //0   1      2             3               4      5       6         7       8          9
                   //#	Idx Basepoint.Serie	Basepoint.Point	OffsetX	OffsetY	OffsetZ	IdxStyle	Alignment	Texte
                   UneAnnotation.Basepoint.eSerie := StrToIntDef(PrmsLn[2], 1);
                   UneAnnotation.Basepoint.ePoint := StrToIntDef(PrmsLn[3], 1);
                   UneAnnotation.Offset.X         := ConvertirEnNombreReel(PrmsLn[4], 0.00);
                   UneAnnotation.Offset.Y         := ConvertirEnNombreReel(PrmsLn[5], 0.00);
                   UneAnnotation.Offset.Z         := ConvertirEnNombreReel(PrmsLn[6], 0.00);
                   UneAnnotation.IdxStyle         := StrToIntDef(PrmsLn[7], 0);
                   UneAnnotation.PosPtDeBase      := StrToIntDef(PrmsLn[8], 0);
                   UneAnnotation.Texte            := Trim(PrmsLn[9]);
                   AddAnnotation(UneAnnotation);
                 end;
            -10: begin // Nouvelle section: secteurs
                   with UnSecteur do
                   begin
                     //IdxSecteur     := StrToIntDef(PrmsLn[1], 0);
                     CouleurSecteur := RGBToColor(StrToIntDef(PrmsLn[2], 0),
                                           StrToIntDef(PrmsLn[3], 0),
                                           StrToIntDef(PrmsLn[4], 0));
                     NomSecteur     := Trim(PrmsLn[5]);
                   end;
                   AddSecteur(UnSecteur);
                 end;
             -9: begin  // visées en antenne
                   with UneViseeAntenne do
                   begin
                     IDViseeAntenne   := StrToIntDef(PrmsLn[1], 0);

                     Reseau           := StrToIntDef(PrmsLn[2], 0);
                     Secteur          := StrToIntDef(PrmsLn[3], 0);

                     SerieDepart      := StrToIntDef(PrmsLn[4], 0);
                     PtDepart         := StrToIntDef(PrmsLn[5], 0);
                     Code             := StrToIntDef(PrmsLn[6], 0);
                     Expe             := StrToIntDef(PrmsLn[7], 0);
                     IDTerrainStation := FormatterIDTerrainStation(PrmsLn[8]);
                     Longueur         := ConvertirEnNombreReel(PrmsLn[9], 0.00);
                     Azimut           := ConvertirEnNombreReel(PrmsLn[10], 0.00);
                     Pente            := ConvertirEnNombreReel(PrmsLn[11], 0.00);
                     Commentaires     := Trim(PrmsLn[12]);
                     //AfficherMessage(Format('  >> Antenne trouvee: %d - %s', [IDViseeAntenne, IDTerrainStation]));
                   end;
                   AddViseeAntenne(UneViseeAntenne);
                 end;
             -8: begin  // réseaux
                   with UnReseau do
                   begin
                     //IdxReseau   := StrToIntDef(PrmsLn[1], 0);
                     ColorReseau := RGBToColor(StrToIntDef(PrmsLn[2], 0),
                                        StrToIntDef(PrmsLn[3], 0),
                                        StrToIntDef(PrmsLn[4], 0));
                     TypeReseau  := StrToIntDef(PrmsLn[5], 0);
                     NomReseau   := Trim(PrmsLn[6]);
                     ObsReseau   := Trim(PrmsLn[7]);
                   end;
                   AddReseau(UnReseau);
                 end;
             -7: ; // Classeurs (désactivé)
             -6: begin  // Section -6: Entrée
                   ProvListeEntrees.Add(PrmsLn[2] + '|' + PrmsLn[3]);
                 end;
             -5: begin  // Entrées
                   //nombre d'entrées nul = on définit l'entrée par défaut
                   if (GetNbEntrees = 0) then begin
                     SetDefaultCoords(ConvertirEnNombreReel(PrmsLn[2], 0.00),
                                      ConvertirEnNombreReel(PrmsLn[3], 0.00),
                                      ConvertirEnNombreReel(PrmsLn[4], 0.00)
                                     );
                     SetRefSeriePoint(StrToIntDef(PrmsLn[5],1),
                                      StrtoIntDef(PrmsLn[6], 0)
                                     );

                   end;
                   //Commentaires:=PrmsLn[7];
                   // Ajouter les entrées
                   // le nombre d'entrées retenu est celui décompté dans
                   // la section -5
                   with UneEntree do
                   begin
                     //eNumEntree:= GetNbEntrees + 1;
                     // On ajoute l'entrée récupérée en -6
                     try
                       //QStr := ProvListeEntrees.Strings[eNumEntree-1];
                       QN := GetNbEntrees();
                       QStr := ProvListeEntrees.Strings[QN];
                       QArrStrg := Split(QStr, '|');
                       eNomEntree := Trim(QArrStrg[0]);
                       eIDTerrain := Trim(QArrStrg[1]);
                     except
                       AfficherMessage(Format(rsWARNINGENTRYADDED,[QN]));
                       eNomEntree := Format(rsRD_TAB_ENTRANCE,[QN]);
                     end;
                     eXEntree  := ConvertirEnNombreReel(PrmsLn[2], 0.00);
                     eYEntree  := ConvertirEnNombreReel(PrmsLn[3], 0.00);
                     eZEntree  := ConvertirEnNombreReel(PrmsLn[4], 0.00);
                     //------------
                     // entrées non géoréférencées ?
                     if (eXEntree < 100.00) or (eYEntree < 100.00) or (eZEntree < 100.00)
                     then WriteWarning(Format(GetResourceString(rsRD_TAB_ENTR_NOGEOREF),
                                             [NoLigneTAB, QN, eNomEntree]));
                     eRefSer   := StrToIntDef(PrmsLn[5],1);
                     eRefSt    := StrtoInt(PrmsLn[6]);
                     if ((eRefSer < 1) or (eRefSt < 0)) then
                        WriteWarning(Format(rsRD_TAB_ENTR_BADLINK, [NoLigneTAB, QN, eNomEntree, eRefSer, eRefSt]));
                     eObserv   := PrmsLn[7];
                   end;
                   AddEntree(UneEntree);
                 end;
             -4: ;
             -3: ;
             -2: begin // Expés
                   with UneExpe do begin
                     IDExpe := StrToInt(PrmsLn[1]);
                     if (IDExpe <= 0) then
                     begin
                        ErrMsg:=rsRD_TAB_BAD_TRIP;
                        raise Exception.Create(ErrMsg);
                     end;
                     JourExpe      := StrToIntDef(PrmsLn[2],0);
                     MoisExpe      := StrToIntDef(PrmsLn[3],0);
                     AnneeExpe     := StrToIntDef(PrmsLn[4],0);
                     if (JourExpe=0) or (MoisExpe=0) or (AnneeExpe=0) then
                     begin
                       //ErrMsg:=Format('Date incorrecte: %d/%d/%d',[JourExpe,MoisExpe,AnneeExpe]);
                       //raise Exception.Create(ErrMsg);
                       WriteWarning(Format(rsRD_TAB_BAD_DATE,
                                          [NoLigneTAB]));
                       SafeDecodeDate(Now, yyy, mmm, ddd);
                       JourExpe := ddd;
                       MoisExpe := mmm;
                       AnneeExpe:= yyy;

                     end;
                     Operateur     := PrmsLn[5];             // spéléomètre
                     ClubSpeleo    := PrmsLn[6];             // spéléographe
                     ModeDecl      := StrToInt(PrmsLn[7]);   // déclinaison auto ?
                     Declinaison   := ConvertirEnNombreReel(PrmsLn[8], 0.00); // déclinaison
                     //Inclinaison   := ConvertirEnNombreReel(PrmsLn[9], 0.00); // correction clino x10
                     Couleur       := StrToInt(PrmsLn[10]);  // couleur
                     Commentaire   := PrmsLn[11];            // commentaire
                   end;
                   AddExpe(UneExpe);
                 end;
             -1: begin // Codes
                   with UnCode do
                   begin
                      IDCode     := StrToInt(PrmsLn[1]);    // ID Code
                      if (IDCode <= 0) then
                      begin
                        //WriteWarning(Format('WARNING ! (%d) - Numéro de Code incorrect (Valeur: %d) - Mis à %d',[NoLigneTAB, IDCode, FNbCodes]));
                        ErrMsg:=rsRD_TAB_BAD_CODE;
                        raise Exception.Create(ErrMsg);
                      end;
                      GradAz     := ConvertirEnNombreReel(PrmsLn[2], 360.00);  // unité boussole
                      GradInc    := ConvertirEnNombreReel(PrmsLn[3], 360.00);  // unite  CLINO
                      PsiL       := ConvertirEnNombreReel(PrmsLn[4], 0.01);  // precision longueur
                      PsiAz      := ConvertirEnNombreReel(PrmsLn[5], 0.1);  // precision azimut
                      PsiP       := ConvertirEnNombreReel(PrmsLn[6], 0.1);  // precision pente
                      // TODO: Support de FactLong a valider
                      FactLong   := 1.00;
                      //FactLong   := ConvertirEnNombreReel(PrmsLn[7], 1.00)/100;  // Facteur des longueurs
                      AngLimite  := ConvertirEnNombreReel(PrmsLn[8], 0.00);  // angle limite
                      Commentaire := PrmsLn[9];              // commentaire
                      //ReservedInt := StrToIntDef(PrmsLn[10],0);
                      ParamsFuncCorrAz  := MakeTParamFoncCorrectionAngulaire(ConvertirEnNombreReel(PrmsLn[11], 0.00),
                                                                             ConvertirEnNombreReel(PrmsLn[12], 0.00),
                                                                             ConvertirEnNombreReel(PrmsLn[13], 0.00));
                      ParamsFuncCorrInc := MakeTParamFoncCorrectionAngulaire(ConvertirEnNombreReel(PrmsLn[14], 0.00),
                                                                             ConvertirEnNombreReel(PrmsLn[15], 0.00),
                                                                             ConvertirEnNombreReel(PrmsLn[16], 0.00));
                      ErreurTourillon   := ConvertirEnNombreReel(PrmsLn[17], 0.00);
                    end;
                    AddCode(UnCode);
                 end;
             0: ; // pas de section 0
           // end
         otherwise  // on est dans les séries !
           // Si le préfixe 2 (2e colonne) =-1 =>nouvelle série
           Prefix1 := StrToInt(PrmsLn[0]);
           if (Prefix1 > 0) then
           begin
             Prefix2 := StrToInt(PrmsLn[1]);

             if (Prefix2 = -1) then
             begin
               // si c'est la première série, on crée
               if (Prefix1 = 1) then
               begin
                 UneSerie := TObjSerie.Create;
                 UneSerie.ClearStations();
               end
               else  // sinon on ferme la série courante et on crée la suivante
               begin
                 self.AddSerie(UneSerie);
                 UneSerie := TObjSerie.Create;
               end;
               with UneSerie do
               begin
                 SetNumeroSerie(TNumeroSerie(Prefix1));
                 SetSeriePtExtremites(StrToInt(PrmsLn[2]),
                                      StrToInt(PrmsLn[3]),
                                      StrToInt(PrmsLn[4]),
                                      StrToInt(PrmsLn[5]));
                 // NbPoints: Non utilisé
                 //NbPoints      :=StrToInt(PrmsLn[6]);
                 SetChanceObstacle(StrToInt(PrmsLn[7]), StrToInt(PrmsLn[8]));
                 SetNomSerie(PrmsLn[9]);
                 SetObsSerie(PrmsLn[10]);
                 SetNumeroReseau(StrToIntDef(PrmsLn[11],0));
                 // raideur de la série
                 SetRaideur(ConvertirEnNombreReel(Prmsln[12], 1.02));
                 // entrée de rattachement
                 SetNumeroEntrance(StrToIntDef(Prmsln[13], 0));
                 // on lit toutes les visées
               end; //with UneSerie do
               //end; // if (Prefix2 = -1) then
             end else
             begin
               with qVisee do
               begin
                 TypeVisee   := tgDEFAULT;
                 Code        := StrToInt(PrmsLn[2]);
                 Expe        := StrToInt(PrmsLn[3]);
                 Longueur    := ConvertirEnNombreReel(PrmsLn[4], 0.00);
                 if (Longueur < 0.00) then
                 begin
                   WriteWarning(Format(rsRD_TAB_NEG_LONG, [NoLigneTAB, Longueur]));
                   Longueur := Abs(Longueur);
                 end;
                 Azimut      := ConvertirEnNombreReel(PrmsLn[5], 0.00);
                 Pente       := ConvertirEnNombreReel(PrmsLn[6], 0.00);
                 LG          := ConvertirEnNombreReel(PrmsLn[7], 0.00);  //LG
                 LD          := ConvertirEnNombreReel(PrmsLn[8], 0.00);  //LD
                 HZ          := ConvertirEnNombreReel(PrmsLn[9], 0.00);
                 HN          := ConvertirEnNombreReel(PrmsLn[10], 0.00);
                 Commentaires:= PrmsLn[11];
                 IDTerrainStation := FormatterIDTerrainStation(PrmsLn[12]);
                 blaireau    := StrToIntDef(PrmsLn[13], 0);
                 TypeVisee   := GetTypeDeVisee(blaireau); // galerie fossile supposée
                 IDSecteur   := StrToIntDef(PrmsLn[14], 0);
               end; // with qVisee do
               UneSerie.AddVisee(qVisee);
             end;
           end   // if Prefix2
         end; // case Prefix1
      except
        Inc(NbErrorsInLoading);
        WriteWarning('');
        WriteWarning(Format(rsRD_ERROR_LN,[NoLigneTab, ErrMsg]));
        WriteWarning(Format(rsRD_CONTNT_LN,[LigneTab]));
        WriteWarning(rsRD_TAB_FIELD_VALUES);
        for Ex:=0 to High(PrmsLn) do
        begin
          WriteWarning(Format(' PrmsLn[%.2d] = "%s"',[Ex, PrmsLn[Ex]]));
          WriteWarning('');
        end;

      end;
    end; // while Not Eof(pTAB) do begin
    // On ferme la dernière série
    self.AddSerie(UneSerie);
    // Cause courante de plantages: Absence de la série 1
    VerifieSerie1();
    //--------------------------------------------------------------------------
    // Recensement des ID terrain
    RecenserLesIDsTerrain();
    // Actions de réparation / vérif
    //CheckerLesDonneesTopo();
    // Passage ici ? -> OK
    Result := 666;
    AfficherMemoryUsage();
    // Recenser les POI
    RecenserPointsOfInterest();
  finally
    CloseFile(pTAB);
    ProvListeEntrees.Free;
  end;
end;

function TToporobotStructure2012.VerifieSerie1(): boolean;
var
  QSR, MySerie: TObjSerie;
  QInternalSerie, i, nb: integer;
begin
  result := false;
  AfficherMessage(Format('%s.VerifieSerie1', [ClassName]));
  AfficherMessageErreur(Format('Entre dans %s.VerifieSerie1: %d séries', [ClassName, GetNbSeries()]));
  //MySerie := TObjSerie.Create;
  (*
  nb := GetNbSeries();
  for i := 0 to nb - 1 do
  begin
    try
      MySerie := GetSerie(i);
      AfficherMessageErreur(Format('%d - %d - %s - %d', [i, MySerie.GetNumeroDeSerie(), MySerie.GetNomSerie(), MySerie.GetNbVisees()]));

    except
      AfficherMessageErreur( Format('%d - ***', [i]));
    end;
  end;
  AfficherMessageErreur('**********************');
  //*)

  if (ExistsSerie(1)) then
  begin
    AfficherMessageErreur('*** Série 1 inexistante ***');
    Exit;
  end;
  // si la série 1 est inexistante, on la crée
  // en supprimant la série bâtarde 1
  try
    self.RemoveSerie(1);
  except
    AfficherMessageErreur('*** Série 1 supprimée mais non libérée ***');
  end;
  AfficherMessageErreur('-> Série 1 inexistante. Recréée');
  MySerie := TObjSerie.Create;
  MySerie.ClearStations();
  MySerie.SetNumeroSerie(1);
  MySerie.SetSeriePtExtremites(1, 0, 1, 1);
  MySerie.SetChanceObstacle(0, 0);
  MySerie.SetNomObsSerie('Serie 1 reconstruite', 'Serie 1 absente dans le fichier original - Recréée');
  MySerie.AddVisee(0, 0, 0, 0, tgDEFAULT , 0.01, 45.00, -45.00, 0.0, 0.0, 0.0, 0.0, '','');
  MySerie.AddVisee(1, 0, 0, 0, tgDEFAULT , 0.0125, 45.00, -45.00, 0.0, 0.0, 0.0, 0.0, '','');
  AfficherMessage('--> Ajout de la série 1');
  self.AddSerie(MySerie);
  // et on trie la table
  self.SortSeries();
  // controle
  nb := GetNbSeries();
  AfficherMessageErreur('Le document comporte une série 1 absente qui a été reconstruite');
  AfficherMessageErreur(format('Liste des %d séries', [nb]));
  for i := 0 to nb - 1 do
  begin
    MySerie := GetSerie(i);
    AfficherMessageErreur(Format('%d - %d - %s - %d', [i, MySerie.GetNumeroDeSerie(), MySerie.GetNomSerie(), MySerie.GetNbVisees()]));
  end;
end;

// Statut: Opérationnel mais les commentaires des séries et stations sont ignorés
// Le format .Text est deprecated et supporté en lecture seule par GHTopo
// Il est surtout destiné à PocketTopo
// DONE: Entrées sans série de départ sont ignorées
// @deprecated
// Shuanghe: OK
// Charentais OK
// Carrière de Bellegarde avec antennes: OK
function TToporobotStructure2012.LoadFichierText(const FichierText: TStringDirectoryFilename): integer; deprecated;
const
  // colonnes des champs
  COL_COMMENT     = 00; // parenthèse de commentaire éventuel
  COL_NUM_SECTION = 01; // numéro de section ou numéro de série
  COL_NUM_ITEM    = 02; // numéro d'item (expé, code, point topo)

  COL_SESSION     = 03; // numéro de session (inutilisé)
  COL_CODE        = 04; // code attaché à la station
  COL_EXPE        = 05; // numéro de séance

  COL_VAR_DATA    = 06; // colonne suivant les colonnes de préfixes et index
  COL_NOMENTREE   = COL_VAR_DATA; // nom de l'entrée
  COL_XENTREE     = COL_VAR_DATA;
  COL_YENTREE     = COL_XENTREE + 1;
  COL_ZENTREE     = COL_YENTREE + 1;
  COL_ENT_SER     = COL_ZENTREE + 1;
  COL_ENT_ST      = COL_ENT_SER + 1;

  COL_DATE_EXPE   = COL_VAR_DATA;
  COL_SPELEOMETRE = COL_DATE_EXPE + 1;
  COL_SPELEOGRAPH = COL_SPELEOMETRE + 1;
  COL_MODEDECL    = COL_SPELEOGRAPH + 1;
  COL_DECL_VALUE  = COL_MODEDECL + 1;
  COL_INCL_VALUE  = COL_DECL_VALUE + 1;
  COL_COULEUR     = COL_INCL_VALUE + 1;

  COL_U_AZIMUT    = COL_VAR_DATA;
  COL_U_PENTE     = COL_U_AZIMUT + 1;
  COL_PSI_L       = COL_U_PENTE + 1;
  COL_PSI_AZ      = COL_PSI_L +1;
  COL_PSI_P       = COL_PSI_AZ + 1;
  //COL_OLD_DECL    = COL_PSI_P  + 1;
  COL_FACT_LONG   = COL_PSI_P + 1;
  COL_ANG_LIMITE  = COL_FACT_LONG + 1;

  COL_SER_DEP     = COL_VAR_DATA;
  COL_PT_DEP      = COL_SER_DEP + 1;
  COL_SER_ARR     = COL_PT_DEP + 1;
  COL_PT_ARR      = COL_SER_ARR + 1;
  COL_NB_PTS      = COL_PT_ARR + 1;
  COL_CHANCE      = COL_NB_PTS + 1;
  COL_OBSTACLE    = COL_CHANCE + 1;

  COL_LONG        = COL_VAR_DATA;
  COL_AZ          = COL_LONG + 1;
  COL_P           = COL_AZ + 1;
  COL_LG          = COL_P + 1;
  COL_LD          = COL_LG + 1;
  COL_HZ          = COL_LD + 1;
  COL_HN          = COL_HZ + 1;

  TEMPORARYTABFILE = '~GhtopoTemp.txt';

var
  EPSGDefault : TLabelSystemesCoordsEPSG;
  // fichier d'erreur
  pTEXT        : TextFile;
  // liste provisoire pour entrées
  ProvListeEntrees: TStringList;
  // séries, réseaux, etc
  UneEntree : TEntrance;
  UnReseau  : TReseau;
  UneExpe : TExpe;
  UnCode  : TCode;
  UneSerie: TObjSerie;
  qVisee  : TUneVisee;
  // ligne de données
  NoLigneTEXT,
  NbErrorsInLoading: integer;
  LigneTEXT: string;
  PrmsLn   : TGHStringArray;
  // la ligne contient une parenthèse de tête ?
  HasParenthesis: boolean;
  // messages d'erreur
  ErrMsg  : string;
  // Préfixes de section
  Prefix1,
  Prefix2 : integer;

  // index de boucles
  i ,n, Ex: integer;
  // divers
  str         : string;
  llanfair_pg : string;
  // dates
  yyy, mmm, ddd: word;
  // Index courant
  //IdxCourant: integer;
  IdxCourantStation: integer;
  NomDeLaSerie: String;
  WU: String;
  mioumiou: Boolean;
  IdxCourant, QN: Integer;

  myViseeAntenne: TViseeAntenne;
  // Conversion de lignes Mac > PC
  procedure ConvertTextMacToPC(var Lin: string);
  var
    pn   : integer;
    ps  : integer;
    s,p : string;
    w1  : integer;
  begin
    // convertir accentués
    Lin := StringReplace(Lin, #136, 'a', [rfReplaceAll]);
    Lin := StringReplace(Lin, #142, 'e', [rfReplaceAll]);
    Lin := StringReplace(Lin, #143, 'e', [rfReplaceAll]);
    Lin := StringReplace(Lin, #148, 'i', [rfReplaceAll]);
    Lin := StringReplace(Lin, #137, 'a', [rfReplaceAll]);
    Lin := StringReplace(Lin, #144, 'e', [rfReplaceAll]);
    Lin := StringReplace(Lin, #144, 'e', [rfReplaceAll]);
    Lin := StringReplace(Lin, #141, 'c', [rfReplaceAll]);
  end;
  // extraire une visée en antenne
  function ExtraireViseeEnAntenne(const qStr: string;
                                  const qIdxReseau: integer;
                                  const qIdxSerie: integer;
                                  const qIdxPoint: integer;
                                  const qMyVisee: TUneVisee;
                                  var qVA : TViseeAntenne): boolean;
  var
    myStr: String;
    CrO, CrF: integer;
    EWE: TGHStringArray;
  begin
    Result := False;
    myStr := Trim(qStr);
    try
      // Visées en antenne sont de la forme [%.2f %.2f %.2f]
      // crochets indispensables
      // recherche de crochets et suppression
      CrO := Pos('[', myStr);
      if (CrO = 0) then Exit;
      if (CrO > 0) then System.Delete(myStr, CrO, 1);
      CrF := Pos(']', myStr);
      if (CrF > 0) then System.Delete(myStr, CrF, 1);
      EWE := Split(myStr, ' ');
      qVA.SerieDepart      := qIdxSerie;
      qVA.PtDepart         := qIdxPoint;
      qVA.IDTerrainStation := qMyVisee.IDTerrainStation;
      qVA.Code             := qMyVisee.Code;
      qVA.Expe             := qMyVisee.Expe;
      qVA.Reseau           := qIdxReseau;
      qVA.Secteur          := qMyVisee.IDSecteur;
      qVA.Commentaires     := '';
      qVA.Longueur         := ConvertirEnNombreReel(Trim(EWE[0]), -1.00);
      qVA.Azimut           := ConvertirEnNombreReel(Trim(EWE[1]), 0.00);
      qVA.Pente            := ConvertirEnNombreReel(Trim(EWE[2]), 0.00);
      //AfficherMessage(Format('%f, %f, %f',[qVA.Longueur, qVA.Azimut, qVA.Pente]));
      Result := (qVA.Longueur > 0.00);
    except
    end;
  end;
  // afficher avertissement
  procedure WriteWarning(const wm: string);
  begin
    AfficherMessage(wm);
  end;
begin
  //self.ClearListeSeries();
  EPSGDefault.CodeEPSG := DEFAULT_SYSTEME_COORDONNEES_CODE_EPSG;
  EPSGDefault.NomEPSG  := DEFAULT_SYSTEME_COORDONNEES_NOM;

  WU := Format('%s.LoadFichierText: %s', [ClassName, FichierText]);
  AfficherMessage(WU);
  AfficherMessageErreur(WU);
  Result := -1;
  // Fichier introuvable ? OK, On sort
  if not FileExists(FichierText) then  begin Result:=-32; Exit; end;
  // nom étude par défaut
  SetNomEtude('Etude001');
  SetNameSpace('Namespace00');
  SetCommentairesEtude('Issued from Toporobot Text format (fixed columns)');
  SetCodeEPSGSystemeCoordonnees(EPSGDefault);


  try // try..finally niveau 0
     //*************************************************************************
     // Pour pallier à tout problème de formats de fichiers textes,
     // et notamments ceux liés aux fins de ligne,
     // on fait une conversion.
     // Pour économiser une variable, on utilise une variable utilisée + tard
     LigneTEXT := AnsiToUtf8(ExtractFilePath(FichierText)+ TEMPORARYTABFILE);
     ConvertTextFile(FichierText, LigneTEXT, tfMAC, tfWINDOWS);
     AfficherMessage(GetResourceString(rsSEPARATOR_LINE));
     // nom de la base
     SetDatabaseName(ExtractFileNameWithoutExt(FichierText)); // Laz 1.8: ExtractFileNameWithoutExt() remplace ExtractFileNameOnly()
     // Initialisation des listes
     ReInitialiser(True);
     AfficherMessage(Format('%s: %s', [{$I %FILE%}, {$I %LINE%}]));
     // code EPSG par défaut
     SetCodeEPSGSystemeCoordonnees(EPSGDefault);

     // début du véritable traitement
     // liste provisoire pour lecture des sections -6
     ProvListeEntrees:=TStringList.Create;
     // ouverture fichier Text
     AssignFile(pTEXT, LigneTEXT);
     AfficherMessage('--> Lecture du fichier');

     try // niveau 1: pour le fichier Text
       Reset(pTEXT);
       // lire tout le fichier
       NoLigneTEXT:=0;
       NbErrorsInLoading:=0;
       ErrMsg:= rsRD_TAB_MSG_ERR;
       IdxCourant := -1;
       while Not Eof(pTEXT) do
       begin
         try // traitement local d'exceptions dans la lecture des lignes
           Prefix1:=0;
           Inc(NoLigneTEXT);
           ReadLn(pTEXT, LigneTEXT);
           AfficherMessage('--> ' + LigneTEXT);
           //AfficherMessage(Format('Ligne %d - %s', [NoLigneTEXT, LigneTEXT]));
           //AfficherMessage(LigneTEXT);
           //ConvertTextMacToPC(LigneTEXT);
           if Trim(LigneTEXT)=''    then Prefix1:=-100       // lignes vides
           else if LigneTEXT[1]='#' then Prefix1:=-1000     // commentaires sur une ligne
           else
           begin
             // Commentaire = parenthèse de tête présente
             HasParenthesis := (LigneTEXT[1] = '(');
             //if (HasParenthesis) then Continue;
             // attraper le préfixe de section (en évitant une décomposition)
             str := Trim(Copy(LigneTEXT, 2, 5));
             Prefix1   := StrToIntDef(str, -110);

             if (HasParenthesis) then  // On zappe les commentaires (difficultés d'implémentation)
             begin
               //Continue;
               // classique décomposition ...
               PrmsLn      := SplitFixedColLine(LigneTEXT, [1, 2, 7, 13, 17, 21, 25, 80,100]);
               llanfair_pg := Trim(PrmsLn[COL_VAR_DATA]);
                case Prefix1 of
                 -5: begin // entrées

                       UneEntree := GetEntree(IdxCourant);
                       // on concatène les lignes
                       UneEntree.eObserv := UneEntree.eObserv + llanfair_pg;

                       PutEntree(IdxCourant, UneEntree);
                     end;
                 -4: ;     // sessions (obsolète)
                 -3: ;     // section non affectée
                 -2: begin // expés
                       IdxCourant := GetNbExpes - 1;
                       UneExpe   := GetExpe(IdxCourant);
                       // on concatène les lignes

                       UneExpe.Commentaire := UneExpe.Commentaire + llanfair_pg;
                       PutExpe(IdxCourant, UneExpe);
                     end;
                 -1: begin // code
                       IdxCourant := GetNbCodes - 1;
                       UnCode    := GetCode(IdxCourant);
                       // on concatène les lignes
                       UnCode.Commentaire := UnCode.Commentaire + llanfair_pg;
                       PutCode(IdxCourant, UnCode);
                     end;
               else  // séries
                 // Zapper cette section
                 begin
                   //Continue;
                   if (Prefix1 > 0) then
                   begin
                     Prefix2:=StrToInt(PrmsLn[COL_NUM_ITEM]);
                     if (Prefix2 = -1) then
                     begin // tête de série
                       IdxCourant := GetNbSeries - 1;
                       UneSerie := GetSerie(IdxCourant);
                       UneSerie.SetObsSerie(UneSerie.GetObsSerie + llanfair_pg);
                       //PutSerie(IdxCourant, UneSerie);

                     end else
                     begin // sinon, station
                       //Continue;
                       IdxCourant := GetNbSeries - 1;
                       IdxCourantStation := UneSerie.GetNbVisees - 1;
                       qVisee := UneSerie.GetVisee(IdxCourantStation);
                       if (ExtraireViseeEnAntenne(llanfair_pg,
                                                  UneSerie.GetNumeroReseau(),
                                                  UneSerie.GetNumeroDeSerie(),
                                                  IdxCourantStation,
                                                  qVisee,
                                                  myViseeAntenne)) then
                       begin
                         AddViseeAntenne(myViseeAntenne);
                         llanfair_pg := '';
                       end;



                       qVisee.Commentaires := qVisee.Commentaires + llanfair_pg;
                       UneSerie.PutVisee(IdxCourantStation, qVisee);
                       //PutSerie(IdxCourant + 1, UneSerie);
                       AfficherMessage(Format('IndexSerieCourant: %d', [IdxCourant]));
                     end;
                   end;
                 end;
               end;
               // on zappe tout le reste ...
               Continue;
             end;  // if HasParenthesis
             //*)
             case Prefix1 of

                -9999: begin // Arrêt forcé du traitement
                             // utiliser cette balise avec de grandes précautions
                             // et de préférence à la fin d'une série
                         WriteWarning('');
                         WriteWarning(Format(rsRD_TAB_STOP_9999, [NoLigneTEXT]));
                         WriteWarning('');
                         Break;
                       end;
                -1000: AfficherMessage(Format(GetResourceString(rsRD_TAB_LN_OBS),[NoLigneTEXT]));  // commentaire
                 -900: Continue; // balise de pause
                 -110: Continue; // Ligne invalide; ignorée
                 -100: Continue; // ignorer les lignes vides
                   -7: AfficherMessage(GetResourceString(rsRD_TAB_CLASSEURS)); // Section -7 : Classeurs -- Ignoré par GHTopo
                   -6: begin  // Section -6: Entrée
                         PrmsLn := SplitFixedColLine(LigneTEXT, [1, 2, 7, 13, 17, 21, 25, 80, 100]);
                         ProvListeEntrees.Add(PrmsLn[COL_NOMENTREE]);

                         Continue;
                       end;
                   -5: begin  // Entrées
                         PrmsLn := SplitFixedColLine(LigneTEXT, [1, 2, 7, 13, 17, 21, 25, 37, 49, 61, 67, 73, 80]);
                         //nombre d'entrées nul = on définit l'entrée par défaut
                         // TODO: Revoir cette zone
                         if (self.GetNbEntrees = 0) then
                         begin
                           SetDefaultCoords(ConvertirEnNombreReel(PrmsLn[COL_XENTREE], 0.00),
                                            ConvertirEnNombreReel(PrmsLn[COL_YENTREE], 0.00),
                                            ConvertirEnNombreReel(PrmsLn[COL_ZENTREE], 0.00)
                                           );

                           SetRefSeriePoint(StrToIntDef(PrmsLn[COL_ENT_SER],1),
                                            StrtoIntDef(PrmsLn[COL_ENT_ST], 0)
                                           );

                         end;
                         // Ajouter les entrées
                         // le nombre d'entrées retenu est celui décompté dans
                         // la section -5
                         with UneEntree do
                         begin
                           //eNumEntree := ProvListeEntrees.Count; //GetNbEntrees + 1;
                           QN := GetNbEntrees();
                           //eNumEntree := GetNbEntrees + 1;
                           eIDTerrain := '';  // non supporté par TOPOROBOT
                           // On ajoute l'entrée récupérée en -6
                           try
                             eNomEntree:=ProvListeEntrees.Strings[QN];
                           except
                             AfficherMessage(Format(rsWARNINGENTRYADDED,[QN]));
                             eNomEntree:= Format(rsRD_TAB_ENTRANCE,[QN]);
                           end;
                           eXEntree  := ConvertirEnNombreReel(PrmsLn[COL_XENTREE], 0.00);
                           eYEntree  := ConvertirEnNombreReel(PrmsLn[COL_YENTREE], 0.00);
                           eZEntree  := ConvertirEnNombreReel(PrmsLn[COL_ZENTREE], 0.00);

                           //------------
                           // entrées non géoréférencées ?
                           if (eXEntree < 100.00) or
                              (eYEntree < 100.00) or
                              (eZEntree < 100.00)
                           then WriteWarning(Format(GetResourceString(rsRD_TAB_ENTR_NOGEOREF),
                                                   [NoLigneTEXT,QN, eNomEntree]));
                           eRefSer   := StrToIntDef(PrmsLn[COL_ENT_SER],1);
                           eRefSt    := StrtoInt(PrmsLn[COL_ENT_ST]);
                           if ((eRefSer < 1) or (eRefSt < 0)) then
                                  WriteWarning(Format(rsRD_TAB_ENTR_BADLINK,
                                       [NoLigneTEXT, QN, eNomEntree, eRefSer, eRefSt]));
                           eObserv   := '';
                         end; // with UneEntree
                         // n'ajouter l'entrée que si elle comporte une RefSerie non nulle
                         mioumiou := (UneEntree.eRefSer > 0);
                         if (mioumiou) then self.AddEntree(UneEntree);
                         // index courant
                         IdxCourant := GetNbEntrees - 1;
                         Continue;
                       end;
                   -4: begin
                         AfficherMessage('-4'+ rsRD_TAB_IGNORED_SEC); // sessions; ignoré
                         Continue;
                       end;
                   -3: begin
                         AfficherMessage('-3'+ rsRD_TAB_IGNORED_SEC); // réservé; ignoré
                         Continue;
                       end;
                   -2: begin // Expés
                         //Ajoute l'expé dans la liste
                         PrmsLn := SplitFixedColLine(LigneTEXT, [1, 2, 7, 13, 17, 21, 25, 36, 50, 64, 65, 73, 77, 81,100]);
                         with UneExpe do
                         begin
                           IDExpe        := StrToInt(PrmsLn[COL_NUM_ITEM]);
                           if (IDExpe<=0) then begin
                              ErrMsg:=rsRD_TAB_BAD_TRIP;
                              raise Exception.Create(ErrMsg);
                           end;
                           // décomposition de la date
                           str := Trim(PrmsLn[COL_DATE_EXPE]);
                           JourExpe      := StrToIntDef(Copy(STR, 0, 2), 0);
                           MoisExpe      := StrToIntDef(Copy(STR, 4, 2), 1);
                           AnneeExpe     := StrToIntDef(Copy(STR, 7, 2), 1);
                           if (JourExpe = 0) or (MoisExpe = 0) or (AnneeExpe = 0) then
                           begin
                             WriteWarning(Format(rsRD_TAB_BAD_DATE,
                                                [NoLigneTEXT]));
                             SafeDecodeDate(Now, yyy, mmm, ddd);
                             JourExpe := ddd;
                             MoisExpe := mmm;
                             AnneeExpe:= yyy;
                           end;
                           Operateur     := PrmsLn[COL_SPELEOMETRE];             // spéléomètre
                           ClubSpeleo    := PrmsLn[COL_SPELEOGRAPH];             // spéléographe
                           ModeDecl      := StrToIntDef(PrmsLn[COL_MODEDECL], 0);   // déclinaison auto ?
                           Declinaison   := ConvertirEnNombreReel(PrmsLn[COL_DECL_VALUE], 0.00); // déclinaison
                           //Inclinaison   := ConvertirEnNombreReel(PrmsLn[COL_INCL_VALUE], 0.00); // correction clino x10
                           Couleur       := StrToIntDef(PrmsLn[COL_COULEUR], 0);  // couleur
                           Commentaire   := ''; //PrmsLn[11];            // commentaire

                         end;  // with UneExpe
                         AddExpe(UneExpe);
                         Continue;
                       end;
                   -1: begin // Codes
                         //                                                           U_AZ
                         //                                                            |   U_P
                         //                                                            |   |   PSI_L
                         //                                                            |   |   |   PSI_A
                         //                                                            |   |   |   |   PSI_P
                         //                                                            |   |   |   |   |   FACT
                         //                                                            |   |   |   |   |   |   ANG_LIM
                         //                                                            |   |   |   |   |   |   |
                         PrmsLn := SplitFixedColLine(LigneTEXT, [1, 2, 7, 13, 17, 21, 25, 33, 41, 49, 57, 65, 73, 81, 100]);
                         with UnCode do
                         begin
                           IDCode     := StrToInt(PrmsLn[COL_NUM_ITEM]);    // ID Code
                           if (IDCode<=0) then
                           begin
                             ErrMsg:=rsRD_TAB_BAD_CODE;
                             raise Exception.Create(ErrMsg);
                           end;
                           GradAz     := ConvertirEnNombreReel(PrmsLn[COL_U_AZIMUT], 360.00);  // unité boussole
                           GradInc    := ConvertirEnNombreReel(PrmsLn[COL_U_PENTE], 360);  // unite  CLINO
                           PsiL       := ConvertirEnNombreReel(PrmsLn[COL_PSI_L], 0.01);  // precision longueur
                           PsiAz      := ConvertirEnNombreReel(PrmsLn[COL_PSI_AZ], 0.5);  // precision azimut
                           PsiP       := ConvertirEnNombreReel(PrmsLn[COL_PSI_P], 0.5);  // precision pente
                           // TODO: Support de FactLong a valider
                           FactLong   := 1.00; // ConvertirEnNombreReel(PrmsLn[COL_FACT_LONG], 1.00)/100.00;  // Facteur des longueurs
                           AngLimite  := ConvertirEnNombreReel(PrmsLn[COL_ANG_LIMITE], 0.00);  // angle limite
                           ErreurTourillon  := 0.00;
                           ParamsFuncCorrAz := MakeTParamFoncCorrectionAngulaire(0.00, 0.00, 0.00);   // non supporté par LimeLight
                           ParamsFuncCorrAz := MakeTParamFoncCorrectionAngulaire(0.00, 0.00, 0.00);

                           Commentaire:= ''; //PrmsLn[9];              // commentaire
                         end; // with UnCode
                         AddCode(UnCode);
                         Continue;
                       end;
                    0: begin
                         Continue;  // pas de section 0
                       end
             else // on est dans les séries
               begin
                 PrmsLn := SplitFixedColLine(LigneTEXT, [1, 2, 7, 13, 17, 21, 25, 80, 100]);
                 // Si le préfixe 2 (2e colonne) =-1 =>nouvelle série
                 Prefix1 := StrToInt(PrmsLn[COL_NUM_SECTION]);
                 if (Prefix1 > 0) then
                 begin
                   Prefix2 := StrToInt(PrmsLn[COL_NUM_ITEM]);
                   // C'est le header de série ?
                   // les headers de série comportent deux lignes !
                   if (Prefix2 = -2) then
                   begin
                     // première ligne
                     NomDeLaSerie := Trim(PrmsLn[COL_VAR_DATA]);
                     AfficherMessage(Format('Une serie a ajouter: Prefix1: %d - Prefix2 = %d - %s',[Prefix1, Prefix2, NomDeLaSerie]));
                     // Lecture de la 2e ligne
                     Inc(NoLigneTEXT);
                     ReadLn(pTEXT, LigneTEXT);

                     // décomposition
                     //*******************************************************************************
                     //                                                           SD
                     //                                                            |   PD
                     //                                                            |   |   SA
                     //                                                            |   |   |   PA
                     //                                                            |   |   |   |   NB
                     //                                      Serie                 |   |   |   |   |   CH
                     //                                                            |   |   |   |   |   |   OBS
                     //                                                            |   |   |   |   |   |   |
                     PrmsLn := SplitFixedColLine(LigneTEXT, [1, 2, 7, 13, 17, 21, 25, 33, 41, 49, 57, 65, 73, 81, 100]);
                     //AfficherMessage(Format('%d %d - %s', [Prefix1, Prefix2,  UneSerie.GetNomSerie]));
                     // si c'est la première série, on crée
                     AfficherMessage(LigneTEXT);
                     AfficherMessage(PrmsLn[1]);
                     Prefix1 := StrToInt(PrmsLn[1]);
                     if (Prefix1 = 1) then
                     begin
                       UneSerie := TObjSerie.Create;
                       UneSerie.SetNumeroSerie(1);
                       AfficherMessage(Format('Prefix = %d - Serie 1 creee',[Prefix1]));
                     end
                     else  // sinon on ferme la série courante et on crée la suivante
                     begin
                       self.AddSerie(UneSerie);
                       UneSerie := TObjSerie.Create;
                     end;
                     with UneSerie do
                     begin
                       SetNumeroSerie(TNumeroSerie(Prefix1));
                       SetNomSerie(NomDeLaSerie);
                       SetSeriePtExtremites(StrToInt(PrmsLn[COL_SER_DEP]),
                                            StrToInt(PrmsLn[COL_PT_DEP]),
                                            StrToInt(PrmsLn[COL_SER_ARR]),
                                            StrToInt(PrmsLn[COL_PT_ARR])
                                           );
                       SetChanceObstacle(StrToInt(PrmsLn[COL_CHANCE]),
                                         StrToInt(PrmsLn[COL_OBSTACLE]));
                       SetNumeroReseau(0); // non supporté par TOPOROBOT
                       SetObsSerie('');
                       SetRaideur(1.00); // Raideur: Non supporté par TOPOROBOT
                     end; //with UneSerie do
                   end else
                   begin // on lit les visées
                     //                 ('-->'+LigneTab);
                     //*******************************************************************************
                     //                                                           Long
                     //                                                            |   Azimut
                     //                                                            |   |   Pente
                     //                                      Parenthèse            |   |   |   LG
                     //                                      | Serie               |   |   |   |   LD
                     //                                      | |   Station         |   |   |   |   |   HZ
                     //                                      | |   |  Code         |   |   |   |   |   |   HN
                     //                                      | |   |  |    |Session|   |   |   |   |   |   |
                     //                                      | |   |  |    | Expé  |   |   |   |   |   |   |
                     PrmsLn := SplitFixedColLine(LigneTEXT, [1, 2, 7, 13, 17, 21, 25, 33, 41, 49, 57, 65, 73, 81, 100]);

                     with qVisee do
                     begin
                       TypeVisee   := tgDEFAULT;
                       Code        := StrToInt(PrmsLn[COL_CODE]);
                       Expe        := StrToInt(PrmsLn[COL_EXPE]);
                       Longueur    := ConvertirEnNombreReel(PrmsLn[COL_LONG], 0.00);
                       if (Longueur < 0.00) then
                       begin
                         WriteWarning(Format(rsRD_TAB_NEG_LONG, [NoLigneTEXT, Longueur]));
                         Longueur:=Abs(Longueur);
                       end;
                       Azimut      := ConvertirEnNombreReel(PrmsLn[COL_AZ], 0.00);
                       Pente       := ConvertirEnNombreReel(PrmsLn[COL_P], 0.00);
                       LG          := ConvertirEnNombreReel(PrmsLn[COL_LG], 0.00);  //LG
                       LD          := ConvertirEnNombreReel(PrmsLn[COL_LD], 0.00);  //LD
                       HZ          := ConvertirEnNombreReel(PrmsLn[COL_HZ], 0.00);
                       HN          := ConvertirEnNombreReel(PrmsLn[COL_HN], 0.00);
                       Commentaires:= '';
                       IDSecteur   := 0;
                       IDTerrainStation := '';   // Non supporté par TOPOROBOT
                       TypeVisee        := tgDEFAULT; // Non supporté par TOPOROBOT
                     end; // with qVisee do
                     UneSerie.AddVisee(qVisee);
                   end;   // if Prefix2
                 end; // case
               end;
             end; //  case Prefix1 of
           end;
           IdxCourantStation := UneSerie.GetNbVisees - 1;
           // imprimer la ligne traitée
           //AfficherMessage('Ligne traitée');
         except
           AfficherMessageErreur(Format('[%d] %s', [NoLigneTEXT, LigneTEXT]));
         end;
       end;  //while Not Eof(pTEXT) do
       // On ferme la dernière série
       AddSerie(UneSerie);
       //--------------------------------------------------------------------------
       // Cause courante de plantages: Absence de la série 1
       VerifieSerie1();
       // Actions de réparation / vérif
       CheckerLesDonneesTopo();
       // recenser IDs terrain
       RecenserLesIDsTerrain();
       Result := 1;
       AfficherMessageErreur('Fichier TEXT OK');
       RecenserPointsOfInterest();
     finally
       CloseFile(pTEXT);
     end; // niveau 1: pour le fichier Text
  finally
    ProvListeEntrees.Free;
  end; //try..finally niveau 0
end;


//******************************************************************************
procedure TToporobotStructure2012.SaveToFile(const FichierTAB: TStringDirectoryFilename;
  const ModeSaveTAB: TModeSaveTAB; const TextFileFormat: TTextFileFormat);
const
  LINE_CODE    = '%d'+TAB+'%d'+TAB+
                 '%.2f'+TAB+'%.2f'+TAB+
                 '%.2f'+TAB+'%.2f'+TAB+'%.2f'+TAB+
                 '%.2f'+TAB+'%.2f'+TAB+
                 '%s';
  LINE_STATION = '%d'+TAB+'%d'+TAB+
                 '%d'+TAB+'%d'+TAB+
                 '%.2f'+TAB+'%.2f'+TAB+'%.2f'+TAB+
                 '%.2f'+TAB+'%.2f'+TAB+'%.2f'+TAB+'%.2f'+TAB+
                 '%s';
  FMT_PARAMS_CORR_FUNC = '%.6f' + TAB + '%.6f' + TAB + '%.6f';
var
  ENDL     : string;
  Sr, St, i, QNbNotesVersions, QNbAnnotations, QNbFiltresPerso: integer;
  pTAB: TextFile;
  Entrance: TEntrance;
  Expe: TExpe;
  Code: TCode;
  Serie: TObjSerie;
  Station: TUneVisee;
  s, s1: string;
  R: TReseau;
  SC: TSecteur;
  VA: TViseeAntenne;
  WU: string;
  EPSG: TLabelSystemesCoordsEPSG;
  ANN: TAnnotation;
  NV: TReleaseNotes;
  MyFiltrePerso: TFiltrePersonnalise;
  procedure WrtLn(const Str: string);
  begin
    Write(pTAB, Str+ENDL);
  end;
  procedure WrtCommentaire(const Str: string);
  begin
    if (ModeSaveTAB = mtabEXTENDEDTAB) then
      Write(pTAB, '#### '+Str+ENDL);
  end;
begin
  case TextFileFormat of  // mise en place des fins de lignes
    tfWINDOWS: begin s:='Windows';   ENDL := #13+#10; end;
    tfMAC    : begin s:='Macintosh'; ENDL := #13    ; end;
    tfUNIX   : begin s:='Unix';      ENDL := #10    ; end;
  end;
  AfficherMessage(Format('TToporobotStructure.SaveFile(%s) as %s format',[FichierTAB, s]));
  try
    // Sauvegarder ancien fichier

    AssignFile(pTAB, FichierTAB);
    ReWrite(pTAB);
    AfficherMessage('   Saving header');
    // Sections ajoutées au format TOPOROBOT
    // Fichiers XTB uniquement
    if (ModeSaveTAB = mtabEXTENDEDTAB) then
    begin
      // Section -20: Date de dernières modifications, version du fichier et du logiciel
      AfficherMessage('--> Horodating saves at #20 section');

      WrtLn(Format('%d'+TAB+'%s'+TAB+'%s'+TAB+'%s',
                   [-20,
                    DateToStr(Now), TimeToStr(Now),
                    self.FDataBaseName
                   ]));
      // Sections -19 : Libellé et commentaires de l'étude (XTB uniquement)
      AfficherMessage('--> Writing general infos at #-19 to #-18 sections');
      WrtLn(Format('%d'+TAB+'%s' +TAB+'%s',
                   [-19, self.FNomEtude, self.FCommentaireEtude]));
      // Section -18: Espace de noms
      WrtLn(Format('%d'+TAB+'%s', [-18, self.FNameSpace]));
      //************************************************************************
      // Section -18: Notes de version (XTB uniquement);
      {$IFDEF SUPPORT_NOTES_VERSION}
      AfficherMessage('--> Saving #-18 Release_notes section');

      WrtLn('');
      WrtCommentaire('Release_notes list');
      WrtLn('');
      WrtLn('#' + TAB + 'Date' + TAB + 'Author' + TAB + 'Description');
      WrtLn('');
      QNbNotesVersions := GetNbNotesVersion();
      if (QNbNotesVersions > 0) then
      begin
        for i := 0 to QNbNotesVersions - 1 do
        begin
          NV := GetNoteVersion(i);
          WrtLn(Format('%d'+ TAB + '%s' + TAB + '%s' + TAB + '%s',
                       [-17,
                        DateTimePascalToDateTimeSQL(NV.Date),
                        NV.Author, NV.Description
                        ]));
        end;
      end;
      {$endif SUPPORT_NOTES_VERSION}
      // Section -16: Filtres personnalisés (XTB uniquement);
      AfficherMessage('--> Saving filters at #-16 section');
      WrtLn('');
      QNbFiltresPerso := GetNbFiltresPersos();
      WrtCommentaire('#-16'+ TAB + 'IDFiltre' + TAB + 'NomFiltre' + TAB + 'CouleurFiltre_R' + TAB + 'CouleurFiltre_G' + TAB + 'CouleurFiltre_B' + TAB + 'FiltreExpr' + TAB + 'Description');
      WrtCommentaire(Format('%d filtres perso', [QNbFiltresPerso]));

      if (QNbFiltresPerso > 0) then
      begin
        for i := 0 to QNbFiltresPerso - 1 do
        begin
          MyFiltrePerso := self.GetFiltrePerso(i);

          WrtLn(Format('%d' + TAB + '%d' + TAB +
                       '%d' + TAB + '%d' + TAB + '%d' + TAB +
                       '%s' + TAB +
                       '%s' + TAB + '%s',
                       [-16, i,
                        Red(MyFiltrePerso.CouleurFiltre), Green(MyFiltrePerso.CouleurFiltre), Blue(MyFiltrePerso.CouleurFiltre),
                        MyFiltrePerso.NomFiltre,
                        MyFiltrePerso.Expression, MyFiltrePerso.Description
                        ]));
          //*)
        end;
      end;
      //*)

      // Section -15: Système de coordonnées géographiques (XTB uniquement)
      // TODO: Désactivé pour l'instant - A réimplémenter
      AfficherMessage('--> Writing coordinates system at #-15 section');
      WrtLn('');
      WrtCommentaire('#-15'+ TAB + 'Code EPSG' + TAB + 'Coordinates system description');
      // -15	31563	LT3	Lambert 3 France
      EPSG := GetCodeEPSGSystemeCoordonnees;
      WrtLn('');
      WrtLn('#' + TAB + 'CodeEPSG' + TAB + 'NomEPSG');
      WrtLn('');
      WrtLn(Format('%d'+TAB+'%d'+TAB+'%s',  [-15, EPSG.CodeEPSG, EPSG.NomEPSG]));

      // Section -11 Annotations (XTB uniquement)

      AfficherMessage('--> Saving #-11 Annotations section');

      WrtLn('');
      WrtCommentaire('Annotations list');
      WrtLn('');
      WrtLn('#' + TAB + 'Idx' + TAB + 'Basepoint.Serie' + TAB + 'Basepoint.Point' + TAB +
            'OffsetX' + TAB + 'OffsetY' + TAB + 'OffsetZ' + TAB +
            'IdxStyle' + TAB + 'Alignment' + TAB +
	          'Texte');
      WrtLn('');
      QNbAnnotations := GetNbAnnotations();
      if (QNbAnnotations > 0) then
      begin
        for i := 0 to QNbAnnotations - 1 do
        begin
          ANN := GetAnnotation(i);
          WrtLn(Format('%d'+ TAB + '%d' + TAB +
                       '%d'+ TAB + '%d' + TAB +       // série, point
                       '%.3f'+ TAB + '%.3f' + TAB + '%.3f'+TAB+    // offset
                       '%d'+ TAB + '%d' + TAB + // index style
                       '%s',
                       [-11, i,
                        ANN.Basepoint.eSerie, ANN.Basepoint.ePoint,
                        ANN.Offset.X, ANN.Offset.Y, ANN.Offset.Z,
                        ANN.IdxStyle, ANN.PosPtDeBase,
                        ANN.Texte
                        ]));
        end;
      end;


      //*************************************
      // Section -10: Secteurs (Format XTB uniquement)
      AfficherMessage('--> Saving #-10 Sectors section');

      WrtLn('');
      WrtCommentaire('Sectors list');
      WrtLn('');
      WrtLn('#' + tab + 'IdxSecteur' + tab + 'Color R' + tab + 'Color G' + tab + 'Color B' + tab +
	    'NomSecteur');
      WrtLn('');

      if (GetNbSecteurs > 1) then
      begin
        for i:=1 to GetNbSecteurs - 1 do
        begin
          SC := GetSecteur(i);
          WrtLn(Format('%d'+TAB+'%d'+TAB+
                       '%d'+TAB+'%d'+TAB+'%d'+TAB+
                       '%s',
                       [-10,
                        i,
                        Red(SC.CouleurSecteur),
                        Green(SC.CouleurSecteur),
                        Blue(SC.CouleurSecteur),
                        SC.NomSecteur
                        ]));
        end;
      end;
      // Section -9: Visées en antennes (Format XTB uniquement)
      if (GetNbAntennes > 0) then begin
        AfficherMessage('--> Saving #-9 Antenna shots section');

        WrtLn('');
        WrtCommentaire('Antenna-shots list');
        WrtLn('');
        WrtLn('#' + TAB + 'ID' + TAB + 'ID Reseau' + TAB + 'ID Secteur' + TAB +
              'Serie' + Tab + 'Point' + TAB +
              'Code' + TAB + 'Expe' + TAB + 'IDTerrain' + TAB +
              'Longueur' + TAB + 'Azimut' + TAB + 'Pente' + TAB +
              'Observ.');
        WrtLn('');

        for i := 1 to GetNbAntennes - 1 do
        begin
          VA := GetViseeAntenne(i);
          //-9	1	2	1102	10	101	101	GHLMF	12.56	289.00	-14.00	le chat dans la souricière
          WrtLn(Format('%d'+TAB+'%d'+TAB+'%d'+TAB+'%d'+
                            TAB+'%d'+TAB+'%d'+ // série, point
                            TAB+'%d'+TAB+'%d'+ // code, expé
                            TAB+'%s'+          // ID terrain
                            TAB+'%.2f'+TAB+'%.2f'+TAB+'%.2f'+
                            TAB+'%s',
                   [-9, i, VA.Reseau, VA.Secteur,
                           VA.SerieDepart, VA.PtDepart,
                           VA.Code, VA.Expe,
                           VA.IDTerrainStation,
                           VA.Longueur, VA.Azimut, VA.Pente,
                           VA.Commentaires
                   ]));
        end;

      end;

      // Section -8: Réseaux (Format XTB uniquement)
      AfficherMessage('--> Saving #-8 Networks section');
      WrtLn('');
      WrtCommentaire('Networks list');
      WrtLn('');
      WrtLn('#' + tab + 'IdxReseau' + tab +
                'Color R' + tab + 'Color G' + tab + 'Color B' + tab +
	        'TypeReseau' + tab + 'NomReseau' + tab + 'ObsReseau');
      WrtLn('');
      if (GetNbReseaux > 1) then
      begin
        for i:=1 to GetNbReseaux - 1 do
        begin
          R:=GetReseau(i);
          WrtLn(Format('%d'+TAB+'%d'+TAB+
                       '%d'+TAB+'%d'+TAB+'%d'+TAB+
                       '%d'+TAB+
                       '%s'+TAB+
                       '%s',
                       [-8,
                        i,
                        Red(R.ColorReseau),
                        Green(R.ColorReseau),
                        Blue(R.ColorReseau),
                        R.TypeReseau,
                        R.NomReseau,
                        R.ObsReseau
                        ]));
        end;
      end;
    end; //  if (ModeSaveTAB = mtabEXTENDEDTAB) then begin
    // Classeur (débrayé)

    //WriteLn(pTAB, Format('%d'+TAB+'%d'+TAB+'%d'+TAB+'%d'+TAB+'%d',[-7,1,1,1,1]));
    //WrtLn(Format('%d'+TAB+'%d'+TAB+'%d'+TAB+'%d'+TAB+'%d',[-7,1,1,1,1]));
    // Sauvegarde des entrées
    WrtLn('');
    WrtCommentaire('Entrances list');
    WrtLn('');
    WrtLn('#' + TAB + 'ID Entrance' + TAB + 'NomEntrance' +
                TAB + 'Serie' + TAB + 'Station' +
                TAB + 'X Entrance' + TAB + 'Y Entrance' + TAB + 'Z Entrance' +
                TAB +  'Observ');

    WrtLn('');
    (*
    #	ID Entrance	NomEntrance

-6	1	Citon I.1
-6	2	Citon II.1
-6	3	Citon II.2	1151.25
-6	4	Citon II.3	1164.3
-6	5	Citon  II.4 Petit Macon
-6	6	Puits 2386.3	2386.3
-6	7	Puits Cornet 1503.2	1503.2
-6	8	Citon II.8	2017.3
#	X Entrance	Y Entrance	Z Entrance	Serie	Station	Observ

-5	1	378467.00	3281285.00	16.00	1	0	Citon 1
-5	2	378373.00	3281021.00	35.00	1000	0	Entrée descenderie
-5	3	378323.17	3280954.94	28.74	1151	25	La plus grande entrée du réseau, végétation et fossé difficile à franchir
-5	4	378328.67	3280911.93	27.32	1164	3	Débouche dans un sous-bois
-5	5	378170.00	3280250.00	16.00	1600	190	Débouche dans une propriété privée
-5	6	378520.72	3280913.56	43.58	2386	3	Puits remontant en surface
-5	7	379180.71	3280631.96	50.61	1503	2	Puits remontant en surface
-5	8	378401.53	3281076.70	26.61	2017	3	Débouche dans une propriété privée

    //*)
    AfficherMessage('--> Saving #-6 Entrances section');

    for i:=0 to GetNbEntrees - 1 do
    begin
      Entrance := GetEntree(i);
      WrtLn(Format('%d' + TAB + '%d' + TAB + '%s' + TAB + '%s' + TAB +
                   '%d' + TAB + '%d' + TAB +
                   '%.2f' + TAB + '%.2f' + TAB + '%.2f' + TAB +
                   '%s',
                  [-6, i+1,
                   Entrance.eNomEntree,
                   Entrance.eIDTerrain,
                   Entrance.eRefSer,  Entrance.eRefSt,
                   Entrance.eXEntree, Entrance.eYEntree, Entrance.eZEntree,
                   Entrance.eObserv ]));
    end;





    AfficherMessage('--> Saving #-5 Deprecated Entrances section');
    WrtLn('#' + TAB + 'X Entrance' + TAB + 'Y Entrance' + TAB + 'Z Entrance' + TAB +
              'Serie' + TAB + 'Station' + TAB +
              'Observ' + TAB + '*** Deprecated section. Data moved to #6 section ***');
    WrtLn('');
    for i:=0 to GetNbEntrees - 1 do
    begin
      Entrance:=GetEntree(i);
      WrtLn(Format('%d' + TAB + '%d' + TAB +
                   '%.2f' + TAB + '%.2f' + TAB + '%.2f' + TAB +
                   '%d' + TAB + '%d' + TAB + '%s',
                          [-5, i+1,
                          Entrance.eXEntree,
                          Entrance.eYEntree,
                          Entrance.eZEntree,
                          Entrance.eRefSer,
                          Entrance.eRefSt,
                          Entrance.eObserv
                          ]));
    end;
    WrtLn('');
    WrtCommentaire('Trips list');
    WrtLn('');
    AfficherMessage('--> Saving Expes');
    if (GetNbExpes > 0) then
    begin
      for i:=1 to GetNbExpes - 1 do
      begin
        Expe := GetExpe(i);
        with Expe do begin
          //WriteLn(pTAB, Format('%d'+TAB+'%d'+TAB+
          WrtLn(Format('%d'+TAB+'%d'+TAB+
                             '%d'+TAB+'%d'+TAB+'%d'+TAB+
                             '%s'+TAB+'%s'+TAB+
                             '%d'+TAB+'%.2f'+TAB+
                             '%d'+TAB+'%d'+TAB+ // anciennement '%f'+TAB+'%d'+TAB+
                             '%s',
                             [-2, IDExpe,
                              JourExpe, MoisExpe, AnneeExpe,
                              Operateur, ClubSpeleo,
                              ModeDecl, Declinaison,
                              0, Couleur, //Inclinaison, Couleur,
                              Commentaire
                             ]));
         end;
      end;
    end;
    // Codes
    //-1	999	400.00	400.00	0.00	0.00	0.00	100.00	-100.00	Fixpunkt
    AfficherMessage('--> Saving Codes');
    WrtLn('');
    WrtCommentaire('Instruments codes list');
    WrtLn('');
    if (GetNbCodes > 1) then
    begin
      for i:=1 to GetNbCodes - 1 do
      begin
        Code := GetCode(i); // Laisser tel quel
        with Code do begin
          case ModeSaveTAB of
            mtabEXTENDEDTAB:
              //-1	1	360.00	360.00	0.05	1.00	1.00	1.00	100.00		5314011	0.000000	0.000000	0.000000	0.000000	0.000000	0.000000
              WrtLn(Format(LINE_CODE + TAB + '%d' + TAB +
                           FMT_PARAMS_CORR_FUNC + TAB +
                           FMT_PARAMS_CORR_FUNC + TAB +
                           '%.3f',
                             [-1, IDCode,
                                  GradAz, GradInc,
                                  PsiL, PsiAz, PsiP,
                                  FactLong,
                                  AngLimite,
                                  Commentaire,
                                  0, //ReservedInt,
                                  ParamsFuncCorrAz.Co, ParamsFuncCorrAz.ErreurMax, ParamsFuncCorrAz.PosErrMax,
                                  ParamsFuncCorrInc.Co, ParamsFuncCorrInc.ErreurMax, ParamsFuncCorrInc.PosErrMax,
                                  ErreurTourillon
                             ]));
            mtabTOPOROBOT:
              WrtLn(Format(LINE_CODE,
                             [-1, IDCode,
                                  GradAz, GradInc,
                                  0.10, 1.0, 1.0, //PsiL, PsiAz, PsiP,
                                  100.00, 0.00, //FactLong, AngLimite,
                                  SafeTruncateString(Commentaire, 50)
                             ]));
          end;
        end; //with Code
      end;
      WrtLn('');
    end;
    // Séries
    case ModeSaveTAB of
      mtabEXTENDEDTAB:
        AfficherMessage('--> Saving Series at mtabEXTENDEDTAB mode');
      mtabTOPOROBOT:
        AfficherMessage('--> Saving Series at mtabTOPOROBOT mode');
    end;
    WrtLn('');
    WrtCommentaire('Series and stations');
    WrtLn('');


    for Sr:=1 to GetNbSeries - 1 do
    begin
      WrtLn('');
      Serie := GetSerie(Sr);
      AfficherMessage(Format('---> %d : %s (%d pts)',[Serie.GetNumeroDeSerie(), Serie.GetNomSerie, Serie.GetNbVisees]));
      with Serie do begin
        // ligne de tête
        // Numero est remplacé par IndexSerie
        // /!\ 'NbPoints' est remplacé par 'NbPoints - 1'
        case ModeSaveTAB of
          mtabEXTENDEDTAB:
            //WriteLn(pTAB, Format('%d'+TAB+'%d'+TAB+
            WrtLn(Format('%d'+TAB+'%d'+TAB+
                                 '%d'+TAB+'%d'+TAB+
                                 '%d'+TAB+'%d'+TAB+
                                 '%d'+TAB+
                                 '%d'+TAB+'%d'+TAB+
                                 '%s'+TAB+'%s'+TAB+
                                 '%d'+TAB+'%.2f'+TAB+
                                 '%d',
                                 [GetNumeroDeSerie(), -1,
                                  GetNoSerieDep(), GetNoPointDep(),
                                  GetNoSerieArr(), GetNoPointArr(),
                                  GetNbVisees() - 1,
                                  GetChance(), GetObstacle(),
                                  GetNomSerie(),
                                  GetObsSerie(),
                                  GetNumeroReseau(),
                                  GetRaideur(),
                                  GetNumeroEntrance()
                                 ]));

          mtabTOPOROBOT:
            //WriteLn(pTAB, Format('%d'+TAB+'%d'+TAB+
            WrtLn(Format('%d'+TAB+'%d'+TAB+
                                 '%d'+TAB+'%d'+TAB+
                                 '%d'+TAB+'%d'+TAB+
                                 '%d'+TAB+
                                 '%d'+TAB+'%d'+TAB+
                                 '%s', //+TAB+'%s',
                                 [GetNumeroDeSerie(), -1,
                                  GetNoSerieDep, GetNoPointDep,
                                  GetNoSerieArr, GetNoPointArr,
                                  GetNbVisees - 1,
                                  GetChance, GetObstacle,
                                  SafeTruncateString(GetNomSerie, 20)
                                 ]));

        end;

        for St:=0 to Serie.GetNbVisees - 1 do
        begin
          Station:=Serie.GetVisee(St);
          with Station do begin
            case ModeSaveTAB of
              mtabEXTENDEDTAB:
                begin
                  WrtLn(Format(LINE_STATION+TAB+'%s'+TAB+'%d'+TAB+'%d',
                                  [Serie.GetNumeroDeSerie(), St,
                                   Code, Expe,
                                   Longueur, Azimut, Pente,
                                   LG, LD, HZ, HN,
                                   Commentaires,
                                   IDTerrainStation,
                                   TypeVisee,
                                   IDSecteur
                                  ]));
                end;
              mtabTOPOROBOT:
                begin
                  WrtLn(Format(LINE_STATION,
                                  [Serie.GetNumeroDeSerie(), St,
                                   Code, Expe,
                                   Longueur, Azimut, Pente,
                                   LG, LD, HZ, HN,
                                   Commentaires
                                  ]));
                end;
            end;
          end;
        end;
      end;
    end;
    AfficherMessage('TToporobotStructure.SaveToFile OK');
  finally
    CloseFile(pTAB);
  end;
end;


//* Méthodes relatives aux entrées
procedure TToporobotStructure2012.AddEntree(const AEntree: TEntrance);
begin
  FTableEntrees.AddElement(AEntree);
end;
function TToporobotStructure2012.GetEntree(const NumEntree: integer): TEntrance;
begin
  Result := FTableEntrees.GetElement(NumEntree);
end;
procedure TToporobotStructure2012.PutEntree(const NoEntree: integer; const AEntree: TEntrance);
begin
  FTableEntrees.PutElement(NoEntree, AEntree);
end;
function TToporobotStructure2012.RemoveEntree(const Idx: integer): boolean; inline;
begin
  FTableEntrees.RemoveElement(Idx);
end;
function TToporobotStructure2012.GetNbEntrees(): integer;
begin
  Result := FTableEntrees.GetNbElements;
end;



function TToporobotStructure2012.CreateNouveauEntree(): boolean;
var
  nb: Integer;
  EWE: TEntrance;
begin
  result := false;
  try
    nb  := GetNbEntrees;
    EWE := GetEntree(nb - 1);
    //EWE.eNumEntree   := getNextValidIdxEntree;
    EWE.eNomEntree   := 'Nouvelle entree';
    self.AddEntree(EWE);
    result := true;
    AfficherMessage(Format('%d entrées', [GetNbEntrees]));
  except
  end;
end;


// méthodes relatives aux réseaux
procedure TToporobotStructure2012.AddReseau(const AReseau: TReseau);
begin
  FTableReseaux.AddElement(AReseau);
end;
function TToporobotStructure2012.GetReseau(const NumReseau: integer): TReseau;
begin
  try
    Result := FTableReseaux.GetElement(NumReseau);
  except
    Result := FTableReseaux.GetElement(0);
  end;
end;
procedure TToporobotStructure2012.PutReseau(const NoReseau: integer; const AReseau: TReseau);
begin
  FTableReseaux.PutElement(NoReseau, AReseau);
end;
function TToporobotStructure2012.RemoveReseau(const Idx: integer): boolean; inline;
begin
  Result := FTableReseaux.RemoveElement(Idx);
end;
function TToporobotStructure2012.GetNbReseaux(): integer;
begin
  Result := FTableReseaux.GetNbElements;
end;

// méthodes relatives aux labels de terrain
procedure TToporobotStructure2012.AddLabelTerrain(const ALabelTerrain: TEtiquetteTerrain);
begin
  FTableIDsTerrain.AddElement(ALabelTerrain);
end;


function TToporobotStructure2012.GetLabelTerrain(const NumLabelTerrain: integer): TEtiquetteTerrain;
begin
  Result := FTableIDsTerrain.GetElement(NumLabelTerrain);
end;



procedure TToporobotStructure2012.PutLabelTerrain(const NoLabelTerrain: integer; const ALabelTerrain: TEtiquetteTerrain);
begin
  FTableIDsTerrain.PutElement(NoLabelTerrain, ALabelTerrain);
end;


function TToporobotStructure2012.RemoveLabelTerrain(const Idx: integer): boolean;
begin
  Result := FTableIDsTerrain.RemoveElement(Idx);
end;


function TToporobotStructure2012.GetNbLabelTerrain: integer;
begin
  Result := FTableIDsTerrain.GetNbElements;
end;





function TToporobotStructure2012.CreateNouveauReseau(): boolean;
var
  nb: Integer;
  EWE: TReseau;
begin
  Result := false;
  try
    nb  := GetNbReseaux;
    EWE := GetReseau(nb - 1);
    EWE.NomReseau   := 'Nouveau reseau';
    EWE.ObsReseau   := '';
    self.AddReseau(EWE);
    result := true;
  except
    ;
  end;
end;

function TToporobotStructure2012.GetLastReseau(): TReseau;
var
  Nb: Integer;
begin
  Nb := self.GetNbReseaux - 1;
  Result := GetReseau(Nb);
end;

function TToporobotStructure2012.HasNumeroReseau(const Idx: integer): boolean;
var
  n: Integer;
begin
  n := GetNbReseaux();
  Result := (Idx < n);
end;


// Méthodes relatives aux secteurs
procedure TToporobotStructure2012.AddSecteur(const ASecteur: TSecteur);
begin
  FTableSecteurs.AddElement(ASecteur);
end;
function TToporobotStructure2012.GetSecteur(const NumSecteur: integer): TSecteur;
begin
  try
    Result := FTableSecteurs.GetElement(NumSecteur);
  except
    Result := FTableSecteurs.GetElement(0);
  end;
end;

procedure TToporobotStructure2012.PutSecteur(const NoSecteur: integer; const ASecteur: TSecteur);
begin
  FTableSecteurs.PutElement(NoSecteur, ASecteur);
end;


function TToporobotStructure2012.RemoveSecteur(const Idx: integer): boolean; inline;
begin
  FTableSecteurs.RemoveElement(Idx);
end;
function TToporobotStructure2012.GetNbSecteurs(): integer;
begin
  Result := FTableSecteurs.GetNbElements;
end;



function TToporobotStructure2012.CreateNouveauSecteur(): boolean;
var
  nb: Integer;
  EWE: TSecteur;
begin
  result := false;
  try
    nb  := GetNbSecteurs();
    EWE := GetSecteur(nb - 1);
    EWE.NomSecteur   := 'Nouveau secteur';
    self.AddSecteur(EWE);
    result := true;
  except
  end;
end;

function TToporobotStructure2012.GetLastSecteur(): TSecteur;
var
  Nb: Integer;
begin
  Nb := self.GetNbSecteurs - 1;
  Result := GetSecteur(Nb);
end;

function TToporobotStructure2012.HasNumeroSecteur(const Idx: integer): boolean;
var
  n: Integer;
begin
  n := GetNbSecteurs();
  Result := (Idx < n);
end;

// méthodes relatives aux expés
//------------------------------------------
procedure TToporobotStructure2012.AddExpe(const LaExpe: TExpe);
begin
  FTableExpes.AddElement(LaExpe);
end;
procedure TToporobotStructure2012.PutExpe(const Idx: integer; const LaExpe: TExpe);
begin
  FTableExpes.PutElement(Idx, LaExpe);
end;
function TToporobotStructure2012.GetExpe(const Idx: integer): TExpe;
begin
  Result := FTableExpes.GetElement(Idx);
end;
function  TToporobotStructure2012.GetExpeByNumero(const NumeroExpe: TNumeroExpe): TExpe;
begin
  Result := FTableExpes.GetElementByIndex(NumeroExpe);
end;
function TToporobotStructure2012.GetNbExpes(): integer;
begin
  Result := FTableExpes.GetNbElements;
end;




function TToporobotStructure2012.CreateNouveauExpe(): boolean;
var
  EWE: TExpe;
  nb: Integer;
begin
  result := false;
  try
    nb  := GetNbExpes;
    EWE := GetExpe(nb - 1);
    EWE.IDExpe       := 1 + getMaxIdxExpe();
    EWE.Commentaire  := 'Nouvelle expe';
    self.AddExpe(EWE);
    result := true;
  except
  end;
end;

function TToporobotStructure2012.GetLastExpe(): TExpe;
var
  Nb: Integer;
begin
  Nb := self.GetNbExpes - 1;
  Result := GetExpe(Nb);
end;

function TToporobotStructure2012.RemoveExpe(const Idx: integer): boolean;
begin
  Result := FTableExpes.RemoveElement(Idx);
end;
// méthodes relatives aux codes
//------------------------------------------
procedure TToporobotStructure2012.AddCode(const LeCode: TCode);
begin
  FTableCodes.AddElement(LeCode);
end;
function TToporobotStructure2012.GetCode(const Idx: integer): TCode;
begin
  Result := FTableCodes.GetElement(Idx);
end;
//------------------------------------------
function  TToporobotStructure2012.GetCodeByNumero(const NumeroCode: TNumeroCode): TCode;
begin
  Result := FTableCodes.GetElementByIndex(NumeroCode);
end;
procedure TToporobotStructure2012.PutCode(const Idx: integer; const LeCode: TCode);
begin
  FTableCodes.PutElement(Idx, LeCode);
end;
function  TToporobotStructure2012.RemoveCode(const Idx: integer): boolean; inline;
begin
  FTableCodes.RemoveElement(Idx);
end;
function TToporobotStructure2012.GetNbCodes(): integer;
begin
  Result := FTableCodes.GetNbElements();
end;




function TToporobotStructure2012.CreateNouveauCode(): boolean;

var
  nb: Integer;
  EWE: TCode;
begin
  result := false;
  try
    nb  := GetNbCodes;
    EWE := GetCode(nb - 1);
    EWE.IDCode      := 1 + getMaxIdxCode;
    EWE.Commentaire := 'Nouveau code';
    self.AddCode(EWE);
    result := true;
  except
  end;
end;

function TToporobotStructure2012.GetLastCode(): TCode;
var
  Nb: Integer;
begin
  Nb := self.GetNbCodes - 1;
  Result := GetCode(Nb);
end;

// méthodes relatives aux antennes
procedure TToporobotStructure2012.AddViseeAntenne(const VA: TViseeAntenne);
begin
  FTableViseesAntenne.AddElement(VA);
end;


function  TToporobotStructure2012.GetViseeAntenne(const No: integer): TViseeAntenne;
begin
  Result := FTableViseesAntenne.GetElement(No);
end;
procedure TToporobotStructure2012.PutViseeAntenne(const No: integer; const VA: TViseeAntenne);

begin
  FTableViseesAntenne.PutElement(No, VA);
end;

function TToporobotStructure2012.ReattribuerCodesExpesViseesAntennes(): boolean;
var
  PreviousViseeAntenne, MyViseeAntenne: TViseeAntenne;
  Nb, i, QIdxSerie: Integer;
  WU, SameBasePt: Boolean;
  MySerie: TObjSerie;
  MyStation: TUneVisee;
  T: TDateTime;
begin
  result := false;
  AfficherMessage(format('%s.ReattribuerCodesExpesViseesAntennes()', [ClassName]));
  Nb := GetNbAntennes();
  if (Nb <= 1) then Exit(false);
  MyViseeAntenne := GetViseeAntenne(0);
  WU := GetSerieByNumeroSerie(MyViseeAntenne.SerieDepart, MySerie, QIdxSerie);
  MyStation := MySerie.GetVisee(MyViseeAntenne.PtDepart);
  T := now();
  AfficherMessage(DateTimeToStr(T));
  for i := 1 to Nb - 1 do
  begin
    try
      PreviousViseeAntenne := GetViseeAntenne(i-1);
      MyViseeAntenne       := GetViseeAntenne(i);
      // les extractions de série/stations sont coûteuses et ne doivent être effectuées
      // que si les basepoint des antennes n et n-1 diffèrent
      SameBasePt := (MyViseeAntenne.SerieDepart = PreviousViseeAntenne.SerieDepart) and
                    (MyViseeAntenne.PtDepart    = PreviousViseeAntenne.PtDepart);
      if ((i = 1) or (not SameBasePt)) then
      begin
        WU := GetSerieByNumeroSerie(MyViseeAntenne.SerieDepart, MySerie, QIdxSerie);
        MyStation := MySerie.GetVisee(MyViseeAntenne.PtDepart);
      end;
      // on reporte les code, expé, réseau et secteur de la station d'accrochage
      MyViseeAntenne.Reseau  := MySerie.GetNumeroReseau();
      MyViseeAntenne.Code    := MyStation.Code;
      MyViseeAntenne.Expe    := MyStation.Expe;
      MyViseeAntenne.Secteur := MyStation.IDSecteur;
      // et on actualise
      PutViseeAntenne(i, MyViseeAntenne);
      Result := True;
    except
      ;
    end;
  end;
  AfficherMessage(TimeToStr(Now() - T));
end;




function TToporobotStructure2012.RemoveViseeAntenne(const No: integer): boolean;
begin
  Result := FTableViseesAntenne.RemoveElement(No);
end;


function TToporobotStructure2012.GetNbAntennes(): integer;
begin
  Result := FTableViseesAntenne.GetNbElements;
end;
// supprimer les antennes orphelines
// et celles dont le numéro serie/station est spécifié (si qIdxSerie > 0)
// longMax: Longueur au-delà de laquelle la visée est supprimée
function TToporobotStructure2012.NettoyerAntennes(const longMax: double; const qIdxSerie, qIdxStation: integer): integer;
var
  EWE: TViseeAntenne;
  n, Nb: Integer;
  B0, B1, B2, B3, B4: boolean;
begin
  Result := 0;
  AfficherMessage(Format('%s.NettoyerAntennes: %.3f (origine: %d.%d)', [self.ClassName, longMax, qIdxSerie, qIdxStation]));
  n := GetNbAntennes;
  Nb := 0;
  //while Nb < FTableViseesAntenne.Count - 1 do
  for Nb := FTableViseesAntenne.Count - 1 downto 0 do
  begin
    EWE := GetViseeAntenne(Nb);
    B0  := (EWE.PtDepart = 0);
    B1  := not IsInRange(EWE.Longueur, 0.001, longMax);
    B2  := Not HasStationForAntennaShot(EWE);
    B3  := false;
    if (Assigned(FProcDispProgression) and (Nb mod 1000 = 0)) then FProcDispProgression('Antennes', Nb, 0, FTableViseesAntenne.Count);
    if (qIdxSerie > 0) then B3 := (EWE.SerieDepart = qIdxSerie) and
                                  (EWE.PtDepart = qIdxStation);
    if (B0 or B1 or B2 or B3) then
    begin
      RemoveViseeAntenne(Nb);
      AfficherMessageErreur(Format('Antenne retiree: %d.%d', [EWE.SerieDepart, EWE.PtDepart]));
      Result += 1;
      //Nb -= 1;
    end;
    //Nb += 1;
  end;
  AfficherMessage(Format('%d antennes supprimées', [Result]));
end;

// existence d'éléments correspondant à un index ?
(*
function TToporobotStructure2012.ExistsIdxEntrance(const Idx: TNumEntrance): boolean;
begin
  Result := FTableEntrees.ExistsElement(Integer(Idx));
end;
//*)


function TToporobotStructure2012.ExistsIdxCode(const Idx: integer): boolean;
begin
  Result := FTableCodes.ExistsElement(Idx);
end;

function TToporobotStructure2012.ExistsIdxExpe(const Idx: integer): boolean;
begin
  Result := FTableExpes.ExistsElement(Idx);
end;

function TToporobotStructure2012.ExistsIdxReseau(const Idx: integer): boolean;
begin
  Result := (Idx < FTableReseaux.GetNbElements());
end;

function TToporobotStructure2012.ExistsIdxEntrance(const Idx: integer): boolean;
begin
  Result := (Idx < FTableEntrees.GetNbElements());
end;



function TToporobotStructure2012.GetLastFindText(): string;
begin
  Result := FLastFindTextWhat;
end;

procedure TToporobotStructure2012.ExportListeEntreesCSV(const FileName: TStringDirectoryFilename);
var
  fp: TextFile;
  i, Nb: Integer;
  myEntrance: TEntrance;
begin
  Nb := GetNbEntrees;
  AfficherMessage(Format('%s.ExportListeEntreesCSV: %d', [ClassName, Nb]));
  if (Nb > 0) then
  begin
    try
      AssignFile(fp, FileName);
      ReWrite(fp);
      WriteLn(fp, 'No' + #9 + 'Nom' + #9 +
                  'X' + #9 + 'Y' + #9 + 'Z' + #9 +
                  'Serie' + #9 + 'Point' + #9 +
                  'Observ');
      for i := 0 to Nb - 1 do
      begin
        myEntrance := GetEntree(i);
        WriteLn(fp, Format('%d' + #9 + '%s' + #9 +
                           '%.2f' + #9 + '%.2f' + #9 + '%.2f' + #9 +
                           '%d' + #9 + '%d' + #9 +
                           '%s',
                           [i, myEntrance.eNomEntree,
                            myEntrance.eXEntree, myEntrance.eYEntree, myEntrance.eZEntree,
                            myEntrance.eRefSer, myEntrance.eRefSt,
                            myEntrance.eObserv
                           ]));
      end;
    finally
      CloseFile(fp);
    end;
  end;
end;

procedure TToporobotStructure2012.ExportListeReseauxCSV(const FileName: TStringDirectoryFilename);
var
  fp: TextFile;
  i, Nb: Integer;
  myReseau: TReseau;
begin
  Nb := GetNbReseaux;
  AfficherMessage(Format('%s.ExportListeReseauxCSV: %d', [ClassName, Nb]));
  if (Nb > 0) then
  begin
    try
      AssignFile(fp, FileName);
      ReWrite(fp);
      WriteLn(fp, 'No' + #9 + 'Type' + #9 + 'Couleur' + #9 +
                  'Nom' + #9 + 'Observ');
      for i := 0 to Nb -1 do
      begin
        myReseau := GetReseau(i);
        WriteLn(fp, Format('%d' + #9 + '%d' + #9 + '#%X' + #9 +
                           '%s' + #9 +
                           '%s',
                           [i,
                            myReseau.TypeReseau, myReseau.ColorReseau,
                            myReseau.NomReseau,
                            myReseau.ObsReseau
                           ]));
      end;
    finally
      CloseFile(fp);
    end;
  end;
end;

procedure TToporobotStructure2012.ExportListeSecteursCSV(const FileName: TStringDirectoryFilename);
var
  fp: TextFile;
  i, Nb: Integer;
  mySecteur: TSecteur;
begin
  Nb := GetNbSecteurs;
  AfficherMessage(Format('%s.ExportListeSecteursCSV: %d', [ClassName, Nb]));
  if (Nb > 0) then
  begin
    try
      AssignFile(fp, FileName);
      ReWrite(fp);
      WriteLn(fp, 'No' + #9 + 'Type' + #9 + 'Couleur' + #9 +
                  'Nom');
      for i := 0 to Nb -1 do
      begin
        mySecteur := GetSecteur(i);
        WriteLn(fp, Format('%d' + #9 + '#%X' + #9 + '%s',
                           [i, mySecteur.CouleurSecteur,
                            mySecteur.NomSecteur
                           ]));
      end;
    finally
      CloseFile(fp);
    end;
  end;
end;

procedure TToporobotStructure2012.ExportListeCodesCSV(const FileName: TStringDirectoryFilename);
var
  fp: TextFile;
  i, Nb: Integer;
  myCode: TCode;
begin
  Nb := GetNbCodes();
  AfficherMessage(Format('%s.ExportListeCodesCSV: %d', [ClassName, Nb]));
  if (Nb > 0) then
  begin
    try
      AssignFile(fp, FileName);
      ReWrite(fp);
      WriteLn(fp, FileName);
      writeln(fp,  'No' + #9 + 'UniteBoussole' + #9 + 'CorrectionAzimut' +  #9 + 'UniteClino' + #9 + 'CorrectionPente' + #9 + 'Observ');
      for i := 0 to Nb -1 do
      begin
        myCode := GetCode(i);
        //TODO: Export codes à faire
        WriteLn(fp, Format('%d' + #9 +
                           '%.2f' + #9 + '%.2f' + #9 +
                           '%.2f' + #9 + '%.2f' + #9 +
                           '%s',
                           [myCode.IDCode,
                            myCode.GradAz , myCode.ParamsFuncCorrAz.Co,
                            myCode.GradInc, myCode.ParamsFuncCorrInc.Co,
                            myCode.Commentaire
                           ]));
      end;
    finally
      CloseFile(fp);
    end;
  end;
end;

procedure TToporobotStructure2012.ExportListeExpesCSV(const FileName: TStringDirectoryFilename);
var
  fp: TextFile;
  i, Nb: Integer;
  myExpe: TExpe;
begin
  Nb := GetNbExpes;
  AfficherMessage(Format('%s.ExportListeExpesCSV: %d', [ClassName, Nb]));

  if (Nb > 0) then
  begin
    try
      AssignFile(fp, FileName);
      ReWrite(fp);
      WriteLn(fp, 'No' + #9 + 'Couleur' + #9 + 'Date' + #9 + 'Equipe' + #9 +
                  'Observ');
      for i := 0 to Nb -1 do
      begin
        myExpe := GetExpe(i);
        WriteLn(fp, Format('%d' + #9 + '%d' + #9 +
                           '%.2d/%.2d/%.4d' + #9 +
                           '%s, %s' + #9 +
                           '%s',
                           [myExpe.IDExpe, myExpe.Couleur,
                            myExpe.JourExpe, myExpe.MoisExpe, myExpe.AnneeExpe,
                            myExpe.Operateur, myExpe.ClubSpeleo,
                            myExpe.Commentaire
                           ]));
      end;
    finally
      CloseFile(fp);
    end;
  end;
end;

procedure TToporobotStructure2012.ExportListeAnnotationsCSV(const FileName: TStringDirectoryFilename);
var
  fp: TextFile;
  Nb, i: Integer;
  myAnnotation: TAnnotation;
begin
  Nb := GetNbAnnotations();
  AfficherMessage(Format('%s.ExportListeAnnotationsCSV: %d', [ClassName, Nb]));
  if (Nb > 0) then
  begin
    try
      AssignFile(fp, FileName);
      ReWrite(fp);
      WriteLn(fp, 'No' + #9 + 'Serie' + #9 + 'Station' + #9 +
                  'OffsetX' + #9 + 'OffsetY' + #9 + 'OffsetZ' + #9 +
                  'Style' + #9 + 'Ancrage' + #9 +
                  'Texte');
      for i := 0 to Nb -1 do
      begin
        myAnnotation := GetAnnotation(i);
        WriteLn(fp, Format('%d' + #9 +
                           '%d' + #9 + '%d' + #9 +
                           '%.3f' + #9 + '%.3f' + #9 + '%.3f' + #9 +
                           '%d' + #9 + '%d' + #9 +
                           '%s',
                           [i,
                            myAnnotation.Basepoint.eSerie, myAnnotation.Basepoint.ePoint,
                            myAnnotation.Offset.X, myAnnotation.Offset.Y, myAnnotation.Offset.Z,
                            myAnnotation.IdxStyle, myAnnotation.PosPtDeBase,
                            myAnnotation.Texte
                           ]));
      end;
    finally
      CloseFile(fp);
    end;
  end;

end;


procedure TToporobotStructure2012.ExportListeEntetesSeries(const FileName: TStringDirectoryFilename);
var
  fp: TextFile;
  i: integer;
  S: TObjSerie;
  procedure Wrtln(const S: string);
  begin
    WriteLn(fp, S);
  end;
  procedure WrtItem(const S: TObjSerie);
  const
    FMT666 = '%d' + #9 + '%s' + #9 +
             '%d' + #9 + '%d' + #9 +
             '%d' + #9 + '%d' + #9 +
             '%d' + #9 + '%d' + #9 +
             '%d';
  begin
    WrtLn(Format(FMT666,
                 [S.GetNumeroDeSerie(),
                  S.GetNomSerie(),
                  S.GetNumeroEntrance(), S.GetNumeroReseau(),
                  S.GetNoSerieDep(), S.GetNoPointDep(),
                  S.GetNoSerieArr(), S.GetNoSerieArr(),
                  S.GetNbVisees()
                 ]));
  end;
begin
  AfficherMessage(Format('%s.ExportListeEntetesSeries', [ClassName]));
  AssignFile(fp, FileName);
  try
    Rewrite(fp);
    Wrtln('Serie' + #9 +
          'Nom serie' + #9 +
          'Entree' + #9 +
          'Reseau' + #9 +
          'Ser Dep' + #9 +
          'Pt Dep' + #9 +
          'Ser Arr' + #9 +
          'Pt Arr' + #9 +
          'NbPts');
    for i := 0 to GetNbSeries - 1 do
    begin
      S := GetSerie(i);
      WrtItem(S);
    end;
  finally
    CloseFile(fp);
  end;
end;



//------------------------------------------------------------------------------
// check des données
procedure TToporobotStructure2012.CheckerLesDonneesTopo();
begin
  AfficherMessage(Format('%s.CheckerLesDonneesTopo',[ClassName]));
  CheckerLesEntrees();        // scan des entrées
  CheckerLesSeries();         // scan des séries
  CheckerLesAntennes();
end;

procedure TToporobotStructure2012.CheckerLesEntrees();
var
  UneEntree: TEntrance;
begin
  AfficherMessage(Format('%s.CheckerLesEntrees',[ClassName]));
  if (GetNbEntrees() = 0) then
  begin
    AfficherMessage('*** Entree ajoutee ***');
    //UneEntree.eNumEntree := 0;
    UneEntree.eNomEntree := 'Entree par default';
    UneEntree.eXEntree   := self.FPositionPointZero.X;
    UneEntree.eYEntree   := self.FPositionPointZero.Y;
    UneEntree.eZEntree   := self.FPositionPointZero.Z;
    UneEntree.eRefSer    := 1;
    UneEntree.eRefSt     := 0;
    UneEntree.eObserv    := 'Entree reparee';
    self.AddEntree(UneEntree);
  end;
end;
// 21/08/2017: Ajout de la détection des doublons dans les numéros de séries
procedure TToporobotStructure2012.CheckerLesSeries();
var
  EWE, WU: TObjSerie;
  i, Nb  , QQ: Integer;
  QNS0, QNS1: TNumeroSerie;
begin
  Nb := GetNbSeries();
  AfficherMessageErreur(Format('-- %s.CheckerLesSeries: %d series',[ClassName, Nb]));
  for i := 1 to Nb - 1 do
  begin
    WU  := GetSerie(i-1);
    EWE := GetSerie(i);
    // détection des doublons dans les numéros de série
    QNS0 := WU.GetNumeroDeSerie();
    QNS1 := EWE.GetNumeroDeSerie();
    if (QNS0 = QNS1) then
      AfficherMessageErreur(Format('--- Doublons dans les séries: (%d - %s) et (%d - %s)',
                            [QNS0, WU.GetNomSerie(),
                             QNS1, EWE.GetNomSerie()]));

    QQ := CheckSerie(EWE, i);
  end;
end;
// // retourne: -1 si série vide; numéro de visée si erreur, 0 si OK
function TToporobotStructure2012.CheckSerie(const MySerie: TObjSerie; const QIdx: integer): integer;
var
  i, N: Integer;
  MyVisee: TUneVisee;
  YaErreur: TErrorVisee;
  MessageRetour: string;
  NS, SD, SA, PD, PA: integer;
  QLongueurSerie: Double;
begin
  Result := -1;
  N := MySerie.GetNbVisees;
  try


    // détection des auto-boucles (séries se refermant sur elles-mêmes)
    NS := MySerie.GetNumeroDeSerie();
    SD := MySerie.GetNoSerieDep;
    SA := MySerie.GetNoSerieArr;
    PD := MySerie.GetNoPointDep;
    PA := MySerie.GetNoPointArr;
    if (NS = SA) then
    begin
      if (PA < (N - 1)) then AfficherMessageErreur(Format('  Auto-loop : station %d.%d linked to %d.%d', [SA, N - 1, NS, PA]));
    end;

    if ((SD = SA) and (SD = NS) and (NS <> 1)) then AfficherMessageErreur(Format('  Série %d orpheline : %d.%d -> %d/%d', [NS, SD, PD, SA, PA]));
    // détection des séries non raccordées

    // scan de la table des visées
    QLongueurSerie := 0.00;
    for i := 0 to N - 1 do
    begin
      MyVisee := MySerie.GetVisee(i);
      QLongueurSerie += MyVisee.Longueur;
      YaErreur := CheckVisee(MyVisee, MessageRetour);
      if (YaErreur <> []) then
      begin
        AfficherMessageErreur('');
        AfficherMessageErreur(Format('*** In serie %d - Shot %d ***', [MySerie.GetNumeroDeSerie(), i]));
        AfficherMessageErreur('------------------------------------------');
        AfficherMessageErreur(MessageRetour);
        if (errvSECTEUR   in YaErreur) then AfficherMessageErreur(Format('  Sector = %d', [MyVisee.IDSecteur]));
        if (errvTYPEVISEE in YaErreur) then AfficherMessageErreur(Format('  Type   = %d', [MyVisee.TypeVisee]));
        if (errvCODE      in YaErreur) then AfficherMessageErreur(Format('  Code   = %d', [MyVisee.Code]));
        if (errvEXPE      in YaErreur) then AfficherMessageErreur(Format('  Expe   = %d', [MyVisee.Expe]));
        if (errvLONGUEUR_NULLE  in YaErreur) then
        begin
          AfficherMessageErreur(Format('  Length = %.3f m too short - Fixed at minimal accepted value', [MyVisee.Longueur]));
          MyVisee.Longueur := SEUIL_LONGUEUR_NULLE + 0.001;
          MySerie.PutVisee(i, MyVisee);
        end;
        if (errvLONGUEUR_OVERFLOW  in YaErreur) then AfficherMessageErreur(Format('  Length = %.3f m too big', [MyVisee.Longueur]));

        if (errvAZIMUT    in YaErreur) then AfficherMessageErreur(Format('  Azimut = %.3f a.u.', [MyVisee.Azimut]));
        if (errvPENTE     in YaErreur) then AfficherMessageErreur(Format('  Inclin = %.3f a.u.', [MyVisee.Pente]));
        if (errvLG        in YaErreur) then AfficherMessageErreur(Format('  Left   = %.3f m', [MyVisee.LG]));
        if (errvLD        in YaErreur) then AfficherMessageErreur(Format('  Right  = %.3f m', [MyVisee.LD]));
        if (errvHZ        in YaErreur) then AfficherMessageErreur(Format('  Upper  = %.3f m', [MyVisee.HZ]));
        if (errvHN        in YaErreur) then AfficherMessageErreur(Format('  Downer = %.3f m', [MyVisee.HN]));
        if (errvLONGUEUR_NULLE in YaErreur) then AfficherMessageErreur(Format('--- Zero length shot: %.3f m', [MyVisee.HN]));
      end;
    end;
    // longueur de série nulle
    if (QLongueurSerie < 0.50) then AfficherMessageErreur(Format('Serie %d: (%d stations): Developpement < 1.00 m', [MySerie.GetNumeroDeSerie(), N]));

  except

  end;
end;

function TToporobotStructure2012.CheckVisee(const V: TUneVisee; out MsgErr: string): TErrorVisee;
var
  MyCode: TCode;
  MyExpe: TExpe;
  WU: Int64;
  miou: Extended;
  CC: Integer;
  EE: Integer;
  AzMax: Extended;
  function EWE(const QMsg: string; const QMini, QMaxi: double; const QUnit: string): string;
  begin
    Result := Format(rsCHECK_VISEE_VALID_INTERVAL + #13#10, [QMsg, QMini, QUnit, QMaxi, QUnit]);
  end;
begin
  MsgErr := '';
  Result := [];
  // on teste d'abord ce qui est indépendant des expés et codes
  // types de visée
  if (not IsInRange(Ord(V.TypeVisee), 0, Ord(tgVISEE_RADIANTE))) then
  begin
    MsgErr := MsgErr + Format(rsCHECK_VISEE_VALID_TYPE_VISEE + #13#10, [0, Ord(tgVISEE_RADIANTE)]);
    Result := Result + [errvTYPEVISEE];
  end;
  //AfficherMessageErreur('--- CheckVisee: 002');
  // visée de longueur nulle
  if (V.Longueur < SEUIL_LONGUEUR_NULLE) then
  begin
    MsgErr := MsgErr + Format('Longueur = %.2f m', [V.Longueur]);
    Result := Result + [errvLONGUEUR_NULLE];
  end;
  //AfficherMessageErreur('--- CheckVisee: 003');

  if (V.Longueur > SEUIL_LONGUEUR_MAXI) then
  begin
    MsgErr := MsgErr + Format('Longueur = %.2f m', [V.Longueur]);
    Result := Result + [errvLONGUEUR_OVERFLOW];
  end;
  //AfficherMessageErreur('--- CheckVisee: 004');
  // LRUD
  if (not IsInRange(V.LG, 0.00, SEUIL_LONGUEUR_MAXI)) then
  begin
    MsgErr := MsgErr + EWE('Left distance', 0.00, SEUIL_LONGUEUR_MAXI, 'm');
    Result := Result + [errvLG];
  end;
  if (not IsInRange(V.LD, 0.00, SEUIL_LONGUEUR_MAXI)) then
  begin
    MsgErr := MsgErr + EWE('Right distance', 0.00, SEUIL_LONGUEUR_MAXI, 'm');
    Result := Result + [errvLD];
  end;
  if (not IsInRange(V.HZ, 0.00, SEUIL_LONGUEUR_MAXI)) then
  begin
    MsgErr := MsgErr + EWE('Upper distance', 0.00, SEUIL_LONGUEUR_MAXI, 'm');
    Result := Result + [errvHZ];
  end;
  if (not IsInRange(V.HN, 0.00, SEUIL_LONGUEUR_MAXI)) then
  begin
    MsgErr := MsgErr + EWE('Downer distance', 0.00, SEUIL_LONGUEUR_MAXI, 'm');
    Result := Result + [errvHN];
  end;
  // existence de secteurs, codes et expés
  if (GetIdxSecteurByNumero(V.IDSecteur) = -1) then Result := Result + [errvSECTEUR];
  CC := GetIdxCodeByNumero(V.Code);
  EE := GetIdxExpeByNumero(V.Expe);
  if ((CC = -1) OR (EE = -1)) then
  begin
    if (CC = -1) then Result := Result + [errvCODE];
    if (EE = -1) then Result := Result + [errvEXPE];
    AfficherMessageErreur(Format('  -- Invalid code (%d) or trip (%d) -> check aborted', [CC, EE]));
    Exit;
  end;
  // codes, secteur et expés
  MyCode := GetCodeByNumero(V.Code);
  MyExpe := GetExpeByNumero(V.Expe);
  // azimuts
  AzMax := 10.0 * round(MyCode.GradAz / 10.0); // pour s'affranchir des codes 359, 361, 399, 401
  if (not IsInRange(V.Azimut, 0.00, MyCode.GradAz)) then
  begin
    MsgErr := MsgErr + EWE('Azimuth', 0.00, AzMax, 'a.u.');
    Result := Result + [errvAZIMUT];
  end;
  // pentes (le gros morceau)
  WU := Trunc(MyCode.GradInc);
  miou := MyCode.GradInc / 4; // quart de cercle dans l'unité de GradInc
  case WU of
    360, 400: // clinos classiques degré et grades
    begin
      if (not IsInRange(V.Pente, -miou, +miou)) then
      begin
        MsgErr := MsgErr + EWE('Inclination', -miou, +miou, 'a.u.');
        Result := Result + [errvPENTE];
      end;
    end;
    790, 800:  // Lasermètres Stanley TLM330, directes et inverses
     begin
      if (not IsInRange(V.Pente, -SEUIL_LONGUEUR_MAXI, +SEUIL_LONGUEUR_MAXI)) then
      begin
        MsgErr := MsgErr + EWE('Difference in height', -miou, +miou, 'a.u.');
        Result := Result + [errvPENTE];
      end;
    end;
  else
    ;
  end;
end;

procedure TToporobotStructure2012.RecalculerDeclinaisonsMagnetiquesExpes();
var
  CalculateurDeclimag : TSAGeoMag;
  i, Nb, NbE: Integer;
  EE: TEntrance;
  EX: TExpe;
  X0, Y0, Z0, Lon, Lat, Decl: double;
  PTI, PTO: TPoint2Df;
  YYYY, MM, DD: word;
  EPSG: TLabelSystemesCoordsEPSG;
begin
  AfficherMessage(Format('%s.recalculerDeclinaisonsMagnetiquesExpes',[ClassName]));
  Lon := 0.00;
  Lat := 43.00;
  CalculateurDeclimag := TSAGeoMag.Create;
  try
    // Etape 0: Démarrage du calculateur de déclinaison
    if (not CalculateurDeclimag.Initialiser('')) then
    begin
      AfficherMessage('Calculateur de déclinaisons magnétiques inopérant - Arrêt');
      exit;
    end;
    // Etape 1: Calcul des coordonnées du point zéro = moyenne des coordonnées des entrées
    NbE := GetNbEntrees;
    AfficherMessage(Format('-- %d entrées', [NbE]));
    Nb := 0;
    X0 := 0.00;
    Y0 := 0.00;
    Z0 := 0.00;
    for i := 0 to NbE - 1 do
    begin
      try
        EE := GetEntree(i);
        if (IsZero(EE.eXEntree) or IsZero(EE.eYEntree)) then Continue;
        X0 += EE.eXEntree;
        Y0 += EE.eYEntree;
        Z0 += EE.eZEntree;
        Nb += 1;
      except
      end;
    end;
    // coordonnées du point moyen
    if (Nb <> 0) then
    begin
      X0 := X0 / Nb;
      Y0 := Y0 / Nb;
      Z0 := Z0 / Nb;
    end
    else
    begin
      X0 := FPositionPointZero.X;
      Y0 := FPositionPointZero.Y;
      Z0 := FPositionPointZero.Z;
    end;
    // qu'on convertit en WGS84
    PTI := MakeTPoint2Df(X0, Y0);
    PTO := MakeTPoint2Df(0.00, 0.00);
    // si la conversion réussit:
    // Etape 2: Calcul des déclinaisons pour les expés
    EPSG := GetCodeEPSGSystemeCoordonnees;
    if (ConversionCoupleCoordonneesIsoleesEPSG(EPSG.CodeEPSG, CODE_EPSG_WGS84, PTI, PTO)) then
    begin
      Lon := PTO.X; Lat := PTO.Y;
      AfficherMessage(Format('-- X = %.2f, Y = %.2f, Longitude = %.8f, Latitude = %.8f, Altitude: %.2f', [X0, Y0, Lon, Lat, Z0]));
      AfficherMessage('');
      NbE := GetNbExpes;
      AfficherMessage(Format('-- Declinaisons pour les %d expes', [Nb]));
      for i := 1 to NbE - 1 do
      begin
        EX := GetExpe(i);
        YYYY := EX.AnneeExpe;
        MM   := EX.MoisExpe;
        DD   := EX.JourExpe;
        if (YYYY < 1920) then   // année incorrecte -> on prend année courante
        begin
          YYYY := YearOf(Now());
          MM   := 1;
          DD   := 1;
        end;
        Decl := CalculateurDeclimag.CalculerDeclinaison(lon, lat, Z0, EncodeDate(YYYY, MM, DD));
        EX.ModeDecl    := 1;                           // on force en automatique
        EX.Declinaison := makeToporobotDeclimag(decl); // conversion de la déclinaison au format Toporobot
        PutExpe(i, EX);                                // et on met à jour la table
      end;
    end;
    CalculateurDeclimag.Finaliser;
  finally
    CalculateurDeclimag.Free;
  end;
end;

//***********************************
// recherche de l'index correspondant à un élément

function TToporobotStructure2012.GetIdxEntreeByNumero(const N: TNumeroEntrance): integer;
begin
  Result := n;
end;
function TToporobotStructure2012.GetIdxReseauByNumero(const N: TNumeroReseau): integer;
begin
  Result := n;
end;


//*)

function TToporobotStructure2012.GetIdxSecteurByNumero(const N: TNumeroSecteur): integer;
begin
  Result := n;
end;
function TToporobotStructure2012.GetIdxCodeByNumero(const N: TNumeroCode): integer;
var
  i  : integer;
  EWE: TCode;
begin
  Result := -1;
  for i := 0 to GetNbCodes - 1 do
  begin
    EWE := GetCode(i);
    if (EWE.IDCode = N) then
    begin
      Result := i;
      Break;
    end;
  end;
end;
function TToporobotStructure2012.GetIdxExpeByNumero(const N: TNumeroExpe): integer;
var
  i  : integer;
  EWE: TExpe;
begin
  Result := -1;
  for i := 0 to GetNbExpes - 1 do
  begin
    EWE := GetExpe(i);
    if (EWE.IDExpe = N) then
    begin
      Result := i;
      Break;
    end;
  end;
end;



function TToporobotStructure2012.GetIdxSerieByNumero(const qIdx: TNumeroSerie): integer;
var
  Nb: Integer;
  i: Integer;
  EWE: TObjSerie;
begin
  Result := -1;
  Nb := self.GetNbSeries;
  if (Nb = 0) then exit;
  try
    for i := 0 to Nb - 1 do
    begin
      EWE := self.GetSerie(i);
      if (qIdx = EWE.GetNumeroDeSerie()) then
      begin
        Result := i;
        Exit;
      end;
    end;
  except
  end;
end;

//------------------------------------------------------------------------------
function TToporobotStructure2012.HasPtTopoBySerSt(const qSr, qPt: integer): boolean;
var
  S, P: Integer;
  mySerie: TObjSerie;
  myVisee: TUneVisee;
  EWE: Integer;
begin
  Result := False;
  for S := 0 to GetNbSeries - 1 do
  begin
    mySerie := GetSerie(S);
    EWE := mySerie.GetNumeroDeSerie;
    for P := 0 to mySerie.GetNbVisees - 1 do
    begin
      myVisee := mySerie.GetVisee(P);
      if ((qSr = EWE) and (qPt = P)) then
      begin
         Result := True;
         AfficherMessage(Format('Station found in %d.%d', [mySerie.GetNumeroDeSerie(), P]));
         Exit;
      end;
    end;
  end;
end;
// recherche du couple série station d'après ID littéral

function  TToporobotStructure2012.FindPtTopoByCle(const DoMatchExact: boolean;
                                                  var Cle: string;
                                                  var Sr, Pt: integer): boolean;
                                                  //out ListeStationsFound: TArrayStationsMatchFound): boolean;
var
  //Clef: string;
  S, P: integer;
  Q, R: string;
  mySerie: TObjSerie;
  myPoint : TUneVisee;
  WU: SizeInt;
  qSer: Integer;
  qSt: Integer;
  N: Integer;
  S1, S2: String;
  QFound: Boolean;
  QListFound: TStringList;
begin
  Result := False;
  Cle := UpperCase(Trim(Cle));
  // analyse de la clé
  // la clé contient un point --> recherche par série/station
  WU := Pos('.', Cle);
  if (WU > 0) then
  begin
    N := Length(Cle);
    S1 := Trim(Copy(Cle, 0, WU - 1));
    S2 := Trim(Copy(Cle, WU + 1, N));
    qSer := StrToIntDef(S1, 0);
    qSt  := StrToIntDef(S2, 0);
    if (HasPtTopoBySerSt(qSer, qSt)) then
    begin
      Sr := qSer;
      Pt := qSt;
      Exit(True);
    end;
  end;
  // station non trouvee ? Tentative de Recherche par clé
  Sr := -1;
  Pt := -1;
  Result := False;
  for S := 0 to GetNbSeries - 1 do
  begin
    mySerie := GetSerie(S);
    for P := 0 to mySerie.GetNbVisees - 1 do
    begin
      myPoint := mySerie.GetVisee(P);
      // recherche sur IDTerrain et commentaires
      Q  := UpperCase(Trim(myPoint.IDTerrainStation + '|' + myPoint.Commentaires));  // recherche sur une partie de la clé
      R  := UpperCase(Trim(myPoint.IDTerrainStation)); // recherche exacte
      QFound := IIF(DoMatchExact, (R = Cle), (Pos(Cle,  Q) > 0));
      if (QFound) then
      begin
         Sr := mySerie.GetNumeroDeSerie();
         Pt := P;
         Cle := Q;
         Exit(True);
      end;
    end;
  end;
end;

function TToporobotStructure2012.FindIndexesSeriesRattachedAtSerie(const Idx: integer; out LS: TListOfIntegers): boolean;
var
  n, i: Integer;
  MySerie, QSerie: TObjSerie;
  QNumSer: TNumeroSerie;
begin
  Result := false;
  AfficherMessage(Format('%s.FindSeriesRattachedAtSerie: %d', [ClassName, Idx]));
  LS.ClearListe(); // par sécurité
  n := self.GetNbSeries();
  MySerie := GetSerie(Idx);  // attrapper la série courante
  QNumSer := MySerie.GetNumeroDeSerie();
  LS.AddElement(QNumSer);    // on accroche en tête la série courante
  for i := 0 to n - 1 do
  begin
    QSerie := GetSerie(i);
    if (QNumSer = QSerie.GetNoSerieDep()) then LS.AddElement(QSerie.GetNumeroDeSerie());
    if (QSerie.GetNoSerieArr() = QNumSer) then LS.AddElement(QSerie.GetNumeroDeSerie());
  end;
  Result := (LS.Count > 0);
end;

// Infos sur les données topo
// ne nécessite pas de calcul topo
// c'est du fichier de log -> messages en anglais uniquement
function TToporobotStructure2012.GetInfosCavite(const LS: TStrings): integer;
const
  FMTLNENTREE = '%d'+TAB+'%.2f'+TAB+'%.2f'+TAB+'%.2f'+TAB+'%s';
  FMTLNSER    = '%d'+TAB+'%d'+TAB+'%d'+TAB+'%d'+TAB+'%d'+TAB+'%-60s'+TAB+'%d'+
                 TAB+'%.2f'+TAB+'%.2f';
var
  i, n: integer;
  E: TEntrance;
  Ser: TObjSerie;
  V: TUneVisee;
  D1, DevTotal: double;
  VolumeTotal : double;
  VolumeSerie : double;
  NbPtsTotal: Integer;
begin
  Result:=-1;
  AfficherMessage(Format('%s.GetCaveInfoSeries()',[ClassName]));
  LS.Clear;
  try
    LS.Add(Format('File created  %s - %s',[DateToStr(Now), TimeToStr(Now)]));
    LS.Add('');
    LS.Add('Database : ' + FDatabaseName);
    LS.Add('');
    LS.Add('Survey: ' + GetNomEtude);
    LS.Add('');
    LS.Add(Format('List of %d entrances', [GetNbEntrees]));
    LS.Add('');
    for i:=0 to GetNbEntrees - 1 do begin
      E:=GetEntree(i);
     // E.eXEntree;
      LS.Add(Format(FMTLNENTREE, [i, E.eXEntree, E.eYEntree, E.eZEntree, E.eNomEntree]));
    end;
    LS.Add('');
    LS.Add(Format('List of %d series', [GetNbSeries]));
    LS.Add('');
    LS.Add('No'+TAB+'S.Dep'+TAB+'Pt.Dep'+TAB+'S.Arr'+TAB+'Pt.Arr'+TAB+
           'Serie name'+TAB+'Nb points'+TAB+'Length'+
           TAB + 'Estimated volume');
    LS.Add('');
    DevTotal:=0.00;
    NbPtsTotal:=0;
    VolumeTotal:=0.00;
    for i:=0 to GetNbSeries - 1 do
    begin
      Ser:=GetSerie(i);
      // calcul de la longueur, de la surface et du volume de la série
      D1:=0.00;
      VolumeSerie:=0.00;
      for n:=0 to Ser.GetNbVisees - 1 do
      begin
        V:=Ser.GetVisee(n);
        // les sections considérées pour les volumes
        // sont des ellipses    (A = pi . a . b)
        case V.TypeVisee of
          tgDEFAULT, tgFOSSILE, tgVADOSE, tgENNOYABLE, tgSIPHON: // galeries naturelles
          begin;
            D1          += V.Longueur; // X += N -> X := X + N
            VolumeSerie += PI * (V.HZ+V.HN) * (V.LD + V.LG) * V.Longueur;
          end;
        else
          ;
        end;
      end;
      VolumeTotal:=VolumeTotal + VolumeSerie;
      DevTotal:=DevTotal+D1;
      NbPtsTotal:=NbPtsTotal + Ser.GetNbVisees - 1;
      LS.Add(Format(FMTLNSER,
                    [Ser.GetNumeroDeSerie(),
                     Ser.GetNoSerieDep, Ser.GetNoPointDep,
                     Ser.GetNoSerieArr, Ser.GetNoPointArr,
                     Ser.GetNomSerie,
                     Ser.GetNbVisees - 1,
                     D1, VolumeSerie]));
    end;  // for i:=0 to TableSeries.Count-1
    LS.Add('');
    LS.Add(Format('Total length: %.2f m (%d stations) - Estimated Volume: %.1f m3',
                  [DevTotal, NbPtsTotal, VolumeTotal]));
    Result:=1;
  except
  end;
end;

// l'export vers Visual Topo est désormais du ressort de cette classe
// Difficultés en raison de l'intolérance de VTopo
// 27/05/2015: Petit fix de l'unité des azimuts (Degrés décimaux = 'Degd' et non 'Deg')
procedure TToporobotStructure2012.ExportVisualTopo(const FichierTRO: TStringDirectoryFilename;
                                                   const DoExportViseesRadiantes: boolean);
const
  DEG_DEC = 'Degd';
  GRADS   = 'Gra';
  VTOPO_VERSION = 'Version 4.9';
  CLITO_VULCAIN = 'Vulc';
  CLITO_STD     = 'Clino';
  CLITO_DENIV   = 'Deniv';
  FMTSERSTVTOPO = '%d_%d';
var
  pTRO: TextFile;
  Palette256: TPalette256;
  Entree: TEntrance;
  DefautCouleur: TColor;
  Serie: TObjSerie;
  CurrStation, PrevStation: TUneVisee;
  FromStationID, ToStationID: String;
  S777: String;
  S666: String;
  NoSer: Integer;
  NoSt, NbA, j: Integer;
  Code: TCode;
  Expe: TExpe;
  EWE: String;
  ArrAntennesFound: TArrayOfTViseeAntenne;
  MyAntenne: TViseeAntenne;
  QNumeroDeSerie: TNumeroSerie;
  Toto: Boolean;
  procedure WriteLnParam(const C: TCode; const E: TExpe);
  var
    UB, UC, Clino: string;
    CC: TColor;
    Decl: Extended;
  begin
    writeLn(pTRO,'');
    //Param Deca Deg Clino Deg 1.0300 Dir,Dir,Dir Std
    case Trunc(C.GradAz) of
      360: UB := DEG_DEC;
      400: UB := GRADS;
    else
      UB := GRADS;
    end;
    case Trunc(C.GradInc) of
       360: begin
              Clino := CLITO_STD; //'Clino';
              UC    := DEG_DEC;
            end;
       361: begin
              Clino := CLITO_VULCAIN; //'Vulc';
              UC    := DEG_DEC;
            end;
       400: begin
              Clino := CLITO_STD; //'Clino';
              UC    := GRADS;
            end;
       401: begin
              Clino := CLITO_VULCAIN; //'Vulc';
              UC    := GRADS;
            end;
       380: begin
              Clino := CLITO_DENIV; //'Deniv'; // anciennement: Prof
              UC    := '';
            end;
       UNITE_CLINO_LASERMETRE_STANLEY_TLM330_DIR,
       UNITE_CLINO_LASERMETRE_STANLEY_TLM330_INV:
            begin
              Clino := CLITO_DENIV; //'Deniv'; // anciennement: Prof
              UC    := '';
            end;
    end;
    Decl:=(E.Declinaison/10) * 400.0 / C.GradAz;
    CC:=Palette256.GetColorByIndex(E.Couleur);
    //if FDefaultColor then CC:=lbColorReseau.Color;
    WriteLn(pTRO, Format('Param Deca %s %s %s %.4f Dir,Dir,Dir %d,%d,%d',
                          [UB, Clino, UC,
                           Decl,
                           Red(CC),
                           Green(CC),
                           Blue(CC)
                          ]));
    writeln(pTRO,'');
  end;
begin
  AfficherMessage(Format('%s.ExportVisualTopo(%s)',[ClassName, FichierTRO]));
  setlength(ArrAntennesFound, 0);
  // démarrage de la palette
  Palette256 := TPalette256.Create;
  // ouverture fichier VTopo
  AssignFile(pTRO, FichierTRO);
  try
    Palette256.GenerateTOPOROBOTPalette();
    ReWrite(pTRO);
    WriteLn(pTRO, VTOPO_VERSION);
    WriteLn(pTRO, '');

    // Entrée
    Entree := GetEntree(0);

    //----------------------------------------------------------------------------------
    Serie := GetSerie(1);
    EWE := Format(FMTSERSTVTOPO,[Serie.GetNoSerieDep, Serie.GetNoPointDep]);
    // Trou CDS03,0.000,0.000,0,
    WriteLn(pTRO, Format('Trou %s,%.3f,%.3f,%f,', [Entree.eNomEntree, Entree.eXEntree / 1000, Entree.eYEntree / 1000, Entree.eZEntree]));
    WriteLn(pTRO, Format('Entree %s',[Format(FMTSERSTVTOPO, [Entree.eRefSer, Entree.eRefSt])]));
    WriteLn(pTRO, Format('Club %s',[rsGHTOPOEXENAME]));
    DefautCouleur:= clRED;
    WriteLn(pTRO, Format('Couleur %d,%d,%d',[Red(DefautCouleur), Green(DefautCouleur), Blue(DefautCouleur)]));

    // première ligne de données
    WriteLn(pTRO,'Param Deca Degd Clino Deg 0.0000 Dir,Dir,Dir Std');
    Serie := GetSerie(1);
    WriteLn(pTRO, Format('%10s %10s %.1f   %.1f %.1f %.1f %.2f %.2f %.2f %s',
                               [EWE, EWE,
                                0.00, 0.00, 0.00, 1.00, 1.00, 1.00, 1.00,
                                'Entrée ' + EWE]));
    // exportation proprement dite
    for NoSer:=1 to GetNbSeries - 1 do
    begin
      Serie := GetSerie(NoSer);
      QNumeroDeSerie := Serie.GetNumeroDeSerie();
      AfficherMessage(Format('--> Export serie: %d - %s',[Serie.GetNumeroDeSerie(), Serie.GetNomSerie]));
      FromStationID:=Format(FMTSERSTVTOPO,[Serie.GetNoSerieDep, Serie.GetNoPointDep]);
      for NoSt:=1 to Serie.GetNbVisees() - 1 do
      begin
        CurrStation := Serie.GetVisee(NoSt);
        Code := GetCodeByNumero(CurrStation.Code);
        Expe := GetExpeByNumero(CurrStation.Expe);
        if ((NoSer=1) and (NoSt=1)) then WriteLnParam(Code, Expe);
        if (CurrStation.Code <> PrevStation.Code) OR (CurrStation.Expe <> PrevStation.Expe)
        then WriteLnParam(Code, Expe);
        ToStationID:=Format(FMTSERSTVTOPO,[Serie.GetNumeroDeSerie(), NoSt]);
        if (NoSt = Serie.GetNbVisees - 1) then ToStationID := Format(FMTSERSTVTOPO,[Serie.GetNoSerieArr, Serie.GetNoPointArr]);
        S777 := IIF((Trim(CurrStation.Commentaires)='')     , '', Format('%s',[CurrStation.Commentaires]));
        S666 := IIF((Trim(CurrStation.IDTerrainStation) =''), '', Format('[%s]',[CurrStation.IDTerrainStation]));

        if (Trim(CurrStation.Commentaires)='') and (Trim(CurrStation.IDTerrainStation)='')
        then S777 := ''
        else S777 := ';' + S666 + S777 + ';';
        WriteLn(pTRO, Format('%10s %10s %.2f   %.2f %.2f %.2f %.2f %.2f %.2f N I * %s',
                             [FromStationID,
                              ToStationID,
                              CurrStation.Longueur,
                              CurrStation.Azimut,
                              CurrStation.Pente,
                              CurrStation.LG,
                              CurrStation.LD,
                              CurrStation.HZ,
                              CurrStation.HN,
                              S777]));
        PrevStation    := CurrStation;
        FromStationID  := ToStationID;
        // Traitement des visées radiantes
        if (DoExportViseesRadiantes) then
        begin
          // ignorer les antennes de la dernière visée dont la série arrive sur un point connu
          Toto := True;
          if (NoSt = Serie.GetNbVisees() - 1) then Toto := (NoSt = Serie.GetNoPointArr());

          NbA := ExtractViseesAntennesOfStation(QNumeroDeSerie, NoSt, ArrAntennesFound);
          //202_7      *                         3.60  266.00  -20.00      *      *      *      * N I M
          //202_7      *                         3.90  211.00  -26.00      *      *      *      * N I M
          if (Toto And (NbA > 0)) then
          begin
            //WriteLn(pTRO, Format('# %d.%d', [QNumeroDeSerie, NoSt]));
            for j := 0 to NbA - 1 do
            begin
              MyAntenne := ArrAntennesFound[j];
              FromStationID:=Format(FMTSERSTVTOPO,[MyAntenne.SerieDepart, MyAntenne.PtDepart]);
              WriteLn(pTRO, Format('%10s %10s %.2f   %.2f %.2f     *      *      *      * N I M %s',
                               [FromStationID,
                                '        *',
                                MyAntenne.Longueur,
                                MyAntenne.Azimut,
                                MyAntenne.Pente,
                                '']));
            end;
            //WriteLn(pTRO, '');
          end;
        end; // if (DoExportViseesRadiantes) then
      end; // for NoSt:=1 to Serie.GetNbVisees() - 1 do
    end; // for NoSer:=1 to GetNbSeries - 1 do
    // antennes

    // footpage
    WriteLn(pTRO,'');
    WriteLn(pTRO,'[Configuration 4.9]');
    WriteLn(pTRO,'');
    WriteLn(pTRO,'Visual Topo=2,3,-1,-1,-1,-1,1,1,801,610');
    WriteLn(pTRO,'Options=1,1');
    WriteLn(pTRO,'Calcul=0,1,-1,-1,-1,-1,22,22,844,449');
    WriteLn(pTRO,'Options=26,1,0,1,0');
    WriteLn(pTRO,'ExportDxf=0,100,391,1,6,7,4,4,3,3,2,7,9');
    WriteLn(pTRO,'Colonnes=8.00,8.00,8.00,8.00,8.00,8.00,8.00,8.00,8.00,'+
                 '8.00,8.00,2.00,4.00,0.38,8.00,8.00,8.00,8.00,8.00,'+
                 '0.38,0.00,0.00');
    WriteLn(pTRO,'Options=384,1.00,4,0,2,6914,1');
    AfficherMessage(Format('%s.ExportVTOPO(%s) OK',[ClassName, FichierTRO]));
  finally
    CloseFile(pTRO);
    Palette256.Free;
  end;
end;

procedure TToporobotStructure2012.GenererThConfig(const THC, FTH: string);
var
  fp: TextFile;
begin
  // encoding  utf-8
  // source Toporabot1_TLM330_accents_secteurs.th
  // export model -fmt loch # Toporabot1_TLM330_accents_secteurs.lox
  assignFile(fp, THC);
  try
    ReWrite(fp);
    WriteLn(fp, 'encoding ' + 'cp1252');
    WriteLn(fp, 'source ' + ExtractFileName(FTH));
    WriteLn(fp, 'export model -fmt loch');
  finally
    CloseFile(fp);
  end;
end;



procedure TToporobotStructure2012.GenererDossierTherion(const DossierTherion: TStringDirectoryFilename);
var
  FichierTH: String;
  EWE: String;
  FichierTHConfig: String;
begin
  AfficherMessage(Format('%s.GenererDossierTherion: %s', [ClassName, DossierTherion]));
  ForceDirectories(DossierTherion);
  EWE := 'Polygonale' + '.th';
  FichierTH       := DossierTherion + PathDelim + EWE;
  FichierTHConfig := DossierTherion + PathDelim + 'thconfig';
  GenererThConfig(FichierTHConfig, FichierTH);
  ExportVersTherion(FichierTH);
end;


// TODO: Implanter le contrôle des azimuts et pentes
function TToporobotStructure2012.CheckAzimutByIndex(const Idx: TNumeroCode; const QAz: double; out MsgErr: string): boolean;
var
  CC: TCode;
  EWE: Boolean;
begin
  MsgErr := '';
  Result := False;
  if (Not ExistsIdxCode(Idx)) then
  begin
    MsgErr := rsMSG_CODE_NOT_FOUND;
  end
  else
  begin
    CC     := GetCodeByNumero(Idx);
    EWE    := IsInRange(QAz, 0.00, CC.GradAz);
    MsgErr := IIF(EWE, '', GetResourceString(Format(rsMSG_ANGLE_OUT_OF_RANGE, ['Azimut', 0.00, CC.GradAz])));
    Result := EWE;
  end;
end;

procedure TToporobotStructure2012.CheckerLesAntennes();
var
  Nb: Integer;
begin
  Nb := GetNbAntennes();
  AfficherMessage(Format('-- %s.CheckerLesAntennes: %d visees rayonnantes',[ClassName, Nb]));
end;

function TToporobotStructure2012.CheckPenteByIndex(const Idx: TNumeroCode; const QP: double; out MsgErr: string): boolean;
var
  CC: TCode;
  EWE: Boolean;
begin
  MsgErr := '';
  Result := False;
  if (Not ExistsIdxCode(Idx)) then
  begin
    MsgErr := rsMSG_CODE_NOT_FOUND;
  end
  else
  begin
    CC := GetCodeByNumero(Idx);
    Result := True;
  end;
end;

// Export vers Thérion
// Pas encore validé
procedure TToporobotStructure2012.ExportVersTherion(const FichierTH: string);
const
  TH_KW_CENTERLINE     = 'centerline';
  TH_KW_ENDCENTERLINE  = 'end' + TH_KW_CENTERLINE;
var
  THF      : TextFile;
  MySerie  : TObjSerie;
  MyFixPt  : TEntrance;
  // Station de rattachement courante
  MaStationDeRattachement: string;
  i           : integer;
  procedure WrtLn(const Line: string);
  begin
    WriteLn(THF, Line);
  end;
  // définition d'un en-tete de séance
  procedure DefineSeance(const E: TExpe);
  begin
    WrtLn(Format('       # Seance %d: %s',[E.IDExpe, E.Commentaire]));
    WrtLn(Format('       date %.4d.%.2d.%.2d',[ E.AnneeExpe,  E.MoisExpe,  E.JourExpe]));
    WrtLn('       #-------------------');
  end;
  // definition d'un code - Valeur de retour par défaut: -1, sinon, un angle droit
  function DefineCode(const C: TCode): double;
  const
    UNITE_ANGULAIRE_GRADES   = 'grads';
    UNITE_ANGULAIRE_DEGRES   = 'degrees';
    DIRECT_VISEE             = '';
    BACKSIDE_VISEE           = 'back';
    MODE_DENIVELE            = 'depthchange';
    MODE_DIVING              = 'diving';
  var
    UnitCompass   : string;
    UnitClino     : string;
    IsBackCompass : string;
    IsBackClino   : string;
    InstrVertical : string; // mesures altimétriques
    DepthMode     : boolean; // si DepthMode=True, ne pas spécifier les unités de clino
    InstrLongueur : string;
    ModeMesures   : string; // normal = longueurs; diving = profondimètres

  begin
    Result := 0.00;
    WrtLn(Format('       # Code %d: %s',[C.IDCode, C.Commentaire]));
    InstrLongueur := 'tape';
    ModeMesures   := 'normal';
    // unités de la boussole
    case Trunc(C.GradAz) of
      359, 360: begin  UnitCompass := UNITE_ANGULAIRE_DEGRES;   IsBackCompass := DIRECT_VISEE; end; // visées directes en degrés
      399, 400: begin  UnitCompass := UNITE_ANGULAIRE_GRADES;   IsBackCompass := DIRECT_VISEE; end; // visées directes en grades
      349, 350: begin  UnitCompass := UNITE_ANGULAIRE_DEGRES;   IsBackCompass := BACKSIDE_VISEE; end; // visées inverses en degrés
      389, 390: begin  UnitCompass := UNITE_ANGULAIRE_GRADES;   IsBackCompass := BACKSIDE_VISEE; end;  // visées inverses en grades
      else      begin UnitCompass  := UNITE_ANGULAIRE_DEGRES;   IsBackCompass := DIRECT_VISEE; end;
    end;
    WrtLn('       units compass ' + UnitCompass);
    // unités du clinomètres
    InstrVertical := 'clino';
    IsBackClino   := DIRECT_VISEE;
    DepthMode     := false;
    case Trunc(C.GradInc) of
      350: begin   UnitClino := UNITE_ANGULAIRE_DEGRES;        IsBackClino := BACKSIDE_VISEE; end; // zéro à l'horizontale degré inverse
      390: begin   UnitClino := UNITE_ANGULAIRE_GRADES;        IsBackClino := BACKSIDE_VISEE; end; // zéro à l'horizontale grade inverse
      360: begin   UnitClino := UNITE_ANGULAIRE_DEGRES;                               end; // zéro à l'horizontale degré direct
      400: begin   UnitClino := UNITE_ANGULAIRE_GRADES;                                 end; // zéro à l'horizontale grade direct
      361: begin   UnitClino := UNITE_ANGULAIRE_DEGRES;        Result    :=   90.00;  end; // zéro zénithal degré
      401: begin   UnitClino := UNITE_ANGULAIRE_GRADES;        Result    :=  100.00;  end; // zéro zénithal grade
      359: begin   UnitClino := UNITE_ANGULAIRE_DEGRES;        Result    :=  -90.00;  end; // zéro nadiral degré
      399: begin   UnitClino := UNITE_ANGULAIRE_GRADES;        Result    := -100.00;  end; // zéro nadiral grade
      370: begin ;                                                       end; // pourcentages
      380: begin
             InstrVertical := MODE_DENIVELE;
             ModeMesures   := MODE_DIVING;
             DepthMode     := True;
           end; // différences d'altitudes
      UNITE_CLINO_LASERMETRE_STANLEY_TLM330_INV:
           begin
             InstrVertical := MODE_DENIVELE;
             ModeMesures   := MODE_DIVING;
             IsBackClino   := BACKSIDE_VISEE;
             DepthMode     := True;
           end;
      UNITE_CLINO_LASERMETRE_STANLEY_TLM330_DIR:
           begin
             InstrVertical := MODE_DENIVELE;
             ModeMesures   := MODE_DIVING;
             DepthMode     := True;
           end;
    else
      begin
        UnitClino   := UNITE_ANGULAIRE_DEGRES;
        IsBackClino := '';
      end;
    end;
    if (Not DepthMode) then WrtLn('       units clino ' + UnitClino);
    WrtLn(Format('       data %s from to %s %scompass %s%s right left ceiling floor',
          [ModeMesures,
           InstrLongueur,
           IsBackCompass,
           IsBackClino,
           InstrVertical
          ]));
    WrtLn('       #-------------------');
    WrtLn('');
  end;
  // définition d'une série
  procedure DefineSerie(const S: TObjSerie; const N: integer; const StationDeRattachement: string);
  var
    foo, bar : string;
    Qdr, Incl: double;
    MyExpe   : TExpe;
    MyCode   : TCode;
    MyStation: TUneVisee;
    St       : integer;
    OldIdxExpe  : integer;
    OldIdxCode   : integer;
    OldIDStation: string;
    CurIDStation: string;
    Tag          : string; // Commentaires et ID littéraux de stations
    miou: Int64;
    lulu: Double;
  begin
    //S.NomSerie
    AfficherMessage(Format('----> Serie %d: %s',[S.GetNumeroDeSerie(), S.GetNomSerie]));
    WrtLn('');
    WrtLn(Format('    # Serie: %d - %s',[S.GetNumeroDeSerie(), S.GetNomSerie]));
    WrtLn('    #========================');
    WrtLn('    ' + TH_KW_CENTERLINE);
    WrtLn('');
    // stations
    OldIdxExpe  := -1;
    OldIdxCode  := -1;

    OldIDStation:= Format(FMTSTS, [S.GetNoSerieDep, S.GetNoPointDep]);
    // première station
    MyStation   := S.GetVisee(1);
    MyExpe      := GetExpeByNumero(MyStation.Expe);
    MyCode      := GetCodeByNumero(MyStation.Code);
    OldIdxExpe  := MyExpe.IDExpe;
    OldIdxCode  := MyCode.IDCode;
    DefineSeance(MyExpe);
    Qdr := DefineCode(MyCode);
    for St := 1 to S.GetNbVisees - 1 do
    begin
      MyStation := S.GetVisee(St);
      //if NoSt=Serie.PointsTopo.Count-1 then ToStationID:=Format(FMTSERSTVTOPO,[S.SerieArr, S.PtArr])
      if (St = S.GetNbVisees - 1) then
        CurIDStation := Format(FMTSTS, [S.GetNoSerieArr, S.GetNoPointArr])
      else
        CurIDStation := Format(FMTSTS, [S.GetNumeroDeSerie(), St]);
      MyExpe      := GetExpeByNumero(MyStation.Expe);
      MyCode      := GetCodeByNumero(MyStation.Code);
      if (MyExpe.IDExpe <> OldIdxExpe) then
      begin
        DefineSeance(MyExpe);
        OldIdxExpe  := MyExpe.IDExpe;
      end;

      if (MyCode.IDCode <> OldIdxCode) then
      begin
        Qdr := DefineCode(MyCode);
        //DefineSeance(MyExpe);
        OldIdxCode  := MyCode.IDCode;
      end;
      if Trunc(Qdr)=0 then
        Incl := MyStation.Pente
      else begin
        if (Qdr>1.00) then Incl := Qdr - MyStation.Pente
                      else Incl := MyStation.Pente + Qdr;
      end;
      // Lasermètres de bâtiment => Utilisation des longueurs projetées
      miou := Trunc(MyCode.GradInc);
      if ((miou = UNITE_CLINO_LASERMETRE_STANLEY_TLM330_INV) or
          (miou = UNITE_CLINO_LASERMETRE_STANLEY_TLM330_DIR)) then
        lulu := Hypot2D(MyStation.Longueur, MyStation.Pente)
      else
        lulu := MyStation.Longueur;
      // ID Littéraux
      Tag := '';
      if (MyStation.IDTerrainStation <> '') then
        Tag := Format('# Label: %s ',[MyStation.IDTerrainStation]);
      if (MyStation.Commentaires <> '') then
        Tag := Tag + Format('# Observ: %s', [MyStation.Commentaires]);

      WrtLn(Format('       %s %s %.2f %.2f %.2f  %.2f %.2f  %.2f %.2f  %s',
                   [OldIDStation, CurIDStation,
                    lulu, MyStation.Azimut, Incl,
                    MyStation.LG, MyStation.LD, MyStation.HZ, MyStation.HN,
                    Tag]));
      //--------------------------------------------------------------
      OldIDStation := CurIDStation;
    end;
    WrtLn('');
    WrtLn('    ' + TH_KW_ENDCENTERLINE);
    foo := Format(FMTSTS, [S.GetNoSerieDep, S.GetNoPointDep]);
    bar := Format('%s@SERIE%d',[foo, S.GetNoSerieDep]);
  end;
begin
  AfficherMessage(Format('%s.ExportToTherion(%s)',[ClassName, FichierTH]));
  AssignFile(THF, FichierTH);
  try
    ReWrite(THF);
    AfficherMessage('--> Header');
    // encodage
    WrtLn('encoding  iso8859-2');
    WrtLn('');
    // date et nom du dossier
    WrtLn(StringOfChar('#', 80));
    WrtLn('## Data file for Therion generated by GHTopo');
    WrtLn('## Folder: ' + FDatabaseName);
    WrtLn('## Date  : ' + DateTimeToStr(Now()));
    WrtLn(StringOfChar('#', 80));
    WrtLn('');
    // centerlines et consorts
    WrtLn(Format('survey Survey001 -title "%s"',[GetNomEtude]));
     // points fixes
    AfficherMessage('--> Fixed points');
    WrtLn('  # Section Entrances and Fixed points');
    WrtLn('  ' + TH_KW_CENTERLINE);

    for i := 0 to GetNbEntrees -1 do begin
      MyFixPt := GetEntree(i);
      WrtLn(Format('    fix %s %.2f %.2f %.2f # %s',
                   [ Format(FMTSTS, [MyFixPt.eRefSer, MyFixPt.eRefSt]),
                     MyFixPt.eXEntree,
                     MyFixPt.eYEntree,
                     MyFixPt.eZEntree,
                     MyFixPt.eNomEntree
                   ]));
    end;
    WrtLn('  ' + TH_KW_ENDCENTERLINE);
    WrtLn('  # End section Entrances and Fixed points');
    WrtLn('  # Section SERIES');
    // séries
    AfficherMessage('--> Series');
    MaStationDeRattachement := '';
    for i := 1 to GetNbSeries-1 do begin
      MySerie := GetSerie(i);
      DefineSerie(MySerie, i, MaStationDeRattachement);
    end;
    WrtLn('  # End section SERIES');
    WrtLn('endsurvey');
  finally
    CloseFile(THF);
  end;
end;
// Export vers Topodroid
// /!\ Non testé
procedure TToporobotStructure2012.ExportVersTopodroid(const FichierCSV: TStringDirectoryFilename);
const
  MY_NAMESPACE = 'CAVITE';
var
  THF      : TextFile;
  MySerie  : TObjSerie;
  MyFixPt  : TEntrance;
  // Station de rattachement courante
  MaStationDeRattachement: string;
  i           : integer;
  procedure WrtLn(const Line: string);
  begin
    WriteLn(THF, Line);
  end;

  // définition d'un en-tete de séance
  procedure DefineSeance(const E: TExpe);
  begin
    //WrtLn(Format('       # Seance %d: %s',[E.IDExpe, E.Commentaire]));
    //WrtLn(Format('       date %.4d.%.2d.%.2d',[ E.AnneeExpe,  E.MoisExpe,  E.JourExpe]));
    //WrtLn('       #-------------------');
  end;
  // definition d'un code - Valeur de retour par défaut: -1, sinon, un angle droit
  function DefineCode(const C: TCode): double;
  var
    UnitCompass   : string;
    UnitClino     : string;
    IsBackCompass : string;
    IsBackClino   : string;
    InstrVertical : string; // mesures altimétriques
    DepthMode     : boolean; // si DepthMode=True, ne pas spécifier les unités de clino
    InstrLongueur : string;
    ModeMesures   : string; // normal = longueurs; diving = profondimètres

  begin
    Result := 0.00;
    WrtLn(Format('       # Code %d: %s',[C.IDCode, C.Commentaire]));
    InstrLongueur := 'tape';
    ModeMesures   := 'normal';
    // unités de la boussole
    case Trunc(C.GradAz) of
      359, 360: begin  UnitCompass := 'degrees';   IsBackCompass :=''; end; // visées directes en degrés
      399, 400: begin  UnitCompass := 'grads';     IsBackCompass :=''; end; // visées directes en grades
      349, 350: begin  UnitCompass := 'degrees';   IsBackCompass :='back'; end; // visées inverses en degrés
      389, 390: begin  UnitCompass := 'grads';     IsBackCompass :='back'; end;  // visées inverses en grades
      else begin UnitCompass    := 'degrees';  IsBackCompass  := ''; end;
    end;
    WrtLn('       units compass ' + UnitCompass);
    // unités du clinomètres
    InstrVertical := 'clino';
    IsBackClino   := '';
    DepthMode     := false;
    case Trunc(C.GradInc) of
      350: begin   UnitClino   := 'degrees';      IsBackClino := 'back'; end; // zéro à l'horizontale degré inverse
      390: begin   UnitClino := 'grads';          IsBackClino := 'back'; end; // zéro à l'horizontale grade inverse
      360: begin   UnitClino := 'degrees';                               end; // zéro à l'horizontale degré direct
      400: begin   UnitClino := 'grads';                                 end; // zéro à l'horizontale grade direct
      361: begin   UnitClino := 'degrees';        Result    :=  90.00;   end; // zéro zénithal degré
      401: begin   UnitClino := 'grads';          Result    := 100.00;   end; // zéro zénithal grade
      359: begin   UnitClino := 'degrees';        Result    := -90.00;   end; // zéro nadiral degré
      399: begin   UnitClino := 'grads';          Result    := -100.00;  end; // zéro nadiral grade
      370: begin ;                                                       end; // pourcentages
      380: begin
             InstrVertical := 'depthchange';
             ModeMesures   := 'diving';
             DepthMode     := True;
           end; // différences d'altitudes
      UNITE_CLINO_LASERMETRE_STANLEY_TLM330_INV:
           begin
             InstrVertical := 'depthchange';
             ModeMesures   := 'diving';
             IsBackClino   := 'back';
             DepthMode     := True;
           end;
      UNITE_CLINO_LASERMETRE_STANLEY_TLM330_DIR:
           begin
             InstrVertical := 'depthchange';
             ModeMesures   := 'diving';
             DepthMode     := True;
           end;
    else
      begin UnitClino   := 'degrees'; IsBackClino :=''; end;
    end;
    if (Not DepthMode) then WrtLn('       units clino ' + UnitClino);
    WrtLn(Format('       data %s from to %s %scompass %s%s right left ceiling floor',
          [ModeMesures,
           InstrLongueur,
           IsBackCompass,
           IsBackClino,
           InstrVertical
          ]));
    WrtLn('       #-------------------');
    WrtLn('');
  end;
  //**********************************
  // définition d'une série
  procedure DefineSerie(const S: TObjSerie; const N: integer; const StationDeRattachement: string);
  var
    foo, bar : string;
    Qdr, Incl: double;
    MyExpe   : TExpe;
    MyCode   : TCode;
    MyStation: TUneVisee;
    St       : integer;
    OldIdxExpe  : integer;
    OldIdxCode   , nva, aa: integer;
    OldIDStation: string;
    CurIDStation: string;
    Tag          : string; // Commentaires et ID littéraux de stations
    miou: Int64;
    QLongueur: Double;
    MyViseeAntenne: TViseeAntenne;
  begin
    //S.NomSerie
    AfficherMessage(Format('----> Serie %d: %s',[S.GetNumeroDeSerie(), S.GetNomSerie]));
    WrtLn('');
    WrtLn(Format('    # Serie: %d - %s',[S.GetNumeroDeSerie(), S.GetNomSerie]));
    WrtLn('# from to tape compass clino (declination 0.0000)');
    WrtLn('# units tape meters compass clino degrees');
    //WrtLn('    #========================');
    //WrtLn('    centerline');
    //WrtLn('');
    // stations
    OldIdxExpe  := -1;
    OldIdxCode  := -1;

    OldIDStation:= Format(FMTSTS, [S.GetNoSerieDep, S.GetNoPointDep]);
    // première station
    MyStation   := S.GetVisee(1);
    MyExpe      := GetExpeByNumero(MyStation.Expe);
    MyCode      := GetCodeByNumero(MyStation.Code);
    OldIdxExpe  := MyExpe.IDExpe;
    OldIdxCode  := MyCode.IDCode;
    DefineSeance(MyExpe);
    Qdr := DefineCode(MyCode);
    for St := 1 to S.GetNbVisees - 1 do
    begin
      MyStation := S.GetVisee(St);
      //if NoSt=Serie.PointsTopo.Count-1 then ToStationID:=Format(FMTSERSTVTOPO,[S.SerieArr, S.PtArr])
      if (St = S.GetNbVisees - 1) then
        CurIDStation := Format(FMTSTS, [S.GetNoSerieArr, S.GetNoPointArr])
      else
        CurIDStation := Format(FMTSTS, [S.GetNumeroDeSerie(), St]);
      MyExpe      := GetExpeByNumero(MyStation.Expe);
      MyCode      := GetCodeByNumero(MyStation.Code);
      if (MyExpe.IDExpe <> OldIdxExpe) then
      begin
        DefineSeance(MyExpe);
        OldIdxExpe  := MyExpe.IDExpe;
      end;

      if (MyCode.IDCode <> OldIdxCode) then
      begin
        Qdr := DefineCode(MyCode);
        //DefineSeance(MyExpe);
        OldIdxCode  := MyCode.IDCode;
      end;
      if Trunc(Qdr)=0 then
        Incl := MyStation.Pente
      else
      begin
        if (Qdr>1.00) then Incl := Qdr - MyStation.Pente
                      else Incl := MyStation.Pente + Qdr;
      end;
      // Lasermètres de bâtiment => Utilisation des longueurs projetées
      miou := Trunc(MyCode.GradInc);
      if ((miou = UNITE_CLINO_LASERMETRE_STANLEY_TLM330_INV) or
          (miou = UNITE_CLINO_LASERMETRE_STANLEY_TLM330_DIR)) then
        QLongueur := Hypot2D(MyStation.Longueur, MyStation.Pente)
      else
        QLongueur := MyStation.Longueur;
      // ID Littéraux
      Tag := '';
      if (MyStation.IDTerrainStation <> '') then
        Tag := Format('# Label: %s ',[MyStation.IDTerrainStation]);
      if (MyStation.Commentaires <> '') then
        Tag := Tag + Format('# Observ: %s', [MyStation.Commentaires]);
      //8.5@miaou,8.6@miaou,4.65,352.2,-0.2
      WrtLn(Format('%s@%s,%s@%s,%.3f,%.3f,%.3f',
                   [OldIDStation, MY_NAMESPACE,
                    CurIDStation, MY_NAMESPACE,
                    QLongueur, MyStation.Azimut, Incl
                    ]));
      // et les antennes ici
      nva := self.GetNbAntennes();
      for aa := 0 to nva - 1 do
      begin
        MyViseeAntenne := GetViseeAntenne(aa);
        if ((MyViseeAntenne.SerieDepart = MySerie.GetNumeroDeSerie()) and (MyViseeAntenne.PtDepart = St)) then
        begin
          // write antenne;
          //0@miaou,-,1.55,257.1,-36.7
          wrtln(Format('** %d_%d@%s,%s,%.3f,%.3f,%.3f', [MyViseeAntenne.SerieDepart, MyViseeAntenne.PtDepart,
                                                         MY_NAMESPACE, '-',
                                                         MyViseeAntenne.Longueur, MyViseeAntenne.Azimut, MyViseeAntenne.Pente]));
        end;
      end;
      //--------------------------------------------------------------
      OldIDStation := CurIDStation;
    end;
    WrtLn('');
    //WrtLn('    endcenterline');
    foo := Format(FMTSTS, [S.GetNoSerieDep, S.GetNoPointDep]);
    bar := Format('%s@SERIE%d',[foo, S.GetNoSerieDep]);
  end;

begin
  AfficherMessage(Format('%s.ExportVersTopodroid(%s)',[ClassName, FichierCSV]));
  AssignFile(THF, FichierCSV);
  //try
    ReWrite(THF);
    AfficherMessage('--> Header');
    // encodage
    // date et nom du dossier
    WrtLn(StringOfChar('#', 80));
    WrtLn('# Data file for Topodroid generated by GHTopo');
    WrtLn('# Folder: ' + FDatabaseName);
    WrtLn('# Date  : ' + DateTimeToStr(Now()));



    //WrtLn(StringOfChar('#', 80));
    //WrtLn('');
    // centerlines et consorts
    //WrtLn(Format('survey Survey001 -title "%s"',[GetNomEtude]));
     // points fixes
    AfficherMessage('--> Fixed points');
    //WrtLn('  # Section Entrances and Fixed points');
    //WrtLn('  centerline');
    for i := 0 to GetNbEntrees -1 do
    begin
      MyFixPt := GetEntree(i);
      (*
      WrtLn(Format('    fix %s %.2f %.2f %.2f # %s',
                   [ Format(FMTSTS, [MyFixPt.eRefSer, MyFixPt.eRefSt]),
                     MyFixPt.eXEntree,
                     MyFixPt.eYEntree,
                     MyFixPt.eZEntree,
                     MyFixPt.eNomEntree
                   ]));
      //*)
    end;

    //WrtLn('  endcenterline');
    //WrtLn('  # End section Entrances and Fixed points');
    //WrtLn('  # Section SERIES');
    // séries
    AfficherMessage('--> Series');
    MaStationDeRattachement := '';
    for i := 1 to GetNbSeries-1 do
    begin
      MySerie := GetSerie(i);
      DefineSerie(MySerie, i, MaStationDeRattachement);
    end;
    //WrtLn('  # End section SERIES');
    //WrtLn('endsurvey');
  //finally
    CloseFile(THF);
  //end;
end;

// Lecture/écriture au format XML
// (futur format standard de GHTopo)
procedure TToporobotStructure2012.SaveToXML(const FichierXML: TStringDirectoryFilename);
const
  IDX_MINI = 1;
var
  i: Integer;
  MyDocXML: TXMLDocument;
  GHTopoRoot: TDOMElement;
  SectionGeneral: TDOMElement;
  SubSection: TDOMElement;
  SubSubSection: TDOMElement;
  SubSubSubSection: TDOMElement;
  SectionEntrances: TDOMElement;
  E: TEntrance;
  SectionReseaux: TDOMElement;
  R: TReseau;
  SectionSecteurs: TDOMElement;
  SC: TSecteur;
  SectionCodes: TDOMElement;
  C: TCode;
  SectionSeances: TDOMElement;
  EX: TExpe;
  SectionSeries: TDOMElement;
  SER: TObjSerie;

  ST: TUneVisee;
  S, QNbReleaseNotes, QNbAntennes, QNbFiltres: Integer;
  SectionAntennes, SectionReleaseNotes, SectionFiltres: TDOMElement;
  ANT: TViseeAntenne;
  EPSG: TLabelSystemesCoordsEPSG;
  NV: TReleaseNotes;
  FF: TFiltrePersonnalise;
  procedure KWYXZ(const S666: string);
  begin
    AfficherMessage(Format('-- Writing "%s" section', [S666]));
  end;
begin
  AfficherMessage(Format('%s.SaveToXML(%s)', [ClassName, FichierXML]));
  MyDocXML := TXMLDocument.Create;
  try
    // racine du document = nom du logiciel
    GHTopoRoot := MyDocXML.CreateElement('GHTopo');
    MyDocXML.AppendChild(GHTopoRoot);
    // section Général
    SectionGeneral := MyDocXML.CreateElement(GTX_KEY_SECTION_GENERAL);
      GHTopoRoot.AppendChild(SectionGeneral);
      SubSection := MyDocXML.CreateElement(GTX_KEY_CAVITE);
      SubSection.SetAttribute(GTX_ATTR_NAMESPACE, PasStrToXMLStr(GetNameSpace()));
      SubSection.SetAttribute(GTX_ATTR_NOM_ETUDE, PasStrToXMLStr(GetNomEtude()));
      EPSG := GetCodeEPSGSystemeCoordonnees;
      SubSection.SetAttribute(GTX_ATTR_COORDS_SYSTEM_EPSG, Format('%d', [EPSG.CodeEPSG]));
      SubSection.SetAttribute(GTX_ATTR_COORDS_SYSTEM_NOM, PasStrToXMLStr(EPSG.NomEPSG));
      SubSection.SetAttribute(GTX_ATTR_COMMENTAIRES_ETUDE, PasStrToXMLStr(GetCommentairesEtude));
      SectionGeneral.AppendChild(SubSection);
    // section Notes de Version
    {$IFDEF SUPPORT_NOTES_VERSION}
    QNbReleaseNotes := GetNbNotesVersion();
    if (QNbReleaseNotes > 0) then
    begin
      SectionReleaseNotes := MyDocXML.CreateElement(GTX_KEY_SECTION_RELEASE_NOTES);
      GHTopoRoot.AppendChild(SectionReleaseNotes);
      KWYXZ(GTX_KEY_SECTION_RELEASE_NOTES);
      for i := 0 to QNbReleaseNotes - 1 do
      begin
        NV := GetNoteVersion(i);
        SubSection := MyDocXML.CreateElement(GTX_KEY_RELEASE_NOTE);
          SubSection.SetAttribute(GTX_ATTR_RELEASE_NOTE_DATE      , DateTimePascalToDateTimeSQL(NV.Date));
          SubSection.SetAttribute(GTX_ATTR_RELEASE_NOTE_AUTHOR    , Trim(NV.Author));
          SubSection.SetAttribute(GTX_ATTR_RELEASE_NOTE_DESCRO    , Trim(NV.Description));
        SectionReleaseNotes.AppendChild(SubSection);
      end;
    end;
    {$endif SUPPORT_NOTES_VERSION}
    // section Filtres perso
    QNbFiltres := self.GetNbFiltresPersos();
    if (QNbFiltres > 0) then
    begin
      SectionFiltres := MyDocXML.CreateElement(GTX_KEY_SECTION_FILTERS);
      GHTopoRoot.AppendChild(SectionFiltres);
      KWYXZ(GTX_KEY_SECTION_FILTERS);
      for i := 0 to QNbFiltres - 1 do
      begin
        FF := GetFiltrePerso(i);
        SubSection := MyDocXML.CreateElement(GTX_KEY_FILTER);
          SubSection.SetAttribute(GTX_ATTR_FILTER_IDX        , Format('%d', [i]));
          SubSection.SetAttribute(GTX_ATTR_FILTER_NAME       , PasStrToXMLStr(FF.NomFiltre));
          SubSection.SetAttribute(GTX_ATTR_FILTER_COLOR_R    , Format('%d', [Red(FF.CouleurFiltre)]));
          SubSection.SetAttribute(GTX_ATTR_FILTER_COLOR_G    , Format('%d', [Green(FF.CouleurFiltre)]));
          SubSection.SetAttribute(GTX_ATTR_FILTER_COLOR_B    , Format('%d', [Blue(FF.CouleurFiltre)]));
          SubSection.SetAttribute(GTX_ATTR_FILTER_EXPRESSION , PasStrToXMLStr(FF.Expression));
          SubSection.SetAttribute(GTX_ATTR_FILTER_DESCRIPTION, PasStrToXMLStr(FF.Description));
        SectionFiltres.AppendChild(SubSection);
      end;
    end;
    // section Entrées
    SectionEntrances := MyDocXML.CreateElement(GTX_KEY_SECTION_ENTRANCES);
      GHTopoRoot.AppendChild(SectionEntrances);
      KWYXZ(GTX_KEY_SECTION_ENTRANCES);
      for i:= 0 to GetNbEntrees - 1 do
      begin
        E := GetEntree(i);
        SubSection := MyDocXML.CreateElement(GTX_KEY_ENTRANCE);
          SubSection.SetAttribute(GTX_ATTR_ENTRANCE_IDX      , Format('%d', [i]));
          SubSection.SetAttribute(GTX_ATTR_ENTRANCE_IDTERRAIN, Format('%s', [PasStrToXMLStr(E.eIDTerrain)]));
          SubSection.SetAttribute(GTX_ATTR_ENTRANCE_NAME     , Format('%s', [PasStrToXMLStr(E.eNomEntree)]));
          SubSection.SetAttribute(GTX_ATTR_ENTRANCE_REFSERIE , Format('%d', [E.eRefSer]));
          SubSection.SetAttribute(GTX_ATTR_ENTRANCE_REFPT    , Format('%d', [E.eRefSt]));
          SubSection.SetAttribute(GTX_ATTR_ENTRANCE_X, Format('%.2f', [E.eXEntree]));
          SubSection.SetAttribute(GTX_ATTR_ENTRANCE_Y, Format('%.2f', [E.eYEntree]));
          SubSection.SetAttribute(GTX_ATTR_ENTRANCE_Z, Format('%.2f', [E.eZEntree]));
          SubSection.SetAttribute(GTX_ATTR_ENTRANCE_OBS, PasStrToXMLStr(E.eObserv));

        SectionEntrances.AppendChild(SubSection);
      end;
    // section Réseaux (5 champs sur 5)
    SectionReseaux := MyDocXML.CreateElement(GTX_KEY_SECTION_RESEAUX);
      KWYXZ(GTX_KEY_SECTION_RESEAUX);
      GHTopoRoot.AppendChild(SectionReseaux);
      for i := IDX_MINI to GetNbReseaux - 1 do
      begin
        R := GetReseau(i);
        SubSection := MyDocXML.CreateElement(GTX_KEY_RESEAU);
          SubSection.SetAttribute(GTX_ATTR_RESEAU_IDX, Format('%d', [i]));
          SubSection.SetAttribute(GTX_ATTR_RESEAU_COLOR_R, Format('%d', [Red(R.ColorReseau)]));
          SubSection.SetAttribute(GTX_ATTR_RESEAU_COLOR_G, Format('%d', [Green(R.ColorReseau)]));
          SubSection.SetAttribute(GTX_ATTR_RESEAU_COLOR_B, Format('%d', [Blue(R.ColorReseau)]));
          SubSection.SetAttribute(GTX_ATTR_RESEAU_TYPE, Format('%d', [R.TypeReseau]));
          SubSection.SetAttribute(GTX_ATTR_RESEAU_NAME,  PasStrToXMLStr(R.NomReseau));
          SubSection.SetAttribute(GTX_ATTR_RESEAU_OBS,  PasStrToXMLStr(R.ObsReseau));
        SectionReseaux.AppendChild(SubSection);
      end;
    // section Secteurs (5 champs sur 5)
    SectionSecteurs := MyDocXML.CreateElement(GTX_KEY_SECTION_SECTEURS);
      KWYXZ(GTX_KEY_SECTION_SECTEURS);
      GHTopoRoot.AppendChild(SectionSecteurs);
      for i := IDX_MINI to GetNbSecteurs - 1 do
      begin
        SC := GetSecteur(i);
        SubSection := MyDocXML.CreateElement(GTX_KEY_SECTEUR);
          SubSection.SetAttribute(GTX_ATTR_SECTEUR_IDX    , Format('%d', [i]));
          SubSection.SetAttribute(GTX_ATTR_SECTEUR_COLOR_R, Format('%d', [Red(SC.CouleurSecteur)]));
          SubSection.SetAttribute(GTX_ATTR_SECTEUR_COLOR_G, Format('%d', [Green(SC.CouleurSecteur)]));
          SubSection.SetAttribute(GTX_ATTR_SECTEUR_COLOR_B, Format('%d', [Blue(SC.CouleurSecteur)]));
          SubSection.SetAttribute(GTX_ATTR_SECTEUR_NAME   , PasStrToXMLStr(SC.NomSecteur));
        SectionSecteurs.AppendChild(SubSection);
      end;

    // section Codes -- Champs OK
    SectionCodes := MyDocXML.CreateElement(GTX_KEY_SECTION_CODES);
      KWYXZ(GTX_KEY_SECTION_CODES);
      GHTopoRoot.AppendChild(SectionCodes);
      for i := IDX_MINI to GetNbCodes - 1 do
      begin
        C := GetCode(i);
        SubSection := MyDocXML.CreateElement(GTX_KEY_CODE);
          SubSection.SetAttribute(GTX_ATTR_CODE_NUMERO           , Format('%d', [C.IDCode]));                   //  IDCode
          SubSection.SetAttribute(GTX_ATTR_CODE_TYPE             , Format('%d', [0])); //C.ReservedInt]));              //  ReservedInt
          // TODO: Support de FactLong a valider
          SubSection.SetAttribute(GTX_ATTR_CODE_FACT_LONG        , Format('%.3f', [1.00])); ///// SubSection.SetAttribute(GTX_ATTR_CODE_FACT_LONG     , Format('%.3f', [C.FactLong]));               //  FactLong    : double; // pour compatibilité ascendante
          SubSection.SetAttribute(GTX_ATTR_CODE_UCOMPASS         , Format('%.2f', [C.GradAz]));                 //  GradAz
          SubSection.SetAttribute(GTX_ATTR_CODE_UCLINO           , Format('%.2f', [C.GradInc]));                //  GradInc     : double;
          SubSection.SetAttribute(GTX_ATTR_CODE_ANGLIMITE        , Format('%.2f', [C.AngLimite]));              //  AngLimite   : double;
          SubSection.SetAttribute(GTX_ATTR_CODE_PSI_L            , Format('%.3f', [C.PsiL]));                   //  PsiL        : double;
          SubSection.SetAttribute(GTX_ATTR_CODE_PSI_A            , Format('%.3f', [C.PsiAz]));                  //  PsiAz       : double;
          SubSection.SetAttribute(GTX_ATTR_CODE_PSI_P            , Format('%.3f', [C.PsiP]));                   //  PsiP        : double;
          SubSection.SetAttribute(GTX_ATTR_CODE_ERROR_TOURILLON  , Format('%.3f', [C.ErreurTourillon]));

          SubSection.SetAttribute(GTX_ATTR_CODE_OBS              , PasStrToXMLStr(C.Commentaire));                              //  Commentaire : string;
          SubSection.SetAttribute(GTX_ATTR_CODE_FUNC_CORR_AZ_CO            , Format('%.6f', [C.ParamsFuncCorrAz.Co]));
          SubSection.SetAttribute(GTX_ATTR_CODE_FUNC_CORR_AZ_ERR_MAX       , Format('%.6f', [C.ParamsFuncCorrAz.ErreurMax]));
          SubSection.SetAttribute(GTX_ATTR_CODE_FUNC_CORR_AZ_POS_ERR_MAX   , Format('%.6f', [C.ParamsFuncCorrAz.PosErrMax]));
          SubSection.SetAttribute(GTX_ATTR_CODE_FUNC_CORR_INC_CO           , Format('%.6f', [C.ParamsFuncCorrInc.Co]));
          SubSection.SetAttribute(GTX_ATTR_CODE_FUNC_CORR_INC_ERR_MAX      , Format('%.6f', [C.ParamsFuncCorrInc.ErreurMax]));
          SubSection.SetAttribute(GTX_ATTR_CODE_FUNC_CORR_INC_POS_ERR_MAX  , Format('%.6f', [C.ParamsFuncCorrInc.PosErrMax]));

        SectionCodes.AppendChild(SubSection);
      end;
    // section Expés - Champs OK
    SectionSeances := MyDocXML.CreateElement(GTX_KEY_SECTION_EXPES);
      KWYXZ(GTX_KEY_SECTION_EXPES);
      GHTopoRoot.AppendChild(SectionSeances);
      for i := IDX_MINI to GetNbExpes - 1 do
      begin
        EX := GetExpe(i);
        SubSection := MyDocXML.CreateElement(GTX_KEY_EXPE);
          SubSection.SetAttribute(GTX_ATTR_EXPE_NUMERO   , Format('%d', [EX.IDExpe]));                                             // IDExpe      : integer;
          SubSection.SetAttribute(GTX_ATTR_EXPE_DATE     , Format('%.4d-%.2d-%.2d', [EX.AnneeExpe, EX.MoisExpe, EX.JourExpe]));    // JourExpe    : integer;MoisExpe    : integer;AnneeExpe   : integer;
          SubSection.SetAttribute(GTX_ATTR_EXPE_IDXCOLOR , Format('%d', [EX.Couleur]));                                            // //Couleur     : Integer;
          SubSection.SetAttribute(GTX_ATTR_EXPE_DECLINAT , Format('%.4f', [EX.Declinaison]));                                      // Declinaison : double;
          SubSection.SetAttribute(GTX_ATTR_EXPE_MODEDECL , Format('%d', [EX.ModeDecl]));                                           // ModeDecl    : integer; // deprecated;
          SubSection.SetAttribute(GTX_ATTR_EXPE_INCLINAT , Format('%.4f', [0.00])); //EX.Inclinaison]));                                      //  Inclinaison : double;
          SubSection.SetAttribute(GTX_ATTR_EXPE_SURVEY1  , PasStrToXMLStr(EX.Operateur));                                                          // Speleometre : String;
          SubSection.SetAttribute(GTX_ATTR_EXPE_SURVEY2  , PasStrToXMLStr(EX.ClubSpeleo));                                                         //  Speleographe: string;
          SubSection.SetAttribute(GTX_ATTR_EXPE_OBS      , PasStrToXMLStr(EX.Commentaire));                                                        //  Commentaire : string;
        SectionSeances.AppendChild(SubSection);
      end;
    // section Séries
    SectionSeries := MyDocXML.CreateElement(GTX_KEY_SECTION_SERIES);
      GHTopoRoot.AppendChild(SectionSeries);
      KWYXZ(GTX_KEY_SECTION_SERIES);
      for i := IDX_MINI to GetNbSeries - 1 do
      begin
        SER := GetSerie(i);
        SubSection := MyDocXML.CreateElement(GTX_KEY_SERIE);
          SubSection.SetAttribute(GTX_ATTR_SERIE_Numero  , Format('%d', [SER.GetNumeroDeSerie()]));
          SubSection.SetAttribute(GTX_ATTR_SERIE_NAME    , SER.GetNomSerie);
          SubSection.SetAttribute(GTX_ATTR_SERIE_ENTRANCE, Format('%d', [SER.GetNumeroEntrance()]));

          SubSection.SetAttribute(GTX_ATTR_SERIE_RESEAU  , Format('%d', [SER.GetNumeroReseau]));

          SubSection.SetAttribute(GTX_ATTR_SERIE_SERDEP  , Format('%d', [SER.GetNoSerieDep()]));
          SubSection.SetAttribute(GTX_ATTR_SERIE_PTDEP   , Format('%d', [SER.GetNoPointDep()]));
          SubSection.SetAttribute(GTX_ATTR_SERIE_SERARR  , Format('%d', [SER.GetNoSerieArr()]));
          SubSection.SetAttribute(GTX_ATTR_SERIE_PTARR   , Format('%d', [SER.GetNoPointArr()]));

          SubSection.SetAttribute(GTX_ATTR_SERIE_COLOR   , Format('%s', [ColorToHTMLColor(SER.GetCouleur())]));

          SubSection.SetAttribute(GTX_ATTR_SERIE_CHANCE  , Format('%d', [SER.GetChance]));
          SubSection.SetAttribute(GTX_ATTR_SERIE_OBSTACLE, Format('%d', [SER.GetObstacle]));
          SubSection.SetAttribute(GTX_ATTR_SERIE_OBS     , PasStrToXMLStr(SER.GetObsSerie));

          SubSection.SetAttribute(GTX_ATTR_SERIE_RAIDEUR , Format('%.4f', [SER.GetRaideur]));

        SectionSeries.AppendChild(SubSection);
        SubSubSection := MyDocXML.CreateElement(GTX_KEY_STATIONS);
        SubSection.AppendChild(SubSubSection);
        for S:=0 to SER.GetNbVisees - 1 do
        begin
          ST := SER.GetVisee(S);
          SubSubSubSection := MyDocXML.CreateElement(GTX_KEY_VISEE);
            SubSubSubSection.SetAttribute(GTX_ATTR_VISEE_ID   , Format('%d.%d', [SER.GetNumeroDeSerie(), S]));     //
            SubSubSubSection.SetAttribute(GTX_ATTR_VISEE_LBL  , PasStrToXMLStr(ST.IDTerrainStation));         // IDTerrainStation: string;

            SubSubSubSection.SetAttribute(GTX_ATTR_VISEE_SECTEUR , Format('%d', [ST.IDSecteur]));             // IDSecteur : integer;
            SubSubSubSection.SetAttribute(GTX_ATTR_VISEE_TYPE , Format('%d', [Ord(ST.TypeVisee)]));           // TypeVisee : TTypeDeVisee;

            SubSubSubSection.SetAttribute(GTX_ATTR_VISEE_CODE , Format('%d', [ST.Code]));                     // Code      : integer;
            SubSubSubSection.SetAttribute(GTX_ATTR_VISEE_EXPE , Format('%d', [ST.Expe]));                     // Expe      : integer;

            SubSubSubSection.SetAttribute(GTX_ATTR_VISEE_LONG , Format('%.3f', [ST.Longueur]));               // Longueur  : double;
            SubSubSubSection.SetAttribute(GTX_ATTR_VISEE_AZ   , Format('%.2f', [ST.Azimut]));                 // Azimut    : double;
            SubSubSubSection.SetAttribute(GTX_ATTR_VISEE_P    , Format('%.2f', [ST.Pente]));                  // Pente     : double;

            SubSubSubSection.SetAttribute(GTX_ATTR_VISEE_LG   , Format('%.2f', [ST.LG]));                     // LG        : double;
            SubSubSubSection.SetAttribute(GTX_ATTR_VISEE_LD   , Format('%.2f', [ST.LD]));                     // LD        : double;
            SubSubSubSection.SetAttribute(GTX_ATTR_VISEE_HZ   , Format('%.2f', [ST.HZ]));                     // HZ        : double;
            SubSubSubSection.SetAttribute(GTX_ATTR_VISEE_HN   , Format('%.2f', [ST.HN]));                     // HN        : double;

            SubSubSubSection.SetAttribute(GTX_ATTR_VISEE_OBS  , PasStrToXMLStr(ST.Commentaires));                             // Commentaires    : string;


            SubSubSection.AppendChild(SubSubSubSection);
        end;
      end;
    // section Antennes (à la fin du document XML)
    QNbAntennes := GetNbAntennes();
    if (QNbAntennes > 0) then
    begin
      SectionAntennes := MyDocXML.CreateElement(GTX_KEY_SECTION_ANTENNAS);
      KWYXZ(GTX_KEY_SECTION_ANTENNAS);
      GHTopoRoot.AppendChild(SectionAntennes);
      for i := IDX_MINI to QNbAntennes - 1 do
      begin
        ANT := GetViseeAntenne(i);
        SubSection := MyDocXML.CreateElement(GTX_KEY_ANTENNA_SHOT);
          SubSection.SetAttribute(GTX_KEY_ANTENNA_NUMERO  , Format('%d', [ANT.IDViseeAntenne]));              // IDViseeAntenne      : integer;
          SubSection.SetAttribute(GTX_KEY_ANTENNA_LABEL   , Format('%s', [ANT.IDTerrainStation]));            // IDTerrainStation    : string;
          SubSection.SetAttribute(GTX_KEY_ANTENNA_NETWORK , Format('%d', [ANT.Reseau]));                      // Reseau              : integer;
          SubSection.SetAttribute(GTX_KEY_ANTENNA_SECTEUR , Format('%d', [ANT.Secteur]));                     // Secteur             : integer;
          SubSection.SetAttribute(GTX_KEY_ANTENNA_SERIE   , Format('%d', [ANT.SerieDepart]));                 // SerieDepart         : integer;
          SubSection.SetAttribute(GTX_KEY_ANTENNA_POINT   , Format('%d', [ANT.PtDepart]));                    // PtDepart            : integer;
          SubSection.SetAttribute(GTX_KEY_ANTENNA_CODE    , Format('%d', [ANT.Code]));                        // Code                : integer;
          SubSection.SetAttribute(GTX_KEY_ANTENNA_TRIP    , Format('%d', [ANT.Expe]));                        // Expe                : integer;
          SubSection.SetAttribute(GTX_KEY_ANTENNA_LONG    , Format('%.3f', [ANT.Longueur]));                  // Longueur            : double;      //         'Longueur
          SubSection.SetAttribute(GTX_KEY_ANTENNA_AZIMUT  , Format('%.2f', [ANT.Azimut]));                    // Azimut              : double;      //         'Azimut
          SubSection.SetAttribute(GTX_KEY_ANTENNA_PENTE   , Format('%.2f', [ANT.Pente]));                     // Pente               : double;      //         'Pente
          SubSection.SetAttribute(GTX_KEY_ANTENNA_OBS     , PasStrToXMLStr(ANT.Commentaires));                // Commentaires        : string;
        SectionAntennes.AppendChild(SubSection);
      end;
    end;
    // sérialisation du fichier
    WriteXMLFile(MyDocXML, FichierXML);
  finally
    MyDocXML.Free;
  end;
end;



// ouverture d'un fichier XML (gtx)
function  TToporobotStructure2012.LoadFromXML(const FichierXML: TStringDirectoryFilename): integer;
var
  EPSGDefault: TLabelSystemesCoordsEPSG;
  MyDocXML: TXMLDocument;
  GHTopoRoot: TDOMElement;
  // items de tables simples
  UneNoteVersion        : TReleaseNotes;
  UnFiltrePerso         : TFiltrePersonnalise;
  UneEntree             : TEntrance;
  UnReseau              : TReseau;
  UnSecteur             : TSecteur;
  UneExpe               : TExpe;
  UnCode                : TCode;
  UneViseeAntenne       : TViseeAntenne;
  UneSerie: TObjSerie;
  UneVisee: TUneVisee;

  NbSect: LongWord;
  ListeDesSections: TDOMNodeList;
  MySection: TDOMNode;
  NoSec: Integer;
  SectionReseau: TDOMNode;
  EWE, WU, WOK, QIN, QAT: TDOMNode;
  //--------------
  ListeRubriquesGeneral: TDOMNodeList;
  ListeDesNotesVersion : TDOMNodeList;
  ListeDesFiltres      : TDOMNodeList;
  ListeDesEntrees      : TDOMNodeList;
  ListeDesReseaux      : TDOMNodeList;
  ListeDesSecteurs     : TDOMNodeList;
  ListeDesExpes        : TDOMNodeList;
  ListeDesCodes        : TDOMNodeList;
  ListeDesSeries       : TDOMNodeList;
  ListeDesAntennes     : TDOMNodeList;
  ListeStations: TDOMNodeList;



  QMyDate: TDateTime;
  St, i: Integer;
  qNomSerie, qObsSerie: WideString;
  qSerDepart, qPtDepart, qSerArrivee, qPtArrivee: Integer;
  qChance, qObstacle, qReseau: Integer;
  qRaideur: Extended;


  qR, qG, qB : byte;
  qEntrance: TNumeroEntrance;
  qIdxSerie: TNumeroSerie;
  procedure KWYXZ(const S666: string);
  begin
    AfficherMessage(Format('-- Found "%s" section', [S666]));
  end;
  function QExtractTextContent(const QQQ: TDOMNode): string;
  begin
    Result := Trim(PasStrToXMLStr(QQQ.TextContent));
  end;

begin
  Result := -1;
  AfficherMemoryUsage();
  AfficherMessage(Format('%s.LoadFromXML(%s)', [ClassName, FichierXML]));
   // nom étude par défaut
  SetNomEtude('Etude001');
  SetNameSpace('Namespace00');
  SetCommentairesEtude('Issued from XML format');
  EPSGDefault.CodeEPSG := DEFAULT_SYSTEME_COORDONNEES_CODE_EPSG;
  EPSGDefault.NomEPSG  := DEFAULT_SYSTEME_COORDONNEES_NOM;

  SetCodeEPSGSystemeCoordonnees(EPSGDefault);
  // réinit du document
  self.ReInitialiser(True);
  MyDocXML := TXMLDocument.Create;
  try
    ReadXMLFile(MyDocXML, FichierXML);
    GHTopoRoot := MyDocXML.DocumentElement;
    // section General
    AfficherMessage('-> Reading section: ' + GTX_KEY_SECTION_GENERAL);
    EWE := GHTopoRoot.FindNode(GTX_KEY_SECTION_GENERAL);
    if (EWE <> Nil) then
    begin
      ListeRubriquesGeneral := EWE.ChildNodes;
      WU := ListeRubriquesGeneral.Item[0];
      QIN := WU.Attributes.GetNamedItem(GTX_ATTR_NAMESPACE);
      SetNameSpace(QExtractTextContent(QIN));
      QIN := WU.Attributes.GetNamedItem(GTX_ATTR_NOM_ETUDE);
      SetNomEtude(QExtractTextContent(QIN));
      QIN := WU.Attributes.GetNamedItem(GTX_ATTR_COORDS_SYSTEM_EPSG);
      EPSGDefault.CodeEPSG := StrToIntDef(QIN.TextContent, DEFAULT_SYSTEME_COORDONNEES_CODE_EPSG);
      QIN := WU.Attributes.GetNamedItem(GTX_ATTR_COORDS_SYSTEM_NOM);
      EPSGDefault.NomEPSG := QExtractTextContent(QIN);
      SetCodeEPSGSystemeCoordonnees(EPSGDefault);
      QIN := WU.Attributes.GetNamedItem(GTX_ATTR_COMMENTAIRES_ETUDE);
      SetCommentairesEtude(QExtractTextContent(QIN));
    end;
    {$IFDEF SUPPORT_NOTES_VERSION}
    // section Notes de version
    AfficherMessage('-> Reading section: ' + GTX_KEY_SECTION_RELEASE_NOTES);
    EWE := GHTopoRoot.FindNode(GTX_KEY_SECTION_RELEASE_NOTES);
    if (EWE <> Nil) then
    begin
      ListeDesNotesVersion := EWE.ChildNodes;
      for i := 0 to ListeDesNotesVersion.Count - 1 do
      begin
        try
          WU := ListeDesNotesVersion.Item[i];
          // un item = un attribut
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_RELEASE_NOTE_DATE);
          UneNoteVersion.Date        := DateSQLToDatePascal(QExtractTextContent(QIN));
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_RELEASE_NOTE_AUTHOR);
          UneNoteVersion.Author      := QExtractTextContent(QIN);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_RELEASE_NOTE_DESCRO);
          UneNoteVersion.Description := QExtractTextContent(QIN);

          self.AddNoteVersion(UneNoteVersion);
        except
        end;
      end;
    end;
    {$endif SUPPORT_NOTES_VERSION}
    // filtres personnalisés
    AfficherMessage('======================');
    AfficherMessage('-> Reading section: ' + GTX_KEY_SECTION_FILTERS);
    EWE := GHTopoRoot.FindNode(GTX_KEY_SECTION_FILTERS);
    if (EWE <> Nil) then
    begin
      ListeDesFiltres := EWE.ChildNodes;
      for i := 0 to ListeDesFiltres.Count - 1 do
      begin
        try
          WU := ListeDesFiltres.Item[i];
          // un item = un attribut
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_FILTER_COLOR_R);
          qR := StrToIntDef(QIN.TextContent, 0);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_FILTER_COLOR_G);
          qG := StrToIntDef(QIN.TextContent, 0);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_FILTER_COLOR_B);
          qB := StrToIntDef(QIN.TextContent, 0);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_FILTER_NAME);
          UnFiltrePerso.NomFiltre := QExtractTextContent(QIN);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_FILTER_EXPRESSION);
          UnFiltrePerso.Expression := QExtractTextContent(QIN);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_FILTER_DESCRIPTION);
          UnFiltrePerso.Description := QExtractTextContent(QIN);
          self.AddFiltrePerso(UnFiltrePerso);
        except
        end;
      end;
    end;

    // lister les entrées
    AfficherMessage('======================');
    AfficherMessage('-> Reading section: ' + GTX_KEY_SECTION_ENTRANCES);
    EWE := GHTopoRoot.FindNode(GTX_KEY_SECTION_ENTRANCES);
    if (EWE <> Nil) then
    begin
      ListeDesEntrees := EWE.ChildNodes;
      for i := 0 to ListeDesEntrees.Count - 1 do
      begin
        try
          WU := ListeDesEntrees.Item[i];
          // un item = un attribut
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_ENTRANCE_IDX);
          //UneEntree.eNumEntree := StrToIntDef(QIN.TextContent, -1);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_ENTRANCE_IDTERRAIN);
          UneEntree.eIDTerrain := QExtractTextContent(QIN);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_ENTRANCE_NAME);
          UneEntree.eNomEntree := QExtractTextContent(QIN);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_ENTRANCE_REFSERIE);
          UneEntree.eRefSer    := StrToIntDef(QIN.TextContent, -1);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_ENTRANCE_REFPT);
          UneEntree.eRefSt     := StrToIntDef(QIN.TextContent, -1);

          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_ENTRANCE_X);
          UneEntree.eXEntree   := ConvertirEnNombreReel(QIN.TextContent, -1.00);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_ENTRANCE_Y);
          UneEntree.eYEntree   := ConvertirEnNombreReel(QIN.TextContent, -1.00);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_ENTRANCE_Z);
          UneEntree.eZEntree   := ConvertirEnNombreReel(QIN.TextContent, -1.00);
          self.AddEntree(UneEntree);
        except
        end;
      end;
    end;
    // lister les réseaux
    AfficherMessage('======================');
    AfficherMessage('-> Read section: ' + GTX_KEY_SECTION_RESEAUX);
    EWE := GHTopoRoot.FindNode(GTX_KEY_SECTION_RESEAUX);
    if (EWE <> Nil) then
    begin
      ListeDesReseaux := EWE.ChildNodes;
      for i := 0 to ListeDesReseaux.Count - 1 do
      begin
        try
          WU := ListeDesReseaux.Item[i];
          // <Network Name="Réseau principal" Type="0" Color="#000080FF" Numero="0" Comments=""/>
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_RESEAU_IDX);
          //UnReseau.IdxReseau := StrToIntDef(QIN.TextContent, 0);

          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_RESEAU_COLOR_R);
          qR := StrToIntDef(QIN.TextContent, 0);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_RESEAU_COLOR_G);
          qG := StrToIntDef(QIN.TextContent, 0);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_RESEAU_COLOR_B);
          qB := StrToIntDef(QIN.TextContent, 0);

          UnReseau.ColorReseau := RGBToColor(qR, qG, qB);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_RESEAU_TYPE);
          UnReseau.TypeReseau  := StrToIntDef(QIN.TextContent, 0);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_RESEAU_NAME);
          UnReseau.NomReseau   := QExtractTextContent(QIN);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_RESEAU_OBS);
          UnReseau.ObsReseau   := QExtractTextContent(QIN);
          self.AddReseau(UnReseau);
        except
        end;
      end;
    end;
    // lister les secteurs
    AfficherMessage('======================');
    AfficherMessage('-> Read section: ' + GTX_KEY_SECTION_SECTEURS);
    EWE := GHTopoRoot.FindNode(GTX_KEY_SECTION_SECTEURS);
    if (EWE <> Nil) then
    begin
      ListeDesSecteurs := EWE.ChildNodes;
      for i := 0 to ListeDesSecteurs.Count - 1 do
      begin
        try
          WU := ListeDesSecteurs.Item[i];
          // <Network Name="Réseau principal" Type="0" Color="#000080FF" Numero="0" Comments=""/>
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_SECTEUR_IDX);
          //UnSecteur.IdxSecteur := StrToIntDef(QIN.TextContent, 0);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_SECTEUR_COLOR_R);
          qR := StrToIntDef(QIN.TextContent, 0);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_SECTEUR_COLOR_G);
          qG := StrToIntDef(QIN.TextContent, 0);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_SECTEUR_COLOR_B);
          qB := StrToIntDef(QIN.TextContent, 0);
          UnSecteur.CouleurSecteur := RGBToColor(qR, qG, qB);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_SECTEUR_NAME);
          UnSecteur.NomSecteur   := QExtractTextContent(QIN);
          self.AddSecteur(UnSecteur);
        except
        end;
      end;
    end;

    // lister les codes
    AfficherMessage('======================');
    AfficherMessage('-> Read section: ' + GTX_KEY_SECTION_CODES);
    EWE := GHTopoRoot.FindNode(GTX_KEY_SECTION_CODES);
    if (EWE <> Nil) then
    begin
      ListeDesCodes := EWE.ChildNodes;
      for i := 0 to ListeDesCodes.Count - 1 do
      begin
        try
          WU := ListeDesCodes.Item[i];
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_NUMERO);
          UnCode.IDCode:= StrToIntDef(QIN.TextContent, 0);

          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_TYPE);
          //UnCode.ReservedInt:= StrToIntDef(QIN.TextContent, 0);
          // TODO: Support de FactLong a valider
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_FACT_LONG);
          UnCode.FactLong   := 1.00; //ConvertirEnNombreReel(QIN.TextContent, 1.0000);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_UCOMPASS);
          UnCode.GradAz:= ConvertirEnNombreReel(QIN.TextContent, 360.00);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_UCLINO);
          UnCode.GradInc:= ConvertirEnNombreReel(QIN.TextContent, 360.00);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_ANGLIMITE);
          UnCode.AngLimite:= ConvertirEnNombreReel(QIN.TextContent, 0.00);

          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_PSI_L);
          UnCode.PsiL:= ConvertirEnNombreReel(QIN.TextContent, 0.01);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_PSI_A);
          UnCode.PsiAz:= ConvertirEnNombreReel(QIN.TextContent, 0.01);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_PSI_P);
          UnCode.PsiP:= ConvertirEnNombreReel(QIN.TextContent, 0.01);

          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_OBS);
          UnCode.Commentaire:= QExtractTextContent(QIN);
          // fonctions de correction
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_FUNC_CORR_AZ_CO);
          UnCode.ParamsFuncCorrAz.Co := ConvertirEnNombreReel(QIN.TextContent, 0.00);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_FUNC_CORR_AZ_ERR_MAX);
          UnCode.ParamsFuncCorrAz.ErreurMax := ConvertirEnNombreReel(QIN.TextContent, 0.00);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_FUNC_CORR_AZ_POS_ERR_MAX);
          UnCode.ParamsFuncCorrAz.PosErrMax := ConvertirEnNombreReel(QIN.TextContent, 0.00);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_FUNC_CORR_INC_CO);
          UnCode.ParamsFuncCorrInc.Co := ConvertirEnNombreReel(QIN.TextContent, 0.00);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_FUNC_CORR_INC_ERR_MAX);
          UnCode.ParamsFuncCorrInc.ErreurMax := ConvertirEnNombreReel(QIN.TextContent, 0.00);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_CODE_FUNC_CORR_INC_POS_ERR_MAX);
          UnCode.ParamsFuncCorrInc.PosErrMax := ConvertirEnNombreReel(QIN.TextContent, 0.00);
          self.AddCode(UnCode);
        except
        end;
      end;
    end;
    //AfficherMessage(datetostr(DateSQLToDatePascal('2012-11-25')));

    // lister les expés
    AfficherMessage('======================');
    EWE := GHTopoRoot.FindNode(GTX_KEY_SECTION_EXPES);
    AfficherMessage('-> Read section: ' + GTX_KEY_SECTION_EXPES);
    if (EWE <> Nil) then
    begin
      ListeDesExpes := EWE.ChildNodes;
      for i := 0 to ListeDesExpes.Count - 1 do
      begin
        try
          WU := ListeDesExpes.Item[i];
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_EXPE_NUMERO);
          UneExpe.IDExpe:= StrToIntDef(QIN.TextContent, 0);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_EXPE_DATE);
          QMyDate := DateSQLToDatePascal(QExtractTextContent(QIN));
          UneExpe.AnneeExpe := YearOf(QMyDate);
          UneExpe.MoisExpe  := MonthOf(QMyDate);
          UneExpe.JourExpe  := DayOf(QMyDate);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_EXPE_IDXCOLOR);
          UneExpe.Couleur:= StrToIntDef(QIN.TextContent, 0);

          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_EXPE_SURVEY1);
          UneExpe.Operateur:= QExtractTextContent(QIN);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_EXPE_SURVEY2);
          UneExpe.ClubSpeleo:= QExtractTextContent(QIN);

          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_EXPE_MODEDECL);
          UneExpe.ModeDecl := StrToIntDef(QIN.TextContent, 0);

          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_EXPE_DECLINAT);
          UneExpe.Declinaison:= ConvertirEnNombreReel(QIN.TextContent, 0.00);
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_EXPE_INCLINAT);
          //UneExpe.Inclinaison:= ConvertirEnNombreReel(QIN.TextContent, 0.00);

          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_EXPE_OBS);
          UneExpe.Commentaire:= QExtractTextContent(QIN);
          self.AddExpe(UneExpe);
        except
        end;
      end;
    end;
    // lister les séries
    AfficherMessage('======================');
    AfficherMessage('-> Read section: ' + GTX_KEY_SECTION_SERIES);
    EWE := GHTopoRoot.FindNode(GTX_KEY_SECTION_SERIES);
    if (EWE <> Nil) then
    begin
      ListeDesSeries := EWE.ChildNodes;
      AfficherMessage(Format('%d séries', [ListeDesSeries.Count]));
      for i := 0 to ListeDesSeries.Count - 1 do
      begin
        UneSerie := TObjSerie.Create;
        try
          UneSerie.ClearStations;
          WU  := ListeDesSeries.Item[i];
          // un item = un attribut
          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_SERIE_Numero);
          qIdxSerie    := TNumeroSerie(StrToIntDef(QIN.TextContent, 0));
          UneSerie.SetNumeroSerie(qIdxSerie);

          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_SERIE_NAME);
          qNomSerie    := QExtractTextContent(QIN);
          UneSerie.SetNomSerie(qNomSerie);

          QIN := WU.Attributes.GetNamedItem(GTX_ATTR_SERIE_OBS);
          qObsSerie    := QExtractTextContent(QIN);
          UneSerie.SetObsSerie(qObsSerie);

          QIN         := WU.Attributes.GetNamedItem(GTX_ATTR_SERIE_SERDEP);
          qSerDepart  := StrToIntDef(QIN.TextContent, 0);
          QIN         := WU.Attributes.GetNamedItem(GTX_ATTR_SERIE_PTDEP);
          qPtDepart   := StrToIntDef(QIN.TextContent, 0);
          QIN         := WU.Attributes.GetNamedItem(GTX_ATTR_SERIE_SERARR);
          qSerArrivee := StrToIntDef(QIN.TextContent, 0);
          QIN         := WU.Attributes.GetNamedItem(GTX_ATTR_SERIE_PTARR);
          qPtArrivee  := StrToIntDef(QIN.TextContent, 0);
          UneSerie.SetSeriePtExtremites(qSerDepart, qPtDepart,
                                        qSerArrivee, qPtArrivee);
          QIN         := WU.Attributes.GetNamedItem(GTX_ATTR_SERIE_CHANCE);
          qChance     := StrToIntDef(QIN.TextContent, 0);
          QIN         := WU.Attributes.GetNamedItem(GTX_ATTR_SERIE_OBSTACLE);
          qObstacle   := StrToIntDef(QIN.TextContent, 0);
          UneSerie.SetChanceObstacle(qChance, qObstacle);

          // NEW : Entrées
          QIN         := WU.Attributes.GetNamedItem(GTX_ATTR_SERIE_ENTRANCE);
          qEntrance   := TNumeroEntrance(StrToIntDef(QIN.TextContent, 0));
          UneSerie.SetNumeroEntrance(qEntrance);

          // réseaux
          QIN         := WU.Attributes.GetNamedItem(GTX_ATTR_SERIE_RESEAU);
          qReseau     := StrToIntDef(QIN.TextContent, 0);
          UneSerie.SetNumeroReseau(qReseau);
          // DONE: Raideur implémentée
          QIN         := WU.Attributes.GetNamedItem(GTX_ATTR_SERIE_RAIDEUR);
          qRaideur    := ConvertirEnNombreReel(QIN.TextContent, 1.00);
          UneSerie.SetRaideur(qRaideur);
          UneSerie.SetCouleur(clBlue);
          // les stations
          QAT := WU.FindNode(GTX_KEY_STATIONS);
          if (QAT <> NIL) then
          begin
            ListeStations := QAT.ChildNodes;
            //AfficherMessage(Format('Série %d %s - %d stations', [UneSerie.GetIndexSerie, UneSerie.GetNomSerie, ListeStations.Count]));
            for St := 0 to ListeStations.Count - 1 do
            begin
              try
                WOK := ListeStations.Item[St];

                QIN := WOK.Attributes.GetNamedItem(GTX_ATTR_VISEE_ID);
                UneVisee.NoVisee := St; // Attribut XML 'ID' non utilisé //StrToIntDef(QIN.TextContent, St); // -1 = numérotation auto
                QIN := WOK.Attributes.GetNamedItem(GTX_ATTR_VISEE_LBL);
                UneVisee.IDTerrainStation := QExtractTextContent(QIN);
                QIN := WOK.Attributes.GetNamedItem(GTX_ATTR_VISEE_SECTEUR);
                UneVisee.IDSecteur := StrToIntDef(QIN.TextContent, 0);
                QIN := WOK.Attributes.GetNamedItem(GTX_ATTR_VISEE_TYPE);
                UneVisee.TypeVisee := GetTypeDeVisee(StrToIntDef(QIN.TextContent, 0));
                QIN := WOK.Attributes.GetNamedItem(GTX_ATTR_VISEE_CODE);
                UneVisee.Code := StrToIntDef(QIN.TextContent, 0);
                QIN := WOK.Attributes.GetNamedItem(GTX_ATTR_VISEE_EXPE);
                UneVisee.Expe := StrToIntDef(QIN.TextContent, 0);
                QIN := WOK.Attributes.GetNamedItem(GTX_ATTR_VISEE_LONG);
                UneVisee.Longueur := ConvertirEnNombreReel(QIN.TextContent, 0.001);
                QIN := WOK.Attributes.GetNamedItem(GTX_ATTR_VISEE_AZ);
                UneVisee.Azimut := ConvertirEnNombreReel(QIN.TextContent, 0.00);
                QIN := WOK.Attributes.GetNamedItem(GTX_ATTR_VISEE_P);
                UneVisee.Pente := ConvertirEnNombreReel(QIN.TextContent, 0.00);
                QIN := WOK.Attributes.GetNamedItem(GTX_ATTR_VISEE_LG);
                UneVisee.LG := ConvertirEnNombreReel(QIN.TextContent, 0.00);
                QIN := WOK.Attributes.GetNamedItem(GTX_ATTR_VISEE_LD);
                UneVisee.LD := ConvertirEnNombreReel(QIN.TextContent, 0.00);
                QIN := WOK.Attributes.GetNamedItem(GTX_ATTR_VISEE_HZ);
                UneVisee.HZ := ConvertirEnNombreReel(QIN.TextContent, 0.00);
                QIN := WOK.Attributes.GetNamedItem(GTX_ATTR_VISEE_HN);
                UneVisee.HN := ConvertirEnNombreReel(QIN.TextContent, 0.00);
                QIN := WOK.Attributes.GetNamedItem(GTX_ATTR_VISEE_OBS);
                UneVisee.Commentaires := QExtractTextContent(QIN);
                // ajout visée
                UneSerie.AddVisee(UneVisee);
              except
              end;
            end; // for stations
            self.AddSerie(UneSerie);
          end;
        except // try séries
        end;
      end;  // for séries
    end;
    // lister les antennes
    AfficherMessage('======================');
    EWE := GHTopoRoot.FindNode(GTX_KEY_SECTION_ANTENNAS);
    if (EWE <> Nil) then
    begin
      ListeDesAntennes := EWE.ChildNodes;
      for i := 0 to ListeDesAntennes.Count - 1 do
      begin
        try
          WU := ListeDesAntennes.Item[i];
          //AfficherMessage(Format('Antenne: %s - Valeur: %s', [WU.NodeName, WU.TextContent]));
          QIN := WU.Attributes.GetNamedItem(GTX_KEY_ANTENNA_NUMERO);
          UneViseeAntenne.IDViseeAntenne := StrToIntDef(QIN.TextContent, -1); // -1 = numérotation auto  //  IDViseeAntenne      : integer;
          QIN := WU.Attributes.GetNamedItem(GTX_KEY_ANTENNA_LABEL);
          UneViseeAntenne.IDTerrainStation := QExtractTextContent(QIN);          //  IDTerrainStation    : string;
          QIN := WU.Attributes.GetNamedItem(GTX_KEY_ANTENNA_NETWORK);
          UneViseeAntenne.Reseau  := StrToIntDef(QIN.TextContent, 0);         //  Reseau              : integer;
          QIN := WU.Attributes.GetNamedItem(GTX_KEY_ANTENNA_SECTEUR);
          UneViseeAntenne.Secteur := StrToIntDef(QIN.TextContent, 0);         //  Secteur             : integer;

          QIN := WU.Attributes.GetNamedItem(GTX_KEY_ANTENNA_SERIE);
          UneViseeAntenne.SerieDepart := StrToIntDef(QIN.TextContent, 0);     //  SerieDepart         : integer;
          QIN := WU.Attributes.GetNamedItem(GTX_KEY_ANTENNA_POINT);
          UneViseeAntenne.PtDepart    := StrToIntDef(QIN.TextContent, 0);     //  PtDepart            : integer;

          QIN := WU.Attributes.GetNamedItem(GTX_KEY_ANTENNA_CODE);
          UneViseeAntenne.Code := StrToIntDef(QIN.TextContent, 0);            //  Code                : integer;
          QIN := WU.Attributes.GetNamedItem(GTX_KEY_ANTENNA_TRIP);
          UneViseeAntenne.Expe := StrToIntDef(QIN.TextContent, 0);            //  Expe                : integer;

          QIN := WU.Attributes.GetNamedItem(GTX_KEY_ANTENNA_LONG);
          UneViseeAntenne.Longueur := ConvertirEnNombreReel(QIN.TextContent, 0.001); // Longueur            : double;      //         'Longueur
          QIN := WU.Attributes.GetNamedItem(GTX_KEY_ANTENNA_AZIMUT);
          UneViseeAntenne.Azimut   := ConvertirEnNombreReel(QIN.TextContent, 0.00);  // Azimut              : double;      //         'Azimut
          QIN := WU.Attributes.GetNamedItem(GTX_KEY_ANTENNA_PENTE);
          UneViseeAntenne.Pente    := ConvertirEnNombreReel(QIN.TextContent, 0.00);  //  Pente               : double;      //         'Pente

          QIN := WU.Attributes.GetNamedItem(GTX_KEY_ANTENNA_OBS);
          UneViseeAntenne.Commentaires := QExtractTextContent(QIN);                     // Commentaires        : string;
          if (i mod 200 = 0) then AfficherMessage(Format('Antenne %d of %d', [i, ListeDesAntennes.Count]));
          (*
          AfficherMessage(Format('-- Antenne: %d [%s] - %d - [%d, %d] - L = %.2f, Az = %.2f, P = %.2f - %s',
                                  [UneViseeAntenne.IDViseeAntenne, UneViseeAntenne.IDTerrainStation,
                                   UneViseeAntenne.Reseau, UneViseeAntenne.Code, UneViseeAntenne.Expe,
                                   UneViseeAntenne.Longueur, UneViseeAntenne.Azimut, UneViseeAntenne.Pente,
                                   UneViseeAntenne.Commentaires
                                  ]));
          //*)
          self.AddViseeAntenne(UneViseeAntenne);
        except
        end;
      end;
    end;
    // Cause courante de plantages: Absence de la série 1
    VerifieSerie1();
    // Actions de réparation / vérif
    CheckerLesDonneesTopo();
    // paramètres par défaut
    UneEntree := GetEntree(0);
    self.SetDatabaseName(ExtractFileNameWithoutExt(FichierXML));
    //self.SetCommentairesEtude('');
    self.SetRefSeriePoint(UneEntree.eRefSer, UneEntree.eRefSt);
    self.SetDefaultCoords(UneEntree.eXEntree, UneEntree.eYEntree, UneEntree.eZEntree);
    //--------------------------------------------------------------------------

    // recenser IDs terrain
    RecenserLesIDsTerrain();
    Result := self.GetNbSeries();
    AfficherMemoryUsage();
    RecenserPointsOfInterest();
  finally
    MyDocXML.Free;
  end;
end;
// styles d'annotations
function TToporobotStructure2012.MakeTStyleAnnotation(const QStyleName: string; const QFontName: string; const QFontColor: TColor; const QBackColor: TColor; const QFontStyle: TFontStyles; const QHauteurTexte: double; const QPosition: byte; const QAngleRot: integer): TTexteAttributs;
begin
  Result.StyleName    := QStyleName;
  Result.FontName     := QFontName;
  Result.FontColor    := QFontColor;
  Result.BackColor    := QBackColor;
  Result.FontStyle    := QFontStyle;
  Result.HauteurTexte := QHauteurTexte;
  Result.Position     := QPosition;
  Result.AngleRot     := QAngleRot;
end;
procedure TToporobotStructure2012.InitTableStylesAnnotations();
var
  i: Integer;
begin
  for i := 0 to high(FTableStylesAnnotations) do
    FTableStylesAnnotations[i] := MakeTStyleAnnotation(Format('Style%d', [i]), DEFAULT_FONT_NAME, clBlack   , clWhite  , [], 0.50, 0, 0);
  FTableStylesAnnotations[0] := MakeTStyleAnnotation('Note'     , DEFAULT_FONT_NAME, clBlack   , clYellow  , [], 0.20, 0, 0);
  FTableStylesAnnotations[1] := MakeTStyleAnnotation('Cote'     , DEFAULT_FONT_NAME, clBlue    , clWhite   , [fsBold], 0.250, 0, 0);
  FTableStylesAnnotations[2] := MakeTStyleAnnotation('Titre'    , DEFAULT_FONT_NAME, clBlue, clWhite, [fsBold, fsUnderline], 0.5, 2, 0);
end;
function TToporobotStructure2012.GetStyleAnnotation(const Idx: integer): TTexteAttributs;
begin
  if (IsInRange(Idx, 0, High(FTableStylesAnnotations))) then Result := FTableStylesAnnotations[Idx]
                                                        else Result := FTableStylesAnnotations[0];

end;


procedure TToporobotStructure2012.PutStyleAnnotation(const Idx: integer; const QT: TTexteAttributs);
begin
  FTableStylesAnnotations[Idx] := QT;
end;


function TToporobotStructure2012.GetNbStylesAnnotations(): integer;
begin
  Result := length(FTableStylesAnnotations);
end;



// vérifier si la visée en antenne possède une station de rattachement
// utilise l'IDTerrain, la série et le point
// Modifie la visée passée en paramètre
function TToporobotStructure2012.HasStationForAntennaShot(var VA: TViseeAntenne): boolean;
var
  SR: TObjSerie;
  VS: TUneVisee;
  i, j: Integer;
  WU1, WU2: String;
begin
  Result := False;
  for i := 1 to GetNbSeries - 1 do
  begin
    SR := GetSerie(i);
    for j := 0 to SR.GetNbVisees - 1 do
    begin
      VS := SR.GetVisee(j);
      // test sur l'ID de série et station
      if ((VA.SerieDepart = SR.GetNumeroDeSerie()) and (VA.PtDepart = j)) then
      begin
        //VA.SerieDepart := SR.GetIndexSerie;
        //VA.PtDepart    := j;

        VA.IDTerrainStation := VS.IDTerrainStation;
        VA.Code        := VS.Code;
        VA.Expe        := VS.Expe;
        VA.Secteur     := VS.IDSecteur;
        VA.Reseau      := SR.GetNumeroReseau();
        VA.IDTerrainStation := '';
        Result := True;
        Exit;
      end;
      (*
      else
      begin
        // test sur l'ID terrain
        WU1 := LowerCase(Trim(VS.IDTerrainStation));
        WU2 := LowerCase(Trim(VA.IDTerrainStation));
        //AfficherMessage(Format('%s - %s - %d.%d = %d.%d ?', [WU1, WU2, VA.SerieDepart, VA.PtDepart, i, j]));
        if ((WU1 = WU2) and (Trim(VA.IDTerrainStation) <> '')) then
        begin
          VA.SerieDepart := SR.GetIndexSerie;
          VA.PtDepart    := j;
          VA.Reseau      := SR.GetNoReseau;
          VA.Code        := VS.Code;
          VA.Expe        := VS.Expe;
          VA.Secteur     := VS.IDSecteur;
          VA.IDTerrainStation := '';
          Result := True;
          Exit;
        end;
    //*)
    end; // for j := 0 to SR.GetNbVisees - 1 do
  end; //  for i := 1 to GetNbSeries - 1 do
end;
// extraire les antennes de la station NoSerie.NoStation
// retourne le nombre d'antennes, ou 0 si aucunne antenne
function TToporobotStructure2012.ExtractViseesAntennesOfStation(const NoSerie: TNumeroSerie;
                                                                const NoStation: Integer;
                                                                out   ArrAntennesFound: TArrayOfTViseeAntenne): integer;
var
  i, Nb: Integer;
  MyAntenne: TViseeAntenne;
begin
  Result := 0;
  SetLength(ArrAntennesFound, 0);
  Nb := GetNbAntennes();
  if (Nb = 0) then Exit;
  for i := 0 to Nb - 1 do
  begin
    MyAntenne := GetViseeAntenne(i);
    if ((NoSerie = MyAntenne.SerieDepart) and (NoStation = MyAntenne.PtDepart)) then
    begin
      Result += 1;
      SetLength(ArrAntennesFound, Result);
      ArrAntennesFound[Result - 1] := MyAntenne;
    end;
  end;
end;




// pour le moment, on exporte au format Text de Toporobot
// Statut: OK, mais des détails à revoir.
// Entrées multiples OK
// Nota: Support du TLM330 non implémenté volontairement
//*)
procedure TToporobotStructure2012.ExporterVersToporobotTEXT(const QFileName: TStringDirectoryFilename;
                                                            const LongueurMaxAntenne: double;
                                                            const DoExportAntennaShots: boolean); deprecated;
var
  fp: TextFile;
  EWE: String;
  myEntree: TEntrance;
  i, Nb, j, NbV: Integer;
  myExpe: TExpe;
  myCode: TCode;
  mySerie: TObjSerie;
  myVisee: TUneVisee;
  NbA: Integer;
  k: Integer;
  myAntenne: TViseeAntenne;
  q: Integer;
  zdar: String;
  function makeColonnesEnTete(const IsCommentaire: boolean;
                              const Section, ID1, ID2, ID3, ID4: integer): string;
  begin
    Result := IIF(IsCommentaire, '(', ' ') +
              FormatterFixedColumn(Format('%d', [Section]), 5, taRightJustify) +
              FormatterFixedColumn(Format('%d', [ID1])    , 6, taRightJustify) +
              FormatterFixedColumn(Format('%d', [ID2])    , 4, taRightJustify) +
              FormatterFixedColumn(Format('%d', [ID3])    , 4, taRightJustify) +
              FormatterFixedColumn(Format('%d', [ID4])    , 4, taRightJustify);
  end;
begin
  AfficherMessage(Format('%s.ExporterVersToporobotTEXT: %s', [ClassName, QFileName]));
  AssignFile(fp, QFileName);
  try
    ReWrite(fp);
    // section -6 = entrée
    Nb := GetNbEntrees;
    for i := 0 to Nb - 1 do
    begin
      myEntree := self.GetEntree(i);
      EWE := makeColonnesEnTete(false, -6, i+1, 1, 1, 1) + ' ' + myEntree.eNomEntree;
      WriteLn(fp, EWE);
    end;
    for i := 0 to Nb - 1 do
    begin
      myEntree := self.GetEntree(i);
      // section -5 = Coordonnées
      //-5     1   1   1   1   629000.72   178000.68     1777.71     1     0
      EWE := makeColonnesEnTete(false, -5, i+1, 1, 1, 1) +
             FormatterFixedColumn(Format('%.2f', [myEntree.eXEntree]), 12, taRightJustify) +
             FormatterFixedColumn(Format('%.2f', [myEntree.eYEntree]), 12, taRightJustify) +
             FormatterFixedColumn(Format('%.2f', [myEntree.eZEntree]), 12, taRightJustify) +
             FormatterFixedColumn(Format('%d', [myEntree.eRefSer]), 6, taRightJustify) +
             FormatterFixedColumn(Format('%d', [myEntree.eRefSt]), 6, taRightJustify);
      WriteLn(fp, EWE);
    end;

    // section -54: Horodatage
    //-4     1   1   1   1 14/12/08 11:47:22  Guest
    for i := 1 to 2 do
    begin
      EWE := makeColonnesEnTete(false, -4, i, 1, 1, 1) +
             FormatterFixedColumn(FormatterDateHeureEnFrancais(Now), 18, taRightJustify) +
             '  ' + 'Guest';
      WriteLn(fp, EWE);
    end;
    // Section -3: Inutilisée
    EWE := makeColonnesEnTete(false, -3, 1, 1, 1, 1);
    WriteLn(fp, EWE);
    // Section -2: Expés
    // -2     1   1   1   1 25/08/07  MCH           AH            1    0.00   0   1
    Nb := self.GetNbExpes;
    for i := 1 to Nb - 1 do
    begin
      myExpe := GetExpe(i);
      EWE := makeColonnesEnTete(false, -2, myExpe.IDExpe, 1, 1, 1) +
             FormatterFixedColumn(Format('%.2d/%.2d/%.2d', [myExpe.JourExpe, myExpe.MoisExpe, myExpe.AnneeExpe mod 100]), 9, taRightJustify) +
             '  ' +
             FormatterFixedColumn(myExpe.Operateur, 13, taLeftJustify) + ' ' +
             FormatterFixedColumn(myExpe.ClubSpeleo, 13, taLeftJustify) + ' ' +
             FormatterFixedColumn(Format('%d', [0]), 2, taLeftJustify) +
             FormatterFixedColumn(Format('%.2f', [0.00]), 7, taRightJustify) +
             FormatterFixedColumn(Format('%d', [0]), 4, taRightJustify) +
             FormatterFixedColumn(Format('%d', [1]), 4, taRightJustify);
      WriteLn(fp, EWE);
    end;
    // Section -1: Codes
    //-1     1   1   1   1  360.00  360.00    0.05    1.00    1.00  100.00  100.00
    Nb := self.GetNbCodes;
    for i := 1 to Nb - 1 do
    begin
      myCode := GetCode(i);
      EWE := makeColonnesEnTete(false, -1, myCode.IDCode, 1, 1, 1) +
             FormatterFixedColumn(Format('%.2f', [myCode.GradAz]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.2f', [myCode.GradInc]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.2f', [0.05]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.2f', [1.00]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.2f', [1.00]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.2f', [100.00]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.2f', [100.00]), 8, taRightJustify);
      WriteLn(fp, EWE);
    end;
    // Sections > 0 = séries
    Nb := self.GetNbSeries;
    for i := 1 to Nb - 1 do
    begin
      mySerie := GetSerie(i);
      //1    -2   1   1   1 GALERIE D'ENTREE D10H10
      EWE := makeColonnesEnTete(false, mySerie.GetNumeroDeSerie(), -2, 1, 1, 1) +
             ' ' + mySerie.GetNomSerie;
      WriteLn(fp, EWE);
      //1    -1   1   1   1       1       0       1       6       6       3       0
      EWE := makeColonnesEnTete(false, mySerie.GetNumeroDeSerie(), -1, 1, 1, 1) +
             FormatterFixedColumn(Format('%.d', [mySerie.GetNoSerieDep]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.d', [mySerie.GetNoPointDep]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.d', [mySerie.GetNoSerieArr]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.d', [mySerie.GetNoPointArr]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.d', [mySerie.GetNbVisees - 1]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.d', [mySerie.GetChance]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.d', [mySerie.GetObstacle]), 8, taRightJustify);
      WriteLn(fp, EWE);
      nbV := mySerie.GetNbVisees;
      EWE := makeColonnesEnTete(false, mySerie.GetNumeroDeSerie(), 0, 1, 1, 1) +
             FormatterFixedColumn(Format('%.2f', [0.00]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.2f', [0.00]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.2f', [0.00]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.2f', [0.00]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.2f', [0.00]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.2f', [0.00]), 8, taRightJustify) +
             FormatterFixedColumn(Format('%.2f', [0.00]), 8, taRightJustify);
      WriteLn(fp, EWE);
      for j := 1 to nbV -1 do   // et non 'for j = 0 to nbV - 1'
      begin
        myVisee := mySerie.GetVisee(j);
        EWE := makeColonnesEnTete(false, mySerie.GetNumeroDeSerie(), j, 1, myVisee.Code, myVisee.Expe) +
               FormatterFixedColumn(Format('%.2f', [myVisee.Longueur]), 8, taRightJustify) +
               FormatterFixedColumn(Format('%.2f', [myVisee.Azimut]), 8, taRightJustify) +
               FormatterFixedColumn(Format('%.2f', [myVisee.Pente]), 8, taRightJustify) +
               FormatterFixedColumn(Format('%.2f', [myVisee.LG]), 8, taRightJustify) +
               FormatterFixedColumn(Format('%.2f', [myVisee.LD]), 8, taRightJustify) +
               FormatterFixedColumn(Format('%.2f', [myVisee.HZ]), 8, taRightJustify) +
               FormatterFixedColumn(Format('%.2f', [myVisee.HN]), 8, taRightJustify);
        WriteLn(fp, EWE);

        // export des commentaires
        q := 0;
        zdar := Trim(myVisee.Commentaires);
        if (zdar <> '') then
        begin
          EWE := makeColonnesEnTete(true, mySerie.GetNumeroDeSerie(), j, 0, myVisee.Code, myVisee.Expe) + ' ' + zdar;
          WriteLn(fp, EWE);
          q+=1;
        end;

        // exporter ici les visées en antenne
        if (DoExportAntennaShots) then
        begin
          NbA := GetNbAntennes;
          if (NbA > 0) then
          begin
            for k := 0 to NbA - 1 do
            begin
              myAntenne := GetViseeAntenne(k);
              if ((myAntenne.SerieDepart = mySerie.GetNumeroDeSerie()) and
                  (myAntenne.PtDepart    = j) and
                  (myAntenne.Longueur < LongueurMaxAntenne)) then
              begin
                // PocketTopo stocke les antennes dans les commentaires du point d'accrochage
                EWE := makeColonnesEnTete(true, myAntenne.SerieDepart, myAntenne.PtDepart, q, myAntenne.Code, myAntenne.Expe) +
                       Format(' [%.2f %.1f %.1f]', [myAntenne.Longueur, myAntenne.Azimut, myAntenne.Pente]);
                WriteLn(fp, EWE);
                q += 1;
              end;
            end;
          end;
        end;
      end;
    end;
  finally
    CloseFile(fp);
  end;
end;
// export brut des séries (sans entête) pour construction de tableaux croisés dynamiques sur tableur
procedure TToporobotStructure2012.ExportSeriesCSV(const QFileName: TStringDirectoryFilename);
const
  FMT666 = '%d' + #9 +
           '%s' + #9 +
           '%s' + #9 +
           '%s' + #9 +
           '%s' + #9 +
           '%d' + #9 +
           '%d' + #9 +
           '%s' + #9 +
           '%s' + #9 +
           '%s' + #9 +
           '%s' + #9 +
           '%s' + #9 +
           '%s' + #9 +
           '%s' + #9 +
           '%s' + #9 +
           '%s';

var
  fp: TextFile;
  EWE, QNomSerie: String;
  NbSeries, i, nv, v: Integer;
  MySerie: TObjSerie;
  MyReseau: TReseau;
  MyEntrance: TEntrance;
  MyVisee: TUneVisee;
  MySecteur: TSecteur;
begin

  AfficherMessage(Format('%s.ExportSeriesCSV: %s', [ClassName, QFileName]));
  AssignFile(fp, QFileName);
  try
    ReWrite(fp);
    NbSeries := GetNbSeries();
    // entête
    EWE := 'Serie' + #9 +
           'Entree' + #9 +
           'Reseau' + #9 +
           'Secteur'  + #9 +
           'Type'  + #9 +
           'Code' + #9 +
           'Expe' + #9 +
           'Longueur' + #9 +
           'Azimut' + #9 +
           'Pente' + #9 +
           'LG' + #9 +
           'LD' + #9 +
           'HZ' + #9 +
           'HN' + #9 +
           'IDTerrain' + #9 +
           'Commentaires';
    WriteLn(fp, EWE);
    for i := 0 to NbSeries - 1 do
    begin
      MySerie    := GetSerie(i);
      QNomSerie  := MySerie.GetNomSerie();
      MyReseau   := GetReseau(MySerie.GetNumeroReseau());
      MyEntrance := GetEntree(MySerie.GetNumeroEntrance());
      nv := MySerie.GetNbVisees();
      for v := 0 to nv - 1 do
      begin
        MyVisee := MySerie.GetVisee(v);
        MySecteur := GetSecteur(MyVisee.IDSecteur);
        EWE := Format(FMT666, [
                  MySerie.GetNumeroDeSerie(),         //       EWE := 'Serie' + #9 +
                  MyEntrance.eNomEntree,              //'Entree' + #9 +
                  MyReseau.NomReseau,                 //'Reseau' + #9 +
                  MySecteur.NomSecteur,               //'Secteur'  + #9 +
                  GetDescTypeVisee(MyVisee.TypeVisee), // 'Type'  + #9 +
                  MyVisee.Code, MyVisee.Expe,
                  FormatterNombreOOo(MyVisee.Longueur, 3, false),
                  FormatterNombreOOo(MyVisee.Azimut, 3, false),
                  FormatterNombreOOo(MyVisee.Pente, 3, false),
                  FormatterNombreOOo(MyVisee.LG, 3, false),
                  FormatterNombreOOo(MyVisee.LD, 3, false),
                  FormatterNombreOOo(MyVisee.HZ, 3, false),
                  FormatterNombreOOo(MyVisee.HN, 3, false),
                  MyVisee.IDTerrainStation,
                  MyVisee.Commentaires
                  ]);
        // remplacer les doubles quotes par une paire de simples quotes
        // (doubles quotes sont interprétées par Calc)
        EWE := StringReplace(EWE, '"', '''''', [rfReplaceAll]);
        WriteLn(fp, EWE);
      end;
    end;
  finally
    CloseFile(fp);
  end;

end;

// vérifie si le document GHTopo comporte des codes instruments étendus
function TToporobotStructure2012.IsGHTopoCodesInstrumentsExtended(): boolean;
var
  i, Nb: Integer;
  myCode: TCode;
  EWE: Boolean;
  miou: Int64;
begin
  AfficherMessage(Format('%s.IsGHTopoCodesInstrumentsExtended', [ClassName]));
  Result := false;
  Nb := GetNbCodes;
  for i := 1 to Nb - 1 do
  begin
    myCode := GetCode(i);
    miou   := Trunc(myCode.GradInc);
    EWE := (miou = 350) or
           (miou = 359) or
           (miou = 360) or
           (miou = 361) or
           (miou = 380) or
           (miou = 390) or
           (miou = 399) or
           (miou = 400) or
           (miou = 401);
    EWE := Not EWE;
    AfficherMessage(Format('%d: %d - %.2f - %s', [i, myCode.IDCode, myCode.GradInc, BoolToStr(EWE, 'VRAI', 'FAUX')]));

    Result := Result OR EWE;
  end;
  AfficherMessage(Format('Le document %s des codes instruments GHTopo', [BoolToStr(Result, 'CONTIENT', 'NE CONTIENT PAS')]));
end;
// supprimer les visées en antenne associées à une station,
function TToporobotStructure2012.RemoveViseesRayonnantesAssocieesAStation(const qSer, qSt: integer): integer;
var
  EWE: TViseeAntenne;
  n, Nb: Integer;

  B3: Boolean;
begin
  Result := 0;
  AfficherMessage(Format('%s.RemoveViseeRayonnanteAssocieeAStation: %d.%d', [self.ClassName, qSer, qSt]));
  n := GetNbAntennes;
  Nb := 0;
  while Nb < FTableViseesAntenne.Count - 1 do
  begin
    EWE := GetViseeAntenne(Nb);
    B3 := False;
    if (qSer > 0) then B3 := (EWE.SerieDepart = qSer) and (EWE.PtDepart = qSt);
    if (B3) then
    begin
      RemoveViseeAntenne(Nb);
      AfficherMessageErreur(Format('Antenne retiree: %d.%d', [EWE.SerieDepart, EWE.PtDepart]));
      Result += 1;
      Nb -= 1;
    end;
    Nb += 1;
  end;
  AfficherMessage(Format('%d antennes supprimées', [Result]));
end;
// renommer les points d'origine des visées rayonnantes associées à une station
function TToporobotStructure2012.DeplacerOrigineViseesRayonnantes(const OldSerie, OldPoint, NewSerie, NewPoint: integer): integer;
var
  i, n: Integer;
  EWE : TViseeAntenne;
  WU  : Boolean;
begin
  AfficherMessage(Format('%s.DeplacerOrigineViseesRayonnantes: %d.%d -> %d.%d', [self.ClassName, OldSerie, OldPoint, NewSerie, NewPoint]));
  n := GetNbAntennes;
  result := 0;
  for i := 0 to n-1 do
  begin
    EWE := GetViseeAntenne(i);
    WU  := (EWE.SerieDepart = OldSerie) and (EWE.PtDepart = OldPoint);
    if (WU) then
    begin
      EWE.SerieDepart := NewSerie;
      EWE.PtDepart    := NewPoint;
      PutViseeAntenne(i, EWE);
      result +=1;
    end;
  end;
end;


procedure TToporobotStructure2012.RecenserLesIDsTerrain();
var
  i, j, NbV: Integer;
  MySerie: TObjSerie;
  EWE: TUneVisee;
  WU: String;
  NbSts: Integer;
  MyLabelTerrain: TEtiquetteTerrain;
  LL: Extended;
begin
  AfficherMessage(Format('%s.RecenserLesIDsTerrain', [ClassName]));
  FTableIDsTerrain.ClearListe();
  // Cause fréquente d'erreurs: absence de la série 1
  NbSts := 0;
  AfficherMessage(Format('%d séries', [GetNbSeries()]));
  for i := 0 to GetNbSeries - 1 do
  begin
    MySerie := GetSerie(i);
    NbV := MySerie.GetNbVisees;
    // calcul des longueurs de séries
    LL := 0.00;
    for j := 0 to NbV - 1 do
    begin
      EWE := MySerie.GetVisee(j);
      LL += EWE.Longueur;
    end;

    AfficherMessage(Format('%d: %d: %s - %d visées (%.2f m)', [i, MySerie.GetNumeroDeSerie(), MySerie.GetNomSerie(), NbV, LL]));
    NbSts += NbV;
    if (NbV > 0) then
    begin
      for j := 0 to NbV - 1 do
      begin
        EWE := MySerie.GetVisee(j);
        WU  := Trim(UpperCase(EWE.IDTerrainStation));
        MyLabelTerrain.LabelTerrain := WU;
        MyLabelTerrain.NoPoint := j;
        MyLabelTerrain.NoSerie := MySerie.GetNumeroDeSerie();
        if (WU <> '') then AddLabelTerrain(MyLabelTerrain);
      end;
      //*)
    end;
  end;
end;

function TToporobotStructure2012.FindIdxIDTerrainByText(const Str: string): integer;
var
  i, p, Nb: Integer;
  S: String;
  LT: TEtiquetteTerrain;
begin
  Result := -1;
  S := UpperCase(Trim(Str));
  Nb := FTableIDsTerrain.Count;
  if (Nb = 0) then Exit;
  for i := 0 to Nb - 1 do
  begin
    LT := GetLabelTerrain(i);
    p := Pos(S, UpperCase(LT.LabelTerrain));
    if (p > 0) then
    begin
      Result := i;
      Exit;
    end;
  end;
end;
//------------------------------------------------------------------------------
function TToporobotStructure2012.ReAttribuerViseesRayonnantesDeUneSerie(const SerieCourante: TObjSerie; const OldNumSerie: integer): integer;
var
  i, Nb: Integer;
  EWE: TUneVisee;
begin
  Result := 0;
  Nb := SerieCourante.GetNbVisees;
  if (Nb = 0) then Exit;
  for i := 0 to Nb-1 do
  begin
    EWE := SerieCourante.GetVisee(i);
    Result += DeplacerOrigineViseesRayonnantes(OldNumSerie, i, SerieCourante.GetNumeroDeSerie(), i);
  end;
end;

// réattribuer les codes série.station d'une série modifiée
function TToporobotStructure2012.UpdateSerPtExtremsSeriesRattachees(const OldNumSerie, NewNumSerie: integer): integer;
var
  i: Integer;
  MySerie: TObjSerie;
  SerieAModifier: Boolean;
  MyEntree: TEntrance;
begin
  Result := 0;
  AfficherMessage(Format('%s.UpdateSerPtExtremsSeriesRattachees(%d -> %d)', [ClassName, OldNumSerie, NewNumSerie]));
  // entrées
  for i := 0 to GetNbEntrees - 1 do
  begin
    MyEntree := GetEntree(i);
    if (OldNumSerie = 1) then Continue;
    if (OldNumSerie = MyEntree.eRefSer) then
    begin
      AfficherMessage(Format('Entree: %d - %s: Noeud %d.%d renomme en %.d.%d',
                             [i, //MyEntree.eNumEntree,
                              MyEntree.eNomEntree,
                              MyEntree.eRefSer, MyEntree.eRefSt,
                              NewNumSerie, MyEntree.eRefSt
                             ]));
      MyEntree.eRefSer := NewNumSerie;
      PutEntree(i, MyEntree);
    end;
  end;
  // séries
  for i := 1 to GetNbSeries - 1 do
  begin
    SerieAModifier := false;
    MySerie := GetSerie(i);
    if (OldNumSerie = MySerie.GetNoSerieDep) then
    begin
      AfficherMessage(Format('Serie: %d - %s: Depart %d.%d renomme en %.d.%d',
                             [MySerie.GetNumeroDeSerie(),
                              MySerie.GetNomSerie,
                              MySerie.GetNoSerieDep, MySerie.GetNoPointDep,
                              NewNumSerie, MySerie.GetNoPointDep
                             ]));
      MySerie.SetNoSerieDep(NewNumSerie);
      SerieAModifier := True;
    end;
    if (OldNumSerie = MySerie.GetNoSerieArr) then
    begin
      AfficherMessage(Format('Serie: %d - %s: Arrivee %d.%d renomme en %.d.%d',
                             [MySerie.GetNumeroDeSerie(),
                              MySerie.GetNomSerie,
                              MySerie.GetNoSerieArr, MySerie.GetNoPointArr,
                              NewNumSerie, MySerie.GetNoPointArr
                             ]));
      MySerie.SetNoSerieArr(NewNumSerie);
      SerieAModifier := True;
    end;
    if (SerieAModifier) then
    begin
      //PutSerie(i, MySerie);
      Result += 1;
    end;
  end;
end;
// liste des annotations
procedure TToporobotStructure2012.AddAnnotation(const Annotation: TAnnotation);
begin
  FTableAnnotations.AddElement(Annotation);
end;


function TToporobotStructure2012.RemoveAnnotation(const Idx: integer): boolean;
begin
  Result := FTableAnnotations.RemoveElement(Idx);
end;
function TToporobotStructure2012.GetNbAnnotations(): integer;
begin
  Result := FTableAnnotations.GetNbElements;
end;

function TToporobotStructure2012.GetLastAnnotation(): TAnnotation;
var
  Nb: Integer;
begin
  Nb := self.GetNbAnnotations() - 1;
  Result := GetAnnotation(Nb);
end;



function TToporobotStructure2012.GetAnnotation(const NumAnnotation: integer): TAnnotation;
begin
  try
    Result := FTableAnnotations.GetElement(NumAnnotation);
  except
    Result := FTableAnnotations.GetElement(0);
  end;
end;

procedure TToporobotStructure2012.PutAnnotation(const NoAnnotation: integer; const AAnnotation: TAnnotation);
begin
  FTableAnnotations.PutElement(NoAnnotation, AAnnotation);
end;
// filtres perso
procedure TToporobotStructure2012.AddFiltrePerso(const FiltrePerso: TFiltrePersonnalise);
begin
  FTableFiltresPersonnalises.AddElement(FiltrePerso);
end;

function TToporobotStructure2012.GetFiltrePerso(const NumFiltrePerso: integer): TFiltrePersonnalise;
begin
  Result := FTableFiltresPersonnalises.GetElement(NumFiltrePerso);
  AfficherMessageErreur('GetFiltrePerso(): ' + Result.NomFiltre);
end;

procedure TToporobotStructure2012.PutFiltrePerso(const NoFiltrePerso: integer; const AFiltrePerso: TFiltrePersonnalise);
begin
  FTableFiltresPersonnalises.PutElement(NoFiltrePerso, AFiltrePerso);
end;

function TToporobotStructure2012.RemoveFiltrePerso(const Idx: integer): boolean;
begin
  FTableFiltresPersonnalises.RemoveElement(Idx);
end;

function TToporobotStructure2012.GetNbFiltresPersos(): integer;
begin
  Result := FTableFiltresPersonnalises.GetNbElements();
end;

function TToporobotStructure2012.GetLastFiltrePerso(): TFiltrePersonnalise;
var
  Nb: Integer;
begin
  Nb := self.GetNbFiltresPersos() - 1;
  Result := GetFiltrePerso(Nb);
end;

procedure TToporobotStructure2012.ViderListeFiltres();
begin
  FTablePointsOfInterest.ClearListe();
end;
// Points d'intérêt
procedure TToporobotStructure2012.AddPointOfInterest(const APOI: TPointOfInterest);
begin
  FTablePointsOfInterest.AddElement(APOI);
end;

function TToporobotStructure2012.GetPointOfInterest(const Idx: integer): TPointOfInterest;
begin
  Result := FTablePointsOfInterest.GetElement(Idx);
end;

procedure TToporobotStructure2012.PutPointOfInterest(const Idx: integer; const APOI: TPointOfInterest);
begin
  FTablePointsOfInterest.PutElement(Idx, APOI);
end;

function TToporobotStructure2012.RemovePointOfInterest(const Idx: integer): boolean;
begin
  FTablePointsOfInterest.RemoveElement(Idx);
end;

procedure TToporobotStructure2012.RecenserPointsOfInterest();
var
  POI: TPointOfInterest;
  NbSts, i, j, NbV: Integer;
  MySerie: TObjSerie;
  MyVisee: TUneVisee;
  MyPOI  : TPointOfInterest;
  WU: String;
  P: SizeInt;
begin

  AfficherMessage(Format('%s.RecenserPointsOfInterest', [ClassName]));
  FTablePointsOfInterest.ClearListe();
  // Cause fréquente d'erreurs: absence de la série 1
  NbSts := 0;
  AfficherMessage(Format('%d séries', [GetNbSeries()]));
  for i := 0 to GetNbSeries() - 1 do
  begin
    MySerie := GetSerie(i);
    NbV := MySerie.GetNbVisees();
    if (NbV > 0) then
    begin
      for j := 0 to NbV - 1 do
      begin
        MyVisee := MySerie.GetVisee(j);
        WU := Trim(LowerCase(MyVisee.IDTerrainStation + MyVisee.Commentaires));
        P  := Pos(KEYWORD_POI, WU);
        if (P > 0) then
        begin
          MyPOI.LabelTerrain := MyVisee.IDTerrainStation;
          MyPOI.Serie        := MySerie.GetNumeroDeSerie();
          MyPOI.Station      := j;
          MyPOI.Couleur      := clGray;
          MyPOI.Coordinates  := MakeTPoint3Df(0.00, 0.00, 0.00);
          MyPOI.Description  := StringReplace(MyVisee.Commentaires, KEYWORD_POI, '', [rfReplaceAll, rfIgnoreCase]);
          self.AddPointOfInterest(MyPOI);
        end;
      end;
      //*)
    end;
  end;
end;

function TToporobotStructure2012.GetNbPointsOfInterests(): integer;
begin
  Result := FTablePointsOfInterest.GetNbElements();
end;


//******************************************************************************
// Attrapper le premier IDx libre
function TToporobotStructure2012.getMaxIdxExpe(): TNumeroExpe;
var
  nb, i: Integer;
  EWE: TExpe;
begin
  result := -1;
  nb     := GetNbExpes();
  for i := 1 to nb - 1 do
  begin
    EWE := GetExpe(i);
    Result := Max(Result, EWE.IDExpe);
  end;
end;
function TToporobotStructure2012.getMaxIdxCode(): TNumeroCode;
var
  nb, i: Integer;
  EWE  : TCode;
begin
  result := -1;
  nb     := getNbCodes();
  for i := 1 to nb - 1 do
  begin
    EWE := getCode(i);
    Result := Max(Result, EWE.IDCode);
  end;
end;
// retourne le prochain index de série libre
function TToporobotStructure2012.getMaxIdxSerie(): TNumeroSerie;
var
  nb, i: Integer;
  EWE: TObjSerie;
  WU: TEntrance;
begin
  Result := -1;
  nb     := self.GetNbSeries();
  for i  := 0 to nb -1 do
  begin
    EWE := GetSerie(i);
    Result := Max(Result, EWE.GetNumeroDeSerie());
  end;
  (*
  nb     := self.GetNbEntrees();
  if (nb > 0) then
  begin
    for i  := 0 to nb -1 do
    begin
      WU := GetEntree(i);
      IdxS := WU.eRefSer;
      if (IdxS > Result) then Result := IdxS;
    end;
  end;
  //*)

end;


function TToporobotStructure2012.getMaxIdxReseau(): integer;
begin
  result := GetNbReseaux();
end;

function TToporobotStructure2012.getMaxIdxSecteur(): integer;
begin
  result := GetNbSecteurs();
end;



{$IFDEF SUPPORT_NOTES_VERSION}
// méthodes relatives aux notes de version

procedure TToporobotStructure2012.AddNoteVersion(const VA: TReleaseNotes);
begin
  FTableNotesVersion.AddElement(VA);
end;

function TToporobotStructure2012.GetNoteVersion(const No: integer): TReleaseNotes;
begin
  Result := FTableNotesVersion.GetElement(No);
end;

procedure TToporobotStructure2012.PutNoteVersion(const No: integer; const VA: TReleaseNotes);
begin
   FTableNotesVersion.PutElement(No, VA);
end;


function TToporobotStructure2012.GetNbNotesVersion: integer;
begin
  Result :=  FTableNotesVersion.GetNbElements();
end;
function TToporobotStructure2012.RemoveNoteVersion(const Idx: integer): boolean;
begin
  Result := FTableNotesVersion.RemoveElement(Idx);
end;

{$endif SUPPORT_NOTES_VERSION}


// En attente
// import de données depuis des fichiers CSV
procedure TToporobotStructure2012.AddEntrancesFromFile(const FileName: TStringDirectoryFilename; const DoReplaceExisting: boolean);
begin
  AfficherMessage(Format('%s.AddEntrancesFromFile: %s', [ClassName, FileName]));
end;
procedure TToporobotStructure2012.AddCodesFromFile(const FileName: TStringDirectoryFilename; const DoReplaceExisting: boolean);
begin
  AfficherMessage(Format('%s.AddCodesFromFile: %s', [ClassName, FileName]));
end;
procedure TToporobotStructure2012.AddExpesFromFile(const FileName: TStringDirectoryFilename; const DoReplaceExisting: boolean);
begin
  AfficherMessage(Format('%s.AddExpesFromFile: %s', [ClassName, FileName]));
end;
procedure TToporobotStructure2012.AddReseauxFromFile(const FileName: TStringDirectoryFilename; const DoReplaceExisting: boolean);
begin
  AfficherMessage(Format('%s.AddReseauxFromFile: %s', [ClassName, FileName]));
end;
procedure TToporobotStructure2012.AddSecteursFromFile(const FileName: TStringDirectoryFilename; const DoReplaceExisting: boolean);
begin
  AfficherMessage(Format('%s.AddSecteursFromFile: %s', [ClassName, FileName]));
end;
procedure TToporobotStructure2012.AddAntennesFromFile(const FileName: TStringDirectoryFilename; const DoReplaceExisting: boolean);
begin
  AfficherMessage(Format('%s.AddAntennesFromFile: %s', [ClassName, FileName]));
end;

// remplacement d'index
procedure TToporobotStructure2012.ReplaceIndexSeriesInEntrances(const OldIndex, NewIndex: TNumeroSerie);
var
  Nb, i: Integer;
  EWE: TEntrance;
begin
  AfficherMessage(Format('%s.ReplaceIndexSeriesInEntrances: %d -> %d', [ClassName, OldIndex, NewIndex]));
  Nb := GetNbEntrees();
  for i := 0 to Nb - 1 do
  begin
    EWE := GetEntree(i);
    if (EWE.eRefSer = OldIndex) then
    begin
      EWE.eRefSer := NewIndex;
    end;
  end;
end;

procedure TToporobotStructure2012.ReplaceIndexInStations(const OldIndex, NewIndex: integer; const ModeRemplacement: TModeRemplacementStations);
var
  EWE: String;
  NbSeries, i, NbVisees, j: Integer;
  MySerie: TObjSerie;
  MyVisee: TUneVisee;
begin
  EWE := ChooseString(Ord(ModeRemplacement), ['mrpSECTEURS', 'mrpCODES', 'mrpEXPES']);
  AfficherMessage(Format('%s.ReplaceIndexInStations: %d -> %d - %s', [ClassName, OldIndex, NewIndex, EWE]));
  NbSeries := GetNbSeries();
  for i := 0 to NbSeries - 1 do
  begin
    MySerie := GetSerie(i);
    NbVisees := MySerie.GetNbVisees();
    for j := 0 to NbVisees - 1 do
    begin
      MyVisee := MySerie.GetVisee(j);
      case ModeRemplacement of
        mrsSECTEURS: if (MyVisee.IDSecteur = OldIndex) then MyVisee.IDSecteur := NewIndex;
        mrsCODES   : if (MyVisee.Code = OldIndex)      then MyVisee.Code      := NewIndex;
        mrsEXPES   : if (MyVisee.Expe = OldIndex)      then MyVisee.Expe      := NewIndex;
      end;
      MySerie.PutVisee(j, MyVisee);
    end;
  end;
end;
procedure TToporobotStructure2012.ReplaceIndexInAntennes(const OldIndex, NewIndex: integer; const ModeRemplacement: TModeRemplacementAntennes);
var
  EWE: String;
  Nb, i: Integer;
  MyAntenne: TViseeAntenne;
begin
  AfficherMessage(Format('%s.ReplaceIndexInAntennes: %d -> %d', [Classname, OldIndex, NewIndex]));
  Nb := GetNbAntennes();
  if (Nb = 0) then Exit;
  for i := 0 to Nb -1 do
  begin
    MyAntenne := GetViseeAntenne(i);
    if (mraRESEAUX       in ModeRemplacement) then
      if (MyAntenne.Reseau = OldIndex) then  MyAntenne.Reseau   := NewIndex;
    if (mraSECTEURS      in ModeRemplacement) then
      if (MyAntenne.Secteur = OldIndex) then MyAntenne.Secteur  := NewIndex;
    if (mraSERIES        in ModeRemplacement) then
      if (MyAntenne.SerieDepart = OldIndex) then MyAntenne.SerieDepart   := NewIndex;
    if (mraCODES         in ModeRemplacement) then
      if (MyAntenne.Code = OldIndex) then MyAntenne.Code   := NewIndex;
    if (mraEXPES         in ModeRemplacement) then
      if (MyAntenne.Expe = OldIndex) then MyAntenne.Expe   := NewIndex;
    PutViseeAntenne(i, MyAntenne);
  end;
end;
procedure TToporobotStructure2012.ReplaceIdxCodes(const OldIndex, NewIndex: TNumeroCode);
var
  Nb, i: Integer;
  EWE: TCode;
begin
  AfficherMessage(Format('%s.ReplaceIdxCodes: %d -> %d', [Classname, OldIndex, NewIndex]));
  Nb := GetNbCodes();
  for i := 0 to Nb - 1 do
  begin
    EWE := GetCode(i);
    if (EWE.IDCode = OldIndex) then
    begin
      EWE.IDCode := NewIndex;
      PutCode(i, EWE);
    end;
  end;
end;

procedure TToporobotStructure2012.ReplaceIdxExpes(const OldIndex, NewIndex: TNumeroExpe);
var
  EWE: TExpe;
  Nb, i: Integer;
begin
  AfficherMessage(Format('%s.ReplaceIdxExpes: %d -> %d', [ClassName, OldIndex, NewIndex]));
  Nb := GetNbCodes();
  for i := 0 to Nb - 1 do
  begin
    EWE := GetExpe(i);
    if (EWE.IDExpe = OldIndex) then
    begin
      EWE.IDExpe := NewIndex;
      PutExpe(i, EWE);
    end;
  end;
end;



procedure TToporobotStructure2012.ReplaceIdxSeries(const OldIndex, NewIndex: TNumeroSerie);
var
  Nb, i: Integer;
  MySerie: TObjSerie;
begin
  AfficherMessage(Format('%s.ReplaceIdxSeries: %d -> %d', [ClassName, OldIndex, NewIndex]));
  Nb := GetNbSeries();
  for i := 0 to Nb - 1 do
  begin
    MySerie := GetSerie(i);
    if (OldIndex = MySerie.GetNumeroDeSerie()) then MySerie.SetNumeroSerie(NewIndex);
    if (OldIndex = MySerie.GetNoSerieDep())    then MySerie.SetNoSerieDep(NewIndex);
    if (OldIndex = MySerie.GetNoSerieArr())    then MySerie.SetNoSerieArr(NewIndex);
  end;
end;



end.

